/*
 * acc_gl.c -
 *  Wrapper for rendering using the OpenGL API
 */

/*
 * File:    accel/acc_gl.c
 * Purpose: Wrapper for OpenGL low-level rendering functions under "accel"
 * Prefix:  Acc_ (suffix _gl)
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"
#if !HAVE_OPENGL
#error Must define HAVE_OPENGL in src/defines.h when compiling for OpenGL!
#endif

#if defined(__WIN32__) || defined(_MSC_VER)
#include "windows.h" /* Win32 requires this before gl.h */
#endif
#include <GL/gl.h>

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "accel.h"
#include "acc_gl.h"
#include "bitmap.h"
#include "graph2d.h"
#include "hud.h"
#include "main.h"
#include "render.h"
#include "shade.h"
#include "surface.h"





/*
 * Local function prototypes
 */
static void draw_fragmented_sprite_notimp (vec *pos, mat *orn,
					   struct tag_ss_bitmap_header *bmp,
					   byte *zarray, int frag_size);
static void draw_fragment_notimp (vec *pos, uint32 rgb, int frag_size);
static void draw_surface_notimp (surface *s);
static void draw_sprite_notimp (surface *s, ss_bitmap_header *bmp);
static void draw_overlay (hudLayer *pLayer, camera *pCam);





/*
 * Initialise the OpenGL subsystem, setting up some sensible initial state
 */
void Acc_init_gl (void)
{

  extern int Screen_width, Screen_height;

  /*
   * Give OpenGL a window to play with
   */
  glViewport (0, 0, Screen_width, Screen_height);

  /*
   * Background colour is black
   */
  glClearColor (0., 0., 0., 0.);

	/* ff1: the following two blocks make it look great */
	/*			but somewhat misplaced, since we wont need it */
	/*			while rendering the real world. */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(1.2f);

	glEnable(GL_FOG);
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_DENSITY, 0.001f);
	glFogf(GL_FOG_START, 1500.0f);
	glFogf(GL_FOG_END, 2000.0f);

	/*
	 * Enable Alpha blending and set the mode
	 * (Needed for draw_overlay)
	 */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

}





/*
 * Prep the OpenGL renderer to render a scene.
 * Send the global state that we need to OpenGL and set the Acc_ prefixed
 *  interface functions
 */
void Acc_setup_scene_gl (camera *c)
{

  GLdouble width_ratio;
  GLdouble height_ratio;
  extern double Min_view_distance;
  extern double Max_view_distance;
  GLdouble transmat [16];

  assert (c->flags & CF_3D_API);

  Camera = c;

/*    printf ("Acc_setup_scene_gl()\n"); */

  /*
   * Clear the screen and the depth buffer
   */
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /*
   * View frustrum goes in the projection matrix
   */
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  /*
   * Set the view frustrum. This is determined by the "screen_z" field of the
   *  camera definition
   * Probably this should be set but once, when the FOV is defined, but since
   *  it's here ...
   */
  width_ratio  = 0.5 * c->window_width  / c->screen_z;
  height_ratio = 0.5 * c->window_height / c->screen_z;
/*    printf ("%g %g\n", Min_view_distance, Max_view_distance); */
  glFrustum (-width_ratio, width_ratio, -height_ratio, height_ratio,
  	     1., 32768);
/*    glFrustum (-1.333, 1.333, -1, 1, 1, 100); */
/*    glOrtho (-1600., 1600., -1200., 1200., -10000., 1000); */

  /*
   * Set the view transform for this scene
   */
  glMatrixMode (GL_MODELVIEW);

  /*
   * Camera orientation. OpenGL uses 4x4 matrices and homogeneous coordinates.
   *  It also uses a right-handed coordinate system (z out of screen) and
   *  specifies its matrices down-then-across (TSSHP uses across-then-down). So
   *  some conversion work is requred ...
   */
  transmat [ 0] =  c->orientation.x.x;
  transmat [ 1] =  c->orientation.y.x;
  transmat [ 2] = -c->orientation.z.x;
  transmat [ 3] =  0.;
  transmat [ 4] =  c->orientation.x.y;
  transmat [ 5] =  c->orientation.y.y;
  transmat [ 6] = -c->orientation.z.y;
  transmat [ 7] =  0.;
  transmat [ 8] =  c->orientation.x.z;
  transmat [ 9] =  c->orientation.y.z;
  transmat [10] = -c->orientation.z.z;
  transmat [11] =  0.;
  transmat [12] =  0.;
  transmat [13] =  0.;
  transmat [14] =  0.;
  transmat [15] =  1.;
  glLoadMatrixd (transmat);

  /*
   * Camera position. The translation should take place in world space; GL
   *   post-multiplies its transforms, so this comes last
   */
  glTranslated (-c->position.x, -c->position.y, -c->position.z);

  /*
   * Set the "accelerated" interface to point at OpenGL
   */
  Acc_draw_line = Acc_draw_line_gl;
  Acc_draw_fragmented_sprite = draw_fragmented_sprite_notimp;
  Acc_draw_fragment          = draw_fragment_notimp;
  Acc_draw_surface           = draw_surface_notimp;
  Acc_draw_sprite            = draw_sprite_notimp;
  Acc_draw_overlay					 = draw_overlay;
}




void Acc_end_scene_gl (void)
{

  /*
   * The scene is drawn, flush it to the screen
   */
  glFlush ();

}



/*
 * A line is a GL primitive, so there's not much work to do on our end
 * It would be more efficient, from an OpenGL point of view at least, to pass
 *  all the lines for a given tile in one go so that they could be handled as
 *  a single glBegin() ... glEnd() block. Since they are already batched in
 *  the tile's line queue that shouldn't be hard.
 */
void Acc_draw_line_gl (line_seg *l, const clipregion_t *frustrum)
{

  glBegin (GL_LINES);

  /*
   * Start point
   */
  glColor3ub (get_red (l->rgb1), get_green (l->rgb1), get_blue (l->rgb1));
  glVertex3d (l->v1.x, l->v1.y, l->v1.z);

  /*
   * End point
   */
  glColor3ub (get_red (l->rgb2), get_green (l->rgb2), get_blue (l->rgb2));
  glVertex3d (l->v2.x, l->v2.y, l->v2.z);

  /*
   * And that's all there is to it
   */
  glEnd ();

# if __DEBUG__
  {
    GLenum err;
    if ((err = glGetError ()) != GL_NO_ERROR)
    {
      printf ("%s\n", glGetString (err));
      Cant_happen ("Acc_draw_line_gl(): OpenGL error");
    }
  }    
# endif

}


void draw_fragmented_sprite_notimp (vec *pos, mat *orn,
				    struct tag_ss_bitmap_header *bmp,
				    byte *zarray, int frag_size)
{
}


void draw_fragment_notimp (vec *pos, uint32 rgb, int frag_size)
{
}


void draw_surface_notimp (surface *s)
{
}


void draw_sprite_notimp (surface *s, ss_bitmap_header *bmp)
{
}


/*
 * Draw overlayed images (typically HUD)
 * Because a) paletted images are damn slow and b) they do not
 * apply to alpha blending anyway, we convert them to TrueColor
 * at first. It is advised to use TrueColor images firsthand.
 */ 

void draw_overlay (hudLayer *pLayer, camera *pCam)
{
	g2dbase gbBuffer = D_G2D_NULLAREA;

	/* reset the matrices for glRasterPos */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
  glOrtho(0, pCam->window_width, 0, pCam->window_height, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	while (pLayer)
	{
		if (pLayer->pG2DB)
		{
			assert(pLayer->pG2DB->bImg.clr != NULL);
			glPixelZoom((double)pLayer->szSize.cx / (double)getRectWidth(pLayer->pG2DB->bBoxArea),
									-((double)pLayer->szSize.cy / (double)getRectHeight(pLayer->pG2DB->bBoxArea)));
			/* luckily GL supports areas within images. does it? */
			glPixelStorei(GL_UNPACK_SKIP_PIXELS, pLayer->pG2DB->bBoxArea.left);
			glPixelStorei(GL_UNPACK_SKIP_ROWS, pLayer->pG2DB->bBoxArea.top);
			glPixelStorei(GL_UNPACK_ROW_LENGTH, pLayer->pG2DB->bSizeImg.cx);

			glRasterPos3f(pLayer->ptPos.x, 
										pCam->window_height - pLayer->ptPos.y, -1);
			switch (pLayer->pG2DB->nType)
			{
				case G2DTrue:
				{	/* that one's really easy */
					glDrawPixels(getRectWidth(pLayer->pG2DB->bBoxArea),
											 getRectHeight(pLayer->pG2DB->bBoxArea),
											 GL_BGRA_EXT, GL_UNSIGNED_BYTE, 
											 &pLayer->pG2DB->bImg.clr[pLayer->pG2DB->bBoxArea.left]);
					break;
				}
				case G2DPal:
				{	/* ff1: i didn't expect that this is actually fast! */
					/*			much faster than paletted draw. but it still */
					/*			might become slow on high-res with big images... */
					G2D_init_clrbase(&gbBuffer, 
													 (uint16)getRectWidth(pLayer->pG2DB->bBoxArea),
													 (uint16)getRectHeight(pLayer->pG2DB->bBoxArea));
					G2Dc_pal_to_clr(&gbBuffer, pLayer->pG2DB);
					glDrawPixels(gbBuffer.bSizeImg.cx, gbBuffer.bSizeImg.cy,
											 GL_BGRA_EXT, GL_UNSIGNED_BYTE, 
											 gbBuffer.bImg.clr);
					G2D_reset_base(&gbBuffer, true);
					break;
				}
				case G2DPixel:
				{	/* basically just a Pal with RGB_trans as palette */
					if (BITS_PER_PIXEL < 24)
					{
						G2D_init_clrbase(&gbBuffer, 
														 (uint16)getRectWidth(pLayer->pG2DB->bBoxArea),
														 (uint16)getRectHeight(pLayer->pG2DB->bBoxArea));
						G2D_pixel_to_color(&gbBuffer, pLayer->pG2DB, RGB_trans);
						glDrawPixels(gbBuffer.bSizeImg.cx, gbBuffer.bSizeImg.cy,
												 GL_BGRA_EXT, GL_UNSIGNED_BYTE, 
												 gbBuffer.bImg.clr);
						G2D_reset_base(&gbBuffer, true);
					}
					else
						glDrawPixels(getRectWidth(pLayer->pG2DB->bBoxArea),
												 getRectHeight(pLayer->pG2DB->bBoxArea),
												 GL_BGRA_EXT, GL_UNSIGNED_BYTE, 
												 &pLayer->pG2DB->bImg.clr[pLayer->pG2DB->bBoxArea.left]);
					break;
				}
				case G2DMask:	
				default:
				{
					Cant_happen("draw_overlay_gl(): Unrenderable Layer");
					break;
				}
			}
		}
		pLayer = pLayer->pNext;
	}
}

