/*
 * accel.c -
 *  3D acceleration (if it's available)
 */

/*
 * File:    accel/model.c
 * Purpose: Wrapper for accelerated (and not) low-level rendering functions
 * Prefix:  Acc_
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdlib.h>

#if HAVE_OPENGL
#include "acc_gl.h"
#endif

#include "accel.h"
#include "main.h"
#include "render.h"
#include "software.h"
#include "surface.h"



/*
 * Globals for the accelerated interface functions
 */
void (*Acc_draw_line) (struct tag_line_seg *l,
		       const struct tag_clip_region *frustrum);
void (*Acc_draw_fragmented_sprite) (vec *pos, mat *orn,
				    struct tag_ss_bitmap_header *bmp,
				    byte *zarray, int frag_size);
void (*Acc_draw_fragment) (vec *pos, uint32 rgb, int frag_size);
void (*Acc_draw_surface) (surface *s);
void (*Acc_draw_sprite)  (surface *s, struct tag_ss_bitmap_header *bmp);
void (*Acc_draw_overlay) (struct tag_hudLayer *pLayer,
													struct tag_camera *pCam);



/*
 * Initialise the accelerated back-ends. This should be called from the GUI
 *  subsystem under GUI_init_screen()
 * The `backends' parameter is a bitfield containing a combination of the
 *  ACCEL_ flags in accel.h . The GUI will set this to whatever back-ends it
 *  understands and are configured
 * Accelerated subsystems are initialised if their flag in backends is set.
 *  Any initialisation that can be done here should be, rather than in the
 *  GUI subsystem, since more than one GUI might support a given acceleration
 *  type (e.g. both GLUT and the SDL support OpenGL).
 * The ACCEL_SOFTWARE flag is ignored. The software renderer will always be
 *  initialised.
 */
void Acc_init (int backends)
{

  Acc_init_software (); /* always init the software renderer */

# if HAVE_OPENGL
  if (backends & ACCEL_OPENGL)
  {
    Acc_init_gl ();
  }
# endif

}




static unsigned int Scene_type = 0;


/*
 * Prep a scene for rendering. This just calls whichever setup_scene() function
 *  is relevant to the camera and to the selected rendering scheme
 */
void Acc_setup_scene (camera *c)
{

  assert (Scene_type == 0);

  if (c->flags & CF_3D_API)
  {
    /*
     * Camera wants accelerated rendering. See what we've got
     */
#   if HAVE_OPENGL
    Acc_setup_scene_gl (c);
    Scene_type = ACCEL_OPENGL;
#   else
    Cant_happen ("Acc_setup_scene(): accelerated rendering selected but no"
		 " renderer available");
#   endif
  }
  else
  {
    /*
     * Software rendering is always allowed
     */
    Acc_setup_scene_software (c);
    Scene_type = ACCEL_SOFTWARE;
  }
}




void Acc_end_scene (void)
{
  assert (Scene_type != 0);

  switch (Scene_type)
  {
  case ACCEL_SOFTWARE:
    Acc_end_scene_software ();
    break;
# if HAVE_OPENGL
  case ACCEL_OPENGL:
    Acc_end_scene_gl ();
    break;
# endif
  default:
    Cant_happen ("Acc_end_scene(): invalid scene type");
  }

  Scene_type = 0;
}
