/*
 * software.c -
 *  "Accelerated" wrapper for software rendering
 */

/*
 * File:    accel/software.c
 * Purpose: Wrapper for software low-level rendering functions under "accel"
 * Prefix:  Acc_ (suffix _software)
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <math.h>
#include <stdlib.h>

#include "accel.h"
#include "clip.h"
#include "main.h"
#include "plot.h"
#include "render.h"
#include "software.h"
#include "surface.h"



/*
 * Should probably be moved here or at least somewhere in accel/
 */
extern Vector Viewpoint;
extern Matrix Orientation;
extern int Origin_x;
extern int Origin_y;
extern int Window_width;
extern int Window_height;




/*
 * Local function prototypes
 */
static float depth_cue_level (const vec *v);
static uint32 depth_cue_rgb (const vec *v, uint32 rgb);





/*
 * Initialise the software renderer.
 * Should probably do something here
 */
void Acc_init_software (void)
{
}




/*
 * Prep the software renderer to render a scene.
 * Set up the globals for framebuffer and window dimensions and the Acc_
 *  prefixed interface functions
 */
void Acc_setup_scene_software (camera *c)
{

  /*
   * Paranoia
   */
  if (c->framebuffer == NULL)
  {
    Cant_happen ("Map_render_scene(): software rendering selected but no"
		 " frame buffer supplied");
  }

  Camera = c;

  /*
   * I'm still not sure all these globals are necessary
   */
  Viewpoint     = Camera->position;
  Orientation   = Camera->orientation;
  Origin_x      = Camera->window_width / 2;
  Origin_y      = Camera->window_height / 2;
  Window_width  = Camera->window_width;
  Window_height = Camera->window_height;

  /*
   * Set the "accelerated" interface to point at the software renderer
   */
  Acc_draw_line = Acc_draw_line_software;
  Acc_draw_fragmented_sprite = Plot_fragmented_sprite;
  Acc_draw_fragment          = Plot_fragment;
  Acc_draw_surface           = Plot_surface;
  Acc_draw_sprite            = Plot_sprite;
  Acc_draw_overlay	     = Plot_overlay;

}



void Acc_end_scene_software (void)
{
  /*
   * Nothing to do here
   */
}



void Acc_draw_line_software (line_seg *l, const clipregion_t *frustrum)
{

  vec tv, view1, view2;
  uint32 rgb1, rgb2;

  /*
   * Clip the line against the view frustrum
   */
  if (!Clip_line_by_region (&(l->v1), &(l->v2), frustrum)) return; 

  /*
   * It's passed the clip, transform and render
   */
  vec_subtract (tv, l->v1, Viewpoint);
  mat_vec_multiply (view1, Orientation, tv);
  vec_subtract (tv, l->v2, Viewpoint);
  mat_vec_multiply (view2, Orientation, tv);

  /*
   * Depth cueing
   */
  rgb1 = depth_cue_rgb (&view1, l->rgb1);
  rgb2 = depth_cue_rgb (&view2, l->rgb2);

  if (rgb1 == rgb2)
  {
    Line_draw (&view1, &view2, Camera->framebuffer, rgb1);
  }
  else
  {
    Line_draw_inter (&view1, &view2, Camera->framebuffer, rgb1, rgb2);
  }

}





/*
 * The distance (squared) at which depth cueing starts
 */
static double Depth_cue_min = 768;

/*
 * The fall-off factor
 */
static double Depth_cue_factor = 0.125 / 256;

/*
 * Depth cueing routines. We handle depth cueing here rather than earlier
 *  because it should normally be done in hardware
 * Vertices passed in should already have been transformed
 */
float depth_cue_level (const vec *v)
{

  double dist2 = vec_magnitude2 (*v);
  if (dist2 < Depth_cue_min * Depth_cue_min) return 1.;

  /*
   * Level at exactly the minimum distance should be 1, thereafter it should
   *  fall off as the inverse square
   */
  dist2 = (dist2 - Depth_cue_min * Depth_cue_min) *
    Depth_cue_factor * Depth_cue_factor + 1;
  return 1. / dist2;

}



uint32 depth_cue_rgb (const vec *v, uint32 rgb)
{
  int r = rgb & 0xff;
  int g = (rgb >> 8) & 0xff;
  int b = (rgb >> 16) & 0xff;
  float level = depth_cue_level (v);
  r *= level;
  g *= level;
  b *= level;
  return r | (g << 8) | (b << 16);
}
