/*
 * sdl.c -
 *  stuff for talking to SDL
 *
 * This file by Damian Gryski 2000 dgryski@uwaterloo.ca
 *    based on x11.c by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 *    adopted and enhanced by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>		/* malloc() */
#include <memory.h>		/* memset() */

#include "SDL.h"

#include "types.h"
#include "graph2d.h"
#include "accel.h"
#include "control.h"
#include "event.h"
#include "gamemem.h"
#include "hud.h"
#include "main.h"           /* Cant_happen() */
#include "option.h"
#include "palette.h"
#include "physics.h"
#include "render.h"
#include "shade.h"
#include "timer.h"
#include "version.h"   /* version no. displayed in title */



/*
 * We want pi!
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif

#define FAST_SCREEN_COPY 1

/*
 * Local function prototypes
 */

uint16	gui_get_button(Uint8 fButtons);

/*
 * Global variables
 */
extern int Screen_width;
extern int Screen_height;

static SDL_Surface *screen = NULL;
static pixel *Screen_bits = NULL;

static const SDL_VideoInfo *Vid_info;

uint16	gKeyMap[SDLK_LAST];



void GUI_init (int *argcp, char *argv [])
{

	uint16 i;
	UNUSED(argcp);
	UNUSED(argv);

  if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) != 0)
  {
    printf ("%s\n", SDL_GetError ());
    Cant_happen ("GUI_init(): SDL initialise failed");
  }

  /*
   * This might be of interest
   */
  Vid_info = SDL_GetVideoInfo ();
  assert (Vid_info != NULL);
  printf ("Got SDL Video information : flags follow\n");
  if (Vid_info->hw_available) printf (" * HW available\n");
  if (Vid_info->wm_available) printf (" * WM available\n");
  if (Vid_info->blit_hw)      printf (" * Hardware blit accn\n");
  if (Vid_info->blit_hw_CC)   printf (" * Hardware CC blit accn\n");
  if (Vid_info->blit_hw_A)    printf (" * Hardware alpha blit accn\n");
  if (Vid_info->blit_sw)      printf (" * Software blit accn\n");
  if (Vid_info->blit_sw_CC)   printf (" * Software CC blit accn\n");
  if (Vid_info->blit_sw_A)    printf (" * Software alpha blit accn\n");
  if (Vid_info->blit_fill)    printf (" * Accelerated fills\n");
  printf ("Total video RAM %uK\n", Vid_info->video_mem);
  printf ("Preferred pixel format is :\n");
  printf (" * %d bits\n", Vid_info->vfmt->BitsPerPixel);
  printf (" * %d bytes\n", Vid_info->vfmt->BytesPerPixel);
  printf (" * Red mask    %08x\n", Vid_info->vfmt->Rmask);
  printf (" * Green mask  %08x\n", Vid_info->vfmt->Gmask);
  printf (" * Blue mask   %08x\n", Vid_info->vfmt->Bmask);
  printf (" * Alpha mask  %08x\n", Vid_info->vfmt->Amask);
  printf (" * Red shift   %d\n", Vid_info->vfmt->Rshift);
  printf (" * Green shift %d\n", Vid_info->vfmt->Gshift);
  printf (" * Blue shift  %d\n", Vid_info->vfmt->Bshift);
  printf (" * Alpha shift %d\n", Vid_info->vfmt->Ashift);
  printf (" * Red loss    %d\n", Vid_info->vfmt->Rloss);
  printf (" * Green loss  %d\n", Vid_info->vfmt->Gloss);
  printf (" * Blue loss   %d\n", Vid_info->vfmt->Bloss);
  printf (" * Alpha loss  %d\n", Vid_info->vfmt->Aloss);
  printf (" * CKey %08x\n", Vid_info->vfmt->colorkey);
  printf (" * Surface alpha %02x\n", Vid_info->vfmt->alpha);

	/* init keymappings */
	memset(gKeyMap, 0x00, sizeof(uint16) * SDLK_LAST);
	for (i = 0; i < 128; i++)
		gKeyMap[i] = i;			/* first 128 are ident to ASCII table... */
	for (i = 'A'; i <= 'Z'; i++)
		gKeyMap[i] = 'a' + (i - 'A');
	for (i = 0; i < 16; i++)
		gKeyMap[SDLK_KP0 + i] = KM_KP_0 + i;
	for (i = 0; i < 12; i++)
		gKeyMap[SDLK_F1 + i] = KM_F01 + i;
	gKeyMap[SDLK_UP]			= KM_UP;
	gKeyMap[SDLK_DOWN]		= KM_DOWN;
	gKeyMap[SDLK_RIGHT]		= KM_RIGHT;
	gKeyMap[SDLK_LEFT]		= KM_LEFT;
	gKeyMap[SDLK_INSERT]	= KM_INSERT;
	gKeyMap[SDLK_HOME]		= KM_HOME;
	gKeyMap[SDLK_END]			= KM_END;
	gKeyMap[SDLK_PAGEUP]	= KM_PGUP;
	gKeyMap[SDLK_PAGEDOWN] = KM_PGDN;

	gKeyMap[SDLK_NUMLOCK] = KM_MOD_NUM;
	gKeyMap[SDLK_CAPSLOCK] = KM_MOD_CAPS;
	gKeyMap[SDLK_SCROLLOCK] = KM_MOD_SCROLL;
	gKeyMap[SDLK_RSHIFT] = KM_MOD_RSHIFT;
	gKeyMap[SDLK_LSHIFT] = KM_MOD_LSHIFT;
	gKeyMap[SDLK_RCTRL] = KM_MOD_RCTRL;
	gKeyMap[SDLK_LCTRL] = KM_MOD_LCTRL;
	gKeyMap[SDLK_RALT] = KM_MOD_RALT;
	gKeyMap[SDLK_LALT] = KM_MOD_LALT;
}

void GUI_init_screen (void)
{

  Uint32 flags = 0x00000000;
  static const char title [] = "shack (TSSHP v" TSSHP_VERSION_STRING ")";

  /*
   * Something other than "SDL_app" ;)
   */
  SDL_WM_SetCaption (title, "");

	if (Option_fullscreen)
		flags |= SDL_FULLSCREEN;

# if HAVE_OPENGL
  flags |= SDL_OPENGL;
# endif

  /*
   * Need this for X otherwise shack goes psychedelic (look at the colours,
   *  man!)
   */
	if (BITS_PER_PIXEL < 24) flags |= SDL_HWPALETTE;
  screen = SDL_SetVideoMode (Screen_width, Screen_height, 
														 BITS_PER_PIXEL, flags);
# if (BITS_PER_PIXEL < 24)
  {
    SDL_Color cmap[256];
    int i;
		for (i = 0; i < 256; i++)
			{
				cmap[i].r = Palette[i][0];
				cmap[i].g = Palette[i][1];
				cmap[i].b = Palette[i][2];
			}

		SDL_SetPalette (screen, SDL_PHYSPAL, cmap, 0, 256);
	}
# endif

# if HAVE_OPENGL
  Acc_init (ACCEL_SOFTWARE | ACCEL_OPENGL);
# else
  Acc_init (ACCEL_SOFTWARE);
# endif

  /*
   * jim - the GUI subsystem now has responsibility over the framebuffer
   */
# if FAST_SCREEN_COPY
  Screen_bits = screen->pixels;
# else
  Screen_bits = Alloc_init (Screen_width * Screen_height * sizeof (pixel),
                            "GUI_init", "framebuffer");
# endif
	SDL_ShowCursor(0);
	HUD_show_cursor(0);
}



pixel *GUI_get_framebuffer (void)
{
  return Screen_bits;
}


void GUI_loop (void)
{
  SDL_Event ev;
	ss_controlset ssControl;
	uint32 now, then, nFrames = 0, nFrameStart, nFrameStop;
  double elapsed, dblFrames;
	point ptMouse;
	int mousex, mousey;
	char strFps[50];
	short nFpsMsg = -1;
	bool quit = false, bFps = true;
	size_t sImageSize;

  nFrameStart = then = SDL_GetTicks();

	sImageSize = Screen_width * Screen_height * sizeof(pixel);
	HUD_set_cursor_shade(0xB0);

	HUD_get_cursor_pos(&ptMouse);
	SDL_WarpMouse((Uint16)ptMouse.x, (Uint16)ptMouse.y);
  while (!quit)
  {
		now = SDL_GetTicks();
		elapsed = (now - then) / 1.e3;

		Timer_work(now);
		Evt_work();
		Phy_frame(elapsed);

    if (Screen_bits != NULL) /* might be NULL if we are using hardware */
    {
      memset(Screen_bits, 0x00, sImageSize);
    }
		Ren_frame(elapsed, Screen_bits);

#   if HAVE_OPENGL
    SDL_GL_SwapBuffers ();
#   else
#   if !FAST_SCREEN_COPY
		memcpy(screen->pixels, Screen_bits, sImageSize);
#   endif
		SDL_UpdateRect(screen, 0, 0, 0, 0);
#   endif

    while (SDL_PollEvent (&ev))
		{	/* loop through the events */
			switch (ev.type)
			{
				case SDL_KEYDOWN:
					if (ev.key.keysym.sym >= SDLK_NUMLOCK && 
							ev.key.keysym.sym <= SDLK_COMPOSE)
						Con_key_modifier(gKeyMap[ev.key.keysym.sym], true);
					else
						Con_key_activate(gKeyMap[ev.key.keysym.sym], now);
					break;
				case SDL_KEYUP:
					if (ev.key.keysym.sym >= SDLK_NUMLOCK && 
							ev.key.keysym.sym <= SDLK_COMPOSE)
						Con_key_modifier(gKeyMap[ev.key.keysym.sym], false);
					else
						Con_key_deactivate(gKeyMap[ev.key.keysym.sym], now);
					break;
#ifndef NO_MOUSE
				case SDL_MOUSEMOTION:
					HUD_get_cursor_pos(&ptMouse);
					SDL_GetMouseState(&mousex, &mousey);
					if ((ptMouse.x != mousex || ptMouse.y != mousey) &&
							(ev.motion.xrel || ev.motion.yrel))
					{	/* no endless loop */
						Con_move(MOVE_MOUSEX, ev.motion.xrel, now);
						Con_move(MOVE_MOUSEY, ev.motion.yrel, now);
						HUD_get_cursor_pos(&ptMouse);
						SDL_WarpMouse((Uint16)ptMouse.x, (Uint16)ptMouse.y);
					}
		      break;
				case SDL_MOUSEBUTTONUP:
					ptMouse.x = ev.motion.x;
					ptMouse.y = ev.motion.y;
					Con_mouse_deactivate(gui_get_button(ev.button.button), 
															 ptMouse, now);
					break;
				case SDL_MOUSEBUTTONDOWN:
					ptMouse.x = ev.motion.x;
					ptMouse.y = ev.motion.y;
					Con_mouse_activate(gui_get_button(ev.button.button), 
														 ptMouse, now);
					break;
#endif
		    case SDL_QUIT:
					quit = true;
					break;
		    default:		/* nothing */
					break;
	    }
		}
		Con_snapshot(&ssControl, false);
		if (ssControl.quit)
			quit = true;
		nFrameStop = SDL_GetTicks();
		nFrames++;
		if ((nFrameStop - nFrameStart) >= 200)
		{
			if (bFps)
			{
				dblFrames = 1000.0 / ((double)(nFrameStop - nFrameStart) / (double)(nFrames));
				sprintf(strFps, "fps: %g", dblFrames);
				if (nFpsMsg < 0)
					nFpsMsg = MSG_add_message(F_MSG_STATIC, strFps, 0, 0, 
																		rgbt(0x00, 0x00, 0xFF, 0xFF),
																		rgbt(0x00, 0x00, 0xFF, 0x40));
				else
					MSG_edit_message(nFpsMsg, F_MSG_MSK_TEXT, 0,
													 strFps, 0, 0, 0, 0);
			}
			nFrameStart = nFrameStop;
			nFrames = 0;
		}
		then = now;
  }
	SDL_ShowCursor(1);

	/*
	 * jim - dump the framebuffer (hope nothing tries to use it now!)
	 */
# if !FAST_SCREEN_COPY
	free (Screen_bits);
# endif
  SDL_Quit();
}

uint16	gui_get_button(Uint8 fButtons)
{
	uint16 fNewButtons = 0x0000;

  switch (fButtons)
  {
    case SDL_BUTTON_LEFT:		fNewButtons = KM_MB_LEFT;		break;
    case SDL_BUTTON_MIDDLE: fNewButtons = KM_MB_CENTER; break;
    case SDL_BUTTON_RIGHT:	fNewButtons = KM_MB_RIGHT;	break;
    default: break;
  }

	return fNewButtons;
}



/*
 * End of gui/sdl/sdl.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
