/*
 * x11.c -
 *  stuff for talking to X
 */

/*
 * File:    gui/x11/x11.c
 * Purpose: Front-end GUI interface for X11 / XFree86
 * Prefix:  GUI_
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"
#include "types.h"
#include "graph2d.h"

#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>  /* malloc() */

#include <sys/time.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>

#include "shade.h"
#include "control.h"
#include "game.h"
#include "gamemem.h"
#include "hud.h"
#include "main.h"     /* Cant_happen() */
#include "map.h"
#include "mfd.h"
#include "option.h"
#include "palette.h"
#include "physics.h"
#include "render.h"
#include "resource.h"
#include "texture.h"
#include "timer.h"
#include "event.h"

/*
 * We want pi!
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif



/*
 * Just testing
 */
extern ss_bitmap_header *Texture_128;


/*
 * Local function prototypes
 */
static void quit_func (Widget w, XtPointer client_data, XtPointer call_data);
static void screen_func (Widget w, XtPointer client_data, XEvent *event,
			 Boolean *cont);
#if MOUSE
static void mouse_func (Widget w, XtPointer client_data, XEvent *event,
			Boolean *cont);
#endif
static void key_func (Widget w, XtPointer client_data, XEvent *event,
		      Boolean *cont);
static void focus_func (Widget w, XtPointer client_data, XEvent *event,
			Boolean *cont);

static void refresh_timer_func (XtPointer client_data, XtIntervalId *id);


uint16	gui_get_modifier(KeySym nKeySym);

/*
 * Global variables
 */
static XtAppContext App_context;
static Widget Top_level_widget;
static Widget Screen_widget;
static Display *Screen_display;

GC Screen_gc;


static pixel *Screen_bits;
static XImage *Screen_image;

XtIntervalId Refresh_timer;
int Refresh_delay = 29;

char Title [300];

int Coltrans [256];

/*
 * X keysym -> TSSHP keymap table
 * Bits 8-15 of an X keysym are (in my keysym.h)
 *  0x00 ISO Latin-1
 *  0x01 ISO Latin-2
 *  0x02 ISO Latin-3
 *  0x03 ISO Latin-4
 *  0x04 Katakana
 *  0x05 Arabic
 *  0x06 Cyrillic (top half of KOI8-R)
 *  0x07 Greek
 *  0x08 Technical
 *  0x09 Special
 *  0x0A Publishing
 *  0x0B APL
 *  0x0C Hebrew
 *  0x0D Thai
 *  0x0E Hangul
 *  0x20 Currency
 *  0xFD 3270 Terminal Keys
 *  0xFE ISO 9995 function and modifier keys
 *  0xFF Function keys
 *
 * Wow.
 * The only ones of the above that are actually interesting from our point of
 *  view are the function keys 0xFFxx, so a 256-byte table will suffice for
 *  the mapping.
 */
uint16 X_to_tsshp_fkey [256];

/*
 * Was auto-repeat allowed before we started fiddling with it?
 */
int Orig_auto_repeat;


extern int Screen_width;
extern int Screen_height;

static struct timeval now = {0, 0}, tvStart = { 0, 0 };

void GUI_init (int *argcp, char *argv [])
{
	uint16 i;

  /*
   * X!
   */
  Top_level_widget = XtAppInitialize (&App_context,
				      "shack",
				      NULL,
				      0,
				      argcp,
				      argv,
				      NULL,
				      NULL,
				      0);

  /*
   * Initialise the timer
   */
  gettimeofday (&now, NULL);
	
  /*
   * Intialise the keymap from X to TSSHP function keys
   */
  memset(X_to_tsshp_fkey, 0x00, sizeof X_to_tsshp_fkey);

  /*
   * Standard keys
   */
  X_to_tsshp_fkey [XK_BackSpace    & 0xFF] = KM_BACKSPACE;
  X_to_tsshp_fkey [XK_Tab          & 0xFF] = KM_TAB;
  X_to_tsshp_fkey [XK_Linefeed     & 0xFF] = KM_RETURN;
  X_to_tsshp_fkey [XK_Return       & 0xFF] = KM_RETURN;
  X_to_tsshp_fkey [XK_Pause        & 0xFF] = KM_BREAK;
  X_to_tsshp_fkey [XK_Scroll_Lock  & 0xFF] = KM_SCROLLLOCK;
  X_to_tsshp_fkey [XK_Sys_Req      & 0xFF] = KM_PRINTSCRN;
  X_to_tsshp_fkey [XK_Escape       & 0xFF] = KM_ESCAPE;
  X_to_tsshp_fkey [XK_Delete       & 0xFF] = KM_DEL;

  /*
   * Keypad numerals
   */
  for (i = 0; i < 10; i++)
  {
    X_to_tsshp_fkey [(XK_KP_0 & 0xFF) + i] = KM_KP_0 + i;
  }

  /*
   * Other keypad functions
   */
  X_to_tsshp_fkey [XK_KP_Space     & 0xFF] = ' ';
  X_to_tsshp_fkey [XK_KP_Tab       & 0xFF] = KM_TAB;
  X_to_tsshp_fkey [XK_KP_Enter     & 0xFF] = KM_KP_ENTER;
  X_to_tsshp_fkey [XK_KP_Home      & 0xFF] = KM_HOME;
  X_to_tsshp_fkey [XK_KP_Left      & 0xFF] = KM_LEFT;
  X_to_tsshp_fkey [XK_KP_Up        & 0xFF] = KM_UP;
  X_to_tsshp_fkey [XK_KP_Right     & 0xFF] = KM_RIGHT;
  X_to_tsshp_fkey [XK_KP_Down      & 0xFF] = KM_DOWN;
  X_to_tsshp_fkey [XK_KP_Page_Up   & 0xFF] = KM_PGUP;
  X_to_tsshp_fkey [XK_KP_Page_Down & 0xFF] = KM_PGDN;
  X_to_tsshp_fkey [XK_KP_End       & 0xFF] = KM_END;
  X_to_tsshp_fkey [XK_KP_Begin     & 0xFF] = KM_HOME;
  X_to_tsshp_fkey [XK_KP_Insert    & 0xFF] = KM_INSERT;
  X_to_tsshp_fkey [XK_KP_Delete	   & 0xFF] = KM_KP_PERIOD;
  X_to_tsshp_fkey [XK_KP_Equal     & 0xFF] = '=';
  X_to_tsshp_fkey [XK_KP_Multiply  & 0xFF] = KM_KP_MUL;
  X_to_tsshp_fkey [XK_KP_Add       & 0xFF] = KM_KP_PLUS;
  X_to_tsshp_fkey [XK_KP_Separator & 0xFF] = KM_KP_PERIOD;
  X_to_tsshp_fkey [XK_KP_Subtract  & 0xFF] = KM_KP_MINUS;
  X_to_tsshp_fkey [XK_KP_Decimal   & 0xFF] = KM_KP_PERIOD;
  X_to_tsshp_fkey [XK_KP_Divide	   & 0xFF] = KM_KP_DIV;

  /*
   * Function keys (only 12 defined for TSSHP at the moment)
   */
  for (i = 0; i < 12; i++)
  {
    X_to_tsshp_fkey [(XK_F1 & 0xFF) + i] = KM_F01 + i;
  }

  /*
   * Other functions
   */
  X_to_tsshp_fkey [XK_Home         & 0xFF] = KM_HOME;
  X_to_tsshp_fkey [XK_Left         & 0xFF] = KM_LEFT;
  X_to_tsshp_fkey [XK_Up           & 0xFF] = KM_UP;
  X_to_tsshp_fkey [XK_Right        & 0xFF] = KM_RIGHT;
  X_to_tsshp_fkey [XK_Down         & 0xFF] = KM_DOWN;
  X_to_tsshp_fkey [XK_Page_Up      & 0xFF] = KM_PGUP;
  X_to_tsshp_fkey [XK_Page_Down    & 0xFF] = KM_PGDN;
  X_to_tsshp_fkey [XK_End          & 0xFF] = KM_END;
  X_to_tsshp_fkey [XK_Begin        & 0xFF] = KM_HOME;
  X_to_tsshp_fkey [XK_Select       & 0xFF] = KM_NONE;
  X_to_tsshp_fkey [XK_Print        & 0xFF] = KM_PRINTSCRN;
  X_to_tsshp_fkey [XK_Execute      & 0xFF] = KM_EXECUTE;
  X_to_tsshp_fkey [XK_Insert       & 0xFF] = KM_INSERT;
  X_to_tsshp_fkey [XK_Undo         & 0xFF] = KM_NONE;
  X_to_tsshp_fkey [XK_Redo         & 0xFF] = KM_NONE;
  X_to_tsshp_fkey [XK_Menu         & 0xFF] = KM_MENU;
  X_to_tsshp_fkey [XK_Find         & 0xFF] = KM_NONE;
  X_to_tsshp_fkey [XK_Cancel       & 0xFF] = KM_NONE;
  X_to_tsshp_fkey [XK_Help         & 0xFF] = KM_HELP;
  X_to_tsshp_fkey [XK_Break        & 0xFF] = KM_BREAK;
  X_to_tsshp_fkey [XK_Mode_switch  & 0xFF] = KM_NONE;
  X_to_tsshp_fkey [XK_Num_Lock     & 0xFF] = KM_NUMLOCK;

  /*
   * Modifier keys
   */
  X_to_tsshp_fkey [XK_Shift_L      & 0xFF] = KM_LSHIFT;
  X_to_tsshp_fkey [XK_Shift_R      & 0xFF] = KM_RSHIFT;
  X_to_tsshp_fkey [XK_Control_L    & 0xFF] = KM_LCTRL;
  X_to_tsshp_fkey [XK_Control_R    & 0xFF] = KM_RCTRL;
  X_to_tsshp_fkey [XK_Caps_Lock    & 0xFF] = KM_CAPSLOCK;
  X_to_tsshp_fkey [XK_Shift_Lock   & 0xFF] = KM_CAPSLOCK;
  X_to_tsshp_fkey [XK_Meta_L       & 0xFF] = KM_LALT;
  X_to_tsshp_fkey [XK_Meta_R       & 0xFF] = KM_RALT;
  X_to_tsshp_fkey [XK_Alt_L        & 0xFF] = KM_LALT;
  X_to_tsshp_fkey [XK_Alt_R        & 0xFF] = KM_RALT;

}





void GUI_init_screen (void)
{

  XKeyboardState keyboard_state;
# if (BITS_PER_PIXEL < 24)
  Colormap colmap;
  int i;
  XColor xcolor;
# endif

  Screen_widget = XtVaCreateManagedWidget ("box", coreWidgetClass,
					   Top_level_widget,
					   XtNwidth,  Screen_width,
					   XtNheight, Screen_height,
					   XtNbackground, "black",
					   XtNforeground, "white",
					   NULL);

  /*
   * Handle redraw requests
   */
  XtAddEventHandler (Screen_widget, ExposureMask, False, screen_func, NULL);

  /*
   * Handle mouse events
   */
# if MOUSE
  XtAddEventHandler (Screen_widget,
		     ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		     False, mouse_func, NULL);
# endif
  XtAddEventHandler (Screen_widget, KeyPressMask |
		                    KeyReleaseMask, False, key_func, NULL);

  /*
   * Focus change handler, so we can turn off auto-repeat when we get the
   *  focus
   */
  Screen_display = XtDisplay (Screen_widget);
  XGetKeyboardControl (Screen_display, &keyboard_state);
  printf ("Auto-repeat is %d\n", keyboard_state.global_auto_repeat);
  Orig_auto_repeat = keyboard_state.global_auto_repeat;
  XtAddEventHandler (Screen_widget, FocusChangeMask, False, focus_func, NULL);

  XtRealizeWidget (Top_level_widget);

  /*
   * Allocate some space for the main framebuffer
   */
  Screen_bits = malloc (Screen_width * Screen_height * sizeof (pixel));
  if (Screen_bits == NULL)
  {
    Cant_happen ("Failed to allocate space for the screen");
  }
# if MEM_INIT_MESSAGES
  printf ("GUI_init(): %d bytes allocated for framebuffer\n",
	  Screen_width * Screen_height * sizeof (pixel));
# endif

  Screen_image = XCreateImage (Screen_display,
			       DefaultVisual (Screen_display,
					      DefaultScreen (Screen_display)),
			       DefaultDepth (Screen_display,
					      DefaultScreen (Screen_display)),
			       ZPixmap, 0, (char *) Screen_bits,
			       Screen_width, Screen_height, BITS_PER_PIXEL, 0);

  printf ("%d\n", DefaultDepth (Screen_display, DefaultScreen (Screen_display)));

# if(BITS_PER_PIXEL < 24)
  colmap = XCreateColormap (Screen_display, XtWindow (Screen_widget),
			    DefaultVisual (Screen_display, DefaultScreen (Screen_display)),
			    AllocAll);

  for (i = 0; i < 256; ++i)
  {
    xcolor.red   = 257 * Palette [i] [0];
    xcolor.green = 257 * Palette [i] [1];
    xcolor.blue  = 257 * Palette [i] [2];
    xcolor.pixel = i;
    xcolor.flags = DoRed | DoGreen | DoBlue;
    XStoreColor (Screen_display, colmap, &xcolor);
    Coltrans [i] = i;
  }
  XSetWindowColormap (Screen_display, XtWindow (Top_level_widget), colmap);
# endif

	/* ff1: added start time */
	gettimeofday(&tvStart, NULL);
}



pixel *GUI_get_framebuffer (void)
{
  return Screen_bits;
}



void GUI_loop (void)
{
	HUD_set_cursor_shade(0xA0);
  refresh_timer_func (NULL, NULL);

  /*
   * Pass control over to Intrinsics
   */
  XtAppMainLoop (App_context);

  /*
   * XtAppMainLoop() has returned, so presumably something happened to cause
   *  an XtAppSetExitFlag()
   * Intrinsics doesn't expect to return, though, so we get a segfault if we
   *  drop back to main()
   */
  exit (0);

}



ss_bitmap_header *Bitmap = NULL;



static void quit_func (Widget w, XtPointer client_data, XtPointer call_data)
{

  XKeyboardState xkbs;

  UNUSED (w); UNUSED (client_data); UNUSED (call_data);
  printf ("quit_func(): turning auto-repeat back on\n");

  /* ff1: i still don't know if everything is handled here... */
  Game_deinit();
  Res_free_all ();
  Timer_shutdown();
  Evt_deinit();

  /* Got to free the framebuffer before shutting down the memory subsystem! */
  free (Screen_bits);

  Mem_deinit();

  /*
   * In case auto-repeat was left off
   */
  XAutoRepeatOn (Screen_display);
  XGetKeyboardControl (Screen_display, &xkbs);
  if (!xkbs.global_auto_repeat) printf ("Blerh!\n");

  XtAppSetExitFlag (App_context);
}


static void screen_func (Widget w, XtPointer client_data, XEvent *event,
			 Boolean *cont)
{

  XExposeEvent *expose = (XExposeEvent *) event;
  UNUSED (w); UNUSED (client_data); UNUSED (cont);

  if (expose->count == 0)
  {
/* ff1: redraw request was here - any use for that function now?*/
    /* jim - probably not actually. I'll expunge it some time */
  }

}


#if MOUSE

/*  int Ship_button_down; */
/*  int Light_button_down; */
int Last_mouse_x;
int Last_mouse_y;
Time Last_drag_time;

int Mouse_grabbed;

static void mouse_func (Widget w, XtPointer client_data, XEvent *event,
			Boolean *cont)
{

  XButtonEvent *button_event;
  XMotionEvent *motion_event;

  Matrix rot;
  Matrix tm;

	struct timeval curTime;
	point ptMouse;
	uint32 curTick;

  double ct;
  double st;
  int dx;
  int dy;

  UNUSED (w); UNUSED (client_data); UNUSED (cont);

	/* ff1: i love those milliseconds... */
	gettimeofday(&curTime, NULL);
  curTick = (curTime.tv_sec - tvStart.tv_sec) * 1000 + 
						curTime.tv_usec / 1000;
  switch (event->type)
  {

  case ButtonPress:
		/* ff1: uhh... before i do something wrong i leave it... */
    button_event = (XButtonEvent *) event;
    if (button_event->button == Button1)
    {
#     if DEBUG_EVENTS
      printf ("Button 1 (ship) down\n");
#     endif

      /*
       * We have been clicked on. Grab or ungrab the mouse
       */
      if (!Mouse_grabbed)
      {

	XGrabPointer (XtDisplay (Screen_widget),
		      XtWindow (Screen_widget), True,
		      ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		      GrabModeAsync, GrabModeAsync, XtWindow (Screen_widget),
		      None, CurrentTime);

	Last_mouse_x = button_event->x;
	Last_mouse_y = button_event->y;
	Mouse_grabbed = 1;

      }
      else
      {
	XUngrabPointer (XtDisplay (Screen_widget), CurrentTime);
	Mouse_grabbed = 0;
      }

    }
    break;

  case MotionNotify:

    /*
     * The mouse has moved. See if we need to rotate anything
     */
    if (!Mouse_grabbed)
    {
      break;
    }

    motion_event = (XMotionEvent *) event;

    if (motion_event->time - Last_drag_time < 10)
    {
      break;
    }
    Last_drag_time = motion_event->time;

		/* ff1: i hope i'm doing it right... */
		ptMouse.x = motion_event->x;
		ptMouse.y = motion_event->y;
#if 0 /* ff1: bryann reported major slowdown... */
		Con_move(MOVE_MOUSEX, motion_event->x, curTick);
		Con_move(MOVE_MOUSEY, motion_event->y, curTick);
#endif

#   if DEBUG_EVENTS
    printf ("Mouse drag (%d, %d)\n", motion_event->x, motion_event->y);
#   endif

    /*
     * Calculate x-based rotation
     */
    if (motion_event->x != Last_mouse_x)
    {
      dx = motion_event->x - Last_mouse_x;
      ct = cos (dx * M_PI / (Screen_width));
      st = sin (dx * M_PI / (Screen_width));
      rot.x.x = ct;
      rot.x.y = -st;
      rot.x.z = 0;
      rot.y.x = st;
      rot.y.y = ct;
      rot.y.z = 0;
      rot.z.x = 0;
      rot.z.y = 0;
      rot.z.z = 1;
      Mat_multiply (tm, Camera->orientation, rot);
      Camera->orientation = tm;
      Last_mouse_x = motion_event->x;
    }

    /*
     * Calculate y-based rotation
     */
    if (motion_event->y != Last_mouse_y)
    {
      dy = motion_event->y - Last_mouse_y;
      ct = cos (dy * M_PI / (Screen_height));
      st = sin (dy * M_PI / (Screen_height));
      rot.x.x = 1;
      rot.x.y = 0;
      rot.x.z = 0;
      rot.y.x = 0;
      rot.y.y = ct;
      rot.y.z = st;
      rot.z.x = 0;
      rot.z.y = -st;
      rot.z.z = ct;
      Mat_multiply (tm, rot, Camera->orientation);
      Camera->orientation = tm;
      Last_mouse_y = motion_event->y;
    }

    break;

  }

}

#endif



static uint16 keysym_to_km (KeySym ks)
{
  uint16 low  = ks & 0xff;
  uint16 high = ks >> 8;
  if (high < 0x10) return low; /* ordinary keysym */
  else if (high == 0xFF)
  {
    /*
     * Function key ... look it up in the table
     */
    return X_to_tsshp_fkey [low];
  }
  else return KM_NONE;

}


static void key_func (Widget w, XtPointer client_data, XEvent *event,
		      Boolean *cont)
{

  struct timeval now;
  XKeyEvent *key = (XKeyEvent *) event;
	ss_controlset ssControl;
	uint32 curtick;
  KeySym ks = XLookupKeysym (key, 0);

  UNUSED (w); UNUSED (client_data); UNUSED (cont);

	/* get millisecond tick... */
	gettimeofday(&now, NULL);
  curtick = (now.tv_sec - tvStart.tv_sec) * 1000 + now.tv_usec / 1000;
	/*
		fModifier = XYetAnotherXWindowFunctionToGetThoseModifiers... ;)
	*/
  switch (key->type)
  {

  case KeyPress:
		if (IsModifierKey(ks))
			Con_key_modifier(gui_get_modifier(ks), true);
		else
		{
		  Con_key_activate (keysym_to_km (ks), curtick);
		}
		Con_snapshot(&ssControl, false);
		if (ssControl.quit)
		{	/* quit signal... */
      XAutoRepeatOn (XtDisplay (Screen_widget));
/* ff1: d'oh! memory leaks ahead... call Game_deinit() and the like...*/
/*        Res_free_all (); */
/*        exit (0); */
      quit_func (w, client_data, NULL);
		}

#if 0	/* ff1: fingers crossed... */
    switch (ks)
    {
														/* ff1: ;) */
    case 'w': Con_key_activate('s', KM_MOD_LSHIFT, curtick);
      break;

    case 's':	Con_key_activate('s', fModifier, curtick);	break;
    case XK_Up: Con_key_activate(KM_UP, fModifier, curtick);	break;
    case XK_KP_8: Con_key_activate(KM_KP_8, fModifier, curtick);	break;

    case 'x': Con_key_activate('x', fModifier, curtick);	break;
    case XK_Down: Con_key_activate(KM_DOWN, fModifier, curtick);	break;
    case XK_KP_2: Con_key_activate(KM_KP_2, fModifier, curtick);	break;

    case 'd': Con_key_activate('d', fModifier, curtick);	break;
    case XK_Right: Con_key_activate(KM_RIGHT, fModifier, curtick);	break;
    case XK_KP_6: Con_key_activate(KM_KP_6, fModifier, curtick);	break;

    case 'a': Con_key_activate('a', fModifier, curtick);	break;
    case XK_Left: Con_key_activate(KM_LEFT, fModifier, curtick);	break;
    case XK_KP_4: Con_key_activate(KM_KP_4, fModifier, curtick);	break;

    case XK_Escape: 
    case 'q':	Con_key_activate(KM_ESCAPE, fModifier, curtick);
      XAutoRepeatOn (XtDisplay (Screen_widget));
/* ff1: d'oh! memory leaks ahead... call Game_deinit() and the like...*/
      Res_free_all ();
      exit (0);
      break;

    case '#':
      file = fopen ("screenshot.dat", "wb");
      fwrite (Screen_bits, sizeof (pixel), Screen_width * Screen_height, file);
      fclose (file);
      break;

    case '-':
      printf ("%04X %04X\n", (int) Camera->position.x, (int) Camera->position.y);
      break;

		case '3': Con_key_activate('3', fModifier, curtick);	break;

		case XK_Tab: Con_key_activate(KM_TAB, fModifier, curtick);	break;
    }
#endif

    break;

  case KeyRelease:
    if (IsModifierKey(ks))
      Con_key_modifier(gui_get_modifier(ks), false);
    else
    {
      Con_key_deactivate (keysym_to_km (ks), curtick);
    }

  }

}

uint16	gui_get_modifier(KeySym nKeySym)
{
	uint16 fNewModifier = 0x0000;
	uint32 fOldModifier = 0x00000000;

	fOldModifier = (uint32)nKeySym;	/* ff1: trick this gcc... */
	
	switch (fOldModifier)
	{
		case XK_Shift_L:			fNewModifier = KM_MOD_LSHIFT;	break;
		case XK_Shift_R:			fNewModifier = KM_MOD_RSHIFT;	break;
		case XK_Control_L:		fNewModifier = KM_MOD_LCTRL;	break;
		case XK_Control_R:		fNewModifier = KM_MOD_RCTRL;	break;
		case XK_Caps_Lock:		
		case XK_Shift_Lock:		fNewModifier = KM_MOD_CAPS;		break;
		case XK_Alt_L:				fNewModifier = KM_MOD_LALT;		break;
		case XK_Alt_R:				fNewModifier = KM_MOD_RALT;		break;
		case XK_Scroll_Lock:	fNewModifier = KM_MOD_SCROLL;	break;
		case XK_Num_Lock:			fNewModifier = KM_MOD_NUM;		break;

		default:																break;
	}

	return fNewModifier;
}

static void focus_func (Widget w, XtPointer client_data, XEvent *event,
			Boolean *cont)
{

  XFocusChangeEvent *fce = (XFocusChangeEvent *) event;

  UNUSED (client_data); UNUSED (cont);

  switch (fce->type)
  {

  case FocusIn:
/*      printf ("Focus handler got event FocusIn\n"); */
    XAutoRepeatOff (XtDisplay (w));
    break;

  case FocusOut:
/*      printf ("Focus handler got event FocusOut\n"); */
/*      if (Orig_auto_repeat) */
/*      { */
    XAutoRepeatOn (XtDisplay (w));
/*      } */
    break;

  }

}


static void refresh_timer_func (XtPointer client_data, XtIntervalId *id)
{
  size screensize;
  struct timeval then = now;
	uint32 curtick;
  double elapsed;

  UNUSED (client_data); UNUSED (id);

  screensize.cx = Screen_width;
  screensize.cy = Screen_height;
  /*
   * Set up for the next one
   */
  Refresh_timer = XtAppAddTimeOut (App_context, Refresh_delay,
				   refresh_timer_func, NULL);

  gettimeofday (&now, NULL);
  /*
   * jim - frankly, I hadn't considered wrapping effects. A struct timeval
   *  only contains fields for seconds and microseconds. Under Linux
   *  gettimeofday() returns in tv.sec the number of seconds since Jan 1, 1970
   *  (the usual Unix measure of time) which will wrap round in 2038. That's
   *  nearer the time when System Shock is set than it is to today, so I
   *  really shouldn't worry ...
   */

  /* ff1: again changed. searched the internet on this struct */
  /* and now i'm wiser ;) -- should work now...		      */
  curtick = (now.tv_sec - tvStart.tv_sec) * 1000 + now.tv_usec / 1000;

	Timer_work(curtick);
	Evt_work();
  elapsed = now.tv_sec - then.tv_sec +
    ((double) (now.tv_usec - then.tv_usec)) / 1.e6;


  Phy_frame (elapsed);

  if (Option_frame_times)
  {
    printf ("%f\n", 1. / elapsed);
  }

  /*
   * Not really necessary but it helps to see where bounding box calculations
   *  are wrong, for example
   */
  memset (Screen_bits, 0, Screen_width * Screen_height * sizeof (pixel));
  Ren_frame (elapsed, Screen_bits);
		
  Screen_gc = XCreateGC (XtDisplay (Screen_widget),
			 XtWindow (Screen_widget), 0, NULL);
  
  XPutImage (XtDisplay (Screen_widget), XtWindow (Screen_widget),
	     Screen_gc, Screen_image, 0, 0, 0, 0,
	     Screen_width, Screen_height);

  XFreeGC (XtDisplay (Screen_widget), Screen_gc);

}
