/*
 *  sound_driver_fmod.c - FMOD Sound Routines
 *  Sound routines
 */

/*
 * File:    sound/sound_driver_fmod.c
 * Purpose: Audio handling
 * Prefix:  Sound_
 * Notes:	SDL Audio Handling!
 * Music:   Olive, "Outlaw"
 *
 * This file by Bryan Nolen 2002 bryan@r3v3ng.net
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */

#include <stdio.h>
#include <stdlib.h>
#include "gamemem.h"
#include <math.h>
#include <string.h>										/* strncpy() - msgs */

#include "defines.h"
#include "types.h"
#include "option.h"
#include "event.h"

#include "sound.h"
#include "sound_core.h"
					
#include "main.h"											/* Can't_happen */
#include "resource.h"

#ifdef WIN32
  #include <conio.h>
  #include <windows.h>
#else
  #include "wincompat.h"
#endif

#include "fmod.h"
#include "fmod_errors.h"


/*
 * GLOBAL
 */

/*
 *  This holds the state of initalization of the sound card
 *  if it is 0 then the card has not been initalized and so do
 *  not play any sounds.
 *  This is to clear a known bug with SDL ..... that I missed :)
 */
int sound_init_state=0;


int sound_driver_init(void)
{
	char			key;
	int				driver, i = 0;

	if (FSOUND_GetVersion() < FMOD_VERSION)
	{
		printf("Error : You are using the wrong DLL version!  You should be using FMOD %.02f\n", FMOD_VERSION);
		return 1;
	}

	// ==========================================================================================
	// SELECT OUTPUT METHOD
	// ==========================================================================================
	
	printf("---------------------------------------------------------\n");	
	printf("Output Type\n");	
	printf("---------------------------------------------------------\n");	
#ifdef WIN32
	printf("1 - Direct Sound\n");
	printf("2 - Windows Multimedia Waveout\n");
	printf("3 - A3D\n");
#else
    printf("1 - Open Sound System (OSS) (Linux, Solaris, freebsd)\n");
    printf("2 - Elightment Sound Daemon (ESD, Linux, more ...)\n");
    printf("3 - Alsa Sound System (Linux)\n");   
#endif
	printf("4 - NoSound\n");
	printf("---------------------------------------------------------\n");	// print driver names
	printf("Press a corresponding number or ESC to for no sound\n");

	do
	{
		key = getch();
	} while (key != 27 && key < '1' && key > '4');
	
	switch (key)
	{

#ifdef WIN32
		case '1' :	FSOUND_SetOutput(FSOUND_OUTPUT_DSOUND);
					break;
		case '2' :	FSOUND_SetOutput(FSOUND_OUTPUT_WINMM);
					break;
		case '3' :	FSOUND_SetOutput(FSOUND_OUTPUT_A3D);
					break;
#else      
  	    case '1' :  FSOUND_SetOutput(FSOUND_OUTPUT_OSS);
	                       break;
	    case '2' :  FSOUND_SetOutput(FSOUND_OUTPUT_ESD);
	                       break;
	    case '3' :  FSOUND_SetOutput(FSOUND_OUTPUT_ALSA);
	                       break;
#endif      
		 
		case '4' :	FSOUND_SetOutput(FSOUND_OUTPUT_NOSOUND);
					break;
		default :	return 1; 
	}
	

	// ==========================================================================================
	// SELECT DRIVER
	// ==========================================================================================

	// The following list are the drivers for the output method selected above.
	printf("---------------------------------------------------------\n");	
	switch (FSOUND_GetOutput())
	{
		case FSOUND_OUTPUT_NOSOUND:	printf("NoSound"); break;
#ifdef WIN32
		case FSOUND_OUTPUT_WINMM:	printf("Windows Multimedia Waveout"); break;
		case FSOUND_OUTPUT_DSOUND:	printf("Direct Sound"); break;
		case FSOUND_OUTPUT_A3D:		printf("A3D"); break;
#else
   	    case FSOUND_OUTPUT_OSS:     printf("Open Sound System"); break;
	    case FSOUND_OUTPUT_ESD:     printf("Enlightment Sound Daemon"); break;
	    case FSOUND_OUTPUT_ALSA:    printf("Alsa"); break;
#endif
		 
	};
	printf(" Driver list\n");	
	printf("---------------------------------------------------------\n");	

	for (i=0; i < FSOUND_GetNumDrivers(); i++) 
	{
		printf("%d - %s\n", i+1, FSOUND_GetDriverName(i));	// print driver names
		{
			unsigned int caps = 0;

			FSOUND_GetDriverCaps(i, &caps);
			
			if (caps & FSOUND_CAPS_HARDWARE)
				printf("  * Driver supports hardware 3D sound!\n");
		}
	}
	printf("---------------------------------------------------------\n");	// print driver names
	printf("Press a corresponding number or ESC for no sound\n");

	do
	{
		key = getch();
		if (key == 27) 
			return 0;
		driver = key - '1';
	} while (driver < 0 || driver >= FSOUND_GetNumDrivers());
	FSOUND_SetDriver(driver);					// Select sound card (0 = default)


	// ==========================================================================================
	// INITIALIZE
	// ==========================================================================================
	if (!FSOUND_Init(44100, 32, 0))
	{
		printf("Init: %s\n", FMOD_ErrorString(FSOUND_GetError()));
		return 1;
	}
	else 
	{
		sound_init_state = 1;
		printf("Sound: sound_driver_init() Done\n");
		sound_driver_playmusic(1);
	}

	return 0;
}

int sound_driver_deinit(void)
{
	FSOUND_Close();
    sound_init_state = 0;
	return 0;
}


int sound_driver_play( size_t size, unsigned char * buf, int samrate )
{

int Sample_rate = samrate;
unsigned char * buffer = buf;
int sound_len = size;
FSOUND_SAMPLE	*logaudio = NULL;

/*
 * Only proceed if we have been initalized...
 */
    if( sound_init_state==0 )
    {
        printf("Sound: System initalise failed - no play possible\n");
        return(1);
    }

/*
 * ok, now we need to pass to fmod the buffer(contains the data), Sample_rate and size
 * This will make a sample that has middle volume, medium priority, mono, 2d, 8bits, middle pan, non looping
 * in the first free sample slot.
 */
 
logaudio = FSOUND_Sample_Alloc(FSOUND_FREE, sound_len, 
								(FSOUND_LOOP_OFF | FSOUND_8BITS | FSOUND_MONO | FSOUND_LOADRAW | FSOUND_2D | FSOUND_UNSIGNED),
								Sample_rate, 140, FSOUND_STEREOPAN,	140);

if (!FSOUND_Sample_Upload(logaudio, buffer, (FSOUND_8BITS | FSOUND_MONO | FSOUND_LOADRAW | FSOUND_UNSIGNED)))
{
	printf("Sound_driver_fmod: %s\n", FMOD_ErrorString(FSOUND_GetError()));
	return 1;
}

FSOUND_PlaySound(FSOUND_FREE, logaudio);

/* **
do { Sleep(INTERFACE_UPDATETIME-1); } while (FSOUND_IsPlaying(channelplay));
FSOUND_Sample_Free(logaudio);
** */
return 0;
}


int sound_driver_playmusic(int level)
{
#ifdef WIN32
	FMUSIC_MODULE *levelTune;

	if (level == 0)
	{
		levelTune = FMUSIC_LoadSong("/temp/midi/intro.mid");
	}
	else
	{
		levelTune = FMUSIC_LoadSong("/temp/midi/sshock03.mid");
	}


	if (levelTune != NULL) 
	{
		FMUSIC_PlaySong(levelTune);
		return(0);
	}
	else
	{
		return(1);
	}
#else
	UNUSED(level);
	return(1);
#endif

}