/*
 * anim.c -
 *  Animations
 */

/*
 * File:    src/anim.c
 * Purpose: Generic animation timing
 * Prefix:  Anim_
 * Note:
 * Music:   Van Halen, "Fair Warning"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "gamemem.h"
#include "anim.h"
#include "main.h"     /* Cant_happen() */
#include "map.h"
#include "object.h"
#include "option.h"
#include "resource.h"


#define MAX_ANIM 64

#define ANIM_CONTROL_ARCHIVE_SIZE 15


/*
 * Global variables
 */
anim_control *Anim_list;
int Num_anims;

#define UW_NUM_ANIMOBJS 12
uw_animobj *UW_anim_info;




void Anim_init (void)
{

  Anim_list = Alloc_init (MAX_ANIM * sizeof *Anim_list, "Anim_init",
			  "animations list");

  if (Option_underworld)
  {
    /*
     * Load up the Underworld animation info from objects.dat
     * This is a bit nasty ... we'll sort out something proper when we are
     *  using a few more properties
     */
    int i;
    FILE *propfile = fopen (Res_make_path (UW_DATA_PATH, "objects.dat"), "rb");
    if (propfile == NULL)
    {
      Cant_happen ("Anim_init(): couldn't open object properties file");
    }

    UW_anim_info = Alloc_init (UW_NUM_ANIMOBJS * sizeof *UW_anim_info,
			       "Anim_init", "Underworld animation properties");
    fseek (propfile, 0x0da2, SEEK_SET); /* skip to anim properties */
    for (i = 0; i < UW_NUM_ANIMOBJS; ++i)
    {
      READ_LE_SHORT (propfile, UW_anim_info [i].stuff1);
      UW_anim_info [i].base_frame = fgetc (propfile);
      UW_anim_info [i].num_frames = fgetc (propfile);
      printf ("Anim %d (%d), base %d, %d frames\n", i, UW_anim_info [i].stuff1,
	      UW_anim_info [i].base_frame, UW_anim_info [i].num_frames);
    }
    fclose (propfile);
  }
}

void Anim_deinit(void)
{
	free(Anim_list);
	Anim_list = NULL;
}


void Anim_init_level (int level)
{

# if 0
  size_t size;
  block_handle bh;
  byte *ap;
  int map_anims; /* how many animations defined in the map */
  int i;
# endif

  Num_anims = 0;

# if 0
  if (Option_underworld) return; /* no anim chunk for Underworld */

  bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_ANIM_CONTROL, &size);
  if ((bh == NULL) || (*bh == NULL))
    Cant_happen ("Anim_init_level(): couldn't load anim control chunk");
  if (size % ANIM_CONTROL_ARCHIVE_SIZE != 0)
    Cant_happen ("Anim_init_level(): not a whole number of anims");

  ap = *bh;
  map_anims = size / ANIM_CONTROL_ARCHIVE_SIZE;

  for (i = 0; i < map_anims; ++i)
  {
    int objno;
    int flags;

    objno = *ap + (ap [1] << 8); ap += 2;
    flags = *ap;
    ap += ANIM_CONTROL_ARCHIVE_SIZE - 2;

    if (flags & AF_ACTIVE)
    {
      Anim_add (0.25, objno, flags);
    }
  }
# endif

}




/*
 * Add an animation to the active animations list
 */
anim_control *Anim_add (double delay, map_object *obj, int nframes, int flags)
{
 
  anim_control *a;

  if (Num_anims >= MAX_ANIM)
  {
    Cant_happen ("Anim_add(): too many animations, increase MAX_ANIM");
  }

  a = &Anim_list [Num_anims++];
  a->delay        = delay;
  a->elapsed      = 0.;
  a->num_frames   = nframes;
  a->flags        = flags;
  a->obj          = obj;
  a->advance_hook = NULL;

/*    Obj_init_anim (a); */

  if (a->flags & AF_BIDIRECTIONAL)
  {
    if (obj->state == a->num_frames - 1)
    {
      a->flags |= AF_BACKWARDS;
    }
  }

  return a;
}



/*
 * A frame has happened, advance all active animations
 */
void Anim_frame (double elapsed)
{

  int i;
  for (i = 0; i < Num_anims; ++i)
  {
    Anim_list [i].elapsed += elapsed;
    while (Anim_list [i].elapsed > Anim_list [i].delay)
    {
      Anim_list [i].elapsed -= Anim_list [i].delay;
      if (!(Anim_list [i].flags & AF_PAUSED))
      {
	Anim_advance (&Anim_list [i]);
      }
      /*
       * Call the advance hook, if it has one, for special processing (even
       *  when it's paused as it might want to unpause itself)
       */
      if (Anim_list [i].advance_hook != NULL)
      {
	Anim_list [i].advance_hook (&Anim_list [i]);
      }

    }
  }

}



/*
 * Advance an animation by a single frame
 */
void Anim_advance (anim_control *a)
{

  map_object *mo = a->obj;

  if (a->flags & AF_BACKWARDS)
  {
    if (mo->state == 0)
    {
      a->flags &= ~AF_BACKWARDS;
    }
    else mo->state--;
    return;
  }

  mo->state++;

  if (a->flags & AF_BIDIRECTIONAL)
  {
    if (mo->state == a->num_frames)
    {
      a->flags |= AF_BACKWARDS;
      mo->state--;
    }
  }
  else if (mo->state >= a->num_frames)
  {
    mo->state = 0;
  }

}


uw_animobj *Anim_get_uw_info (int objid)
{
  unsigned int animno = objid - UW_OBJID_ANIM_BASE;
  assert (Option_underworld);
  assert (animno < UW_NUM_ANIMOBJS);
  return &(UW_anim_info [animno]);
}
