/*
 * bitmap.c -
 *  Bitmap loading and caching
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#if NEED_MEMORY_H
#include <memory.h>		/*memset*/
#endif

#include "types.h"
#include "gamemem.h"
#include "bitmap.h"
#include "critter.h"
#include "main.h"
#include "option.h"
#include "resource.h"
#include "texture.h"


/*
 * Debug
 */
#define BITMAP_MESSAGES 0


/*
 * Underworld bitmap (.gr) files, in an arbitrary order
 */
static const struct
{
  int chunk;
  const char *fname;
}
UW_bitmap_file [] =
{
  {    0, "3dwin.gr"    },      /*  0 */
  {    1, "animo.gr"    },      /*  1 */
  {    2, "armor_f.gr"  },      /*  2 */
  {    3, "armor_m.gr"  },      /*  3 */
  {    4, "bodies.gr"   },      /*  4 */
  {    5, "buttons.gr"  },      /*  5 */
  {    6, "chains.gr"   },      /*  6 */
  {    7, "charhead.gr" },      /*  7 */
  {    8, "chrbtns.gr"  },      /*  8 */
  {    9, "compass.gr"  },      /*  9 */
  {   10, "converse.gr" },      /* 10 */
  {   11, "cursors.gr"  },      /* 11 */
  {   12, "doors.gr"    },      /* 12 */
  {   13, "dragons.gr"  },      /* 13 */
  {   14, "eyes.gr"     },      /* 14 */
  {   15, "flasks.gr"   },      /* 15 */
  {   16, "genhead.gr"  },      /* 16 */
  {   17, "heads.gr"    },      /* 17 */
  {   18, "inv.gr"      },      /* 18 */
  {   19, "lfti.gr"     },      /* 19 */
  {   20, "objects.gr"  },      /* 20 */
  {   21, "opbtn.gr"    },      /* 21 */
  {   22, "optb.gr"     },      /* 22 */
  {   23, "optbtns.gr"  },      /* 23 */
  {   24, "panels.gr"   },      /* 24 */   /* fixed size, no headers */
  {   25, "power.gr"    },      /* 25 */
  {   26, "question.gr" },      /* 26 */
  {   27, "scrledge.gr" },      /* 27 */
  {   28, "spells.gr"   },      /* 28 */
  {   29, "tmflat.gr"   },      /* 29 */
  {   30, "tmobj.gr"    },      /* 30 */
  {   31, "views.gr"    },      /* 31 */
  {   32, "weapons.gr"  },      /* 32 */
  { 1362, "cursors.gr"  },      /* alternate chunk for cursors to match SS */
};

static int Num_uw_bitmap_entries =
  sizeof UW_bitmap_file / sizeof *UW_bitmap_file;



/*
 * Local function prototypes
 */
bool load_uw_bitmap_header (ss_bitmap_header *bmp, int chunk, int subchunk);
bool load_uw_bitmap_bits (ss_bitmap_header *bmp, unsigned char *bits,
			  size_t bufsize);


static FILE *find_uw_bitmap_file (int chunk)
{
  int i;
  for (i = 0; i < Num_uw_bitmap_entries; ++i)
  {
    if (UW_bitmap_file [i].chunk == chunk)
    {
      /*
       * Found it
       */
      return fopen (Res_make_path (UW_DATA_PATH, UW_bitmap_file [i].fname),
		    "rb");
    }
  }
  return NULL;
}

/*
 * Find the number of bitmaps in a chunk
 */
int Bmp_get_num_bmps (int chunk)
{

  int num_bmps;

  if (Option_underworld)
  {
    FILE *grfile = find_uw_bitmap_file (chunk);
    if (grfile == NULL)
    {
      Cant_happen ("Bmp_get_num_bmps(): can't open UW graphics file");
    }

    /*
     * I don't know what the first byte means, it seems always to be 1
     */
    fseek (grfile, 1, SEEK_SET);
    READ_LE_SHORT (grfile, num_bmps);
    fclose (grfile);
  }
  else
  {
    size_t size;
    num_bmps = Res_get_numSubchunks_by_id (chunk, &size);
  }

  return num_bmps;

}


bool Bmp_load_bitmap_header (ss_bitmap_header *bmp, int chunk, int subchunk)
{

  unsigned char *bmp_bits;
  block_handle bh;
  size_t size;

  if (Option_underworld)
  {
    return load_uw_bitmap_header (bmp, chunk, subchunk);
  }

  bh = Res_load_subchunk (chunk, subchunk, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    return false;
  }

  bmp_bits = *bh;
  bmp->chunk    = chunk;
  bmp->subchunk = subchunk;
  bmp->type     = bmp_bits [4]  + (bmp_bits [5] << 8);
  bmp->width    = bmp_bits [8]  + (bmp_bits [9] << 8);
  bmp->height   = bmp_bits [10] + (bmp_bits [11] << 8);
  bmp->x1       = bmp_bits [16] + (bmp_bits [17] << 8);
  bmp->y1       = bmp_bits [18] + (bmp_bits [19] << 8);
  bmp->x2       = bmp_bits [20] + (bmp_bits [21] << 8);
  bmp->y2       = bmp_bits [22] + (bmp_bits [23] << 8);

# if BITMAP_MESSAGES
  printf ("Bmp_load_bitmap_header() got type %d, %dx%d\n", bmp->type,
	  bmp->width, bmp->height);
  printf ("Bmp_load_bitmap_header() %d %d %d %d", bmp->x1, bmp->y1,
	  bmp->x2, bmp->y2);
# endif

  return true;

}


bool load_uw_bitmap_header (ss_bitmap_header *bmp, int chunk, int subchunk)
{

  int num_bmps;
  int offs;
  int offs_end;
  FILE *grfile = find_uw_bitmap_file (chunk);
  if (grfile == NULL)
  {
    printf ("%d\n", chunk);
    Cant_happen ("load_uw_bitmap_header(): can't open UW graphics file");
  }

  /*
   * I don't know what the first byte means, it seems always to be 1
   */
  fseek (grfile, 1, SEEK_SET);
  READ_LE_SHORT (grfile, num_bmps);
  if (subchunk >= num_bmps)
  {
    Cant_happen ("load_uw_bitmap_header(): invalid bitmap number");
  }

  fseek (grfile, 3 + 4 * subchunk, SEEK_SET);
  READ_LE_LONG (grfile, offs);
  READ_LE_LONG (grfile, offs_end);

  /*
   * Get the bitmap header (skip the storage type)
   */
  fseek (grfile, offs + 1, SEEK_SET);

  bmp->chunk    = chunk;
  bmp->subchunk = subchunk;
  bmp->type     = 0;
  bmp->width    = fgetc (grfile);
  bmp->height   = fgetc (grfile);
  bmp->x1       = 0;
  bmp->y1       = 0;
  bmp->x2       = 0;
  bmp->y2       = 0;

  /*
   * Should probably check the bitmap length field
   */

  fclose (grfile);
  return true;

}





bool Bmp_load_bitmap_bits (ss_bitmap_header *bmp, unsigned char *bits,
			   size_t bufsize)
{

  block_handle bh;
  size_t size;
  unsigned char *chunk_bits;
  unsigned char *bits_end;
  int i;
  int xc;
# if 0
  FILE *f;
# endif

  /*
   * Paranoia check
   */
  if (bmp == NULL)
  {
    Cant_happen ("Bmp_load_bitmap_bits() passed null bitmap header");
  }
  if (bits == NULL)
  {
    Cant_happen ("Bmp_load_bitmap_bits() passed null bitmap buffer");
  }

  /*
   * Cast to clobber `comparison between signed and unsigned' warning
   */
  if (bufsize < (size_t) bmp->width * bmp->height)
  {
    Cant_happen ("Bmp_load_bitmap_bits() buffer too small");
  }

  if (Option_underworld)
  {
    return load_uw_bitmap_bits (bmp, bits, bufsize);
  }

  /*
   * Make sure we still have the chunk
   */
  bh = Res_load_subchunk (bmp->chunk, bmp->subchunk, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Bmp_load_bitmap_bits(): invalid bitmap chunk");
  }

  bmp->bits = bits;
  bits_end  = bits + bmp->width * bmp->height;
  chunk_bits = ((unsigned char *) *bh) + 0x1c;
  memset (bits, 0, bufsize);

  if (bmp->type == 4)
  {

#   if BITMAP_MESSAGES
    printf ("Bmp_load_bitmap_bits(): got compressed bitmap\n");
#   endif

    while (bits < bits_end)
    {
      xc = *chunk_bits++;
      if (xc == 0)
      {
	xc = *chunk_bits++;
	for (i = 0; i < xc && bits < bits_end; ++i)
	{
	  *bits++ = *chunk_bits;
	}
	++chunk_bits;
      }
      else if (xc < 0x81)
      {
	if (xc == 0x80)
	{
	  xc = *chunk_bits++;
	  if (xc == 0)
	  {
	    break;
	  }
	  if (*chunk_bits < 0x80)
	  {
	    bits += xc + (*chunk_bits << 8);
	    xc = 0;
	  }
	  else if (*chunk_bits > 0x80)
	  {
/*  	    bits += xc + ((*chunk_bits & 0x3f) << 8); */
/*  	    xc = 0; */
  	    xc += ((*chunk_bits & 0x3f) << 8);
	  }
/*  	  xc = *chunk_bits++; */
	  ++chunk_bits;
	}
	for (i = 0; i < xc && bits < bits_end; ++i)
	{
	  *bits++ = *chunk_bits++;
	}
      }
      else
      {
	bits += (xc & 0x7f);
      }
    }

    /*
     * Debug
     */
#   if 0
    if (bits != bits_end)
    {
      printf ("Bmp_load_bitmap_bits(): chunk size mismatch, dumping\n");
      f = fopen ("failbmp.dat", "wb");
      fwrite (*bh, 1, 1000, f);
      fclose (f);
    }
#   endif

  }
  else
  {
#   if BITMAP_MESSAGES
    printf ("Bmp_load_bitmap_bits(): got uncompressed bitmap\n");
#   endif

    memcpy (bits, chunk_bits, bmp->width * bmp->height);
    bmp->bits = bits;
  }
/*    bmp->bits = bits; */

  return true;

}



#define GET_NYBBLE(in,which) ((which)?(which=0,(*in++)&0xf):(which=1,(*in)>>4))
#define PUT_NYBBLE(n,out,which)\
  ((which)?(which=0,*(out++)|=((n)&0xf)):(which=1,(*out)=((n)&0xf)<<4))
#define MAX_UW_PACKED_BMP_SIZE 1024

bool load_uw_bitmap_bits (ss_bitmap_header *bmp, unsigned char *bits,
			  size_t bufsize)
{

  int num_bmps;
  int offs;
  int offs_end;
  unsigned char *ip, *op;
  int w1 = 0;
  int n;
  static unsigned char scratch [MAX_UW_PACKED_BMP_SIZE];
  static unsigned char spal [16];
  FILE *grfile = find_uw_bitmap_file (bmp->chunk);
  FILE *palfile;
  size_t packsize;  /* size of (packed) bitmap data in file */

  if (grfile == NULL)
  {
    Cant_happen ("load_uw_bitmap_bits(): can't open UW graphics file");
  }

  /*
   * I don't know what the first byte means, it seems always to be 1
   */
  fseek (grfile, 1, SEEK_SET);
  READ_LE_SHORT (grfile, num_bmps);
  if (bmp->subchunk >= num_bmps)
  {
    Cant_happen ("load_uw_bitmap_bits(): invalid bitmap number");
  }

  fseek (grfile, 3 + 4 * bmp->subchunk, SEEK_SET);
  READ_LE_LONG (grfile, offs);
  READ_LE_LONG (grfile, offs_end);
  if (offs == offs_end)
  {
    return false;
  }
  memset (bits, 0, bufsize);

  /*
   * Get the bitmap bits
   */
  fseek (grfile, offs, SEEK_SET);

  /*
   * Again, don't know what this byte is, it's always 4
   * It might be the offset to the bitmap data proper
   */
  switch (fgetc (grfile))
  {
  case 4:
    fseek (grfile, 4, SEEK_CUR);
    packsize = bmp->width * bmp->height;
    if (fread (bits, 1, packsize, grfile) != packsize)
    {
      Cant_happen ("load_uw_bitmap_bits(): failed to load bitmap");
    }
    break;

  case 8:
  {

    /*
     * Compressed bitmap - get the auxiliary palette
     */
    fseek (grfile, 2, SEEK_CUR);
    n = fgetc (grfile);
    fseek (grfile, 2, SEEK_CUR);
    palfile = fopen (Res_make_path (UW_DATA_PATH, "allpals.dat"), "rb");
    if (palfile == NULL)
    {
      Cant_happen ("bitmap.c:load_uw_bitmap_bits(): couldn't open UW auxiliary"
		   " palette file");
    }
    fseek (palfile, 16 * n, SEEK_SET);
    fread (spal, 1, 16, palfile);
    fclose (palfile);

    bmp->bits = bits;
    packsize = offs_end - offs - 6;
    if (fread (scratch, 1, packsize, grfile) != packsize)
    {
      Cant_happen ("load_uw_bitmap_bits(): failed to load packed bitmap");
    }

    Bmp_unpack_uw_rle (bmp, scratch, 4, spal);

    break;
  }

  case 10:  /* 4-bit uncompressed bitmap */
  {

    fseek (grfile, 2, SEEK_CUR);
    n = fgetc (grfile);
    fseek (grfile, 2, SEEK_CUR);
    palfile = fopen (Res_make_path (UW_DATA_PATH, "allpals.dat"), "rb");
    if (palfile == NULL)
    {
      Cant_happen ("bitmap.c:load_uw_bitmap_bits(): couldn't open UW auxiliary"
		   " palette file");
    }
    fseek (palfile, 16 * n, SEEK_SET);
    fread (spal, 1, 16, palfile);
    fclose (palfile);

    packsize = offs_end - offs - 6;
    if (fread (scratch, 1, packsize, grfile) != packsize)
    {
      Cant_happen ("load_uw_bitmap_bits(): failed to load packed bitmap");
    }
    ip = scratch;
    op = bits;
    do
    {
      n = GET_NYBBLE (ip, w1);
      *op++ = spal [n];
    } while (op - bits < bmp->width * bmp->height);
    break;
  }

  default:
    Cant_happen ("load_uw_bitmap_bits(): unknown bitmap type");
    break;
  }
  bmp->bits = bits;

  /*
   * Should probably check the bitmap length field
   */

  fclose (grfile);
  return true;

}





ss_bitmap_header *Bmp_load_bitmap (int chunk, int subchunk)
{

  ss_bitmap_header *bmp;
  unsigned char *bits;

  /*
   * This feels a bit kludgy
   */
  if (Option_underworld && (chunk >= FAKE_CRIT_FRAME_CHUNK))
  {
    return Crit_get_frame (chunk - FAKE_CRIT_FRAME_CHUNK + CRIT_UW_BASE_OBJID,
			   subchunk);
  }

  /*
   * OK, it's DEFINITELY kludgy
   */
  if (Option_underworld && (chunk == BITMAP_CHUNK_UW_TMAP))
  {
    /*
     * These should always come direct from the textures cache
     */
    Cant_happen ("Bmp_load_bitmap(): illegal texture map load");
  }

  bmp = malloc (sizeof (ss_bitmap_header));
  if (bmp == NULL)
  {
    Cant_happen ("Bmp_load_bitmap(): couldn't allocate space for bitmap");
  }

  if (!Bmp_load_bitmap_header (bmp, chunk, subchunk))
  {
    free (bmp);
    return NULL;
  }

  bits = malloc (2 * bmp->width * bmp->height);
  if (bits == NULL)
  {
    Cant_happen ("Bmp_load_bitmap(): couldn't allocate space for bitmap");
  }
  if (!Bmp_load_bitmap_bits (bmp, bits, bmp->width * bmp->height))
  {
    free (bmp);
    free (bits);
    return NULL;
  }

  return bmp;

}



ss_bitmap_header *Bmp_load_thingy (int chunk)
{

  ss_bitmap_header *bmp = malloc (sizeof (ss_bitmap_header));
  unsigned char *bits;
  block_handle bh;
  unsigned char *chunk_bits;
  size_t size;

  if (bmp == NULL)
  {
    Cant_happen ("Bmp_load_thingy(): couldn't allocate space for bitmap");
  }

  bh = Res_load_chunk_by_id (chunk, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Bmp_load_thingy(): couldn't load chunk");
  }
  chunk_bits = (unsigned char *) *bh;

  printf ("%d %d %d %d %d %d %d %d\n",
	  chunk_bits [64], chunk_bits [65], chunk_bits [66], chunk_bits [67],
	  chunk_bits [68], chunk_bits [69], chunk_bits [70], chunk_bits [71]);

  bmp->width  = chunk_bits [80] + 256 * chunk_bits [81];
  bmp->height = chunk_bits [82] + 256 * chunk_bits [83];

  bits = malloc (bmp->width * bmp->height);
  if (bits == NULL)
  {
    Cant_happen ("Bmp_load_bitmap(): couldn't allocate space for bitmap");
  }
  memcpy (bits, chunk_bits + chunk_bits [76] + 256 * chunk_bits [77],
	  bmp->width * bmp->height);
  bmp->bits = bits;
  return bmp;

}




ss_bitmap_header *Bmp_create (int width, int height)
{

  ss_bitmap_header *bmp;
  size_t bmp_size = width * height;

  bmp = malloc (sizeof (ss_bitmap_header));
  if (bmp == NULL)
  {
    Cant_happen ("Bmp_create(): couldn't allocate space for bitmap");
  }

	/* ff1: added init ops, so that this function works like the */
	/*			G2D_create_ ones																		 */
	memset(bmp, 0x00, sizeof(ss_bitmap_header)); 
  bmp->x2 = bmp->width  = width;
  bmp->y2 = bmp->height = height;

  /*
   * Creating a bitmap with no pixels in it is allowed; the bitmap header will
   *  be populated accordingly but the bits pointer will be NULL
   */
  if (bmp_size > 0)
  {
    bmp->bits = malloc (bmp_size);
    if (bmp->bits == NULL)
    {
      Cant_happen ("Bmp_create(): couldn't allocate space for bitmap");
    }
  }
  else bmp->bits = NULL;

  return bmp;

}

ss_bitmap_header *Bmp_clone (ss_bitmap_header *pHead)
{
	ss_bitmap_header *pClone = malloc (sizeof (ss_bitmap_header));

	memcpy(pClone, pHead, sizeof (ss_bitmap_header));

  pClone->bits = (byte *)malloc(pHead->width * pHead->height);
	memcpy(pClone->bits, pHead->bits, pHead->width * pHead->height);

	return pClone;
}





/*
 * Only call this for bitmaps allocated with Bmp_load_bitmap(): if you load
 *  into preallocated space using Bmp_load_bitmap_header() and
 *  Bmp_load_bitmap_bits() directly you are responsible for the memory
 */
void Bmp_free_bitmap (ss_bitmap_header *bmp)
{
  if (bmp != NULL)
  {
    free (bmp->bits);
    free (bmp);
  }
}




/*
 * Unpack an Underworld run-length compressed bitmap with variable word length
 *  (4 or 5 bits). Header needs to have been populated already. pack is the
 *  packed bitmap, wsize the number of bits per word, auxpal the auxiliary
 *  palette (must have 2^wsize entries)
 */
void Bmp_unpack_uw_rle (ss_bitmap_header *bmp, byte *pack, int wsize,
			byte *auxpal)
{

  byte *ip = pack;       /* "in pointer"       */
  byte *op = bmp->bits;  /* "out pointer"      */
  unsigned buf = 0;      /* bit buffer         */
  int nbits = 0;         /* no. bits in buffer */
  unsigned mask = (1<<wsize) - 1;

  int run = 1;  /* packed data always starts with a run */
  int c;        /* count */
  int n;
  int scratch;
  int i;

  /*
   * A macro to get a word from the packed data
   */
# define GET_WORD(v) do { \
  while (nbits < wsize) { buf = (buf << 8) + *ip++; nbits += 8; } \
  nbits -= wsize; \
  v = (buf >> nbits) & mask; \
  } while (0)

  /*
   * Compressed bitmap
   */
  do
  {
    GET_WORD (c);
    if (run)
    {
      if (c == 1)
      {
	run = 0;
	GET_WORD (c);
      }
      else if (c == 2)
      {
	GET_WORD (run);
	if (run == 0)
	{
	  GET_WORD (run);
	  GET_WORD (scratch); run = (run << 4) + scratch;
	}
	GET_WORD (c);
      }
    }
    if (c == 0)
    {
      GET_WORD (c);
      GET_WORD (scratch); c = (c << 4) + scratch;
    }
    if (c == 0)
    {
      GET_WORD (c);
      GET_WORD (scratch); c = (c << 4) + scratch;
      GET_WORD (scratch); c = (c << 4) + scratch;
    }

    if (run)
    {
      GET_WORD (n);
      for (i = c; i > 0; --i)
      {
	*op++ = auxpal [n];
      }
      --run;
    }
    else
    {
      for (i = c; i > 0; --i)
      {
	GET_WORD (n);
	*op++ = auxpal [n];
      }
      run = 1;
    }
  } while (op - bmp->bits < bmp->width * bmp->height);

  /*
   * Tidy up after us
   */
# undef GET_WORD

}
