/*
 * bsp.c -
 *  Binary Space Partitioning
 */

/*
 * File:    src/bsp.c
 * Purpose: BSP trees, building and navigating
 * Prefix:  BSP_
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <limits.h>
#include <math.h>
#if NEED_MEMORY_H
#include <memory.h>
#endif
#include <stdlib.h>

#include "bsp.h"
#include "gamemem.h"
#include "main.h"
#include "map.h"



/*
 * Debug
 */
#ifdef __DEBUG__
#define BSP_MESSAGES 0
#endif


/*
 * Global variables
 */
bsp_node *Node_list;
static unsigned int Max_bsp_node;
static unsigned int Num_bsp_nodes;

/*
 * The root of the level BSP tree. Cyberspace only, will go away when we get
 *  proper sectors
 */
bsp_node *BSP_root;


/*
 * Local function prototypes
 */





void BSP_init (void)
{
}



void BSP_exit (void)
{
  free (Node_list);
}


/*
 * Teensy weensy function to find a free BSP node
 */
bsp_node *new_node (void)
{
  if (Num_bsp_nodes >= Max_bsp_node)
  {
    /*
     * This REALLY shouldn't happen, since we preallocate enough space for an
     *  entire level full of BSP nodes
     */
    Cant_happen ("bsp.c:new_node(): too many BSP nodes");
  }
  return &(Node_list [Num_bsp_nodes++]);
}



/*
 * Scratch workspace for the cyberspace tree generation algorithm
 */
struct bsp_scratch
{
  bsp_node *node; /* The node in the BSP tree. May be NULL           */
  double top;     /* The highest z coordinate (for bounding volumes) */
  double bottom;  /* The lowest z coordinate  (for bounding volumes) */
} *scratch_node;




void make_parent (unsigned parent, unsigned child1, unsigned child2, 
		  unsigned node_width, unsigned node_height,
		  const plane_t *sp, const vec *centre)
{

  double cube_radius;
  double top, bottom;
  bsp_node *node;

  /*
   * Reject nodes with no drawable children, promote only children
   */
  if ((scratch_node [child1].node == NULL) &&
      (scratch_node [child2].node == NULL))
  {
    scratch_node [parent].node = NULL;
#   if BSP_MESSAGES
    printf ("Node %u rejected\n", parent);
#   endif
    return;
  }
  else if (scratch_node [child1].node == NULL)
  {
    scratch_node [parent] = scratch_node [child2];
#   if BSP_MESSAGES
    printf ("Node %u promoted (front)\n", parent);
#   endif
    return;
  }
  else if (scratch_node [child2].node == NULL)
  {
    scratch_node [parent] = scratch_node [child1];
#   if BSP_MESSAGES
    printf ("Node %u promoted (back)\n", parent);
#   endif
    return;
  }

  /*
   * The node has 2 drawable children, and therefore must be combined. Make a
   *  new node for it
   */
  node = new_node ();
  scratch_node [parent].node = node;

  node->back  = scratch_node [child1].node;
  node->front = scratch_node [child2].node;

  /*
   * And its bounding spheroid
   */
  top = max (scratch_node [child1].top, scratch_node [child2].top);
  scratch_node [parent].top = top;
  bottom = min (scratch_node [child1].bottom, scratch_node [child2].bottom);
  scratch_node [parent].bottom = bottom;

  cube_radius =
    0.5 * sqrt (2 * node_width * node_width + node_height * node_height);
  node->z_scale = (top - bottom) / node_width;
  node->radius = cube_radius;

  /*
   * Ellipsoid centre. This is the iterative x and y passed in plus the
   *  average of top and bottom in the z direction
   */
  node->centre.x = centre->x;
  node->centre.y = centre->y;
  node->centre.z = (top + bottom) / 2;

  /*
   * Split plane. This is passed straight in
   */
  node->split_plane = *sp;

# if BSP_MESSAGES
  printf ("Node %u radius %g (%g, %g, %g) scale %g (%g-%g)\n", parent,
	  cube_radius, node->centre.x, node->centre.y, node->centre.z,
	  node->z_scale, bottom, top);
# endif

}




/*
 * Should be called after the map is set up (it uses Map_info), to initialise
 *  any BSP trees needed by the level (only a big one for cyberspace, so far)
 */
void BSP_init_level (void)
{

  unsigned int n;            /* node index                      */
  int i;                     /* misc. loop index                */
  unsigned int depth_nodes;  /* no. nodes at this tree depth    */
  unsigned int depth_width;  /* map width in nodes, this depth  */
  unsigned int depth_height; /* map height in nodes, this depth */
  unsigned int node_width  = TILE_GRID_SIZE;
  unsigned int node_height = TILE_GRID_SIZE;

  Num_bsp_nodes = 0;

  if (!Map_info.cyberspace)
  {
    /*
     * No BSP trees used for real-world levels (yet)
     */
    return;
  }

# if BSP_MESSAGES
  printf ("BSP_init_level(): creating BSP tree for cyberspace\n");
# endif

  /*
   * Allocate space for the initial BSP tree (unpruned)
   */
  Max_bsp_node = Map_info.xsize * Map_info.ysize * 2; /* binary tree */
  free (Node_list);
  Node_list = Alloc_init (Max_bsp_node * sizeof *Node_list, "BSP_init_level",
  			  "BSP nodes");

  /*
   * And allocate some scratch to work with
   */
  scratch_node = malloc (Max_bsp_node * sizeof *scratch_node);
  if (scratch_node == NULL)
  {
    Cant_happen ("BSP_init_level(): couldn't allocate workspace");
  }
# if BSP_MESSAGES
  printf ("BSP_init_level(): %d bytes allocated for workspace\n",
	  Max_bsp_node * sizeof *scratch_node);
# endif

  /*
   * First stage (lowest level): create a leaf node for each drawable tile
   *  (i.e. each tile not of type SOLID)
   */
  depth_nodes  = Map_info.xsize * Map_info.ysize;
  depth_width  = Map_info.xsize;
  depth_height = Map_info.ysize;
  for (n = 0; n < depth_nodes; ++n)
  {
    bsp_node *node;
    map_tile *tile = &(Tile [n]);
    unsigned int s = depth_nodes + n; /* pointer to node in scratch list */

    unsigned int tile_x = n % Map_info.xsize;
    unsigned int tile_y = n / Map_info.ysize;

    double cube_radius = 0.5 * sqrt (3*TILE_GRID_SIZE*TILE_GRID_SIZE);

    if (tile->type == TILE_SOLID)
    {
      scratch_node [s].node = NULL; /* solid tiles don't get a node */
      continue;
    }

    /*
     * The tile is drawable, so get a node for it
     */
    node = new_node ();
    scratch_node [s].node = node;
    node->front = NULL;
    node->back  = NULL;
    node->tile  = tile;

    /*
     * Find the highest and lowest points of the tile for the bounding volume
     */
    scratch_node [s].top    = INT_MIN;
    scratch_node [s].bottom = INT_MAX;
    for (i = 0; i < tile->num_vertices; ++i)
    {
      int z = Vertex [tile->vertex [i]].z;
      if (z > scratch_node [s].top)    scratch_node [s].top    = z;
      if (z < scratch_node [s].bottom) scratch_node [s].bottom = z;
    }
#   if BSP_MESSAGES
    printf ("Tile %d top %g bottom %g\n", n, scratch_node [s].top,
	    scratch_node [s].bottom);
#   endif

    node->centre.x = TILE_GRID_SIZE / 2 + tile_x * TILE_GRID_SIZE;
    node->centre.y = TILE_GRID_SIZE / 2 + tile_y * TILE_GRID_SIZE;
    node->centre.z = (scratch_node [s].top + scratch_node [s].bottom) / 2.;
    node->radius = cube_radius; /* basic bounding sphere assumes cube */
    node->z_scale = (scratch_node [s].top - scratch_node [s].bottom)
      / TILE_GRID_SIZE;
#   if BSP_MESSAGES
    printf ("Tile %d (%d,%d) centre (%g,%g,%g) radius %g scale %g\n", n,
	    tile_x, tile_y, node->centre.x, node->centre.y, node->centre.z,
	    node->radius, node->z_scale);
#   endif

  }

  /*
   * Now move up the tree, combining 2 nodes at each stage
   */
  while (depth_nodes > 1)
  {

    unsigned int node_x;
    unsigned int node_y;
    plane_t sp; /* split plane */
    vec centre; /* of current parent node */
      
    /*
     * First combine nodes adjacent in x
     */
    depth_nodes /= 2;
    depth_width /= 2;
    node_width *= 2;
#   if BSP_MESSAGES
    printf ("%u nodes %ux%u\n", depth_nodes, depth_width, depth_height);
#   endif
    sp.normal.x = 1; sp.normal.y = 0; sp.normal.z = 0;
    sp.np = node_width / 2;
    centre.x = sp.np;
    for (node_x = 0; node_x < depth_width; ++node_x)
    {
      unsigned this_node = depth_nodes + node_x;
      centre.y = node_height / 2;
      for (node_y = 0; node_y < depth_height; ++node_y)
      {
	/*
	 * Find the positions of the new node's children
	 */
	unsigned child1 = 2 * this_node;
	unsigned child2 = child1 + 1;

	make_parent (this_node, child1, child2, node_width, node_height,
		     &sp, &centre);
	this_node += depth_width;
	centre.y += node_height;

      }
      sp.np += node_width;
      centre.x += node_width;
    }

    /*
     * And now y, separately
     */
    depth_nodes /= 2;
    depth_height /= 2;
    node_height *= 2;
#   if BSP_MESSAGES
    printf ("%u nodes %ux%u\n", depth_nodes, depth_width, depth_height);
#   endif
    sp.normal.x = 0; sp.normal.y = 1; sp.normal.z = 0;
    sp.np = node_height / 2; centre.y = sp.np;
    for (node_y = 0; node_y < depth_height; ++node_y)
    {
      unsigned this_node = depth_nodes + node_y * depth_width;
      unsigned child1 = 2 * this_node;
      centre.x = node_width / 2;
      for (node_x = 0; node_x < depth_width; ++node_x, ++child1, ++this_node)
      {
	/*
	 * Find the positions of the new node's children
	 */
	unsigned child2 = child1 + depth_width;
	make_parent (this_node, child1, child2, node_width, node_height,
		     &sp, &centre);
	centre.x += node_width;

      }
      sp.np += node_height;
      centre.y += node_height;
    }

  }

# if BSP_MESSAGES
  printf ("BSP_init_level(): freeing scratch workspace\n");
# endif
  free (scratch_node);

# if BSP_MESSAGES
  printf ("%u BSP nodes this level\n", Num_bsp_nodes);
# endif

  /*
   * The root of the level BSP tree must be the last node allocated
   */
  BSP_root = &(Node_list [Num_bsp_nodes - 1]);

  /*
   * Reallocate the BSP tree to save memory (it's static)
   * Oops, it's got internal pointers. Well, not to worry about it at the
   *  moment, but we'll want to free up some space later
   */

}

