/*
 * clip.c -
 *  Routines for general clipping stuff
 */

/*
 * File:    src/clip.c
 * Purpose: 3-D clipping of surfaces
 * Prefix:  Clip_
 * Note:    
 * Music:   Ugly Kid Joe, "America's Least Wanted"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "clip.h"
#include "gamemem.h"
#include "main.h"    /* Cant_happen() */
#include "map.h"
#include "render.h"
#include "shade.h"
#include "surface.h"


/*
 * Debugging
 */
#if __DEBUG__
#define CLIP_MESSAGES 1
#define CLIP_PARANOID 1
#endif




/*
 * Global variables
 */
const double Min_view_distance = -A_VERY_SMALL_THING;
const double Max_view_distance = 32768.;




/*
 * Local typedefs and structs
 */
struct clipentry
{
  clipplane_t p;
  int next;
};


/*
 * Local function prototypes
 */
int find_free_clipplane (void);


static size_t Clipplane_list_size = 2048;
static unsigned int Clipplane_list_end;
static int Clipplane_list_free;

static struct clipentry *Clipplane_list;




void Clip_init (void)
{

  size_t clipplane_list_bytes = Clipplane_list_size * sizeof *Clipplane_list;

  Clipplane_list = malloc (clipplane_list_bytes);
  if (Clipplane_list == NULL)
  {
    Cant_happen ("Clip_init(): failed to allocate space for clipplane cache");
  }

# if CLIP_MESSAGES
  printf ("Clip_init(): %lu bytes allocated for cache\n",
	  (unsigned long) clipplane_list_bytes);
# endif

  Clipplane_list_end = 0;
  Clipplane_list_free = -1;

}

void Clip_deinit		 (void)
{
	free(Clipplane_list);
	Clipplane_list = NULL;
}

void Clip_init_frame (void)
{
/*    Clipplane_list_end = 0; */
/*    Clipplane_list_free = -1; */
}



void Clip_init_camera_frustrum (camera *c)
{

  clipplane_t view_plane;
  clipregion_t *frustrum = &(c->view_frustrum);

  /*
   * Create the view frustrum
   */
  Clip_init_region (frustrum);

  view_plane.normal.x = 0.;
  view_plane.normal.y = 0.;
  view_plane.normal.z = 1.;
  view_plane.np       = Min_view_distance;
  Clip_add_to_region (frustrum, &view_plane);

  view_plane.normal.z = -1.;
  view_plane.np       = -Max_view_distance;
  Clip_add_to_region (frustrum, &view_plane);

  view_plane.normal.x = -c->screen_z;
  view_plane.normal.z =  c->window_width / 2;
  vec_normalise (view_plane.normal);
  view_plane.normal.z -= A_VERY_SMALL_THING;
  view_plane.np       = 0.;  /* sides go through the origin, by definition */
  Clip_add_to_region (frustrum, &view_plane);

  view_plane.normal.x =  c->screen_z;
  view_plane.normal.z =  c->window_width / 2;
  vec_normalise (view_plane.normal);
  view_plane.normal.z -= A_VERY_SMALL_THING;
  view_plane.np       = 0.;
  Clip_add_to_region (frustrum, &view_plane);

  view_plane.normal.x =  0.;
  view_plane.normal.y = -c->screen_z;
  view_plane.normal.z =  c->window_height / 2;
  vec_normalise (view_plane.normal);
  view_plane.normal.z -= A_VERY_SMALL_THING;
  view_plane.np       = 0.;
  Clip_add_to_region (frustrum, &view_plane);

  view_plane.normal.x =  0.;
  view_plane.normal.y =  c->screen_z;
  view_plane.normal.z =  c->window_height / 2;
  vec_normalise (view_plane.normal);
  view_plane.normal.z -= A_VERY_SMALL_THING;
  view_plane.np       = 0.;
  Clip_add_to_region (frustrum, &view_plane);

}


void Clip_init_region (clipregion_t *r)
{
  r->num_planes = 0;
  r->first = -1;
}



void Clip_add_to_region (clipregion_t *r, const clipplane_t *p)
{
  int newplane = find_free_clipplane ();
  Clipplane_list [newplane].p = *p;
  Clipplane_list [newplane].next = r->first;
  r->first = newplane;
  ++r->num_planes;
}



void Clip_free_region (clipregion_t *r)
{
  int i;
  int next;
  for (i = r->first; i >= 0;)
  {
    next = Clipplane_list [i].next;
    Clipplane_list [i].next = Clipplane_list_free;
    Clipplane_list_free = i;    
    i = next;
    r->num_planes--;
    assert (r->num_planes >= 0);
  }
  assert (r->num_planes == 0);
  r->first = -1;
}


void Clip_copy_region (clipregion_t *dst, const clipregion_t *src)
{
  int i;
  Clip_init_region (dst);
  for (i = src->first; i >= 0;)
  {
    Clip_add_to_region (dst, &(Clipplane_list [i].p));
    i = Clipplane_list [i].next;
  }
}


bool Clip_point_in_region (const map_vertex *v, double radius,
			   const clipregion_t *r)
{
  int i = r->first;
  while (i >= 0)
  {
    struct clipentry *ce = &(Clipplane_list [i]);
    if (vec_dot (*v, ce->p.normal) + radius < ce->p.np) return false;
    i = ce->next;
  }
  return true;
}



/*
 * Clip a surface by a single plane, assuming that the surface is a convex
 *  polygon with vertices in circular order.
 * Amends the surface by adding or deleting vertices so that it is entirely
 *  in front of the plane.
 * Return values as discussed in clip.h
 */
int Clip_surface_by_plane_convex (surface *s, const clipplane_t *p)
{

  int i;
  int next;
  int out;
  int num_vertices_out;
  float dist;
  vec vertex [MAX_VERTEX_SURFACE];
  float dark [MAX_VERTEX_SURFACE];
  vec edge;
  vec newpt;
  int reject [MAX_VERTEX_SURFACE];
  int num_rejects = 0;

  for (i = 0; i < s->num_vertices; ++i)
  {
    reject [i] = POINT_IS_BEHIND (s->vertex [i], *p);
    if (reject [i]) ++num_rejects;
  }

  if (num_rejects == 0)                      return CLIP_ACCEPT;
  if (num_rejects == s->num_vertices)        return CLIP_REJECT;

  num_vertices_out = s->num_vertices + 2 - num_rejects;
  if (num_vertices_out > MAX_VERTEX_SURFACE) return CLIP_FAIL;

  for (out = 0, i = 0; i < s->num_vertices; ++i)
  {

    next = (i + 1) % s->num_vertices;

    if (!reject [i])
    {
      /*
       * Vertex is in front of the clip plane; output it unchanged
       */
      vertex [out] = s->vertex [i];
      dark [out] = s->dark [i];
      ++out;

#     if CLIP_PARANOID
      if (out > MAX_VERTEX_SURFACE)
      {
	Cant_happen ("Clip_surface_by_plane_convex(): too many points");
      }
#     endif


      if (reject [next])
      {
	/*
	 * Edge intersects plane going out of the clip region; insert a point
	 *  at the intersection of the edge and the clip plane
	 */
	double dp;
	vec_subtract (edge, s->vertex [next], s->vertex [i]);
	dp = vec_dot (edge, p->normal);

	/*
	 * It is in fact quite possible for the dot product to be zero here:
	 *  if a face is being clipped against a portal with which it shares
	 *  an edge (not a rare occurrence) floating-point inaccuracies might
	 *  cause one vertex to be rejected and the other accepted. Trying to
	 *  insert a new point on the edge in this case will cause a crash. We
	 *  can't eliminate the pathological case so must test for it here.
	 */
  	if ((dp < A_VERY_SMALL_THING) && (dp > -A_VERY_SMALL_THING)) continue;
	dist = (p->np - vec_dot (s->vertex [i], p->normal)) / dp;
/*  	if ((dist > 1.) || (dist < 0.)) continue; */
	vec_multiply (edge, dist);
	vec_add (newpt, edge, s->vertex [i]);
	vertex [out] = newpt;
	dark [out] = s->dark [i] + dist * (s->dark [next] - s->dark [i]);

	++out;

#       if CLIP_PARANOID
        if (out > MAX_VERTEX_SURFACE)
        {
          Cant_happen ("Clip_surface_by_plane_convex(): too many points");
        }
#       endif

      }
    }
    else if (!reject [next])
    {
      /*
       * Edge intersects plane going into the clip region; insert a point at
       *  the intersection of the edge and the clip plane
       */
      double dp;
      vec_subtract (edge, s->vertex [next], s->vertex [i]);
      dp = vec_dot (edge, p->normal);
      if ((dp < A_VERY_SMALL_THING) && (dp > -A_VERY_SMALL_THING)) continue;
      dist = (p->np - vec_dot (s->vertex [i], p->normal)) / dp;
      vec_multiply (edge, dist);
      vec_add (newpt, edge, s->vertex [i]);
      vertex [out] = newpt;
      dark [out] = s->dark [i] + dist * (s->dark [next] - s->dark [i]);
      ++out;

#     if CLIP_PARANOID
      if (out > MAX_VERTEX_SURFACE)
      {
	Cant_happen ("Clip_surface_by_plane_convex(): too many points");
      }
#     endif

    }
  }

  /*
   * Put the vertices back in the surface, deduping as we go
   */
  s->num_vertices = 0;
  i = out - 1;
  for (next = 0; next < out; ++next)
  {
    if (!vec_coincides (vertex [i], vertex [next]))
    {
      /*
       * In case of inaccuracies in the floating-point
       */
/*        if (dark [i] >= DARK_SHADES) dark [i] = DARK_SHADES - 1; */
/*        if (dark [i] < 0) dark [i] = 0; */
      assert ((dark [i] >= 0.) && (dark [i] < DARK_SHADES));
      s->dark [s->num_vertices]   = dark [i];
      s->vertex [s->num_vertices++] = vertex [i];
    }
    i = next;    
  }

  /*
   * Check that the new surface really is a poly i.e. has at least 3 sides
   */
  if (s->num_vertices > 2) return CLIP_ACCEPT; else return CLIP_REJECT;

}




/*
 * Clip a line by a single plane, amending end vertices if necessary
 * Return values as discussed in clip.h
 */
int Clip_line_by_plane (vec *v1, vec *v2, const clipplane_t *p)
{

  double dist;
  double dp;
  vec edge;
  bool r1, r2;

  r1 = POINT_IS_BEHIND (*v1, *p);
  r2 = POINT_IS_BEHIND (*v2, *p);

  if (r1 && r2)    return CLIP_REJECT; /* both endpoints rejected */
  if (!(r1 || r2)) return CLIP_ACCEPT; /* both endpoints accepted */

  /*
   * The line intersects the plane. Find the point of intersection and replace
   *  one or other endpoint
   */
  vec_subtract (edge, *v2, *v1);
  dp = vec_dot (edge, p->normal);

  /*
   * See discussion in Clip_surface_by_plane_convex()
   */
  if (dp == 0.) return CLIP_ACCEPT;
  dist = (p->np - vec_dot (*v1, p->normal)) / dp;
  vec_multiply (edge, dist);
  if (r1) vec_add (*v1, edge, *v1);
  else    vec_add (*v2, edge, *v1);

  return CLIP_ACCEPT;

}




int Clip_ellipsoid_z_by_plane (const vec *centre, double radius,
			       double z_scale, const clipplane_t *p)
{

  double mag2;

  /*
   * This determines on which side of the plane is the centre of the ellipsoid.
   *  Effectively we are translating everything so that this centre is at the
   *  origin
   */
  double adj_np = p->np - vec_dot (*centre, p->normal);

  /*
   * Test for intersection. First scale in the z direction so that the
   *  ellipsoid becomes a sphere
   * Imagine that we are describing the plane parametrically (we don't, but it
   *  makes the transform easier to describe). If both vectors within the plane
   *  which describe it (two non-coincident vectors describe a plane) are
   *  scaled in z (z -> z/h) then their cross product has x -> x/h and y -> y/h
   *  however we multiply up the z coordinate, effectively multiplying both
   *  sides of the plane equation by h.
   */
  vec adj_normal;
  adj_normal.x = p->normal.x; adj_normal.y = p->normal.y;
  adj_normal.z = z_scale * p->normal.z;

  /*
   * Now the clipplane's normal is no longer a unit vector, so correct for this
   */
  mag2 = vec_magnitude2 (adj_normal);

  /*
   * And check for intersection. Sphere / plane intersection is trivial.
   * To avoid a time-consuming square root, square both sides.
   * Multiply radius by normal magnitude to normalise
   */
  if (adj_np * adj_np <= radius * radius * mag2)
  {
    return CLIP_INTERSECT;
  }
  else if (adj_np < 0) return CLIP_WITHIN;
  else return CLIP_WITHOUT;

}





/*
 * Clip a surface so that it is entirely within a clip region. Assumes that
 *  both the surface and the region are convex (and will stay so)
 */
int Clip_surface_by_region_convex (surface *s, const clipregion_t *r)
{

  int i;
  bool fail = false;

  for (i = r->first; i >= 0; i = Clipplane_list [i].next)
  {
    switch (Clip_surface_by_plane_convex (s, &(Clipplane_list [i].p)))
    {
    case CLIP_ACCEPT:
      break;
    case CLIP_REJECT:
      return CLIP_REJECT;
    case CLIP_FAIL:
      fail = true;
      break;
    default:
      Cant_happen ("Clip_surface_by_region_convex(): illegal clip status");
    }
  }

  return fail ? CLIP_FAIL : CLIP_ACCEPT;

}





/*
 * Clip a line so that it is entirely within a (convex) clip region.
 */
int Clip_line_by_region (vec *v1, vec *v2, const clipregion_t *r)
{

  int i;

  for (i = r->first; i >= 0; i = Clipplane_list [i].next)
  {
    switch (Clip_line_by_plane (v1, v2, &(Clipplane_list [i].p)))
    {
    case CLIP_ACCEPT:
      break;
    case CLIP_REJECT:
      return CLIP_REJECT;
      /*
       * CLIP_FAIL is impossible for a line
       */
    default:
      Cant_happen ("Clip_line_by_region(): illegal clip status");
    }
  }

  return CLIP_ACCEPT;

}





/*
 * Test intersection of a bounding ellipsoid scaled in z only (used for
 *  culling of BSP trees) with a clip region
 * This is more or less the same as Clip_surface_by_region_convex()
 */
int Clip_ellipsoid_z_by_region (const vec *centre, double radius,
				double z_scale, const clipregion_t *r)
{

  int i;
  bool intersect = false;

  for (i = r->first; i >= 0; i = Clipplane_list [i].next)
  {
    switch (Clip_ellipsoid_z_by_plane (centre, radius, z_scale,
				       &(Clipplane_list [i].p)))
    {
    case CLIP_WITHIN:
      break;
    case CLIP_WITHOUT:
      return CLIP_WITHOUT;
    case CLIP_INTERSECT:
      intersect = true;
      break;
    default:
      Cant_happen ("Clip_ellipsoid_z_by_region(): illegal clip status");
    }
  }

  return intersect ? CLIP_INTERSECT : CLIP_WITHIN;

}





int Clip_surface_by_view_frustrum_convex (surface *s)
{
  return Clip_surface_by_region_convex (s, &(Camera->view_frustrum));
}




/*
 * Transform a camera frustrum to world space (used by render.c)
 */
void Clip_transform_camera_frustrum (clipregion_t *r, const camera *c)
{

  int i = c->view_frustrum.first;

  Clip_init_region (r);
  while (i >= 0)
  {
    struct clipentry *ce = &(Clipplane_list [i]);
    clipplane_t cp;
    mat_vec_multiply_transpose (cp.normal, c->orientation, ce->p.normal);
    cp.np = ce->p.np + vec_dot (cp.normal, c->position);
    Clip_add_to_region (r, &cp);    
    i = ce->next;
  }

}



/*
 * Transform a clipregion from world to view space for the current camera
 */
void Clip_region_world_to_view (clipregion_t *dst, const clipregion_t *src)
{

  int i;

  assert (dst != NULL);
  assert (src != NULL);
  Clip_init_region (dst);

  for (i = src->first; i >= 0;)
  {
    struct clipentry *ce = &(Clipplane_list [i]);
    clipplane_t cp;
    mat_vec_multiply (cp.normal, Camera->orientation, ce->p.normal);
    cp.np = ce->p.np - vec_dot (ce->p.normal, Camera->position);
    Clip_add_to_region (dst, &cp);    
    i = ce->next;
  }

}




/*
 * Returns false if the point is coplanar with the surface, in which case no
 *  meaningful clip region can be made from it (and we should fall back on the
 *  including region)
 */
bool Clip_region_from_surface_convex (clipregion_t *r, const surface *s,
				      const vec *viewpoint)
{

  int i;
  int prev = s->num_vertices - 1;
  clipplane_t cp;

  double ts = vec_dot (s->normal, *viewpoint) -
    vec_dot (s->vertex [0], s->normal);
  if ((ts < A_VERY_SMALL_THING) && (ts > -A_VERY_SMALL_THING))
  {
    return false;
  }

  /*
   * Making a clipregion out of a convex surface is very simple; we just
   *  make a clip plane out of each pair of contiguous vertices and the
   *  viewpoint
   */
  for (i = 0; i < s->num_vertices; prev = i++)
  {
    vec side1, side2;
    vec_subtract (side1, s->vertex [prev], *viewpoint);
    vec_subtract (side2, s->vertex [i   ], *viewpoint);
    vec_cross (cp.normal, side1, side2);
    vec_normalise (cp.normal);
    cp.np = vec_dot (cp.normal, s->vertex [i]);
    Clip_add_to_region (r, &cp);
  }

  return true;

}



/*
 * For debugging
 */
void Clip_dump_region (clipregion_t *r)
{
  int i;
  printf ("Clip region has %d planes, first %d\n", r->num_planes, r->first);
  for (i = r->first; i != -1; i = Clipplane_list [i].next)
  {
    printf ("\t(%g, %g, %g)\t%g\n", Clipplane_list [i].p.normal.x,
	    Clipplane_list [i].p.normal.y, Clipplane_list [i].p.normal.z,
	    Clipplane_list [i].p.np);
  }
  printf ("\n");
}



int find_free_clipplane (void)
{
  int found;
  if (Clipplane_list_free >= 0)
  {
    assert (Clipplane_list_free < (int) Clipplane_list_size);
    found = Clipplane_list_free;
    Clipplane_list_free = Clipplane_list [found].next;
    assert (Clipplane_list_free < (int) Clipplane_list_size);
  }
  else if (Clipplane_list_end < Clipplane_list_size)
  {
    found = Clipplane_list_end++;
  }
  else
  {
    Cant_happen ("find_free_clipplane(): no free clipplanes found");
    /*
     * Never get here, but keeps the compiler happy
     */
    return -1;
  }
  return found;
}
