/*
 * clip.h -
 *  Clipping
 */



#ifndef CLIP_H
#define CLIP_H



/*
 * #includes
 */
#include "vec.h"




/*
 * #defines
 */
#define MAX_CLIPS_PER_REGION 16


/*
 * Return codes for clipping functions.
 * CLIP_ACCEPT means that the entity was successfully clipped and that at
 *  least part of it is within the region being tested.
 * CLIP_REJECT means that the entity is entirely outside the region and should
 *  be rejected. It is not guaranteed to be in a meaningful state after the
 *  return.
 * CLIP_FAIL means that the entity could not be clipped (a surface would need
 *  extra vertices added which would cause it to overflow the vertex maximum).
 *  It is guaranteed to be a sensible shape but might poke out of the
 *  clipping region in arbitrary ways.
 * CLIP_REJECT will always be zero, so code of the form if (!clip_function)
 *  will work.
 */
#define CLIP_FAIL  -1
#define CLIP_REJECT 0
#define CLIP_ACCEPT 1


/*
 * The following serve a similar purpose for bounding volumes (spheres and
 *  ellipsoids).
 * CLIP_WITHIN means that the bounding volume is entirely within the clipping
 *  region (and hence its associated object need not be clipped, nor its
 *  children if hierarchical).
 * CLIP_WITHOUT means that the bounding volume is entirely outside the clipping
 *  region (and hence the object and children should be rejected).
 * CLIP_INTERSECT means that at least one of the clipplanes intersects the
 *  bounding volume, so that further clipping work is needed to determine
 *  exactly what should be drawn.
 * As above, CLIP_WITHOUT is defined as zero so the simple if() test will work.
 */
#define CLIP_WITHIN    1
#define CLIP_WITHOUT   0
#define CLIP_INTERSECT 2



/*
 * typedefs and structs
 */


/*
 * Planet?
 */
typedef struct
{
  vec    normal;   /* normal vector, not surprisingly                    */
  double np;       /* distance (normal dotted with a point on the plane) */
} plane_t;



/*
 * For clarity of code
 */
typedef plane_t clipplane_t;


/*
 * A collection of clip planes, nothing more
 */
typedef struct tag_clip_region
{
  int num_planes;
  int first;          /* index into the main clip plane list */
} clipregion_t;



/*
 * For unit plane normals ... beware if you set this
 * Positive values are more lenient
 */
#define PLANE_CLIP_THRESHOLD 0



/*
 * Point is behind plane (i.e. not on side of positive normal)
 */
#define POINT_IS_BEHIND(point,plane) \
( vec_dot ((point), (plane).normal) < (plane).np + PLANE_CLIP_THRESHOLD )



#ifdef __cplusplus
extern "C"
{
#endif



/*
 * Function prototypes
 */
void Clip_init       (void);
void Clip_deinit		 (void);
void Clip_init_frame (void);
void Clip_init_region   (clipregion_t *r);

struct tag_camera;
void Clip_init_camera_frustrum (struct tag_camera *c);
void Clip_add_to_region (clipregion_t *r, const clipplane_t *p);
void Clip_free_region   (clipregion_t *r);
void Clip_copy_region (clipregion_t *dst, const clipregion_t *src);
void Clip_transform_camera_frustrum (clipregion_t *r,
				     const struct tag_camera *c);
void Clip_region_world_to_view (clipregion_t *dst, const clipregion_t *src);

struct tag_map_vertex;
bool Clip_point_in_region (const struct tag_map_vertex *v, double radius,
			   const clipregion_t *r);

struct tag_surface;
int Clip_surface_by_plane_convex (struct tag_surface *s, const clipplane_t *p);

/*
 * It is almost certainly the case that we will never allow concave polygons
 *  within a single surface ...
 */
#define Clip_surface_by_plane Clip_surface_by_plane_convex

int Clip_line_by_plane (vec *v1, vec *v2, const clipplane_t *p);
int Clip_ellipsoid_z_by_plane (const vec *centre, double radius,
			       double z_scale, const clipplane_t *p);


int Clip_surface_by_region_convex (struct tag_surface *s,
				   const clipregion_t *r);

/*
 * And we're VERY unlikely to have non-convex clipping regions
 */
#define Clip_surface_by_region Clip_surface_by_region_convex

int Clip_line_by_region (vec *v1, vec *v2, const clipregion_t *r);

int Clip_ellipsoid_z_by_region (const vec *centre, double radius,
				double z_scale, const clipregion_t *r);


/*
 * similarly for the view frustrum
 */
int Clip_surface_by_view_frustrum_convex (struct tag_surface *s);
#define Clip_surface_by_view_frustrum Clip_surface_by_view_frustrum_convex



/*
 * Note that r is not inited by the function; the caller must call
 *  Clip_init_region() on it first. Caller inits, caller frees, always.
 */
bool Clip_region_from_surface_convex (clipregion_t *r,
				      const struct tag_surface *s,
				      const vec *viewpoint);


/*
 * For debugging purposes
 */
void Clip_dump_region (clipregion_t *r);


#ifdef __cplusplus
}
#endif


#endif /* CLIP_H */
