/*
 * critter.c -
 *  Critters! (They Bite)
 */

/*
 * File:    src/critter.c
 * Purpose: General critter object handling
 * Prefix:  Crit_
 * Note:    
 * Music:   Fairport Convention, "Jewel In The Crown"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */




/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "gamemem.h"
#include "bitmap.h"
#include "critter.h"
#include "main.h"     /* Cant_happen() */
#include "object.h"
#include "option.h"
#include "resource.h"
#include "texture.h"



/*
 * #defines
 */
#define CRITTER_BASE_OBJID      439


#define NUM_CRITTERS 37

#define CRITTER_SUBCLASS_MUTANT 0
#define CRITTER_SUBCLASS_BOT    1
#define CRITTER_SUBCLASS_CYBORG 2
#define CRITTER_SUBCLASS_CYBER  3
#define CRITTER_SUBCLASS_BOSS   4

#define CRITTER_NUM_CYBER 7


/*
 * Global variables
 */
int Critter_idle_base_chunk [37] =
{
  1585,   /* humanoid mutant     */
  1593,   /* gor-tiger mutant    */
  1601,   /* insectoid mutant    */
  1609,   /* avian mutant        */
  1617,   /* plant mutant        */
  1625,   /* zero-g creature     */
  1633,   /* Sonic the Hedgehog  */
  1641,   /* invisible mutant    */
  1649,   /* virus mutant        */
  1657,   /* lifter              */
  1665,   /* repairbot           */
  1673,   /* serv-bot            */
  1681,   /* exec-bot            */
  1689,   /* large turret        */
  1690,   /* hopper              */
  1698,   /* security-1-bot      */
  1706,   /* security-2-bot      */
  1714,   /* autobomb            */
  1722,   /* maint-bot           */
  1730,   /* flier               */
  1738,   /* security-3-bot      */
  1746,   /* drone cyborg        */
  1754,   /* warrior cyborg      */
  1762,   /* cyborg assassin     */
  1770,   /* cyborg enforcer     */
  1778,   /* elite cyborg        */
  1786,   /* cortex reaver       */
  1794,   /* mutated cyborg      */
     0,   /* cyberdog            */
     0,   /* cyberguard          */
     0,   /* cyber cortex reaver */
     0,   /* dynamic ice         */
     0,   /* cyber hunter killer */
     0,   /* SHODAN              */
     0,   /* cyberguard          */
  1809,   /* SHODAN              */
  1810,   /* Edward Diego        */
};






/*
 * Local function prototypes
 */
static void init_uw_anim_segments (void);
static void init_uw_crit_anims    (void);
static void init_uw_assoc_table   (void);
static void init_uw_page_map      (void);





void Crit_init (void)
{
  if (Option_underworld)
  {
    init_uw_crit_anims ();
  }
}




/*
 * File offset for class 14 in objprop.dat
 */
#define OBJPROP_OFFSET_CRITTER 0x08b9

byte Cyber_critter_colour [CRITTER_NUM_CYBER] [6];

/*
 * Load up critter data from objprop.dat / objects.dat
 * We are passed the file with its pointer at the beginning of the critter
 *  properties, and must leave it at the end for the next class (not a class
 *  as such, but the generic object data follows critters in the file)
 */
void Crit_load_objprop (FILE *propfile)
{

  int i; /* misc. loop index */

  /*
   * Trust No One ...
   */
  assert (ftell (propfile) == OBJPROP_OFFSET_CRITTER);

  /*
   * First comes the generic object properties, 75 bytes per critter
   */
  for (i = 0; i < NUM_CRITTERS; ++i)
  {
    /*
     * but I don't know what they do
     */
    fseek (propfile, 75, SEEK_CUR);
  }
  fseek (propfile, 9, SEEK_CUR); /* mutants 14/0 */
  fseek (propfile, 24, SEEK_CUR); /* bots 14/1 */
  fseek (propfile, 14, SEEK_CUR); /* cyborgs 14/2 */
  for (i = 0; i < CRITTER_NUM_CYBER; ++i)
  {
    fread (Cyber_critter_colour [i], 1, 6, propfile);
#   if CRITTER_MESSAGES
    printf ("colour %02x %02x %02x %02x %02x %02x\n",
	    Cyber_critter_colour [i] [0], Cyber_critter_colour [i] [1],
	    Cyber_critter_colour [i] [2], Cyber_critter_colour [i] [3],
	    Cyber_critter_colour [i] [4], Cyber_critter_colour [i] [5]);
#   endif
  }
  fseek (propfile, 2, SEEK_CUR); /* boss 14/4 */
/*    printf ("%04lx\n", ftell (propfile)); */

}



byte *Crit_get_colour (map_object *obj)
{

  assert (obj->class == OBJCLASS_CRITTER);

  /*
   * Only cyberspace objects have colour schemes (as colour schemes aren't
   *  used except for c/space)
   */
  if (obj->subclass != CRITTER_SUBCLASS_CYBER) return NULL;

  assert (obj->type < CRITTER_NUM_CYBER);
  return Cyber_critter_colour [obj->type];      

}



/*
 * A critter appears to the object renderer as an orientable sprite, hence
 *  `bmpno' is the orientation
 */
int Crit_fetch_texture (map_object *obj)
{

  /*
   * In future we will return a bitmap from our own cache depending on the
   *  critter AI state and frame. But for now, just return the idle bitmap
   *  for this orientation
   */
  int critno = obj->object_id - CRITTER_BASE_OBJID;

  /*
   * Crit_fetch_texture() should only ever be called once for a given object
   */
  assert (obj->base_texid < 0);

  if (critno == 35)
  {
    /*
     * SHODAN only has one bitmap
     */
    obj->base_texid = Tex_get_cache_id (Critter_idle_base_chunk [critno], 0);
    obj->fetch_texture_hook = Obj_fetch_texture_single;
    return obj->base_texid;
  }
  else if (critno >= 37)
  {
    Cant_happen ("Crit_fetch_texture(): not a critter");
  }

  obj->base_texid = Tex_get_cache_block (Critter_idle_base_chunk [critno],
					 0, 8, TEXTURE_BLOCK_CHUNK);
  obj->fetch_texture_hook = Obj_fetch_texture_osprite;
  return Obj_fetch_texture_osprite (obj);

}





#ifndef MAX_PATH
#define MAX_PATH 512
#endif

extern bool Underworld2;
char UW_crit_dir [MAX_PATH];

#define MAX_ANIM_PAGES      8
#define MAX_CRIT_ANIMS     32
#define MAX_ANIM_SEGMENTS 256
#define MAX_SEGMENT_FRAME   8

int Num_crit_anims;

/*
 * The format string to pass to sprintf to get the name of an animation page
 *  file out. It takes 3 arguments; path, animation number and page number
 */
const char *anim_page_format;

int *UW_page_map;

/*
 * This has type `pointer to array [MAX_ANIM_SEGMENTS] of struct foo'. Use it
 *  as a 2-dimensional array
 */
static struct
{
  int num_frames;
  int frame [MAX_SEGMENT_FRAME];
} (*UW_anim_segment_map) [MAX_ANIM_SEGMENTS];

static struct
{
  int anim;
  int variant;
} *UW_assoc_table;





void init_uw_anim_segments (void)
{
  char segname [MAX_PATH];
  FILE *segfile;
  int anim;
  int segment;
  int fno;
  int page = 0;

  /*
   * Allocate space for the segment table
   */
  UW_anim_segment_map =
    Alloc_init (Num_crit_anims * sizeof *UW_anim_segment_map, "Crit_init",
		"segment map");

  if (Underworld2)
  {
    /*
     * The animation segment map is in its own file, crit/cr.an, in more or
     *  less the format we want it
     */
    sprintf (segname, "%s/cr.an", UW_crit_dir);
    segfile = fopen (segname, "rb");
    if (segfile == NULL)
    {
      Cant_happen ("Crit_init(): can't open segment map file");
    }
    for (anim = 0; anim < Num_crit_anims; ++anim)
    {
      for (segment = 0; segment < 64; ++segment)
      {
	for (fno = 0; fno < 7; ++fno)
	{
	  UW_anim_segment_map [anim] [segment].frame [fno] = fgetc (segfile);
	}
	UW_anim_segment_map [anim] [segment].num_frames = fgetc (segfile);
      }
    }
    fclose (segfile);
  }
  else
  {

    /*
     * Compile a segment table from the segment directories in the page files
     *  themselves
     */
    for (anim = 0; anim < Num_crit_anims; ++anim)
    {
      int baseframe = 0;

      for (page = 0;; ++page)
      {
	int nsegs;
	int segbase;
	int segdir;

	sprintf (segname, anim_page_format, UW_crit_dir, anim, page);
	segfile = fopen (segname, "rb");
	if (segfile == NULL) break;      /* run out of page files */

	segbase = fgetc (segfile);
	nsegs   = fgetc (segfile);
	for (segment = 0; segment < nsegs; ++segment)
	{
	  /*
	   * Get the segment directory offset
	   */
	  fseek (segfile, 2 + segment, SEEK_SET);
	  segdir = fgetc (segfile);
	  if (segdir == 255)
	  {
	    /*
	     * No such segment
	     */
	    UW_anim_segment_map [anim] [segment + segbase].num_frames = 0;
	    continue;
	  }

	  /*
	   * And seek to it ...
	   */
	  fseek (segfile, 2 + nsegs, SEEK_SET);
	  if (segdir >= fgetc (segfile))
	  {
	    Cant_happen ("Crit_init(): bad segment table");
	  }
	  fseek (segfile, 8 * segdir, SEEK_CUR);
	  for (fno = 0; fno < 8; ++fno)
	  {
	    int frame = fgetc (segfile);
	    if (frame == 255)
	    {
	      /*
	       * End of animation
	       */
	      break;
	    }

	    /*
	     * We store absolute not relative frame numbers
	     */
	    UW_anim_segment_map [anim] [segment + segbase].frame [fno]
	      = frame + baseframe;
	  }
	  UW_anim_segment_map [anim] [segment + segbase].num_frames = fno;
	}

	fclose (segfile);
	baseframe = UW_page_map [anim * MAX_ANIM_PAGES + page] + 1;
      }
    }
  }

}



void init_uw_crit_anims (void)
{

  /*
   * Figure out where UW critter stuff lives
   */
  sprintf (UW_crit_dir, "%s/crit", UW_data_path);
  init_uw_page_map ();
  init_uw_assoc_table ();
  init_uw_anim_segments ();

}




void init_uw_assoc_table (void)
{

  char assocname [MAX_PATH];
  FILE *assocfile;
  int i;

  if (Underworld2)
  {
    sprintf (assocname, "%s/as.an", UW_crit_dir);
  }
  else
  {
    sprintf (assocname, "%s/assoc.anm", UW_crit_dir);
  }
  assocfile = fopen (assocname, "rb");
  if (assocfile == NULL)
  {
    Cant_happen ("Crit_init(): failed to open associations file");
  }

  if (!Underworld2)
  {
    /*
     * Ultima Underworld's associations file has the animation names at the
     *  start - skip them
     */
    fseek (assocfile, 8 * Num_crit_anims, SEEK_SET);
  }

  /*
   * There are 2 bytes per critter, animation and variant. Read it into the
   *  associations table
   */
  UW_assoc_table = Alloc_init (CRIT_UW_NUM_CRITS * sizeof *UW_assoc_table,
			       "Crit_init", "association table");

  for (i = 0; i < CRIT_UW_NUM_CRITS; ++i)
  {
    UW_assoc_table [i].anim    = fgetc (assocfile);
    UW_assoc_table [i].variant = fgetc (assocfile);
  }

  fclose (assocfile);

}



void init_uw_page_map (void)
{

  char pgmpname [MAX_PATH];
  FILE *pgmpfile;
  int i, j;

  /*
   * Get the page map file, if it is an Underworld II crit directory
   */
  sprintf (pgmpname, "%s/pg.mp", UW_crit_dir);
  pgmpfile = fopen (pgmpname, "rb");
  if (pgmpfile != NULL)
  {

    /*
     * Got it ... we're Underworld II
     */
    Underworld2 = true;
    anim_page_format = "%s/cr%02o.%02o";

    /*
     * Get the number of critter animations from the size of the page map file
     */
    fseek (pgmpfile, 0, SEEK_END);
    Num_crit_anims = ftell (pgmpfile) / MAX_ANIM_PAGES;

    /*
     * Allocate and load up the page map. This has one entry per page per
     *  critter, giving the LAST animation frame in that page file
     */
    UW_page_map =
      Alloc_init (Num_crit_anims * MAX_ANIM_PAGES * sizeof *UW_page_map,
		  "Crit_init", "page map");

    rewind (pgmpfile);
    for (i = 0; i < Num_crit_anims; ++i)
    {
      for (j = 0; j < MAX_ANIM_PAGES; ++j)
      {
	UW_page_map [i * MAX_ANIM_PAGES + j] = fgetc (pgmpfile);
      }
    }
    fclose (pgmpfile);

  }
  else
  {

    /*
     * We're Underworld ... probably
     */
    anim_page_format = "%s/cr%02opage.n%02o";
    Num_crit_anims = 0;
    UW_page_map =
      Alloc_init (MAX_CRIT_ANIMS * MAX_ANIM_PAGES * sizeof *UW_page_map,
		  "Crit_init", "page map");

    for (i = 0;; ++i)
    {
      int last = -1;

      sprintf (pgmpname, anim_page_format, UW_crit_dir, i, 0);
      pgmpfile = fopen (pgmpname, "rb");
      if (pgmpfile == NULL)
      {
	break;
      }
      ++Num_crit_anims;

      /*
       * This critter animation exists and we have its first page. Find our
       *  list of last frames
       */
      j = 0;
      do
      {
	int skip;

	/*
	 * Find the frame directory
	 */
	fgetc (pgmpfile);         /* start anim slot */
	skip = fgetc (pgmpfile);  /* num. anim slots */
	fseek (pgmpfile, skip, SEEK_CUR);

	skip = fgetc (pgmpfile);  /* num. anim segments */
	fseek (pgmpfile, 8 * skip, SEEK_CUR);

	skip = fgetc (pgmpfile);  /* num. auxiliary palettes */
	fseek (pgmpfile, 32 * skip, SEEK_CUR);

	last += fgetc (pgmpfile); /* at last! */
	UW_page_map [i * MAX_ANIM_PAGES + j] = last;

	/*
	 * And get the next page file
	 */
	fclose (pgmpfile);
	++j;
	sprintf (pgmpname, anim_page_format, UW_crit_dir, i, j);
	pgmpfile = fopen (pgmpname, "rb");

      } while (pgmpfile != NULL);

      /*
       * Fill in the rest with 255s
       */
      for (; j < MAX_ANIM_PAGES; ++j)
      {
	UW_page_map [i * MAX_ANIM_PAGES + j] = 255;
      }

    }

    if (Num_crit_anims == 0)
    {
      /*
       * We didn't find ANY page files at all - panic!
       */
      Cant_happen ("Crit_init(): no animation page files found");
    }
    else
    {
      Underworld2 = false;
    }

  }

}




#define MAX_PACKED_FRAME_SIZE 4096

int Crit_get_uw_segment_frame (int objid, int segment, int frame, int *variant)
{

  int critno = objid - CRIT_UW_BASE_OBJID;
  int anim;

  if ((critno < 0) || (critno >= CRIT_UW_NUM_CRITS))
  {
    Cant_happen ("Crit_get_uw_segment_frame(): illegal object ID");
  }
  anim = UW_assoc_table [critno].anim;

  if (frame >= UW_anim_segment_map [anim] [segment].num_frames)
  {
    Cant_happen ("Crit_get_uw_segment_frame(): illegal frame number");
  }

  *variant = UW_assoc_table [critno].variant;
  return UW_anim_segment_map [anim] [segment].frame [frame];

}

ss_bitmap_header *Crit_get_frame (int objid, int frame)
{

  int critno = objid - CRIT_UW_BASE_OBJID;
  if ((critno < 0) || (critno >= CRIT_UW_NUM_CRITS))
  {
    Cant_happen ("Crit_get_frame(): illegal object ID");
  }

  return Crit_get_anim_frame (UW_assoc_table [critno].anim, frame,
			      UW_assoc_table [critno].variant);

}

ss_bitmap_header *Crit_get_anim_frame (int anim, int frame, int variant)
{

  int page;                    /* number of pagefile where frame is found */
  char animname [MAX_PATH];    /* name of page file */
  FILE *animfile;              /* file handle */
  byte auxpal [32];            /* auxiliary palette (passed to unpack) */
  int npals;                   /* no. of auxiliary palettes in file */
  int skip;                    /* size of field to skip */
  int i;                       /* misc. loop index */
  int offset;                  /* offset of frame within page file */
  ss_bitmap_header *bmp;
  byte pack [MAX_PACKED_FRAME_SIZE];
  int wsize;

  if (anim > Num_crit_anims)
  {
    return NULL; /* animation out of range */
  }

  /*
   * Find out which page file our frame lives in
   */
  for (page = 0; frame > UW_page_map [anim * MAX_ANIM_PAGES + page]
	 && page < MAX_ANIM_PAGES; ++page);
  if ((page >= MAX_ANIM_PAGES) ||
      (UW_page_map [anim * MAX_ANIM_PAGES + page] == 255))
  {
    return NULL; /* frame out of range */
  }

  sprintf (animname, anim_page_format, UW_crit_dir, anim, page);
  animfile = fopen (animname, "rb");
  if (animfile == NULL)
  {
    return NULL;
  }

  /*
   * Now we have a page file ... get the auxiliary palette and unpack the
   *  frame
   */
  if (Underworld2)
  {
    /*
     * Always 4 auxiliary palettes
     */
    npals = 4;
  }
  else
  {
    /*
     * For an Underworld file we must skip the initial tables
     */
    fgetc (animfile);
    skip = fgetc (animfile);
    fseek (animfile, skip, SEEK_CUR);

    skip = fgetc (animfile);
    fseek (animfile, 8 * skip, SEEK_CUR);

    npals = fgetc (animfile);
  }

  /*
   * The animfile pointer should now be at the start of the auxiliary palettes
   *  with npals the number present
   */
  if (variant >= npals) variant = 0;
  for (i = 0; i < npals; ++i)
  {
    if (variant == i)
    {
      fread (auxpal, 1, 32, animfile);
    }
    else
    {
      fseek (animfile, 32, SEEK_CUR);
    }
  }

  /*
   * Having (hopefully) grabbed the correct auxpal in passing, we should now be
   *  positioned at the frame directory
   */
  if (Underworld2)
  {
    /*
     * Underworld II anim page files always have a complete frame directory
     *  with null entries for frames not in that particular file
     */
    fseek (animfile, 2 * frame, SEEK_CUR);
    READ_LE_SHORT (animfile, offset);
    if (offset == 0)
    {
      fclose (animfile);
      return NULL;         /* frame absent, don't know why */
    }
  }
  else
  {
    /*
     * Underworld I, as usual, is awkward
     * Get the number of anim pages, subtract from the last frame
     */
    skip = UW_page_map [anim * MAX_ANIM_PAGES + page] - fgetc (animfile) + 1;
    skip = frame - skip;

    fgetc (animfile); /* don't know what this is ... throw it away */
    fseek (animfile, 2 * skip, SEEK_CUR);
    READ_LE_SHORT (animfile, offset);
  }

  /*
   * Right, it seems that the frame exists (phew!) Populate a bitmap header
   * Fortuitously, offsets are absolute within the file
   */
  fseek (animfile, offset, SEEK_SET);
  bmp = malloc (sizeof *bmp);
  if (bmp == NULL)
  {
    return NULL; /* couldn't allocate bitmap header */
  }
  bmp->chunk    = FAKE_CRIT_FRAME_CHUNK + anim;
  bmp->subchunk = frame;
  bmp->width    = fgetc (animfile);
  bmp->height   = fgetc (animfile);
  bmp->x1       = fgetc (animfile);   /* hotspot x */
  bmp->y1       = fgetc (animfile);   /* hotspot y */
  bmp->type     = fgetc (animfile);   /* don't know, might as well go here */

  if (bmp->type == 6) wsize = 5;
  else wsize = 4;

  bmp->bits = malloc (2 * bmp->width * bmp->height);
  if (bmp->bits == NULL)
  {
    free (bmp);
    return NULL; /* couldn't allocate bitmap bits */
  }
  READ_LE_SHORT (animfile, skip);
  skip = (skip * wsize + 7) / 8;
  fread (pack, 1, skip, animfile);

  fclose (animfile); /* it has served its usefulness */

  /*
   * Now, finally, we get to unpack the bitmap itself
   */
  Bmp_unpack_uw_rle (bmp, pack, wsize, auxpal);
  return bmp;

}
