/*
 * dark.c -
 *  Dark suckers!
 */

/*
 * File:    src/dark.c
 * Purpose: Dynamic lighting (darkening)
 * Prefix:  Dark_
 * Note:    
 * Music:   Def Leppard, "Slang"
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "dark.h"
#include "gamemem.h"
#include "main.h"    /* Cant_happen()     */
#include "map.h"
#include "object.h"
#include "option.h"  /* Option_underworld */
#include "physics.h"
#include "render.h"
#include "shade.h"
#include "surface.h"
#include "vec.h"




/*
 * Local function prototypes
 */
static void add_to_tile (dark_sucker *ds, map_tile *tile);
static dark_sucker *add_avatar_sucker (float intensity);
static void search_tile (dark_sucker *ds, map_tile *tile, const vec *start);


/*
 * Global variables
 */

/*
 * The master dark sucker list
 */
#define MAX_DARK_SUCKER 32
static dark_sucker *Dark_sucker_list;
static dark_sucker *Dark_sucker_head;  /* head of the linked list of suckers */
static dark_sucker *Dark_sucker_free;  /* head of the free list              */

#define MAX_SUCKER_XREF 256            /* Probably too small                 */
static sucker_xref *Dark_sucker_xref;  /* Dark sucker<->tile xref list       */
static int Sucker_xref_free;           /* head of freelist                   */

static int Avatar_dark_index;          /* dummy sucker xref index for avatar */



#define MIN_DARK 0.1 /* brightest allowed unsaturated value */


void Dark_init (void)
{

  /*
   * Allocate space for the dark sucker list
   */
  int i;
  Dark_sucker_list = Alloc_init (MAX_DARK_SUCKER * sizeof *Dark_sucker_list,
				 "Dark_init", "dark suckers");

  /*
   * Initialise the free list
   */
  for (i = 0; i < MAX_DARK_SUCKER - 1; ++i)
  {
    Dark_sucker_list [i].flags = DS_FREE;
    Dark_sucker_list [i].next  = &(Dark_sucker_list [i + 1]);
  }
  Dark_sucker_list [MAX_DARK_SUCKER - 1].flags = DS_FREE;
  Dark_sucker_list [MAX_DARK_SUCKER - 1].next  = NULL;
  Dark_sucker_free = Dark_sucker_list;

  Dark_sucker_head = NULL;

  /*
   * Dark sucker xref list
   */
  Dark_sucker_xref = Alloc_init (MAX_SUCKER_XREF * sizeof *Dark_sucker_xref,
				 "Dark_init", "sucker xref table");

  /*
   * And its freelist
   */
  Sucker_xref_free = 0;
  for (i = 1; i < MAX_SUCKER_XREF; ++i)
  {
    Dark_sucker_xref [i - 1].sucker_chain = i;
  }
  Dark_sucker_xref [MAX_SUCKER_XREF - 1].sucker_chain = -1;

  Avatar_dark_index = -1; /* start off with no avatar lights */

  /*
   * Avatar lantern for Underworld
   */
  if (Option_underworld)
  {
    dark_sucker *avatar_ds = add_avatar_sucker (UW_LANTERN_INTENSITY);
    avatar_ds->rel_pos.x =  10.; /* simulating a shoulder inventory slot */
    avatar_ds->rel_pos.y = 140.; /* relative to avatar position          */
    avatar_ds->rel_pos.z =   0.; /* hope you don't have long hair!       */
  }

}



void Dark_exit (void)
{
  free (Dark_sucker_list);
  free (Dark_sucker_xref);
}



/*
 * Grab a new dark sucker off the free list
 */
dark_sucker *Dark_new_sucker (void)
{
  dark_sucker *ds = Dark_sucker_free;
  if (ds == NULL)
  {
    Cant_happen ("Dark_new_sucker(): no free dark suckers");
  }

  /*
   * Paranoia
   */
  if (!(ds->flags & DS_FREE))
  {
    Cant_happen ("Dark_new_sucker(): non-free sucker on free list");
  }

  Dark_sucker_free = ds->next;
  ds->next = Dark_sucker_head;
  Dark_sucker_head = ds;
  ds->flags = 0;

  return ds;

}



void Dark_free_sucker (dark_sucker *ds)
{

  if (ds->flags & DS_FREE)
  {
    Cant_happen ("Dark_free_sucker(): trying to free already-free sucker");
  }

  ds->flags = DS_FREE;
  Dark_sucker_head = ds->next;
  ds->next = Dark_sucker_free;
  Dark_sucker_free = ds;

}



/*
 * Add a static dark sucker at a (stationary) map object
 */
dark_sucker *Dark_add_static_sucker (map_object *obj, float intensity,
				     float max_radius)
{

  /*
   * First, catch your dark sucker ...
   */
  dark_sucker *ds = Dark_new_sucker ();

  /*
   * Fill in the required fields
   */
  ds->obj        = obj;
  ds->position   = obj->pos;      /* static, so not updated each frame */
  ds->intensity  = intensity;
  ds->max_radius = max_radius;
  ds->flags      = 0;             /* used, static, at obj position     */
  ds->tile_index = -1;

  /*
   * Recurse through the tile map, starting at the tile the object is (mostly)
   *  in
   * This adds the dark sucker to the xref list of any tile encountered by it
   */
  search_tile (ds, Map_get_tile_v (&(obj->pos)), &(obj->pos));

  /*
   * Return the new dark sucker in case the caller wants to play with it
   */
  return ds;

}




/*
 * Called at the start of a frame
 * Clear out all dynamic dark suckers and get their positions
 */
void Dark_frame (void)
{

  dark_sucker *ds = Dark_sucker_head;
  while (ds != NULL)
  {
    if (ds->flags & DS_DYNAMIC)
    {
      /*
       * Remove it from all tile xref lists if it is not avatar-based
       */
      if (ds->flags & DS_RELATIVE)
      {
	Phy_get_rel_pos (ds->obj, &(ds->rel_pos), &(ds->position));
      }
      else
      {
	/*
	 * It's dynamic, but doesn't have a rel_pos: use the object position
	 *  directly
	 */
	ds->position = ds->obj->pos;
      }
    }
    ds = ds->next;
  }

}



/*
 * Apply any relevant dark suckers to a Gouraud-shaded surface
 */
void Dark_do_gouraud (const map_tile *tile, surface *s)
{

  int xref;
  int i;

  /*
   * Avatar suckers are always applied
   */
  xref = Avatar_dark_index;
  while (xref >= 0)
  {
    Dark_apply_dynamic_sucker (Dark_sucker_xref [xref].ds, s);
    xref = Dark_sucker_xref [xref].tile_chain;
  }

  /*
   * And this tile's own xref list
   */
  xref = tile->dark_index;
  while (xref >= 0)
  {
    Dark_apply_dynamic_sucker (Dark_sucker_xref [xref].ds, s);
    xref = Dark_sucker_xref [xref].tile_chain;
  }

  /*
   * Cleanup of dark values
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    assert (s->dark [i] < DARK_SHADES); /* can't get darker! */
    if (s->dark [i] < MIN_DARK) s->dark [i] = MIN_DARK;
  }

}



float Dark_do_plane (const map_tile *tile, const vec *v, const vec *normal)
{

  float light = 0;
  int xref;

  /*
   * Cut-and-paste again
   */
  xref = Avatar_dark_index;
  while (xref >= 0)
  {
    light += Dark_get_sucker_plane (Dark_sucker_xref [xref].ds, v, normal);
    xref = Dark_sucker_xref [xref].tile_chain;
  }

  xref = tile->dark_index;
  while (xref >= 0)
  {
    light += Dark_get_sucker_plane (Dark_sucker_xref [xref].ds, v, normal);
    xref = Dark_sucker_xref [xref].tile_chain;
  }

  return light;

}






float Dark_do_point (const map_tile *tile, const vec *v)
{

  float light = 0;
  int xref;

  /*
   * Cut-and-paste again
   */
  xref = Avatar_dark_index;
  while (xref >= 0)
  {
    light += Dark_get_sucker_point (Dark_sucker_xref [xref].ds, v);
    xref = Dark_sucker_xref [xref].tile_chain;
  }

  xref = tile->dark_index;
  while (xref >= 0)
  {
    light += Dark_get_sucker_point (Dark_sucker_xref [xref].ds, v);
    xref = Dark_sucker_xref [xref].tile_chain;
  }

  return light;

}






void Dark_apply_dynamic_sucker (const dark_sucker *ds, surface *s)
{

  int v;    /* vertex index */

  /*
   * Applying a dark sucker to a Gouraud surface is very simple; for each
   *  vertex we subtract darkness according to the distance from the sucker
   *  and the dot product with the normal vector
   */
  for (v = 0; v < s->num_vertices; ++v)
  {

    float intensity = Dark_get_sucker_plane (ds, &(s->vertex [v]),
					     &(s->normal));
    s->dark [v] -= intensity * DARK_SHADES;
/*      if (s->dark [v] < 0) */
/*      { */
/*        s->dark [v] = FULL_BRIGHT; */
/*      } */

  }

}



/*
 * Get a sucker intensity measured at a point for a plane at a given angle
 */
float Dark_get_sucker_plane (const dark_sucker *ds, const Vector *v,
			     const Vector *normal)
{

  Vector r;
  double dp; /* dot product */
  double distance;

  /*
   * position was set up for us at the start of the frame
   */
  vec_subtract (r, ds->position, *v);

  dp = vec_dot (r, *normal);
  if (dp < 0.) return 0.;      /* light is behind surface */

  /*
   * jim - I don't like having these fabs' lurking about, but they're
   *  probably better than an extra sqrt()
   */
/*    r.x = fabs (r.x); r.y = fabs (r.y); r.z = fabs (r.z); */
  distance = vec_magnitude2 (r);
  distance *= sqrt (distance);

  /*
   * Dot product divided by square of distance magnitude actually gives
   *  inverse linear falloff ... if we then divide by largest distance
   *  component we get a reasonable approximation to the true inverse-square
   *  falloff with distance
   */
/*    if (r.x > r.y) */
/*      if (r.x > r.z) distance *= r.x; else distance *= r.z; */
/*    else */
/*      if (r.y > r.z) distance *= r.y; else distance *= r.z; */

  if (distance < A_VERY_SMALL_THING) distance = A_VERY_SMALL_THING;
  return ds->intensity * dp / distance;

}



float Dark_get_sucker_point (const dark_sucker *ds, const Vector *v)
{

  Vector r;  /* distance sucker to point */
  double distance;

  vec_subtract (r, *v, ds->position);
  distance = vec_magnitude2 (r);
  if (distance < 1.)
  {
    /*
     * Cap it to avoid divides-by-zero
     */
    return ds->intensity;
  }
  return (float) (ds->intensity / distance);

}




/*
 * Add a dark sucker to a tile
 * Note: doesn't check whether the sucker is already cross-referenced with that
 *  tile!
 */
void add_to_tile (dark_sucker *ds, map_tile *tile)
{

  /*
   * Find a free dark sucker xref entry to use
   */
  int xref = Sucker_xref_free;
  if (xref < 0) Cant_happen ("dark.c:add_to_tile(): no free xrefs");

  /*
   * Pull it off the free list
   */
  Sucker_xref_free = Dark_sucker_xref [xref].sucker_chain;

  /*
   * Set its tile and sucker pointers
   */
  Dark_sucker_xref [xref].ds           = ds;
  Dark_sucker_xref [xref].tile         = tile;

  /*
   * And link it in to the two xref chains for tile and sucker
   */
  Dark_sucker_xref [xref].sucker_chain = ds->tile_index;
  Dark_sucker_xref [xref].tile_chain   = tile->dark_index;

  ds->tile_index   = xref;
  tile->dark_index = xref;

}



/*
 * Add an avatar dark sucker, and make an xref for it in the avatar xref list
 * Sets object ref, flags and intensity, but not position
 */
dark_sucker *add_avatar_sucker (float intensity)
{

  dark_sucker *ds = Dark_new_sucker ();  /* new dark sucker */
  int xref        = Sucker_xref_free;    /* new sucker xref */

  if (xref < 0) Cant_happen ("dark.c:add_avatar_sucker(): no free xrefs");
  Sucker_xref_free = Dark_sucker_xref [xref].sucker_chain;

  /*
   * Set up the dark sucker
   */
  ds->obj        = AVATAR_MAP_OBJECT;
  ds->intensity  = intensity;
  /* max_radius isn't used for an avatar dark sucker */
  ds->flags      = DS_AVATAR | DS_DYNAMIC | DS_RELATIVE;
  ds->tile_index = xref;

  /*
   * And link it into the avatar xref list
   */
  Dark_sucker_xref [xref].ds           = ds;
  Dark_sucker_xref [xref].tile         = NULL; /* not used */
  Dark_sucker_xref [xref].sucker_chain = -1;   /* not used */
  Dark_sucker_xref [xref].tile_chain   = Avatar_dark_index;
  Avatar_dark_index = xref;

  return ds;

}




/*
 * Recursive tile search for a dark sucker
 * This is essentially the same algorithm as render.c:recurse_pvt() (tile
 *  search first path) except without the view frustrum clipping.
 * When it encounters a tile it adds the dark sucker to its xref list using
 *  add_to_tile()
 */
void search_tile (dark_sucker *ds, map_tile *tile, const vec *start)
{

  int sno = tile->first_surface;

  /*
   * Add the dark sucker to the tile
   */
  add_to_tile (ds, tile);

  /*
   * For this search, a portal (and its back tile) is considered if it passes
   *  the backface test and is closer than the sucker's max radius
   */
  while (sno != -1)
  {

    extern map_surface *Surface;        /* the main surface list from map.c */
    map_surface *ms = &(Surface [sno]);
    Vector tv;
    double dist;

    sno = ms->tile_next;
    if (!(ms->flags & MSF_PORTAL)) continue; /* only interested in portals */

    /*
     * Distance test and back-face cull
     */
    vec_subtract (tv, ms->anchor, *start);
    dist = vec_dot (tv, ms->normal);
    if ((dist < A_VERY_SMALL_THING) && (dist > -ds->max_radius))
    {

      /*
       * Got one! Recurse into it
       */
      search_tile (ds, ms->behind, start);

    }

  }

}




