/*
 * door.c -
 *  Storage and retrieval of doors and gratings
 */

/*
 * File:    src/door.c
 * Purpose: Code specific to doors class 0A
 * Prefix:  Door_
 * Note:
 * Music:   Ozzy Osbourne, "The Ultimate Sin"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "graph2d.h"
#include "gamemem.h"
#include "bitmap.h"
#include "chunks.h"
#include "door.h"
#include "main.h"     /* Cant_happen() */
#include "map.h"
#include "model.h"
#include "object.h"
#include "resource.h"
#include "surface.h"
#include "texture.h"



/*
 * Debug
 */
#define DOOR_MESSAGES 0


#define DOOR_ARCHIVE_SIZE 14


/*
 * Local typedefs and structs
 */



static door *Door_list;
int Num_doors;


/*
 * Local function prototypes
 */
#if 0
static int check_index (map_object *mo);
#endif
static int fetch_texture_door3d (map_object *obj);


/*
 * For hacking purposes ...
 */
extern map_object *Map_object_list;



/*
 * Load the door table for a level. Called from Obj_init_level()
 */
void Door_init_level (int level)
{

  block_handle bh;
  size_t size;
  int i;
  unsigned char *dp;

  /*
   * Load up the fixture list
   */
  bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_DOOR, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Door_init_level(): couldn't load door data");
  }
  if (size % DOOR_ARCHIVE_SIZE != 0)
  {
    Cant_happen ("Door_init_level(): not exact number of doors");
  }
  Num_doors = size / DOOR_ARCHIVE_SIZE;
  Door_list = malloc (Num_doors * sizeof *Door_list);
  if (Door_list == NULL)
  {
    Cant_happen ("Door_init_level(): couldn't allocate space for door list");
  }

# if DOOR_MESSAGES
  printf ("Door_init_level(): %d bytes allocated for door list\n",
	  Num_doors * sizeof *Door_list);
# endif

  dp = (unsigned char *) *bh;
  for (i = 0; i < Num_doors; ++i)
  {
    Door_list [i].map_index   = *dp + 256 * dp [1]; dp += 2;
    Door_list [i].prev        = *dp + 256 * dp [1]; dp += 2;
    Door_list [i].next        = *dp + 256 * dp [1]; dp += 2;
    Door_list [i].trigger     = *dp + 256 * dp [1]; dp += 2;
    Door_list [i].message     = *dp++;
    Door_list [i].colour      = *dp++;
    Door_list [i].access      = *dp++;
    Door_list [i].stuff       = *dp++;
    Door_list [i].door_xref   = *dp + 256 * dp [1]; dp += 2;
  }

  Res_free_chunk_by_id (4000 + 100*level + MAP_CHUNK_DOOR);

}

void Door_deinit(void)
{
	free(Door_list);
	Door_list = NULL;
}


void Door_init_object (map_object *mo)
{

  /*
   * Set up the bitmap fetch hook
   */
  mo->fetch_texture_hook = Door_get_texture;

  switch (mo->subclass)
  {
  case DOOR_SUBCLASS_3D:
/*      mo->draw_type = OBJDRAW_SPECIAL; */
    mo->fetch_texture_hook = fetch_texture_door3d;
    break;

  case DOOR_SUBCLASS_FORCE:
    /*
     * Force doors don't have permanent map surfaces
     */
    if (mo->type < 4) break;

    /* else FALL THROUGH! */

  default:
    /*
     * A door is sized to tile width, and is square
     */
    Obj_make_decal_surface (mo, TILE_GRID_SIZE, TILE_GRID_SIZE,
			    SF_TEXTURED | SF_MASKED | SF_SHADED | MSF_2SIDED);
    break;

  }

}




/*
 * Get the bitmap associated with a door
 */
extern int Object_index [OBJ_MAXCLASS] [OBJ_MAXSUBCLASS];
int Door_get_texture (map_object *mo)
{

/*    int di = check_index (mo); */

# if DOOR_MESSAGES
  printf ("Door_get_texture(): finding texture for door %d\n", di);
# endif

  return Tex_get_cache_id (CHUNK_TEXTURE_DOOR_BASE + mo->object_id -
			   Object_index [10] [0], mo->state);

}


/*
 * Render a 3D door which opens outwards. Underworld doors are of this type
 */
#define MODEL_DOORFRAME  1
#define MODEL_DOOR      18

void Door_render_3d (map_object *obj)
{
  /*
   * The door frame
   */
  Mod_render_id (obj, MODEL_DOORFRAME, NULL);

  /*
   * The door proper
   */
/*    Mod_render (obj, MODEL_DOOR, box, bitmap); */
}


#if 0
int check_index (map_object *mo)
{
  int di = mo->class_index;
  if (mo->class != OBJCLASS_DOOR)
  {
    Cant_happen ("non-door object");
  }
  if ((di < 1) || (di >= Num_doors))
  {
    Cant_happen ("door index out of range");
  }
  return di;
}
#endif


door		*Door_get_door(int index)
{
	door *pD = NULL;

	if(index >= 0 && index < Num_doors)
		pD = &Door_list[index];

	return pD;
}



int fetch_texture_door3d (map_object *obj)
{

  /*
   * Just use the wall texture for the tile
   */
  map_tile *tile = Map_get_tile_v (&(obj->pos));
  return GET_WALL_TEXTURE (*tile);

}


