/*
 * event.h -
 *  defines and prototypes for the event system
 */

/*
 * File:    src/event.h
 * Purpose: events
 * Prefix:	Evt_
 * Note:    
 * Music:   Global Youth 'Connected Views'
 */

/**********************************************************************/
/* List of Events currently thrown                                    */
/*                                                                    */
/* CLASS          EVENT         DATA, DESC                            */
/*                                                                    */
/* MFD            TargetNew     NULL                                  */
/*                BigmapFade    (void *)int: >0: fadeIn, <0: fadeOut  */
/*                BigmapZoom    &byte: zoomlevel                      */
/*                CompassShow   &byte: hw level [1..3], 0 = off       */
/*                HWInstall     NULL                                  */
/*                HWActivate    (void *)bool: activated/deactivated   */
/*                SideShow      (void *)uint32: mfd id                */
/*                CenterShow    (void *)uint32: mfd id                */
/*                SideFired     (void *)uint32: mfd id                */
/*                CenterFired   (void *)uint32: mfd id                */
/*                SensaroundShow &byte: sa level [1..3], 0 = off      */
/*                VidmailShow   &byte: id                             */
/*                PlayEMail     (void *)uint32: rel. chunk id of text */
/*                PlayLog       (void *)uint32: rel. chunk id of text */
/*                PlayVMail     (void *)uint32: rel. chunk id of text */
/*                                                                    */
/* CONTROL        ModeActivate  (void *)ModeId                        */
/*                ModeDeactivate (void *)ModeId                       */
/*                                                                    */
/**********************************************************************/


#ifndef EVENT_SYSTEM_H
#define	EVENT_SYSTEM_H

/* maximum size of names - excluding \0																*/
#define	C_MAX_EVT_CLASSNAME		30
#define	C_MAX_EVT_EVENTNAME		30

/* the internal events																								*/
#define	C_EVT_INTERN_REMOVE	1			 /* listener gets removed						*/

#ifdef __cplusplus
extern "C"
{
#endif

typedef void *		HEVENTCLASS;
typedef void *		HEVENT;

/* this function listens to events.																		*/
/* if hClass and hEvent are both NULL, pData points to an internal		*/
/* structure with an uint32 as its first member which determines what	*/
/* structure it is.																										*/
typedef void (*EVENT_LISTENER)(HEVENTCLASS hClass, 
															 HEVENT hEvent, void *pData);

/* receiving this structure means that the function will no longer		*/
/* be called for the given event in the given class										*/
typedef struct tag_EventRemover
{	/* type: C_EVT_INTERN_REMOVE	*/
	uint32					erEventType;
	HEVENTCLASS			erClass;
	HEVENT					erEvent;
} EventRemover;


/* init and deinit - straight forward																	*/
void				Evt_init(void);
void				Evt_deinit(void);

/* get handle for event class - if not existing it will be registered	*/
/* automatically																											*/
HEVENTCLASS	Evt_get_class(const char *strEventClass);
/* remove an event class.																							*/
bool				Evt_remove_class(HEVENTCLASS hEventClass);

/* get handle for event - if not existing it will be registered				*/
/* automatically																											*/
HEVENT			Evt_get_event(HEVENTCLASS hEventClass, const char *strEvent);
bool				Evt_remove_event(HEVENT hEvent);

/* register and deregister a function for an event or class						*/
/* if hEvent == NULL, evtListener will be signaled for any event in		*/
/* the given class. (this happens before the single listener get			*/
/* signaled)																													*/
bool				Evt_register(HEVENTCLASS hEventClass, HEVENT hEvent,
												 EVENT_LISTENER evtListener);
void				Evt_unregister(HEVENTCLASS hEventClass, HEVENT hEvent,
													 EVENT_LISTENER evtListener);

/* pulse an event: will signal all listeners and return afterwards		*/
/* returns amount of listeners signaled																*/
int					Evt_pulse_event(HEVENT hEvent, void *pData);
/* mark an event for execution: listeners will be signaled later on;	*/
/* if bAutoDel is set, pData will be freed afterwards.								*/
void				Evt_fire_event(HEVENT hEvent, void *pData, bool bAutoDel);
/* signal all listeners of fired events																*/
void				Evt_work(void);

/* helper																															*/
const char *Evt_get_event_name(HEVENT hEvent);
const char *Evt_get_class_name(HEVENTCLASS hEventClass);
HEVENTCLASS Evt_get_class_from_event(HEVENT hEvent);

#ifdef __cplusplus
}
#endif

#endif /* EVENT_SYSTEM_H */
