/*
 * export.c -
 *  Extract resources as standard format files
 */

/*
 * File:    src/export.c
 * Purpose: Exporting of game resources
 * Prefix:  Exp_
 * Note:
 * Music:
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdlib.h>
#include <string.h>

#include "types.h"
#if HAVE_PNG
#include "png.h"
#endif /* HAVE_PNG */

#include "bitmap.h"
#include "export.h"
#include "main.h"
#include "option.h"
#include "palette.h"
#include "shade.h"
#include "resource.h"




/*
 * Write out all the textures in the game
 */
void Exp_write_all_textures (void)
{
  int i, j, ntexs;

  if (Option_underworld)
  {
    for (i = 0; i < 33; ++i)
    {
      ntexs = Bmp_get_num_bmps (i);
      if (i == 24) continue;
      for (j = 0; j < ntexs; ++j)
      {
	Exp_write_texture (i, j);
      }
    }
  }
	else
	{
		for (i = 0; i < 0x10000; i++)
		{
      ntexs = Bmp_get_num_bmps (i);
			printf("%04d in %04X\n", ntexs, i);
			if (i != 0x24 && (i < 0x3C || i > 0x50) && i != 0x867)
			{
				for (j = 0; j < ntexs; ++j)
				{
					Exp_write_texture (i, j);
				}
			}
		}
	}
}


void Exp_write_texture (int chunk, int subchunk)
{

  char texname [256];
  ss_bitmap_header *bmp = Bmp_load_bitmap (chunk, subchunk);
  if (bmp == NULL)
  {
    printf ("Exp_write_texture(): couldn't load texture %d:%d\n",
	    chunk, subchunk);
    return;
  }

  sprintf (texname, "tex%04X-%04X", chunk, subchunk);
# if HAVE_PNG
  Exp_write_png8 (bmp->width, bmp->height, bmp->bits, texname);
# else
	Exp_write_tga8(bmp->width, bmp->height, bmp->bits, texname, RGB_trans,
								 "TriOptimum Extraction");
# endif

  Bmp_free_bitmap (bmp);

}

void Exp_write_tga8(int width, int height, byte *bits, const char *name,
										const uint32 *pPal, const char *strDesc)
{
	FILE *f = NULL;             
	char strFile[256];
	byte pal[256][3];
	byte tga_header[18];												/*tga header					*/
	int i;

	sprintf(strFile, "%s.tga", name);
	if((f = fopen(strFile, "wb")) != NULL)
	{
		for(i = 0; i < 256; i++)									/*make up palette			*/
		{
			pal[i][0] = get_blue(pPal[i]);
			pal[i][1] = get_green(pPal[i]);
			pal[i][2] = get_red(pPal[i]);
		}
		memset(tga_header, 0, 18);
		tga_header[0] = strlen(strDesc) + 1;			/*len of desc					*/
		tga_header[1] = 1;												/*bmp with color map	*/
		tga_header[2] = 1;												/*mapped image				*/
		tga_header[5] = lo(256);									/*256 colors					*/
		tga_header[6] = hi(256);
		tga_header[7] = 24;												/*24bit palette codes	*/
		tga_header[12] = lo((uint16)(width));			/*dimensions					*/
		tga_header[13] = hi((uint16)(width));
		tga_header[14] = lo((uint16)(height));
		tga_header[15] = hi((uint16)(height));
		tga_header[16] = 8;												/*8bits per index			*/
		tga_header[17] = 0x20;										/*starting top left		*/
		fwrite(tga_header, 18, 1, f);							/*write header				*/
		fwrite(strDesc, 1, tga_header[0], f);			/*write desc					*/
		fwrite(pal, 1, 256 * 3, f);								/*write palette				*/
		fwrite(bits, 1, width * height, f);
		fclose(f);																/*byebye							*/
	}
}

void Exp_write_tga32(int width, int height, uint32 *bits, const char *name,
										 const char *strDesc)
{
	FILE *f = NULL;             
	char strFile[256];
	byte tga_header[18];												/*tga header					*/

	sprintf(strFile, "%s.tga", name);
	if((f = fopen(strFile, "wb")) != NULL)
	{
		memset(tga_header, 0, 18);
		tga_header[0] = strlen(strDesc) + 1;			/*len of desc					*/
		tga_header[1] = 0;												/*no with color map		*/
		tga_header[2] = 2;												/*rgb image						*/
		tga_header[12] = lo((uint16)(width));			/*dimensions					*/
		tga_header[13] = hi((uint16)(width));
		tga_header[14] = lo((uint16)(height));
		tga_header[15] = hi((uint16)(height));
		tga_header[16] = 32;											/*32bits per index		*/
		tga_header[17] = 0x20;										/*starting top left		*/
		fwrite(tga_header, 18, 1, f);							/*write header				*/
		fwrite(strDesc, 1, tga_header[0], f);			/*write desc					*/
		fwrite(bits, 1, width * height * 4, f);
		fclose(f);																/*byebye							*/
	}
}

#if HAVE_PNG
void Exp_write_png8 (int width, int height, byte *bits, const char *name)
{

  FILE *pngfile;
  png_structp png_ptr;
  png_infop info_ptr;
  png_colorp pngpalette;
  png_bytep *rows;
  int i;
  char texname [256];

  /*
   * This is pretty much straight out of example.c in the libpng docs
   */
  sprintf (texname, "%s.png", name);
  pngfile = fopen (texname, "wb");
  if (pngfile == NULL)
  {
    Cant_happen ("Couldn't open png file");
  }

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL)
  {
    fclose (pngfile);
    Cant_happen ("png_create_write_struct() failed");
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (info_ptr == NULL)
  {
    png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
    fclose (pngfile);
    Cant_happen ("png_create_info_struct() failed");
  }

  if (setjmp (png_ptr->jmpbuf))
  {
    png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
    fclose (pngfile);
    Cant_happen ("setjmp() failed");
  }

  png_init_io (png_ptr, pngfile);
# if MONOCHROME
  png_set_IHDR (png_ptr, info_ptr, width * 8, height, 1,
  		PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
  		PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
# else
  png_set_IHDR (png_ptr, info_ptr, width, height, 8,
  		PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
  		PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
# endif /* MONOCHROME */

  /*
   * The palette
   */
  pngpalette = (png_colorp) png_malloc (png_ptr, 256 * sizeof (png_color));
# if MONOCHROME
  for (i = 0; i < 2; ++i)
  {
    pngpalette [i].red   = 255 * i;
    pngpalette [i].green = 255 * i;
    pngpalette [i].blue  = 255 * i;
  }
# else
  for (i = 0; i < 256; ++i)
  {
    pngpalette [i].red   = Palette [i] [0];
    pngpalette [i].green = Palette [i] [1];
    pngpalette [i].blue  = Palette [i] [2];
  }
# endif
  png_set_PLTE (png_ptr, info_ptr, pngpalette, 256);

  /*
   * Write the stuff
   */
  png_write_info (png_ptr, info_ptr);

  rows = malloc (height * sizeof *rows);
  for (i = 0; i < height; ++i)
  {
    rows [i] = bits + i * width;
  }
  png_write_image (png_ptr, rows);
  free (rows);

  png_write_end (png_ptr, info_ptr);
  free (pngpalette);
  png_destroy_info_struct (png_ptr, &info_ptr);
  png_destroy_write_struct (&png_ptr, (png_infopp) NULL);

  /*
   * Like a good little texture extractor
   */
  fclose (pngfile);

}



/*
 * Oops. Heap big cut and paste
 */
void Exp_write_png32 (int width, int height, uint32 *bits, const char *name)
{

  FILE *pngfile;
  png_structp png_ptr;
  png_infop info_ptr;
  png_bytep *rows;
  int i;
  char texname [256];

  /*
   * This is pretty much straight out of example.c in the libpng docs
   */
  sprintf (texname, "%s.png", name);
  pngfile = fopen (texname, "wb");
  if (pngfile == NULL)
  {
    Cant_happen ("Couldn't open png file");
  }

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL)
  {
    fclose (pngfile);
    Cant_happen ("png_create_write_struct() failed");
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (info_ptr == NULL)
  {
    png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
    fclose (pngfile);
    Cant_happen ("png_create_info_struct() failed");
  }

  if (setjmp (png_ptr->jmpbuf))
  {
    png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
    fclose (pngfile);
    Cant_happen ("setjmp() failed");
  }

  png_init_io (png_ptr, pngfile);
  png_set_IHDR (png_ptr, info_ptr, width, height, 8,
  		PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
  		PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

  /*
   * Write the stuff
   */
  png_write_info (png_ptr, info_ptr);

  /*
   * No palette for true-colour!
   * But get rid of the 4th byte
   */
  png_set_filler (png_ptr, 0, PNG_FILLER_AFTER);
  png_set_bgr (png_ptr);

  rows = malloc (height * sizeof *rows);
  for (i = 0; i < height; ++i)
  {
    rows [i] = (png_bytep) (bits + i * width);
  }
  png_write_image (png_ptr, rows);
  free (rows);

  png_write_end (png_ptr, info_ptr);

  png_destroy_info_struct (png_ptr, &info_ptr);
  png_destroy_write_struct (&png_ptr, (png_infopp) NULL);

  /*
   * Like a good little texture extractor
   */
  fclose (pngfile);

}
#endif /* HAVE_PNG */
