/*
 * fixture.c -
 *  Storage and retrieval of fixture data
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "types.h"
#include "chunks.h"
#include "graph2d.h"
#include "gamemem.h"
#include "anim.h"
#include "bitmap.h"
#include "fixture.h"
#include "font.h"
#include "main.h"     /* Cant_happen() */
#include "map.h"
#include "object.h"
#include "resource.h"
#include "screen.h"
#include "surface.h"
#include "texture.h"



/*
 * Debug
 */
#define FIXTURE_MESSAGES 0


#define FIXTURE_ARCHIVE_SIZE 16


#define WORDS_CHUNK 2152

#define BITMAP_CHUNK_ICON     78
#define BITMAP_CHUNK_GRAFFITI 79
#define BITMAP_CHUNK_REPULSOR 80

#define BASE_CHUNK_SCREEN     321

/*
 * Local typedefs and structs
 */



fixture *Fixture_list;
int Num_fixtures;


/*
 * Local function prototypes
 */
static ss_bitmap_header *render_words (int fi);

int fetch_texture_icon     (map_object *obj);
int fetch_texture_graffiti (map_object *obj);
int fetch_texture_words    (map_object *obj);
int fetch_texture_repulsor (map_object *obj);
int fetch_texture_model    (map_object *obj);




/*
 * Load the fixture table for a level. Called from Obj_init_level()
 */
void Fix_init_level (int level)
{

  block_handle bh;
  size_t size;
  int i;
  unsigned char *fp;

  /*
   * Load up the fixture list
   */
  bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_FIXTURE, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Fix_init_level(): couldn't load fixture data");
  }
  if (size % FIXTURE_ARCHIVE_SIZE != 0)
  {
    Cant_happen ("Fix_init_level(): not exact number of fixtures");
  }
  Num_fixtures = size / FIXTURE_ARCHIVE_SIZE;
  Fixture_list = malloc (Num_fixtures * sizeof *Fixture_list);
  if (Fixture_list == NULL)
  {
    Cant_happen ("Fix_init_level(): couldn't allocate space for fixture list");
  }

# if FIXTURE_MESSAGES
  printf ("Fix_init_level(): %d bytes allocated for fixture list\n",
	  Num_fixtures * sizeof *Fixture_list);
# endif

  fp = (unsigned char *) *bh;
  for (i = 0; i < Num_fixtures; ++i)
  {
    Fixture_list [i].map_index   = *fp + 256 * fp [1]; fp += 2;
    Fixture_list [i].prev        = *fp + 256 * fp [1]; fp += 2;
    Fixture_list [i].next        = *fp + 256 * fp [1]; fp += 2;
    Fixture_list [i].u.text      = *fp + 256 * fp [1]; fp += 2;
    Fixture_list [i].flags       = *fp + 256 * fp [1]; fp += 2;
    Fixture_list [i].texture     = *fp + 256 * fp [1]; fp += 2;
    Fixture_list [i].start_frame = *fp + 256 * fp [1]; fp += 2;
    Fixture_list [i].aux         = *fp + 256 * fp [1]; fp += 2;

    Fixture_list [i].bmp_cache   = NULL;
    Fixture_list [i].anim        = NULL;
  }

  Res_free_chunk_by_id (4000 + 100*level + MAP_CHUNK_FIXTURE);

}

void Fix_deinit(void)
{
	free(Fixture_list);
	Fixture_list = NULL;
}

/*
 * Retrieve a fixture, with error checking
 */
fixture *Fix_get_fixture (map_object *mo)
{

  int fi = Fix_check_index (mo);
  extern map_object *Map_object_list;

  /*
   * Ought to check that the fixture links back to the correct map object,
   *  but it's slightly tricky as it stands
   */
  if (mo != Map_object_list + Fixture_list [fi].map_index)
  {
    Cant_happen ("Fix_get_fixture(): bad object link");
  }

  return &(Fixture_list [fi]);

}




/*
 * Special fixture hooks, CST format
 * These are only the ones that directly interest us or we'd be here all day
 */
#define CST_TV               0x070006
#define CST_COMPUTER_MONITOR 0x070007
#define CST_CHAIR1           0x070105
#define CST_TABLE            0x070106
#define CST_COUCH            0x070107
#define CST_CHAIR2           0x070108
#define CST_ICON             0x070201
#define CST_GRAFFITI         0x070202
#define CST_WORDS            0x070203
#define CST_SCREEN           0x070206
#define CST_TMAP             0x070207
#define CST_SUPERSCREEN      0x070208
#define CST_BIGSCREEN        0x070209
#define CST_REPULSOR_LIGHTS  0x07020A
#define CST_CHAIR3           0x070405
#define CST_CONTROL_PEDESTAL 0x070506



/*
 * Fixture object post-initialisation (bitmap hooks, mainly)
 */
void Fix_init_object (map_object *mo)
{

  switch (CST (mo))
  {
  case CST_CHAIR1:
  case CST_TABLE:
  case CST_COUCH:
  case CST_CHAIR2:
  case CST_CHAIR3:
    mo->fetch_texture_hook = fetch_texture_model;
    break;

  case CST_ICON:
    mo->fetch_texture_hook = fetch_texture_icon;
    break;

  case CST_GRAFFITI:
    mo->fetch_texture_hook = fetch_texture_graffiti;
    break;

  case CST_WORDS:
    mo->fetch_texture_hook = fetch_texture_words;
    break;

  case CST_TV:
  case CST_COMPUTER_MONITOR: /* 3D objects with screens */
  case CST_SCREEN:
  case CST_TMAP:
  case CST_SUPERSCREEN:
  case CST_BIGSCREEN:        /* these are handled by screen.c */
  case CST_CONTROL_PEDESTAL: /* has screen on top */
    Scr_init_object (mo);
    break;

  case CST_REPULSOR_LIGHTS:
    mo->fetch_texture_hook = fetch_texture_repulsor;
    break;

  default:
    break;
  }
}



int Words_font [] = { 4, 7, 0, 10 };




int Fix_check_index (map_object *mo)
{
  int fi = mo->class_index;
  if (mo->class != OBJCLASS_FIXTURE)
  {
    Cant_happen ("non-fixture object");
  }
  if ((fi < 1) || (fi >= Num_fixtures))
  {
    Cant_happen ("Fix_check_index(): fixture index out of range");
  }
  return fi;
}





ss_bitmap_header *render_words (int fi)
{

  block_handle bh;
  size_t size;
  int width;
  int height;
  ss_bitmap_header *bmp;

  bh = Res_load_subchunk (WORDS_CHUNK, Fixture_list [fi].u.text, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Fix_get_bitmap(): failed to load WORDS text");
  }
  Font_get_size (Words_font [Fixture_list [fi].flags & 3],
		 *bh, &width, &height);
  bmp = Bmp_create (width, height);
  memset (bmp->bits, 0, width * height);
  Font_render_string (Words_font [Fixture_list [fi].flags & 3], *bh,
		      0, 0, width, height, bmp->bits,
		      (byte)(Fixture_list [fi].texture));
  return bmp;

}



int Fix_get_texture_no (map_object *obj)
{
  int fi = Fix_check_index (obj);
  return Fixture_list [fi].texture;
}


/*
 * Texture hooks
 */
int fetch_texture_icon (map_object *obj)
{
  return Tex_get_cache_id (CHUNK_TEXTURE_DECAL, obj->state);
}

int fetch_texture_graffiti (map_object *obj)
{
  return Tex_get_cache_id (CHUNK_TEXTURE_GRAFFITI, obj->state);
}

int fetch_texture_words (map_object *obj)
{

  if (obj->base_texid < 0)
  {
    /*
     * Render the WORDS into a bitmap, and add to the object bitmap cache to
     *  ensure that it is freed afterwards
     */
    int fi = Fix_check_index (obj);
    ss_bitmap_header *bmp = render_words (fi);
    obj->base_texid = Tex_add_cache (bmp);

    /*
     * Initialise the decal surface here, so that the scale is correct
     */
    if (obj->ms == NULL)
    {
      int logscale = (Fixture_list [fi].flags >> 4) & 3;
      float scale_mult = 4.;
      if (logscale != 0) scale_mult /= (float) (16 >> logscale);
      Obj_make_decal_surface (obj, scale_mult * bmp->width,
			      scale_mult * bmp->height,
			      SF_TEXTURED | SF_MASKED | SF_SHADED);
    }
  }
  assert (obj->base_texid >= 0);

  return obj->base_texid;
}

int fetch_texture_repulsor (map_object *obj)
{
  return Tex_get_cache_id (BITMAP_CHUNK_REPULSOR, obj->state);
}


int fetch_texture_model (map_object *obj)
{

  int fi = Fix_check_index (obj);
  int bmpno = Fixture_list [fi].start_frame;

  return Tex_get_cache_id (CHUNK_TEXTURE_MODEL2_BASE + (bmpno & 0x7f), 0);

}
