/*
 * font.c -
 *  System Shock fonts
 */

/*
 * File:    src/render.c
 * Purpose: Font loading and rendering
 * Prefix:  Font_
 * Note:    
 * Music:   Van Halen, "Fair Warning"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>  /* malloc() */
#include <string.h>  /* memcpy() */

#include "types.h"
#include "gamemem.h"
#include "chunks.h"
#include "font.h"
#include "main.h"    /* Cant_happen() */
#include "option.h"
#include "resource.h"



/*
 * Debug
 */
#define FONT_MESSAGES 1


#define DEFAULT_FONT_COLOUR 54  /* in the middle of the reds */


/*
 * Global variables
 */
static ss_font *Font_list;



/*
 * Local function prototypes
 */
void ensure_cache (int f);
void convert_uw_font (int f);



void Font_init (void)
{

  int i;

  Font_list = Alloc_init (NUM_FONTS * sizeof(*Font_list), "Font_init",
			  "font cache");
  for (i = 0; i < NUM_FONTS; ++i)
  {
    Font_list [i].xpos = NULL;
    Font_list [i].bits = NULL;
  }

}

void Font_deinit(void)
{
	int i;

	for (i = 0; i < NUM_FONTS; i++)
	{
		if (Font_list[i].xpos)
		{
			free(Font_list[i].xpos);
			Font_list[i].xpos = NULL;
		}
		if (Font_list[i].bits)
		{
			free(Font_list[i].bits);
			Font_list[i].bits = NULL;
		}
	}
	free(Font_list);
	Font_list = NULL;
}


void Font_get_size (int f, const char *s, int *w, int *h)
{

  const char *cp;
  int chno;

  ensure_cache (f);

  *h = Font_list [f].height;
  *w = 0;
  for (cp = s; *cp != '\0'; ++cp)
  {
    char c = *cp;

    /*
     * Fudge to allow upper-case only fonts
     */
    if (c > Font_list [f].last) c = toupper (c);
    if ((c < Font_list [f].first) || (c > Font_list [f].last))
    {
      Cant_happen ("Font_get_size(): character out of range");
    }
    chno = c - Font_list [f].first;
    *w += Font_list [f].xpos [chno + 1] - Font_list [f].xpos [chno];
  }

}






void Font_render_string (int f, const char *s, int x, int y,
			 int width, int height, byte *bits, byte col)
{

  const char *cp;
  int chno = 0;
  int sx = x;
  int bot_y;

  ensure_cache (f);
  if (height < Font_list [f].height)
  {
    Cant_happen ("Font_render_string(): buffer is too small");
  }
  bot_y = Font_list [f].height;
  if (bot_y + y > height) bot_y = height - y;

  if (col == 0)
  {
    col = DEFAULT_FONT_COLOUR;
  }

  for (cp = s; *cp != '\0'; ++cp)
  {
    char c = *cp;

    int line;
    int cx, cw;
    int ex;

    /*
     * Fudge to allow upper-case only fonts
     */
    if (c > Font_list [f].last) c = toupper (c);
    if ((c < Font_list [f].first) || (c > Font_list [f].last))
    {
      Cant_happen ("Font_render_string(): character out of range");
    }
    chno = c - Font_list [f].first;
    cx = Font_list [f].xpos [chno];
    cw = Font_list [f].xpos [chno + 1] - cx;
    ex = sx + cw;
    if (ex > width) ex = width;

    /*
     * WTF? Font 2 has 2 zeros ?!?!
     */
    if ((f == 2) && (c > '9')) ++chno;

    for (line = 0; line < bot_y; ++line)
    {

      byte *bp = bits + (y + line) * width + sx;
      int pix;

      for (pix = 0; pix < ex - sx; ++pix)
      {
	if (Font_list [f].colour)
	{
	  *bp = Font_list [f].bits [line * Font_list [f].width + pix + cx];
	}
	else
	{
	  if (Font_list [f].bits [line * Font_list [f].width + (pix + cx) / 8]
	      & (0x80 >> (pix + cx) % 8))
	  {
	    *bp = col;
	  }
	  else
	  {
/*  	    bits [y * width + x + sx] = 0; */
	  }
	}
	++bp;
      }
    }

    sx += cw;
    if (sx >= width) break;

  }

}







void ensure_cache (int f)
{

  block_handle bh;
  size_t size;
  unsigned char *fp;
  int offs_xpos;
  int offs_bits;
  int nxpos;
  int i;

  if (Font_list [f].xpos != NULL)
  {
    return;
  }

  if (Option_underworld)
  {
    convert_uw_font (f);
    return;
  }

  /*
   * Load up the font resource
   */
  bh = Res_load_chunk_by_id (CHUNK_FONT_BASE + f, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("failed to load font");
  }

  /*
   * Load up the font header
   */
  fp = *bh;
  Font_list [f].colour = *fp++;
  fp += 0x23;
  Font_list [f].first  = *fp + 256 * fp [1]; fp += 2;
  Font_list [f].last   = *fp + 256 * fp [1]; fp += 2;
  fp += 0x20;
  offs_xpos            = *fp + 256 * fp [1] + 65536 * fp [2] +
                          16777216 * fp [3]; fp += 4;
  offs_bits            = *fp + 256 * fp [1] + 65536 * fp [2] +
                          16777216 * fp [3]; fp += 4;
  Font_list [f].width  = *fp + 256 * fp [1]; fp += 2;
  Font_list [f].height = *fp + 256 * fp [1]; fp += 2;
  /*
   * Load up the x positions
   */
  nxpos = Font_list [f].last - Font_list [f].first + 2;
  Font_list [f].xpos = malloc (nxpos * sizeof *Font_list [f].xpos);
  if (Font_list [f].xpos == NULL)
  {
    Cant_happen ("failed to allocate space for font positions");
  }
  fp = ((unsigned char *) *bh) + offs_xpos;
  for (i = 0; i < nxpos; ++i)
  {
    Font_list [f].xpos [i] = *fp + 256 * fp [1]; fp += 2;
  }

  /*
   * Load up the bitmap bits proper
   */
  Font_list [f].bits = malloc (Font_list [f].width * Font_list [f].height);
  if (Font_list [f].bits == NULL)
  {
    Cant_happen ("failed to allocate space for font bitmap");
  }
  fp = ((unsigned char *) *bh) + offs_bits;
  memcpy (Font_list [f].bits, fp, Font_list [f].width * Font_list [f].height);

  /*
   * Tidy up after us
   */
  Res_free_chunk_by_id (CHUNK_FONT_BASE + f);

}



const char *UW_font_file [] =
{
  "font4x5p.sys",
  "font5x6i.sys",
  "font5x6p.sys",
  "fontbig.sys",
  "fontbutn.sys",
  "fontchar.sys"
};

#define UW_FONT_HEAD_SIZE 12


/*
 * Convert an Underworld font (individual characters) to a System Shock style
 *  font in one big bitmap
 */
void convert_uw_font (int f)
{

  /*
   * The UW font file header
   */
  int width_bytes;
  int char_size;
  int blank_width;
  int char_height;
  int row_width;
  int max_width;
  int total_width = 0;
  int num_chars;
  int tempx [257];
  int i;

  FILE *fontfile = fopen (Res_make_path (UW_DATA_PATH, UW_font_file [f]),
			  "rb");
  if (fontfile == NULL)
  {
    Cant_happen ("font.c: convert_uw_font(): failed to open font file");
  }

  READ_LE_SHORT (fontfile, width_bytes);
  READ_LE_SHORT (fontfile, char_size);
  READ_LE_SHORT (fontfile, blank_width);
  READ_LE_SHORT (fontfile, char_height);
  READ_LE_SHORT (fontfile, row_width);
  READ_LE_SHORT (fontfile, max_width);

  /*
   * Figure out the width of the font bitmap and populate the X array
   */
  for (num_chars = 0; num_chars <= 256; ++num_chars)
  {
    /*
     * Skip over the bitmap bytes for now, we will come back for them
     */
    fseek (fontfile, char_size, SEEK_CUR);
    tempx [num_chars] = total_width;
    i = fgetc (fontfile);            /* sneaky use of i as a temporary */
    if (i == EOF) break;
    total_width += i;
  }
  if (num_chars > 256) num_chars = 256; /* max 256 characters allowed */

  /*
   * Now we have the size of our bitmap, allocate some space for it. All
   *  Underworld fonts are 1-bit
   */
  Font_list [f].xpos = malloc ((num_chars + 1) * sizeof *(Font_list [f].xpos));
  if (Font_list [f].xpos == NULL)
  {
    Cant_happen ("font.c: convert_uw_font(): couldn't allocate X list");
  }
  Font_list [f].height = char_height;
  Font_list [f].width  = (total_width + 7) / 8;
  Font_list [f].bits = malloc (Font_list [f].width * Font_list [f].height);
  if (Font_list [f].bits == NULL)
  {
    Cant_happen ("font.c: convert_uw_font(): couldn't allocate font bitmap");
  }
  memset (Font_list [f].bits, 0, Font_list [f].width * Font_list [f].height);

  /*
   * Rewind the file and insert the character bitmaps
   * This is a rather inefficient way of packing the bitmap, but we'll only
   *  need to do this when a font is loaded so it probably isn't worth
   *  optimising
   */
  fseek (fontfile, UW_FONT_HEAD_SIZE, SEEK_SET);
  for (i = 0, total_width = 0; i < num_chars; ++i)
  {
    int b;
    byte *bp = Font_list [f].bits + total_width / 8;
    int bit = total_width % 8;
    int y;
    for (y = 0; y < char_height; ++y)
    {
      for (b = 0; b < row_width; ++b)
      {
	int rowbyte = fgetc (fontfile);
	if (bit == 0)
	{
	  bp [b] = rowbyte;
	}
	else
	{
	  bp [b] |= rowbyte >> bit;
	  if (b < row_width - 1) bp [b + 1] = (rowbyte << (8 - bit)) & 0xff;
	}
      }
      bp += Font_list [f].width;
    }
    Font_list [f].xpos [i] = total_width;
    total_width += fgetc (fontfile);
  }
  Font_list [f].xpos [num_chars] = total_width; /* fencepost */

  /*
   * Populate the rest of the font header
   */
  Font_list [f].colour = 0;  /* all 1bit */
  Font_list [f].first  = 0;
  Font_list [f].last   = num_chars - 1;

  printf ("%d chars found in font file\n", num_chars);

  fclose (fontfile);

}
