/*
 * game.c -
 *  main game loop
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>	 /* exit() */
#include <memory.h>
#include <string.h>

#include "graph2d.h"
#include "gamemem.h"
#include "anim.h"
#include "clip.h"
#include "critter.h"
#include "control.h"
#include "dark.h"
#include "export.h"
#include "font.h"
#include "gamestring.h"
#include "gui.h"         /* need to get at the framebuffer for screenshots */
#include "main.h"
#include "map.h"
#include "model.h"
#include "object.h"
#include "option.h"
#include "palette.h"
#include "player.h"
#include "render.h"
#include "resource.h"
#include "surface.h"
#include "texture.h"
#include "shade.h"
#include "event.h"

#include "hud.h"
#include "mfd.h"



extern int Screen_width;
extern int Screen_height;


bool Underworld2 = false;

bool Have_audio_logs = true;
bool Have_video_mail = true;


/*
 * Local function prototypes
 */
void add_resfile (int base, const char *fname);

void Game_debug_dumppal(void);



static void guess_uw_version (void)
{

  /*
   * We detect Underworld2 by the fact that it has a different textures file
   */
  FILE *testfile = fopen (Res_make_path (UW_DATA_PATH, "t64.tr"), "rb");
  if (testfile != NULL)
  {
    fclose (testfile);
    Underworld2 = true;
  }

}




/*
 * Initialise all vital game stuff.
 * Requires Screen_width and Screen_height to have been set up beforehand
 */
void Game_init (void)
{
	Evt_init ();
  Res_init ();
  if (!Option_underworld)
  {
    /*
     * Local add_resfile which bombs on failure
     */
    add_resfile (CD_DATA_PATH, "archive.dat");    /* Level archives     */
    add_resfile (HD_DATA_PATH, "cybstrng.res");   /* Game strings       */
    add_resfile (HD_DATA_PATH, "texture.res");    /* Main textures      */
    add_resfile (HD_DATA_PATH, "sideart.res");		/* Side HW buttons		*/
    add_resfile (HD_DATA_PATH, "mfdart.res");			/* MFD icons/imgs...	*/
    add_resfile (HD_DATA_PATH, "objart2.res");
    add_resfile (HD_DATA_PATH, "objart3.res");
    add_resfile (CD_DATA_PATH, "objart.res");
    if (Option_no_alogs ||
	!Res_add_resfile (Res_make_path (CD_DATA_PATH, "citalog.res")))
    {
      Have_audio_logs = false;
    }
    add_resfile (HD_DATA_PATH, "digifx.res");
    add_resfile (HD_DATA_PATH, "gamescr.res");
    add_resfile (HD_DATA_PATH, "obj3d.res");      /* 3D object models   */
    add_resfile (HD_DATA_PATH, "citmat.res");     /* 3D object textures */
    if (Option_no_vmail ||
	!Res_add_resfile (Res_make_path (CD_DATA_PATH, "vidmail.res")))
    {
      Have_video_mail = false;
    }
/*  add_resfile (CD_DATA_PATH, "splash.res"); */	/* splash init screens*/
    /*    add_resfile (HD_DATA_PATH, "savgam00.dat");  */
/*      Res_add_resfile ("savgam00.dat"); */
/*    Res_print_resfile_info (); */
    if (Option_dump_chunk >= 0)
    {
      if (Option_screen_width < 64)
	Res_dump_chunk (Option_dump_chunk, Option_screen_width);
      else
	Res_dump_chunk (Option_dump_chunk, 16);
      exit (0);
    }

  }
  else
  {
    guess_uw_version ();
  }

  Pal_init ();
  shade_init();  /* jim - requires palette for darkness table */
  Crit_init ();


  if (Option_extract >= 0)
  {
		Game_debug_dumppal();
    Exp_write_texture (Option_extract, Option_extract_sub);
    exit (0);
  }
  if (Option_extract_all)
  {
    Exp_write_all_textures ();
  }

  Str_init ();
  if (Option_dump_strings)
  {
    Str_dump_strings ();
    exit (0);
  }

  Dark_init ();
  Tex_init ();
  Map_init ();
  Sur_init ();
  Anim_init ();
  Obj_init ();
  Clip_init ();
  Ren_init ();
  Mod_init (Screen_width, Screen_height);
  Font_init ();
  Con_init(Screen_width, Screen_height);

	/* ff1: as long as we have no menu... */
  Con_activate_mode(C_CON_MODE_PLAY, true);

  Tex_init_level (Option_start_level);
  Map_init_level (Option_start_level);
  Obj_init_level (Option_start_level);
  Pal_init_level ();

	Player_init();
}

void Game_deinit(void)
{
	Con_deinit();
	Anim_deinit();
	Font_deinit();
	Mod_deinit();
	/* Ren_deinit(); */
	Clip_deinit();
	Obj_deinit();
	Sur_deinit();
	Map_deinit();
	Tex_exit();
	Dark_exit ();
	Str_deinit();
	/* Crit_deinit(); */
	/* shade_deinit(); */
	Pal_deinit();
}


/*
 * Wrapper round Res_add_resfile() for initialisation
 */
void add_resfile (int base, const char *fname)
{
  char err [256];
  if (!Res_add_resfile (Res_make_path (base, fname)))
  {
    sprintf (err, "Couldn't initialise resource file %s :\n"
	          " is your Shock CD mounted?",
	     fname);
    Cant_happen (err);
  }
}


int Screenshot_no;

void Game_screenshot(void)
{
	FILE *pFile = NULL;
  pixel *bmp = GUI_get_framebuffer ();
	char buf[51];
	char shotname [256], testname[256];

	do
	{
		if (pFile)
			fclose(pFile);
# if HAVE_PNG
		sprintf(testname, "shack%02d.png", Screenshot_no);
# else	
		sprintf(testname, "shack%02d.tga", Screenshot_no);
# endif
		sprintf (shotname, "shack%02d", Screenshot_no++);
	}
	while ((pFile = fopen(testname, "r")) != NULL);

#if BITS_PER_PIXEL >= 24
# if HAVE_PNG
	Exp_write_png32 (Screen_width, Screen_height, bmp, shotname);
# else
	Exp_write_tga32(Screen_width, Screen_height, bmp, shotname,
									"TriOptimum Screenshot");
# endif /* HAVE_PNG */
	sprintf(buf, "Screen shot (%s)", shotname);
#elif BITS_PER_PIXEL == 8
# if HAVE_PNG
	Exp_write_png8 (Screen_width, Screen_height, bmp, shotname);
# else
	Exp_write_tga8(Screen_width, Screen_height, bmp, shotname, RGB_trans,
								 "TriOptimum Screenshot");
# endif /* HAVE_PNG */
	sprintf(buf, "Screen shot (%s)", shotname);
#else
	sprintf(buf, "no valid color depth");
#endif
	MSG_add_message(0, buf, 5000, 1, rgbt(0xFF, 0x00, 0x00, 0xFF),
																	 rgbt(0xFF, 0x00, 0x00, 0x40));
}

/* DEBUG output(s) */
extern camera Avatar_camera;	

void Game_debug_output(void)
{
	char strMsg[0x100];

	sprintf(strMsg, "Avatar Camera at: 0x%04X 0x%04X\n", 
				 (int)Avatar_camera.position.x, (int)Avatar_camera.position.y);
	printf(strMsg);
	MSG_add_message(0, strMsg, 5000, 0, 
									rgbt(0x00, 0xFF, 0x00, 0xFF),
									rgbt(0x00, 0xFF, 0x00, 0x40));
}

void Game_debug_dumppal(void)
{
	g2dbase gbPal;
	int i = 0;

	G2D_init_clrbase(&gbPal, 160, 160);
	while (i < 0x100)
	{
		gbPal.bOp.clr = RGB_trans[i];
		gbPal.bBoxArea.right = gbPal.bBoxArea.left + 10;
		gbPal.bBoxArea.bottom = gbPal.bBoxArea.top + 10;
		G2Dc_fill_area(&gbPal);

		i++;
		if (!(i % 16))
		{
			gbPal.bBoxArea.top += 10;
			gbPal.bBoxArea.left = 0;
		}
		else
			gbPal.bBoxArea.left += 10;
	}
#if HAVE_PNG
	Exp_write_png32(160, 160, gbPal.bImg.clr, "palette");
#else
	Exp_write_tga32(160, 160, gbPal.bImg.clr, "palette",
									Option_underworld ? "UW" : "SS");
#endif
	G2D_reset_base(&gbPal, true);
}
