/*
 * gamemem.c -
 *  game memory management
 *
 * This file by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h" /* jim - Always first! */

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>

#include "types.h"
#include "main.h"
#include "gamemem.h"

/* remove nasty recursions ;) */
#ifdef malloc
#undef malloc
#endif
#ifdef realloc
#undef realloc
#endif
#ifdef free
#undef free
#endif

#ifdef _DEBUG
#define	MEM_DEBUG
#endif
/*
#define MEM_LOWLEVEL
*/
#define	MEM_INITMEM		0x00

typedef struct tag_MemoryBlock
{
	void										*mbpData;
	size_t									 mbSize;
	
	char										*mbSourceFile;
	int											 mbLine;

	struct tag_MemoryBlock	*mbpPrev;
	struct tag_MemoryBlock	*mbpNext;
	uint32									 mbId;
} MemoryBlock;

MemoryBlock	*pMemFirst = NULL;
MemoryBlock *pMemLast = NULL;
uint32			 gIDCounter = 0;

void mem_crash(char *strSourceFile, int nLine, size_t stSize)
{
	char strMessage[256];

	sprintf(strMessage, "Failed to allocate %d bytes originated in "
					"%s@%04d", stSize, strSourceFile, nLine);
	Cant_happen(strMessage);
}

MemoryBlock	*mem_find_block(void *pData)
{
	MemoryBlock *pMem = pMemLast, *pFound = NULL;

	while (pMem && !pFound)
	{
		if (pMem->mbpData == pData)
			pFound = pMem;

		pMem = pMem->mbpPrev;
	}

	return pFound;
}

void	 Mem_init(void)
{

}

void	 Mem_deinit(void)
{
	MemoryBlock *pMem;

#ifdef MEM_DEBUG
	uint32 dwLeak = 0;

	if (pMemFirst)
		printf("Mem_deinit() Found memory leaks!\n");
	else
		printf("Mem_deinit() everything clear!\n");
#endif
	pMem = pMemFirst;
	while (pMem)
	{
#ifdef MEM_DEBUG
		printf("%08lx %s@%04d: %6d Bytes\n", pMem->mbId,
					 pMem->mbSourceFile, pMem->mbLine, pMem->mbSize);
		dwLeak += pMem->mbSize;
#endif
		/* free(pMem->mbpData); */
		if (pMem->mbpNext)
		{
			pMem = pMem->mbpNext;
			free(pMem->mbpPrev);
		}
		else
		{
			free(pMem);
			pMem = NULL;
		}
	}
	pMemFirst = pMemLast = NULL;
#ifdef MEM_DEBUG
	if (dwLeak)
		printf("Lost %lu Bytes through leaks\n", dwLeak);
#endif
}

void	*Mem_alloc(size_t stSize, char *strSourceFile, int nLine)
{
	MemoryBlock *pMem = NULL;
	void *pData;

#ifdef MEM_LOWLEVEL
	printf("%08lX %s@%04d wants %6d bytes\n", gIDCounter, 
				 strSourceFile, nLine, stSize);
#endif
	pData = malloc(stSize);
	pMem = mem_find_block(pData);
	if (!pMem)
	{
		pMem = malloc(sizeof(MemoryBlock));
		memset(pMem, 0x00, sizeof(MemoryBlock));

		pMem->mbpPrev = pMemLast;
		if (pMemLast)
			pMemLast->mbpNext = pMem;
		pMemLast = pMem;
		if (!pMemFirst)
			pMemFirst = pMem;
	}
#ifdef MEM_DEBUG
	else
	{
		printf("Mem_alloc() Memory Out of Sync! Lost Track of %08lx\n",
					 pMem->mbId);
	}
#endif

	pMem->mbpData = pData;
	if (!pMem->mbpData)
		mem_crash(strSourceFile, nLine, stSize);
#ifdef MEM_INITMEM
	memset(pMem->mbpData, MEM_INITMEM, stSize);
#endif
	pMem->mbSize = stSize;
	pMem->mbSourceFile = strSourceFile;
	pMem->mbLine = nLine;
	pMem->mbId = gIDCounter++;

	return pMem->mbpData;
}

void	*Mem_realloc(void *pPointer, size_t stNewSize,
									 char *strSourceFile, int nLine)
{
	MemoryBlock *pMem = NULL;
	void *pNewPointer = NULL;

	if (pPointer)
	{
		pMem = mem_find_block(pPointer);
		pNewPointer = realloc(pPointer, stNewSize);
		if (!pNewPointer)
			mem_crash(strSourceFile, nLine, stNewSize);
		if (pMem)
		{			
#ifdef MEM_LOWLEVEL
			printf("%08lX %s@%04d wants %6d new bytes and previously had "
						 "%6d bytes\n", pMem->mbId, strSourceFile, 
						 nLine, stNewSize, pMem->mbSize);
#endif
			pMem->mbpData = pNewPointer;
#ifdef MEM_INITMEM
			if (pMem->mbSize < stNewSize)
				memset(&((byte *)(pMem->mbpData))[pMem->mbSize], MEM_INITMEM, 
							 stNewSize - pMem->mbSize);
#endif
			pMem->mbSize = stNewSize;
			pMem->mbSourceFile = strSourceFile;
			pMem->mbLine = nLine;
		}
		else
		{
#ifdef MEM_LOWLEVEL
			printf("FFFFFFFF %s@%04d wants %6d new bytes of unknown data\n",
						 strSourceFile, nLine, stNewSize);
#endif
		}
	}
	else
		pNewPointer = Mem_alloc(stNewSize, strSourceFile, nLine);

	return pNewPointer;
}

void Mem_free(void *pPointer, char *strSourceFile, int nLine)
{
	MemoryBlock *pMem;

#ifndef MEM_LOWLEVEL
	UNUSED(strSourceFile);
	UNUSED(nLine);
#endif
	if (pPointer)
	{
		pMem = mem_find_block(pPointer);
		if (pMem)
		{
#ifdef MEM_LOWLEVEL
			printf("%08lX %s@%04d frees %6d bytes, originated from %s@%04d\n",
						 pMem->mbId, strSourceFile, nLine, pMem->mbSize,
						 pMem->mbSourceFile, pMem->mbLine);
#endif
			if (pMem->mbpPrev)
				pMem->mbpPrev->mbpNext = pMem->mbpNext;
			if (pMem->mbpNext)
				pMem->mbpNext->mbpPrev = pMem->mbpPrev;
			if (pMem == pMemLast)
				pMemLast = pMem->mbpPrev;
			if (pMem == pMemFirst)
				pMemFirst = pMem->mbpNext;
			free(pMem);
		}
		else
		{
#ifdef MEM_LOWLEVEL
			printf("FFFFFFFF %s@%04d frees unknown data\n", 
						 strSourceFile, nLine);
#endif
		}
		free(pPointer);
	}
}



/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
