/*
 * gamestring.c -
 *  Game strings
 */

/*
 * File:    src/gamestring.c
 * Purpose: Generic string loading and caching
 * Prefix:  Str_
 * Note:    Would be called "string.c" but that would conflict with <string.h>
 * Music:   Skid Row, "Slave To The Grind"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes"
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "gamemem.h"
#include "gamestring.h"
#include "main.h"        /* Cant_happen() */
#include "option.h"
#include "resource.h"



/*
 * Debugging
 */
#define STRING_MESSAGES 1


#define UW_STRING_FILE Res_make_path (UW_DATA_PATH, "strings.pak")



/*
 * Local function prototypes
 */
char **puff_uw_string_chunk (int chunk_no, int *num_strings);




/*
 * Underworld Huffman compression
 */
static struct huff_node
{
  char symbol;
  int parent;
  int left;
  int right;
} *Huff_tree;

static int  Num_huff_nodes;
static int  Num_string_blocks;
static long Offs_strings;



void Str_init (void)
{

  if (Option_underworld)
  {

    int i;

    /*
     * Load up the Huffman tree from strings.pak ... there is only one fixed
     *  tree for all the strings, so we process it at load time
     */
    FILE *file = fopen (UW_STRING_FILE, "rb");
    if (file == NULL)
    {
      Cant_happen ("Str_init(): couldn't open Underworld strings file");
    }
    READ_LE_SHORT (file, Num_huff_nodes);
    Huff_tree = Alloc_init (Num_huff_nodes * sizeof *Huff_tree, "Str_init",
			    "Huffman tree");

    for (i = 0; i < Num_huff_nodes; ++i)
    {
      Huff_tree [i].symbol = fgetc (file);
      Huff_tree [i].parent = fgetc (file);
      Huff_tree [i].left   = fgetc (file);
      Huff_tree [i].right  = fgetc (file);
    }
    READ_LE_SHORT (file, Num_string_blocks);
    Offs_strings = ftell (file);
    fclose (file);

  }

}

void Str_deinit(void)
{
	free(Huff_tree);
	Huff_tree = NULL;
}


/*
 * Load up a string chunk (as if you couldn't guess)
 * Strings are returned in a contiguous string block with the pointers in the
 *  return array pointing into it. This is a new string block containing a
 *  copy of the strings in question and does not point into the chunk cache
 *  from resource.c
 * Always free strings returned by this function using Str_free_string_chunk()
 */
char **Str_load_string_chunk (int chunk_no, int *num_strings)
{

  char **retchunk;
  char *str;
  size_t size;
  size_t totsize;
  int i;

  if (Option_underworld)
  {
    return puff_uw_string_chunk (chunk_no, num_strings);
  }

  *num_strings = Res_get_numSubchunks_by_id (chunk_no, &totsize);
  if (*num_strings != -1) 
  	{	
  		retchunk = malloc (*num_strings * sizeof *retchunk);
  	}
  else retchunk=NULL;
  if (retchunk == NULL)
  {
    Cant_happen ("Str_load_string_chunk(): couldn't allocate string array");
  }

  /*
   * Fudge the total chunk size to take the subchunk directory into account
   */
  totsize -= 2 + 4 + 4 * *num_strings;
  retchunk [0] = malloc (totsize);
  if (retchunk [0] == NULL)
  {
    Cant_happen ("Str_load_string_chunk(): couldn't allocate string buffer");
  }

  str = retchunk [0];
  for (i = 0; i < *num_strings; ++i)
  {
    block_handle bh = Res_load_subchunk (chunk_no, i, &size);
    if ((bh == NULL) || (*bh == NULL))
    {
      Cant_happen ("Str_load_string_chunk(): couldn't load string chunk");
    }
    if (totsize < size)
    {
      Cant_happen ("Str_load_string_chunk(): buffer overflow");
    }
    totsize -= size;
    memcpy (str, *bh, size);
    str [size - 1] = '\0';
    retchunk [i] = str;
    str += size;
  }
  if (totsize != 0)
  {
    printf ("%d %d %s\n", totsize, *num_strings, retchunk [0]);
    Cant_happen ("Str_load_string_chunk(): string size mismatch");
  }

  Res_free_chunk_by_id (chunk_no);
  return retchunk;

}




void Str_free_string_chunk (char **schk)
{
  free (schk [0]);
  free (schk);
}




/*
 * This will only work for Ultima Underworld at the moment ... it sets
 *   Num_string_blocks but System Shock doesn't
 */
void Str_dump_strings (void)
{

  int i;
  char **schk;

  for (i = 0; i < Num_string_blocks; ++i)
  {
    int nstrings;
    int j;
    schk = Str_load_string_chunk (i, &nstrings);
    printf ("\nBlock %d, %d strings\n", i, nstrings);
    for (j = 0; j < nstrings; ++j)
    {
      printf ("%d\t%s\n", j, schk [j]);
    }
    Str_free_string_chunk (schk);
  }

}




#define UW_STRING_CHUNK_BUFSIZE 16384

char **puff_uw_string_chunk (int chunk_no, int *num_strings)
{

  int id;
  long offset;  /* fpos_t is NOT constrained to be a numerical type */

  int stringoffs;

  int i;

  static char str [UW_STRING_CHUNK_BUFSIZE];
  int pos = 0;
  int node;
  char **retchunk;

  FILE *file = fopen (UW_STRING_FILE, "rb");

  if ((chunk_no < 0) || (chunk_no >= Num_string_blocks))
  {
    Cant_happen ("gamestring.c:puff_uw_string_chunk(): illegal chunk number");
  }

  fseek (file, Offs_strings + 6 * chunk_no, SEEK_SET);
  READ_LE_SHORT (file, id);
  READ_LE_LONG  (file, offset);

# if STRING_MESSAGES
  printf ("Block %d at %08lX, ", chunk_no, offset);
# endif
  fseek (file, offset, SEEK_SET);

  READ_LE_SHORT (file, *num_strings);
  retchunk  = malloc (*num_strings * sizeof *retchunk);

# if STRING_MESSAGES
  printf ("\t%d strings\n", *num_strings);
# endif

  for (i = 0; i < *num_strings; ++i)
  {

    int bit = 0, c = 0, raw = 0;

    fseek (file, offset + 2 + 2 * i, SEEK_SET);
    READ_LE_SHORT (file, stringoffs);
    fseek (file, offset + 2 + 2 * *num_strings + stringoffs, SEEK_SET);
    do
    {
      node = Num_huff_nodes - 1;
      while (Huff_tree [node].left != 0xFF)
      {
	if (bit == 0)
	{
	  raw = fgetc (file);
	  bit = 8;
	}
	if (raw & 0x80)
	{
	  node = Huff_tree [node].right;
	}
	else
	{
	  node = Huff_tree [node].left;
	}
	raw <<= 1;
	bit--;
      }
      c = Huff_tree [node].symbol;
      str [pos++] = c;
      if (pos >= UW_STRING_CHUNK_BUFSIZE)
      {
	Cant_happen ("gamestring.c:puff_uw_string_chunk(): buffer overflow");
      }
    } while (c != '|');
    str [pos - 1] = '\0';

  }

  retchunk [0] = malloc (pos);
  memcpy (retchunk [0], str, pos);
  for (i = 1; i < *num_strings; ++i)
  {
    retchunk [i] = retchunk [i - 1] + strlen (retchunk [i - 1]) + 1;
  }

  fclose (file);

  return retchunk;

}

/* String timming on both sides */
char *Str_trim_sides(char *strString, char cTrim)
{
	int i, nLength;

	if (strString)
	{
		nLength = strlen(strString);
		for (i = (nLength - 1); i > 0 && strString[i] == cTrim; i--)
				strString[i] = 0x00;
		nLength = strlen(strString);
		while (nLength && strString[0] == cTrim)
		{
			memmove(strString, &strString[1], nLength);
			nLength--;
		}
	}

	return strString;
}

/* array helper functions. those arrays are treaten NULL terminated */

char	**Str_create_array()
{
	char **pArray;

	pArray = (char **)malloc(sizeof(char *));
	pArray[0] = NULL;

	return pArray;
}

char	**Str_resize_array(char **pArray, uint32 nCurEntries, 
												 uint32 nNewEntries, uint32 nNewChars)
{
	char **pNewArray;
	uint32 i;

	for (i = nNewEntries; i < nCurEntries; i++)
		free(pArray[i]);
	pNewArray = (char **)realloc(pArray, 
															 sizeof(char *) * (nNewEntries + 1));
	for (i = nCurEntries; i < nNewEntries; i++)
	{
		pNewArray[i] = malloc(nNewChars + 1);
		memset(pNewArray[i], 0x00, nNewChars + 1);
	}
	pNewArray[nNewEntries] = NULL;

	return pNewArray;
}

void		Str_free_array(char **pArray)
{
	int i = 0;

	while (pArray[i])
		free(pArray[i++]);
	free(pArray);
}

uint32	Str_size_array(char **pArray)
{
	uint32 nAmount = 0;

	while (pArray[nAmount])
		nAmount++;

	return nAmount;
}

char **Str_extract(const char *strLine, const char *strDelimeter)
{
	char **pResult = NULL, *strBuffer = NULL;
	int i = 0, nLineLength, nDelLength, nCurLen = 0, nCurLines = 0;
	
	nLineLength = strlen(strLine);
	nDelLength = strlen(strDelimeter);
	strBuffer = (char *)malloc(1);
	strBuffer[0] = 0x00;
	pResult = Str_create_array();
	while (i < nLineLength)
	{	/* dig through all characters */
		if (!strncmp(&strLine[i], strDelimeter, nDelLength))		
		{ /* we found our delimeter*/
			nCurLines++;
			pResult = realloc(pResult, sizeof(char *) * (nCurLines + 1));
			pResult[nCurLines] = NULL;
			pResult[nCurLines - 1] = strBuffer;
			i += nDelLength;
			nCurLen = 0;
			strBuffer = (char *)malloc(1);
			strBuffer[0] = 0x00;
		}
		else
		{	/* just an ordinary character */
			nCurLen++;
			strBuffer = (char *)realloc(strBuffer, nCurLen + 1);
			strBuffer[nCurLen] = 0x00;
			strBuffer[nCurLen - 1] = strLine[i];
			i++;
		}
	}
	if (strlen(strBuffer))
	{ /* did the last part contain any characters? */
		nCurLines++;
		pResult = realloc(pResult, sizeof(char *) * (nCurLines + 1));
		pResult[nCurLines] = NULL;
		pResult[nCurLines - 1] = strBuffer;
	}
	else /* no memory leaks... */
		free(strBuffer);

	return pResult;
}
