/*
 * graph2d.c -
 *  processing of 2d types and images
 *
 * This file by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <math.h>

#include "types.h"
#include "vec.h"
#include "graph2d.h"

#include "gamemem.h"

#include "bitmap.h"					
#include "shade.h"					
#include "palette.h"

#include "main.h"
#include "object.h"
#include "render.h"
#include "map.h"
#include "font.h"

#include "gamestring.h"
#include "option.h"


/*
 * We want pi!
 * (ff1: again...)
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif

/* enables sanity checks for the operations;												*/
/* generally this is checked:																				*/
/* - is bBoxArea within bSizeImg																		*/
/* - do the bBoxAreas of multiple area ops have the same size				*/
/* - some additional size checks...																	*/
/*																																	*/
/* should really be disabled in releases since those IFs are death	*/
/* to speed enhancements...																					*/

/*
#define G2D_SANITY_CHECKS
*/

#ifdef G2D_SANITY_CHECKS

#define	isBoxInSize(b, s)															\
				((b).left >= 0 && (b).top >= 0 &&							\
				 (b).right <= (s).cx && (b).bottom <= (s).cy)
#define isBoxSameBox(b1, b2)													\
				(getRectWidth((b1)) == getRectWidth((b2)) &&	\
				 getRectHeight((b1)) == getRectHeight((b2)))
					 
#endif

/* local functions																									*/
void g2d_draw_border(struct tag_ss_bitmap_header *pHead, byte bg);

void	G2D_zoom_rect(const size *orig, const size *target, 
										size *result)
{
	int mscale = 0, oscale = 0;
	
	if (target->cy)													/* get the scales					*/
		mscale = (target->cx * 1000) / target->cy; 
	if (orig->cy)
		oscale = (orig->cx * 1000) / orig->cy;
	if(mscale > oscale)											/* original is thinner		*/
	{
		result->cy = target->cy;							/* set to output height		*/
		result->cx = (target->cy * oscale) / 1000;
	}
	else
	{
		if(mscale < oscale)										/* original is wider			*/
		{
			result->cx = target->cx;						/* set to output width		*/
			if (orig->cx)
				oscale = (orig->cy * 1000) / orig->cx; /* recalculate scale	*/
			else
				oscale = 0;
			result->cy = (target->cx * oscale) / 1000;
		}
		else																	/* both rects same scale	*/
		{
			result->cy = target->cy;										
			result->cx = target->cx;										
		}														
	}
}

void	 G2D_init(void)
{
	srand((unsigned int)time(NULL));
}

void	G2D_reset_base(g2dbase *pG2DB, bool bFreeBuffer)
{
	if (bFreeBuffer)
		free(pG2DB->bImg.clr);
	memset(pG2DB, 0x00, sizeof(g2dbase));
	pG2DB->nType = G2DAmount;
}

void	G2D_resize_base(g2dbase *pG2DB, uint16 nWidth, uint16 nHeight)
{
	size_t nSizes[G2DAmount] = { sizeof(byte), sizeof(pixel),
															 sizeof(pcolor), sizeof(int) };
	byte *pIn = NULL, *pOut = NULL, *pNew = NULL;
	int i, nLen;

	if (pG2DB->nType < G2DAmount &&
			(pG2DB->bSizeImg.cx != nWidth || pG2DB->bSizeImg.cy != nHeight))
	{
		pIn = pG2DB->bImg.bit;
		pOut = pNew = (byte *)malloc(nWidth * nHeight * nSizes[pG2DB->nType]);
		memset(pNew, 0x00, sizeof(nWidth * nHeight * nSizes[pG2DB->nType]));
		nLen = min(pG2DB->bSizeImg.cx, nWidth);
		for (i = 0; i < min(pG2DB->bSizeImg.cy, nHeight); i++)
		{
			memcpy(pOut, pIn, nLen * nSizes[pG2DB->nType]);
			pOut += (nWidth * nSizes[pG2DB->nType]);
			pIn += (pG2DB->bSizeImg.cx * nSizes[pG2DB->nType]);
		}
		pG2DB->bBoxArea.left = pG2DB->bBoxArea.top = 0;
		pG2DB->bBoxArea.right = pG2DB->bSizeImg.cx = nWidth;
		pG2DB->bBoxArea.bottom = pG2DB->bSizeImg.cy = nHeight;
		free(pG2DB->bImg.bit);
		pG2DB->bImg.bit = pNew;
	}
}

bool	G2D_init_clrbase(g2dbase *pG2DB, uint16 width, uint16 height)
{
	pG2DB->nType = G2DTrue;
	pG2DB->bBoxArea.left		=	pG2DB->bBoxArea.top = 0;
	pG2DB->bBoxArea.right		=	pG2DB->bSizeImg.cx	= width;
	pG2DB->bBoxArea.bottom	=	pG2DB->bSizeImg.cy	= height;
	pG2DB->bOp.clr = 0x00000000;
	pG2DB->pPal = NULL;
	pG2DB->nPalSize = 0;
	pG2DB->bImg.clr = malloc(width * height * sizeof(pcolor));

	return (pG2DB->bImg.clr != NULL);
}

bool	G2D_init_pxlbase(g2dbase *pG2DB, uint16 width, uint16 height)
{
	pG2DB->nType = G2DPixel;
	pG2DB->bBoxArea.left		=	pG2DB->bBoxArea.top = 0;
	pG2DB->bBoxArea.right		=	pG2DB->bSizeImg.cx	= width;
	pG2DB->bBoxArea.bottom	=	pG2DB->bSizeImg.cy	= height;
	pG2DB->bOp.clr = 0x00000000;
	pG2DB->pPal = NULL;
	pG2DB->nPalSize = 0;
	pG2DB->bImg.pxl = malloc(width * height * sizeof(pixel));

	return (pG2DB->bImg.pxl != NULL);
}

bool	G2D_init_mskbase(g2dbase *pG2DB, uint16 width, uint16 height)
{
	pG2DB->nType = G2DMask;
	pG2DB->bBoxArea.left		=	pG2DB->bBoxArea.top = 0;
	pG2DB->bBoxArea.right		=	pG2DB->bSizeImg.cx	= width;
	pG2DB->bBoxArea.bottom	=	pG2DB->bSizeImg.cy	= height;
	pG2DB->bOp.trn = 0;
	pG2DB->pPal = NULL;
	pG2DB->nPalSize = 0;
	pG2DB->bImg.msk = malloc(width * height * sizeof(int));

	return (pG2DB->bImg.msk != NULL);
}

bool	G2D_init_bitbase(g2dbase *pG2DB, uint16 width, uint16 height)
{
	pG2DB->nType = G2DPal;
	pG2DB->bBoxArea.left		=	pG2DB->bBoxArea.top = 0;
	pG2DB->bBoxArea.right		=	pG2DB->bSizeImg.cx	= width;
	pG2DB->bBoxArea.bottom	=	pG2DB->bSizeImg.cy	= height;
	pG2DB->bOp.trn = 0;
	pG2DB->pPal = NULL;
	pG2DB->nPalSize = 0;
	pG2DB->bImg.bit = malloc(width * height);

	return (pG2DB->bImg.bit != NULL);
}

void	G2Dc_head_to_base(const g2dbase *pG2DB, 
											  const struct tag_ss_bitmap_header *pHead,
											  pcolor *pPal, bool ignoreNull)
{
	pcolor *pOut;
	byte *pIn;
	uint16 iny, outy, x;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dc_head_to_base: Area out of size");
	if ((getRectWidth(pG2DB->bBoxArea) != (pHead->x2 - pHead->x1)) || 
			(getRectHeight(pG2DB->bBoxArea) != (pHead->y2 - pHead->y1)))
		Cant_happen("G2Dc_head_to_base: size mismatch");
#endif

	for (iny = pHead->y1, outy = pG2DB->bBoxArea.top; 
			 iny < pHead->y2; iny++, outy++)
	{
		pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
														 pG2DB->bSizeImg.cx * outy]);
		pIn = &(pHead->bits[pHead->x1 + pHead->width * iny]);
		for (x = pHead->x1; x < pHead->x2; x++)
		{
			if (*pIn || !ignoreNull)
				*pOut = pPal[*pIn];
			pOut++;
			pIn++;
		}
	}
}

void	G2Dr_head_to_base(const g2dbase *pG2DB, 
											  const struct tag_ss_bitmap_header *pHead,
											  pcolor *pPal)
{
	HSHADE shColor;
	pcolor *pOut;
	byte *pIn;
	uint16 iny, outy, x;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dr_head_to_base: Area out of size");
	if ((getRectWidth(pG2DB->bBoxArea) != (pHead->x2 - pHead->x1)) || 
			(getRectHeight(pG2DB->bBoxArea) != (pHead->y2 - pHead->y1)))
		Cant_happen("G2Dr_head_to_base: size mismatch");
#endif

	for (iny = pHead->y1, outy = pG2DB->bBoxArea.top; 
			 iny < pHead->y2; iny++, outy++)
	{
		pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
														 pG2DB->bSizeImg.cx * outy]);
		pIn = &(pHead->bits[pHead->x1 + pHead->width * iny]);
		for (x = pHead->x1; x < pHead->x2; x++)
		{
			shColor = shade_get(get_trans(pPal[*pIn]));
			*pOut = shade_color(shColor, *pOut, pPal[*pIn]);
			pOut++;
			pIn++;
		}
	}
}

void	G2Dc_pal_to_clr(g2dbase *pDest, g2dbase *pSrc)
{
	byte *pIn;
	pcolor *pOut;
	uint16 iny, outy, x;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dc_pal_to_clr: Area out of size");
	if ((getRectWidth(pSrc->bBoxArea) != getRectWidth(pDest->bBoxArea)) || 
			(getRectHeight(pSrc->bBoxArea) != getRectWidth(pDest->bBoxArea)))
		Cant_happen("G2Dc_pal_to_clr: Size mismatch");
#endif

	for (outy = pDest->bBoxArea.top, iny = pSrc->bBoxArea.top; 
			 outy < pDest->bBoxArea.bottom; iny++, outy++)
	{
		pIn = &(pSrc->bImg.bit[pSrc->bBoxArea.left + 
													 pSrc->bSizeImg.cx * iny]);
		pOut = &(pDest->bImg.clr[pDest->bBoxArea.left + 
														 pDest->bSizeImg.cx * outy]);
		for (x = pDest->bBoxArea.left; x < pDest->bBoxArea.right; x++)
		{
			*pOut = pSrc->pPal[*pIn];
			pOut++;
			pIn++;
		}
	}
}

void	G2D_base_to_head(const struct tag_ss_bitmap_header *pHead, 
											 const g2dbase *pG2DB, pcolor *pPal)
{
	byte *pOut = pHead->bits;
	pcolor *pIn;
	uint16 iny, outy, x;

	UNUSED(pPal);

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2D_base_to_head: Area out of size");
	if ((getRectWidth(pG2DB->bBoxArea) != pHead->width) || 
			(getRectHeight(pG2DB->bBoxArea) != pHead->height))
		Cant_happen("G2D_base_to_head: Size mismatch");
#endif

	for (outy = 0, iny = pG2DB->bBoxArea.top; outy < pHead->height; iny++, outy++)
	{
		pIn = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
														pG2DB->bSizeImg.cx * outy]);
		for (x = 0; x < pHead->width; x++)
		{
			*pOut = (Pal_rgb_to_index(*pIn));
			pOut++;
			pIn++;
		}
	}
}

#if BITS_PER_PIXEL < 32

void	G2D_pixel_to_color(const g2dbase *pTarget, const g2dbase *pSrc,
												 pcolor *pPal)
{
	pcolor *pOut;
	pixel *pIn;
	uint16 iny, outy, x;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2D_pixel_to_color: Source Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2D_pixel_to_color: Target Area out of size");
	if (!isBoxSameBox(pTarget->bBoxArea, pSrc->bBoxArea))
		Cant_happen("G2D_pixel_to_color: Area mismatch");
#endif

	for (iny = pSrc->bBoxArea.top, outy = pTarget->bBoxArea.top; 
			 iny < pSrc->bBoxArea.bottom; iny++, outy++)
	{
		pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left + 
														   pTarget->bSizeImg.cx * outy]);
		pIn = &(pSrc->bImg.pxl[pSrc->bBoxArea.left + 
													 pSrc->bSizeImg.cx * iny]);
		for (x = pSrc->bBoxArea.left; x < pSrc->bBoxArea.right; x++)
		{
			*pOut = pPal[*pIn];
			pOut++;
			pIn++;
		}
	}
}

void	G2D_color_to_pixel(const g2dbase *pTarget, const g2dbase *pSrc,
												 pcolor *pPal)
{
	pcolor *pIn;
	pixel *pOut;
	uint16 iny, outy, x;

	UNUSED(pPal);

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2D_color_to_pixel: Source Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2D_color_to_pixel: Target Area out of size");
	if (!isBoxSameBox(pTarget->bBoxArea, pSrc->bBoxArea))
		Cant_happen("G2D_color_to_pixel: Area mismatch");
#endif

	for (iny = pSrc->bBoxArea.top, outy = pTarget->bBoxArea.top; 
			 iny < pSrc->bBoxArea.bottom; iny++, outy++)
	{
		pOut = &(pTarget->bImg.pxl[pTarget->bBoxArea.left + 
														   pTarget->bSizeImg.cx * outy]);
		pIn = &(pSrc->bImg.clr[pSrc->bBoxArea.left + 
													 pSrc->bSizeImg.cx * iny]);
		for (x = pSrc->bBoxArea.left; x < pSrc->bBoxArea.right; x++)
		{
			*pOut = Pal_rgb_to_pixel(*pIn);
			pOut++;
			pIn++;
		}
	}
}

#endif /* BITS_PER_PIXEL < 32 */ 

void	G2D_set_shades(const g2dbase *pG2DB, byte sh)
{
	byte *pByte;
	uint16 x, y;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2D_set_shades: Area out of size");
#endif

	for (y = pG2DB->bBoxArea.top; y < pG2DB->bBoxArea.bottom; y++)
	{
		pByte = (byte *)(&(pG2DB->bImg.clr[pG2DB->bBoxArea.left +
																			 pG2DB->bSizeImg.cx * y]));
		pByte += (C_SHIFT_TRANS >> 3);	/* i hope this works */
		for (x = pG2DB->bBoxArea.left; x < pG2DB->bBoxArea.right; x++)
		{
			*pByte = sh;
			pByte += 4;
		}
	}
}

void	G2D_add_shades(const g2dbase *pG2DB, int nOffset)
{
	int nTarget;
	byte *pByte;
	uint16 x, y;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2D_add_shades: Area out of size");
#endif

	for (y = pG2DB->bBoxArea.top; y < pG2DB->bBoxArea.bottom; y++)
	{
		pByte = (byte *)(&(pG2DB->bImg.clr[pG2DB->bBoxArea.left +
																			 pG2DB->bSizeImg.cx * y]));
		pByte += (C_SHIFT_TRANS >> 3);	
		for (x = pG2DB->bBoxArea.left; x < pG2DB->bBoxArea.right; x++)
		{
			nTarget = (int)(*pByte) + nOffset;
			*pByte = (byte)min(255, max(0, nTarget));
			pByte += 4;
		}
	}
}

void	G2D_set_mask(const g2dbase *pTarget, const g2dbase *pMask)
{
	byte *pByte;
	int *pIn;
	uint16 x, y;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pMask->bBoxArea, pMask->bSizeImg))
		Cant_happen("G2D_set_mask: Mask Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2D_set_mask: Target Area out of size");
	if (!isBoxSameBox(pTarget->bBoxArea, pMask->bBoxArea))
		Cant_happen("G2D_set_mask: Area mismatch");
#endif

	for (y = pTarget->bBoxArea.top; y < pTarget->bBoxArea.bottom; y++)
	{
		pByte = (byte *)(&(pTarget->bImg.clr[pTarget->bBoxArea.left +
																			   pTarget->bSizeImg.cx * y]));
		pByte += (C_SHIFT_TRANS >> 3);	/* i hope this works */
		pIn = (&(pMask->bImg.msk[pMask->bBoxArea.left +
														 pMask->bSizeImg.cx * y]));
		for (x = pTarget->bBoxArea.left; x < pTarget->bBoxArea.right; x++)
		{
			*pByte = (byte)(*pIn);
			pByte += 4;
			pIn++;
		}
	}
}

void	G2Dm_clip_mask(const g2dbase *pG2DB)
{
	int *pOut;
	uint16 x, y;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2D_clip_mask: Area out of size");
#endif

	for (y = pG2DB->bBoxArea.top; y < pG2DB->bBoxArea.bottom; y++)
	{
		pOut = (&(pG2DB->bImg.msk[pG2DB->bBoxArea.left +
															pG2DB->bSizeImg.cx * y]));
		for (x = pG2DB->bBoxArea.left; x < pG2DB->bBoxArea.right; x++)
		{
			*pOut = min(255, max(*pOut, 0));
			pOut++;
		}
	}
}

/* when not defined, map only draws seen tiles; */
/* ticker is for flag output (if desired)       */
/*
#define D_MAP_DEBUG
int debug_ticker = 0;
*/

/* palette indizes for map buffer																		*/

/* lowest priority; will be drawn over by walls/objects for sure		*/
#define C_MAP_PAL_VOID	0x00		/* clear							*/
#define	C_MAP_PAL_BIO		0x01		/* biohazard					*/
#define	C_MAP_PAL_RAD		0x02		/* radiation					*/
#define	C_MAP_PAL_ELEV	0x03		/* elevator						*/
#define C_MAP_PAL_DEBUG	0x04		/* for debugging			*/
/* walls - from open (no wall) to solid															*/
#define	C_MAP_PAL_BG		0x10		/* normal background	*/
#define C_MAP_PAL_START	0x10		/* start for open wall shades */
				/* the difference between those two values determines the		*/
				/* max available shades; 0xC0 is best since the background	*/
				/* is calculated >> 2  a max of FF (or 0x100) >> 2 -> 0xC0	*/
#define C_MAP_PAL_WALL	0xD0		/* a solid wall								*/
/* objects on the map. shouldn't be drawn over by walls							*/
#define	C_MAP_PAL_DOOR	0xF0		/* door								*/
#define C_MAP_PAL_SEC		0xF1		/* security						*/
#define C_MAP_PAL_CRIT	0xF2		/* critter						*/
#define C_MAP_PAL_MODEL	0xFF		/* player model				*/

void	 G2D_draw_map(const g2dbase *pG2DB, 
										double orientX, double orientY,
										point center,
										byte zoom, byte radius, byte flags)
{
	ss_bitmap_header *pBmp = NULL;									/* the buffer */ 
	pcolor palBuf[0x100];														/* palette */
	point ptStart, ptOut;														/* x/y storage */
	Vector vNormNorth, vOutCenter, vInCenter, vVec, vIn; /* for rot */
	object_xref *pRef = NULL;												/* objects */
	map_object *pObj = NULL;
	object_prop *pProp = NULL;
	map_tile *pTile = NULL, 
					 *pTilePartner[4]; /* left, above, right, below */
	double wallSteps;
	int len = (zoom + 1), dialen = (int)(ceil(sqrt(len * len))), 
			objlenx = 0, objleny = 0, i, outx, outy, tilex, tiley, 
			bottom1, bottom2, top1, top2;
	pcolor *pOutRot;
	byte *pOut = NULL, clrType, wallShades, outerWall[4];

	/* init palette */
	memset(palBuf, 0x00, sizeof(pcolor) * 0x100);
	palBuf[C_MAP_PAL_VOID] = rgbt(0x00, 0x00, 0x00, 0x00);
	palBuf[C_MAP_PAL_BG] = rgbt(get_red(pG2DB->bOp.clr),	/* bg color */
															get_green(pG2DB->bOp.clr),
														  get_blue(pG2DB->bOp.clr),
														  get_trans(pG2DB->bOp.clr) >> 2);
	palBuf[C_MAP_PAL_BIO] = rgbt(0x87, 0x6B, 0x00, 
																get_trans(pG2DB->bOp.clr) >> 2);
	palBuf[C_MAP_PAL_RAD] = rgbt(0x73, 0x07, 0x07, 
																get_trans(pG2DB->bOp.clr) >> 2);
	palBuf[C_MAP_PAL_ELEV] = rgbt(0x5F, 0x37, 0x0B, 
																get_trans(pG2DB->bOp.clr) >> 2);
	palBuf[C_MAP_PAL_DEBUG] = rgbt(0xFF, 0x00, 0xFF,
																 get_trans(pG2DB->bOp.clr) >> 2);
	palBuf[C_MAP_PAL_CRIT] = rgbt(0x00, 0x80, 0xFF, 
																get_trans(pG2DB->bOp.clr) >> 2);
	palBuf[C_MAP_PAL_SEC] = rgbt(0xFF, 0x00, 0x00,
															 get_trans(pG2DB->bOp.clr));
	palBuf[C_MAP_PAL_DOOR] = rgbt(0xFF, 0xFF, 0x00,
																get_trans(pG2DB->bOp.clr));
	palBuf[C_MAP_PAL_MODEL] = rgbt(0x00, 0x00, 0xFF,
																 get_trans(pG2DB->bOp.clr));
	wallShades = get_trans(pG2DB->bOp.clr) - 
									get_trans(palBuf[C_MAP_PAL_BG]);
	wallSteps = 0.0;
	for (i = C_MAP_PAL_START; i <= C_MAP_PAL_WALL; i++)
	{
		palBuf[i] = 
				rgbt(get_red(pG2DB->bOp.clr),	
						 get_green(pG2DB->bOp.clr),
						 get_blue(pG2DB->bOp.clr),
						 get_trans(palBuf[C_MAP_PAL_BG]) +
								min(wallShades, (int)(wallSteps)));
		if (wallShades > 0.) /* jim - causes exception under Borland */
		{
		  wallSteps += ((double)(C_MAP_PAL_WALL - C_MAP_PAL_START) / 
				(double)(wallShades));
		}
	}
	palBuf[C_MAP_PAL_WALL] = pG2DB->bOp.clr; /* for security... */
	/* finished palette */
	/* make up buffer */
	i = (int)(sqrt(getRectWidth(pG2DB->bBoxArea) * 
									getRectWidth(pG2DB->bBoxArea) +
					       getRectHeight(pG2DB->bBoxArea) * 
									getRectHeight(pG2DB->bBoxArea)));
	pBmp = Bmp_create(i, i);

	memset(pBmp->bits, C_MAP_PAL_VOID, 
				 pBmp->width * pBmp->height * sizeof(byte));

	ptStart.x = center.x / TILE_GRID_SIZE * len +
							((center.x % TILE_GRID_SIZE) * 
							 (zoom + 1) / TILE_GRID_SIZE) - pBmp->width / 2;
	ptStart.y = center.y / TILE_GRID_SIZE * len +
							((center.y % TILE_GRID_SIZE) * 
							 (zoom + 1) / TILE_GRID_SIZE) - pBmp->height / 2;

#ifdef D_MAP_DEBUG
	debug_ticker++;
#endif
	/* for the whole map */
	for (tiley = 0; tiley < 64; tiley++)
	{
		for (tilex = 0; tilex < 64; tilex++)
		{
			/* get this tile */
			pTile = Map_get_tile((uint16)(tilex * TILE_GRID_SIZE), 
													 (uint16)(tiley * TILE_GRID_SIZE));
#ifdef D_MAP_DEBUG
if (!(debug_ticker % 20))
{
	if (tiley == (center.y / TILE_GRID_SIZE) &&
			tilex == (center.x / TILE_GRID_SIZE))
		printf("tiley: %02d tilex: %02d flags: %08X rflags: %08X states: %02X%02X%02X%02X\n"
					 "  floor: %04X ceiling: %04X slope: %02X\n", 
					 tiley, tilex, pTile->flags, pTile->rflags,
					 pTile->state[0], pTile->state[1], pTile->state[2], pTile->state[3],
					 pTile->floor, pTile->ceiling, pTile->slope);
}
#endif
#ifndef D_MAP_DEBUG
			if (pTile->flags & TF_AUTOMAPPED)
#endif
			{
				if (pTile->type != TILE_SOLID)	/* solids dont need work */
				{
					pTile->flags |= TF_AUTOMAPPED;	/* mapped it */
					/* get neighbours */
					pTilePartner[0] = Map_get_tile((uint16)((tilex - 1) * TILE_GRID_SIZE), 
																				 (uint16)(tiley * TILE_GRID_SIZE));
					pTilePartner[1] = Map_get_tile((uint16)(tilex * TILE_GRID_SIZE), 
																				 (uint16)((tiley + 1) * TILE_GRID_SIZE));
					pTilePartner[2] = Map_get_tile((uint16)((tilex + 1) * TILE_GRID_SIZE), 
																				 (uint16)(tiley * TILE_GRID_SIZE));
					pTilePartner[3] = Map_get_tile((uint16)(tilex * TILE_GRID_SIZE), 
																				 (uint16)((tiley - 1) * TILE_GRID_SIZE));

					/* determine the solid walls */
					outerWall[0] = outerWall[1] = outerWall[2] = outerWall[3] = 0;
					if (pTilePartner[0] == NULL ||
							pTilePartner[0]->type == TILE_SOLID ||
							pTilePartner[0]->type == TILE_DIAGONAL_NW ||
							pTilePartner[0]->type == TILE_DIAGONAL_SW)
						outerWall[0] = 0x01;
					if (pTilePartner[1] == NULL ||
							pTilePartner[1]->type == TILE_SOLID ||
							pTilePartner[1]->type == TILE_DIAGONAL_NE ||
							pTilePartner[1]->type == TILE_DIAGONAL_NW)
						outerWall[1] = 0x01;
					if (pTilePartner[2] == NULL ||
							pTilePartner[2]->type == TILE_SOLID ||
							pTilePartner[2]->type == TILE_DIAGONAL_NE ||
							pTilePartner[2]->type == TILE_DIAGONAL_SE)
						outerWall[2] = 0x01;
					if (pTilePartner[3] == NULL ||
							pTilePartner[3]->type == TILE_SOLID ||
							pTilePartner[3]->type == TILE_DIAGONAL_SE ||
							pTilePartner[3]->type == TILE_DIAGONAL_SW)
						outerWall[3] = 0x01;

					/* horizontal init */
					ptOut.x = tilex * len - ptStart.x;
					/* vertical init */
					ptOut.y = pBmp->height - (tiley * len - ptStart.y);
					/* start drawing background */
					if (flags & F_DRAWMAP_NORMBG)
						clrType = C_MAP_PAL_BG;
					else
						clrType = C_MAP_PAL_VOID;

/* do when done ;) */

					if (flags & F_DRAWMAP_RAD &&
							pTile->rflags & 0x000000008)
						clrType = C_MAP_PAL_RAD;
					if (flags & F_DRAWMAP_BIO &&
							pTile->rflags & 0x000000080)
						clrType = C_MAP_PAL_BIO;

					/* ...has nearly the best music ever... */
					if ((pTile->flags & 0x0000F000) >= 0x0000E000 &&
							flags & F_DRAWMAP_ELEV) /* test by music ;) */
						clrType = C_MAP_PAL_ELEV;
#ifdef D_MAP_DEBUG
					if (pTile->flags & 0x00000200)
						clrType = C_MAP_PAL_DEBUG;
#endif 
					if (clrType != C_MAP_PAL_VOID)
					{
						for (outy = 0; outy < len; outy++)
						{
							if ((ptOut.y - outy) >= 0 && (ptOut.y - outy) < pBmp->height)
							{
								pOut = &(pBmp->bits[ptOut.x + 
																		pBmp->width * (ptOut.y - outy)]);
								for (outx = 0; outx < len; outx++)
								{
									if ((ptOut.x + outx) >= 0 && (ptOut.x + outx) < pBmp->width)
									{
										if (!*pOut &&
												(pTile->type == TILE_OPEN ||
												 pTile->type > 0x05 ||
												 (pTile->type == TILE_DIAGONAL_SE && outy <= outx) ||
												 (pTile->type == TILE_DIAGONAL_NW && outx <= outy) ||
												 (pTile->type == TILE_DIAGONAL_NE && (len - outy) <= (outx)) ||
												 (pTile->type == TILE_DIAGONAL_SW && (outx) <= (len - outy))
												))
											*pOut = clrType;
									}
									pOut++;
								}
							}
						}
					}
					/* end background */
					/* start draw south wall */
					if (pTile->type != TILE_DIAGONAL_NE &&
							pTile->type != TILE_DIAGONAL_NW &&
							ptOut.y >= 0 && ptOut.y < pBmp->height) 
					{
						clrType = C_MAP_PAL_WALL;
						if (!outerWall[3])
						{
							bottom1 = Map_tile_height(pTile, 
																				TILE_GRID_SIZE / 2, 0, false);
							bottom2 = Map_tile_height(pTilePartner[3],
																				TILE_GRID_SIZE / 2, 
																				TILE_GRID_SIZE, false);
							top1 = Map_tile_height(pTile, 
																		 TILE_GRID_SIZE / 2, 0, true);
							top2 = Map_tile_height(pTilePartner[3],
																		 TILE_GRID_SIZE / 2, 
																		 TILE_GRID_SIZE, true);
							if (bottom2 < top1 && bottom1 < top2)
							{
								clrType = C_MAP_PAL_START + 
											abs(bottom1 - bottom2) * (wallShades - 1) /
												(min(top1, top2) - min(bottom1, bottom2));
							}
						}
						if (clrType > C_MAP_PAL_BG)
						{
							pOut = &(pBmp->bits[ptOut.x + pBmp->width * ptOut.y]);
							for (i = 0; i < (len + 1); i++)
							{
								if ((ptOut.x + i) >= 0 && (ptOut.x + i) < pBmp->width &&
										clrType > *pOut)
									*pOut = clrType;
								pOut++;
							}
						}
					}
					/* end draw south wall */
					/* start diagonal se/nw */
					if (pTile->type == TILE_DIAGONAL_SE ||
							pTile->type == TILE_DIAGONAL_NW)
					{
						for (i = 0; i < dialen; i++)
						{
							if ((ptOut.y - i) >= 0 && (ptOut.y - i) < pBmp->height &&
									(ptOut.x + i) >= 0 && (ptOut.x + i) < pBmp->width)
							{
								pBmp->bits[ptOut.x + i + 
													 pBmp->width * (ptOut.y - i)] = C_MAP_PAL_WALL;
							}
						}
					}
					/* end diagonal */
					/* start draw north wall */
					ptOut.y -= len;
					if (pTile->type != TILE_DIAGONAL_SE &&
							pTile->type != TILE_DIAGONAL_SW &&
							ptOut.y >= 0 && ptOut.y < pBmp->height) 
					{
						clrType = C_MAP_PAL_WALL;
						if (!outerWall[1])
						{
							bottom1 = Map_tile_height(pTile, 
																			  TILE_GRID_SIZE / 2, 
																				TILE_GRID_SIZE, false);
							bottom2 = Map_tile_height(pTilePartner[1],
																				TILE_GRID_SIZE / 2, 0, false);
							top1 = Map_tile_height(pTile, 
																		 TILE_GRID_SIZE / 2, 
																		 TILE_GRID_SIZE, true);
							top2 = Map_tile_height(pTilePartner[1],
																		 TILE_GRID_SIZE / 2, 0, true);
							if (bottom2 < top1 && bottom1 < top2)
							{
								clrType = C_MAP_PAL_START + 
											abs(bottom1 - bottom2) * (wallShades - 1) /
												(min(top1, top2) - min(bottom1, bottom2));
							}
						}
						if (clrType > C_MAP_PAL_BG)
						{
							pOut = &(pBmp->bits[ptOut.x + pBmp->width * ptOut.y]);
							for (i = 0; i < len; i++)
							{
								if ((ptOut.x + i) >= 0 && (ptOut.x + i) < pBmp->width &&
										clrType > *pOut)
									*pOut = clrType;
								pOut++;
							}
						}
					}
					/* end draw north wall */
					/* start draw west wall */
					if (pTile->type != TILE_DIAGONAL_NE &&
							pTile->type != TILE_DIAGONAL_SE &&
							ptOut.x >= 0 && ptOut.x < pBmp->width) 
					{
						clrType = C_MAP_PAL_WALL;
						if (!outerWall[0])
						{
							bottom1 = Map_tile_height(pTile, 0,
																				TILE_GRID_SIZE / 2, false);
							bottom2 = Map_tile_height(pTilePartner[0], 
																				TILE_GRID_SIZE,
																				TILE_GRID_SIZE / 2, false);
							top1 = Map_tile_height(pTile, 0,
																		 TILE_GRID_SIZE / 2, true);
							top2 = Map_tile_height(pTilePartner[0], 
																		 TILE_GRID_SIZE,
																		 TILE_GRID_SIZE / 2, true);
							if (bottom2 < top1 && bottom1 < top2)
							{
								clrType = C_MAP_PAL_START + 
											abs(bottom1 - bottom2) * (wallShades - 1) /
												(min(top1, top2) - min(bottom1, bottom2));
							}
						}
						if (clrType > C_MAP_PAL_BG)
						{
							pOut = &(pBmp->bits[ptOut.x + pBmp->width * ptOut.y]);
							for (i = 0; i < len; i++)
							{
								if ((ptOut.y + i) >= 0 && (ptOut.y + i) < pBmp->height &&
										clrType > *pOut)
									*pOut = clrType;
								pOut += pBmp->width;
							}
						}
					}
					/* end draw west wall */
					/* start diagonal ne/sw */
					if (pTile->type == TILE_DIAGONAL_NE ||
							pTile->type == TILE_DIAGONAL_SW)
					{
						for (i = 0; i < dialen; i++)
						{
							if ((ptOut.y + i) >= 0 && (ptOut.y + i) < pBmp->height &&
									(ptOut.x + i) >= 0 && (ptOut.x + i) < pBmp->width)
							{
								pBmp->bits[ptOut.x + i + 
													 pBmp->width * (ptOut.y + i)] = C_MAP_PAL_WALL;
							}
						}
					}
					/* end diagonal */
					/* start draw east wall */
					ptOut.x += len;
					if (pTile->type != TILE_DIAGONAL_NW &&
							pTile->type != TILE_DIAGONAL_SW &&
							ptOut.x >= 0 && ptOut.x < pBmp->width) 
					{
						clrType = C_MAP_PAL_WALL;
						if (!outerWall[2])
						{
							bottom1 = Map_tile_height(pTile, TILE_GRID_SIZE,
																				TILE_GRID_SIZE / 2, false);
							bottom2 = Map_tile_height(pTilePartner[2], 0,
																				TILE_GRID_SIZE / 2, false);
							top1 = Map_tile_height(pTile, TILE_GRID_SIZE,
																		 TILE_GRID_SIZE / 2, true);
							top2 = Map_tile_height(pTilePartner[2], 0,
																		 TILE_GRID_SIZE / 2, true);
							if (bottom2 < top1 && bottom1 < top2)
							{
								clrType = C_MAP_PAL_START + 
											abs(bottom1 - bottom2) * (wallShades - 1) /
												(min(top1, top2) - min(bottom1, bottom2));
							}
						}
						if (clrType > C_MAP_PAL_BG)
						{
							pOut = &(pBmp->bits[ptOut.x + pBmp->width * ptOut.y]);
							for (i = 0; i < len; i++)
							{
								if ((ptOut.y + i) >= 0 && (ptOut.y + i) < pBmp->height &&
										clrType > *pOut)
									 *pOut = clrType;
								pOut += pBmp->width;
							}
						}
					}
					/* end draw east wall */
					/* start draw objects */
					pRef = Obj_get_reference(pTile->index);
					while (pRef != NULL)
					{
						pObj = Obj_get_object(pRef->index);
						pProp = Obj_get_property(pObj->object_id);

#ifdef D_MAP_DEBUG
if (!(debug_ticker % 20))
{
	if (tiley == (center.y / TILE_GRID_SIZE) &&
			tilex == (center.x / TILE_GRID_SIZE))
	{
		printf("%d, %d, %d\n"
					 " drawtype: %02d mass: %d flags: %08X stuff1: %d stuff3: %d\n", 
					 pObj->class, pObj->subclass, pObj->type,
					 pProp->draw_type, pProp->mass, pObj->flags, pObj->stuff1,
					 pObj->stuff3);
	}
}
#endif

						if (pProp->draw_type != OBJDRAW_INVISIBLE)
						{	/* even the map can't show it ;) */
							clrType = C_MAP_PAL_VOID;
							if (pObj->class == 0x0A)
							{	/* door/grating */		
/*
#ifdef D_MAP_DEBUG
if (!(debug_ticker % 20))
{
	if (tiley == (center.y / TILE_GRID_SIZE) &&
			tilex == (center.x / TILE_GRID_SIZE))
	printf("%g|%g - %g|%g - %g|%g\n", 
				 pObj->orientation.x.x, pObj->orientation.x.y,
				 pObj->orientation.y.x, pObj->orientation.y.y,
				 pObj->orientation.z.x, pObj->orientation.z.y);
}
#endif
*/
								if (abs((int)(floor(pObj->orientation.z.y))) != 1.0)
								{	/* nothing flat please */
									if (abs((int)(floor(pObj->orientation.x.y))) == 0.0)
									{
/* adjust len according to open state... */
										objlenx = len;
										objleny = 1;
									}
									else
									{
										objleny = len;
										objlenx = 1;
									}
									if (pObj->subclass == 0x00 || 
											pObj->subclass == 0x02 ||
											pObj->subclass == 0x03 || 
											pObj->subclass == 0x04)
										clrType = C_MAP_PAL_DOOR;
									if (pObj->subclass == 0x01 || /* grating */
											(pObj->subclass == 0x04 && 
											 (pObj->type == 0x06 || /* hidden door */
												pObj->type == 0x07 || /* hidden door 2 */
												pObj->type == 0x09))) /* space windows */
										clrType = C_MAP_PAL_WALL;
								}
							}
							if (flags & F_DRAWMAP_SEC)
							{
								if (pObj->class == 0x07 && pObj->subclass == 0x00 &&
										(pObj->type == 0x05 || pObj->type == 0x08))
								{	/* sec node big / computer */
									objlenx = objleny = len;
									clrType = C_MAP_PAL_SEC;
								}
								if (pObj->class == 0x07 && pObj->subclass == 0x05 &&
										pObj->type == 0x04)
								{	/* cam */
									objlenx = objleny = max(1, len / 4);
									clrType = C_MAP_PAL_SEC;
								}
							}
							if (flags & F_DRAWMAP_CRIT && pObj->class == 0x0E &&
									radius * TILE_GRID_SIZE >= i)
							{	/* critter */
								i = (int)sqrt(((int)(pObj->pos.x) - center.x) *
																((int)(pObj->pos.x) - center.x) +
															((int)(pObj->pos.y) - center.y) *
																((int)(pObj->pos.y) - center.y));
								if (i <= radius * TILE_GRID_SIZE &&
										(pObj->subclass != 0x02 || pObj->type != 0x04))
								{	/* within radius and no elite cyborg :) */
									objlenx = objleny = max(1, len * 2 / 3);
									clrType = C_MAP_PAL_CRIT;
								}
							}
							if (clrType)
							{
								ptOut.x = ((int)(pObj->pos.x) % TILE_GRID_SIZE) * 
													len / TILE_GRID_SIZE;
								ptOut.x += ((int)(pObj->pos.x / TILE_GRID_SIZE) * len) - 
													 ptStart.x - (objlenx / 2);
								ptOut.y = ((int)(pObj->pos.y) % TILE_GRID_SIZE) * 
													-len / TILE_GRID_SIZE;
								ptOut.y += pBmp->height - 
													 (((int)(pObj->pos.y / TILE_GRID_SIZE) * len) - 
													 ptStart.y + (objleny / 2));
								if (ptOut.x >= 0 && ptOut.x < pBmp->width &&
										ptOut.y >= 0 && ptOut.y < pBmp->height)
								{
									for (outy = 0; outy < objleny; outy++)
									{
										if ((ptOut.y + outy) >= 0 && 
												(ptOut.y + outy) < pBmp->height)
										{
											pOut = &(pBmp->bits[ptOut.x + 
																					pBmp->width * (ptOut.y + outy)]);
											for (outx = 0; outx < objlenx; outx++)
											{
												if ((ptOut.x + outx) >= 0 && 
														(ptOut.x + outx) < pBmp->width &&
														*pOut < clrType)
													*pOut = clrType;
												pOut++;
											}
										}
									}
								}
							}
						} /* not invisible */

						pRef = Obj_get_reference(pRef->obj_chain);
					}	/* while objref */
				} /* not solid */
			}	/* seen tile*/
		} /* for tilex */
	}	/* for tiley */

	vInCenter.x = (double)(pBmp->width) / 2.0;
	vInCenter.y = (double)(pBmp->height) / 2.0;

	if (flags & F_DRAWMAP_MODEL)
	{	/* last finishing touches */
		/* ff1: if we'll do multiplayer support a player model	*/
		/*			will surely be an object on the map; if so, move*/
		/*			the code there...																*/
/* do me better */
		vVec = vInCenter;
		for (i = 0; i < (len / 2); i++)
		{
			ptOut.x = (int)(vVec.x);
			ptOut.y = (int)(vVec.y);
			if (ptOut.y >= 0 && ptOut.y < pBmp->height &&
					ptOut.x >= 0 && ptOut.x < pBmp->width)
				pBmp->bits[ptOut.x + pBmp->width * ptOut.y] = C_MAP_PAL_MODEL;
			vVec.x -= Camera->orientation.x.y;
			vVec.y -= Camera->orientation.x.x;
		}
	}
	/* now for the real cool thing :) */
	vNormNorth.x = orientX;
	vNormNorth.y = orientY;
	vNormNorth.z = 0.0;
	vec_normalize(vNormNorth);

	vOutCenter.x = (double)(getRectWidth(pG2DB->bBoxArea)) / 2.0;
	vOutCenter.y = (double)(getRectHeight(pG2DB->bBoxArea)) / 2.0;

	for (outy = 0; outy < getRectHeight(pG2DB->bBoxArea); outy++)
	{
		pOutRot = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left +
																pG2DB->bSizeImg.cx * 
																	(pG2DB->bBoxArea.top + outy)]);
		for (outx = 0; outx < getRectWidth(pG2DB->bBoxArea); outx++)
		{
			/* move to 0|0 */
			vVec.x = outx - vOutCenter.x;
			vVec.y = outy - vOutCenter.y;
			/* rotate according to North */
			vIn.x = vVec.y * vNormNorth.y + vVec.x * vNormNorth.x;
			vIn.y = vVec.y * vNormNorth.x - vVec.x * vNormNorth.y;
			/* align to center again*/
			vIn.x += vInCenter.x;
			vIn.y += vInCenter.y;
			if (vIn.x >= 0.0 && vIn.x < pBmp->width &&
					vIn.y >= 0.0 && vIn.y < pBmp->height)
			{
				i = (int)(vIn.x) + pBmp->width * (int)(vIn.y);
				if (pBmp->bits[i] != C_MAP_PAL_VOID)
					*pOutRot = palBuf[pBmp->bits[i]];
			}
			pOutRot++;
		}
	}

	if (flags & F_DRAWMAP_RADIUS)
	{
		ptOut.x = getRectWidth(pG2DB->bBoxArea) / 2;
		ptOut.y = getRectHeight(pG2DB->bBoxArea) / 2;
/* daargh - still not working right... */
		G2Dc_draw_circle(pG2DB, ptOut, (uint16)(len * radius), 0.0, 360.0);
/**/
	}

	/* be nice */
	Bmp_free_bitmap(pBmp);
}

void	G2Dc_offset_area(const g2dbase *pG2DB, size off)
{
	g2dbase gbBuffer, gbMirror;
	bounding_box origBox = pG2DB->bBoxArea;
	size trueoff;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2D_offset_area: Area out of size");
#endif

	G2D_init_clrbase(&gbBuffer, 
									 (uint16)getRectWidth(pG2DB->bBoxArea),
									 (uint16)getRectHeight(pG2DB->bBoxArea));

	trueoff.cx = off.cx % pG2DB->bBoxArea.right;
	trueoff.cy = off.cy % pG2DB->bBoxArea.bottom;
	/* create mirror struct */
	memcpy(&gbMirror, pG2DB, sizeof(g2dbase));
	if (trueoff.cy > 0)
	{
		/* copy image */
		G2Dc_merge_areas(&gbBuffer, &gbMirror);	
		/* flip lower part */	
		gbBuffer.bBoxArea.top = gbBuffer.bBoxArea.bottom - trueoff.cy;
		gbMirror.bBoxArea.bottom = gbMirror.bBoxArea.top + trueoff.cy;
		G2Dc_merge_areas(&gbMirror, &gbBuffer);
		/* flip upper part */		
		gbBuffer.bBoxArea.top = 0;
		gbBuffer.bBoxArea.bottom = getRectHeight(gbBuffer.bBoxArea) - trueoff.cy;
		gbMirror.bBoxArea.top = origBox.top + trueoff.cy;
		gbMirror.bBoxArea.bottom = origBox.bottom;
		G2Dc_merge_areas(&gbMirror, &gbBuffer);
		/* reset values */
		gbMirror.bBoxArea = origBox;
		gbBuffer.bBoxArea.top = 0;
		gbBuffer.bBoxArea.bottom = gbBuffer.bSizeImg.cy;
	}
	if (trueoff.cx > 0)
	{
		/* copy image (again) */
		G2Dc_merge_areas(&gbBuffer, &gbMirror);	
		/* flip right part */	
		gbBuffer.bBoxArea.left = gbBuffer.bBoxArea.right - trueoff.cx;
		gbMirror.bBoxArea.right = gbMirror.bBoxArea.left + trueoff.cx;
		G2Dc_merge_areas(&gbMirror, &gbBuffer);
		/* flip left part */		
		gbBuffer.bBoxArea.left = 0;
		gbBuffer.bBoxArea.right = getRectWidth(gbBuffer.bBoxArea) - trueoff.cx;
		gbMirror.bBoxArea.left = origBox.left + trueoff.cx;
		gbMirror.bBoxArea.right = origBox.right;
		G2Dc_merge_areas(&gbMirror, &gbBuffer);
		/* reset values (?) */
		gbMirror.bBoxArea = origBox;
		gbBuffer.bBoxArea.left = 0;
		gbBuffer.bBoxArea.right = gbBuffer.bSizeImg.cx;
	}

	free(gbBuffer.bImg.clr);
}

void	G2Dr_enclose_area(const g2dbase *pG2DB, byte fx, byte fy)
{
	HSHADE shColor = shade_get(get_trans(pG2DB->bOp.clr));
	pcolor *pSide[4];
	int y, x, buf;
	byte truex = fx, truey = fy;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dr_enclose_area: Area out of size");
#endif

	/* clip percentage */	
	if (truex > 100)
		truex	= 100;
	if (truey > 100)
		truey = 100;
	/* the edges themselves; percentage 0 has no effect */
	pSide[0] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + /* top left */
													     pG2DB->bSizeImg.cx * pG2DB->bBoxArea.top]);
	pSide[1] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* top right */
													     pG2DB->bSizeImg.cx * pG2DB->bBoxArea.top]);
	pSide[2] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left +	/* bottom left */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 1)]);
	pSide[3] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* bottom right */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 1)]);
	*pSide[0] = shade_color(shColor, *pSide[0], pG2DB->bOp.clr);
	*pSide[1] = shade_color(shColor, *pSide[1], pG2DB->bOp.clr);
	*pSide[2] = shade_color(shColor, *pSide[2], pG2DB->bOp.clr);
	*pSide[3] = shade_color(shColor, *pSide[3], pG2DB->bOp.clr);
	/* the four horizontal parts */
	buf = getRectWidth(pG2DB->bBoxArea);
	pSide[0]++;
	pSide[1]--;
	pSide[2]++;
	pSide[3]--;
	for (x = 1; x < (buf * truex / 200); x++)
	{
		*pSide[0] = shade_color(shColor, *pSide[0], pG2DB->bOp.clr);
		*pSide[1] = shade_color(shColor, *pSide[1], pG2DB->bOp.clr);
		*pSide[2] = shade_color(shColor, *pSide[2], pG2DB->bOp.clr);
		*pSide[3] = shade_color(shColor, *pSide[3], pG2DB->bOp.clr);

		pSide[0]++;
		pSide[1]--;
		pSide[2]++;
		pSide[3]--;
	}
#if 0 /* doesn't look good in low res */
	if (truex && (buf % 2) && !(truex % 2)) /* one pixel at each side missing */
	{
		*pSide[0] = shade_color(shColor, *pSide[0], pG2DB->bOp.clr);	/* tl */
		*pSide[3] = shade_color(shColor, *pSide[2], pG2DB->bOp.clr);	/* br */
	}
#endif
	/* the four vertical parts */
	buf = getRectHeight(pG2DB->bBoxArea);
	pSide[0] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left +	/* left top */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.top + 1)]);
	pSide[1] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* right top */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.top + 1)]);
	pSide[2] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left +	/* left bottom */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 2)]);
	pSide[3] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* right bottom */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 2)]);
	for (y = 1; y < (buf * truey / 200); y++)
	{
		*pSide[0] = shade_color(shColor, *pSide[0], pG2DB->bOp.clr);
		*pSide[1] = shade_color(shColor, *pSide[1], pG2DB->bOp.clr);
		*pSide[2] = shade_color(shColor, *pSide[2], pG2DB->bOp.clr);
		*pSide[3] = shade_color(shColor, *pSide[3], pG2DB->bOp.clr);

		pSide[0] += pG2DB->bSizeImg.cx;
		pSide[1] += pG2DB->bSizeImg.cx;
		pSide[2] -= pG2DB->bSizeImg.cx;
		pSide[3] -= pG2DB->bSizeImg.cx;
	}
#if 0 /* doesn't look good in low res */
	if (truey && (buf % 2) && !(truey % 2)) /* one pixel at each side missing */
	{
		*pSide[0] = shade_color(shColor, *pSide[0], pG2DB->bOp.clr); /* lt */
		*pSide[3] = shade_color(shColor, *pSide[3], pG2DB->bOp.clr);	/* rb */
	}
#endif
}

void	G2Dc_enclose_area(const g2dbase *pG2DB, byte fx, byte fy)
{
	pcolor *pSide[4];
	int y, x, buf;
	byte truex = fx, truey = fy;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dc_enclose_area: Area out of size");
#endif

	/* clip percentage */	
	if (truex > 100)
		truex	= 100;
	if (truey > 100)
		truey = 100;
	/* the edges themselves; percentage 0 has no effect */
	pSide[0] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + /* top left */
													     pG2DB->bSizeImg.cx * pG2DB->bBoxArea.top]);
	pSide[1] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* top right */
													     pG2DB->bSizeImg.cx * pG2DB->bBoxArea.top]);
	pSide[2] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left +	/* bottom left */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 1)]);
	pSide[3] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* bottom right */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 1)]);
	*pSide[0] = pG2DB->bOp.clr;
	*pSide[1] = pG2DB->bOp.clr;
	*pSide[2] = pG2DB->bOp.clr;
	*pSide[3] = pG2DB->bOp.clr;
	/* the four horizontal parts */
	buf = getRectWidth(pG2DB->bBoxArea);
	pSide[0]++;
	pSide[1]--;
	pSide[2]++;
	pSide[3]--;
	for (x = 1; x < (buf * truex / 200); x++)
	{
		*pSide[0] = pG2DB->bOp.clr;
		*pSide[1] = pG2DB->bOp.clr;
		*pSide[2] = pG2DB->bOp.clr;
		*pSide[3] = pG2DB->bOp.clr;

		pSide[0]++;
		pSide[1]--;
		pSide[2]++;
		pSide[3]--;
	}
#if 0 /* doesn't look good in low res */
	if (truex && (buf % 2) && !(truex % 2)) /* one pixel at each side missing */
	{
		*pSide[0] = pG2DB->bOp.clr;	/* tl */
		*pSide[3] = pG2DB->bOp.clr;	/* br */
	}
#endif
	/* the four vertical parts */
	buf = getRectHeight(pG2DB->bBoxArea);
	pSide[0] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left +	/* left top */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.top + 1)]);
	pSide[1] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* right top */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.top + 1)]);
	pSide[2] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left +	/* left bottom */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 2)]);
	pSide[3] = &(pG2DB->bImg.clr[pG2DB->bBoxArea.right - 1 + /* right bottom */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 2)]);
	for (y = 1; y < (buf * truey / 200); y++)
	{
		*pSide[0] = pG2DB->bOp.clr;
		*pSide[1] = pG2DB->bOp.clr;
		*pSide[2] = pG2DB->bOp.clr;
		*pSide[3] = pG2DB->bOp.clr;

		pSide[0] += pG2DB->bSizeImg.cx;
		pSide[1] += pG2DB->bSizeImg.cx;
		pSide[2] -= pG2DB->bSizeImg.cx;
		pSide[3] -= pG2DB->bSizeImg.cx;
	}
#if 0 /* doesn't look good in low res */
	if (truey && (buf % 2) && !(truey % 2)) /* one pixel at each side missing */
	{
		*pSide[0] = pG2DB->bOp.clr; /* lt */
		*pSide[3] = pG2DB->bOp.clr; /* rb */
	}
#endif
}

void	G2Dm_enclose_area(const g2dbase *pG2DB, byte fx, byte fy)
{
	int *pSide[4];
	int y, x, buf;
	byte truex = fx, truey = fy;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dm_enclose_area: Area out of size");
#endif

	/* clip percentage */	
	if (truex > 100)
		truex	= 100;
	if (truey > 100)
		truey = 100;
	/* the edges themselves; percentage 0 has no effect */
	pSide[0] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.left + /* top left */
													     pG2DB->bSizeImg.cx * pG2DB->bBoxArea.top]);
	pSide[1] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.right - 1 + /* top right */
													     pG2DB->bSizeImg.cx * pG2DB->bBoxArea.top]);
	pSide[2] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.left +	/* bottom left */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 1)]);
	pSide[3] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.right - 1 + /* bottom right */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 1)]);
	*pSide[0] += pG2DB->bOp.trn;
	*pSide[1] += pG2DB->bOp.trn;
	*pSide[2] += pG2DB->bOp.trn;
	*pSide[3] += pG2DB->bOp.trn;
	/* the four horizontal parts */
	buf = getRectWidth(pG2DB->bBoxArea);
	pSide[0]++;
	pSide[1]--;
	pSide[2]++;
	pSide[3]--;
	for (x = 1; x < (buf * truex / 200); x++)
	{
		*pSide[0] += pG2DB->bOp.trn;
		*pSide[1] += pG2DB->bOp.trn;
		*pSide[2] += pG2DB->bOp.trn;
		*pSide[3] += pG2DB->bOp.trn;

		pSide[0]++;
		pSide[1]--;
		pSide[2]++;
		pSide[3]--;
	}
#if 0 /* doesn't look good in low res */
	if (truex && (buf % 2) && !(truex % 2)) /* one pixel at each side missing */
	{
		*pSide[0] += pG2DB->bOp.trn;	/* tl */
		*pSide[3] += pG2DB->bOp.trn;	/* br */
	}
#endif
	/* the four vertical parts */
	buf = getRectHeight(pG2DB->bBoxArea);
	pSide[0] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.left +	/* left top */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.top + 1)]);
	pSide[1] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.right - 1 + /* right top */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.top + 1)]);
	pSide[2] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.left +	/* left bottom */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 2)]);
	pSide[3] = &(pG2DB->bImg.msk[pG2DB->bBoxArea.right - 1 + /* right bottom */
													     pG2DB->bSizeImg.cx * (pG2DB->bBoxArea.bottom - 2)]);
	for (y = 1; y < (buf * truey / 200); y++)
	{
		*pSide[0] += pG2DB->bOp.trn;
		*pSide[1] += pG2DB->bOp.trn;
		*pSide[2] += pG2DB->bOp.trn;
		*pSide[3] += pG2DB->bOp.trn;

		pSide[0] += pG2DB->bSizeImg.cx;
		pSide[1] += pG2DB->bSizeImg.cx;
		pSide[2] -= pG2DB->bSizeImg.cx;
		pSide[3] -= pG2DB->bSizeImg.cx;
	}
#if 0 /* doesn't look good in low res */
	if (truey && (buf % 2) && !(truey % 2)) /* one pixel at each side missing */
	{
		*pSide[0] += pG2DB->bOp.trn; /* lt */
		*pSide[3] += pG2DB->bOp.trn; /* rb */
	}
#endif
}

void	 G2Dr_merge_areas(const g2dbase *pTarget, const g2dbase *pSrc)
{
	HSHADE shColor;
	pcolor *pOut, *pTrueIn;
	byte *pPalIn;
	int outy, iny, x;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dr_merge_areas: Src Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2Dr_merge_areas: Target Area out of size");
	if (!isBoxSameBox(pTarget->bBoxArea, pSrc->bBoxArea))
		Cant_happen("G2Dr_merge_areas: Area mismatch");
#endif

	switch (pSrc->nType)
	{
		case G2DTrue:
		{
			for (outy = pTarget->bBoxArea.top, iny = pSrc->bBoxArea.top; 
					 iny < pSrc->bBoxArea.bottom; outy++, iny++)
			{
				pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left +
																	 pTarget->bSizeImg.cx * outy]);
				pTrueIn = &(pSrc->bImg.clr[pSrc->bBoxArea.left +
															 pSrc->bSizeImg.cx * iny]);
				for(x = pSrc->bBoxArea.left; x < pSrc->bBoxArea.right; x++)
				{
					shColor = shade_get(get_trans(*pTrueIn));
					*pOut = shade_color(shColor, *pOut, *pTrueIn);
					pOut++;
					pTrueIn++;
				}
			}
			break;
		}
		case G2DPal:
		{
			for (outy = pTarget->bBoxArea.top, iny = pSrc->bBoxArea.top; 
					 iny < pSrc->bBoxArea.bottom; outy++, iny++)
			{
				pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left +
																	 pTarget->bSizeImg.cx * outy]);
				pPalIn = &(pSrc->bImg.bit[pSrc->bBoxArea.left +
																	pSrc->bSizeImg.cx * iny]);
				for(x = pSrc->bBoxArea.left; x < pSrc->bBoxArea.right; x++)
				{
					shColor = shade_get(get_trans(pSrc->pPal[*pPalIn]));
					*pOut = shade_color(shColor, *pOut, pSrc->pPal[*pPalIn]);
					pOut++;
					pPalIn++;
				}
			}
			break;
		}
		default:
		{
			break;
		}
	}
}

void	 G2Dc_merge_areas(const g2dbase *pTarget, const g2dbase *pSrc)
{
	pcolor *pOut, *pIn;
	int iny;
	uint16 linesize = (pSrc->bBoxArea.right - pSrc->bBoxArea.left) * 
										sizeof(pcolor);

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dc_merge_areas: Src Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2Dc_merge_areas: Target Area out of size");
	if (!isBoxSameBox(pTarget->bBoxArea, pSrc->bBoxArea))
		Cant_happen("G2Dc_merge_areas: Area mismatch");
#endif

	pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left +
														 pTarget->bSizeImg.cx * pTarget->bBoxArea.top]);
	pIn = &(pSrc->bImg.clr[pSrc->bBoxArea.left +
												 pSrc->bSizeImg.cx * pSrc->bBoxArea.top]);
	for (iny = pSrc->bBoxArea.top; iny < pSrc->bBoxArea.bottom; iny++)
	{
		memmove(pOut, pIn, linesize);
		pOut += pTarget->bSizeImg.cx;
		pIn += pSrc->bSizeImg.cx;
	}
}

void	G2Dm_merge_areas(const g2dbase *pTarget, const g2dbase *pSrc)
{
	int *pOut, *pIn;
	int outy, iny, x;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dm_merge_areas: Src Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2Dm_merge_areas: Target Area out of size");
	if (!isBoxSameBox(pTarget->bBoxArea, pSrc->bBoxArea))
		Cant_happen("G2Dm_merge_areas: Area mismatch");
#endif

	for (outy = pTarget->bBoxArea.top, iny = pSrc->bBoxArea.top; 
			 iny < pSrc->bBoxArea.bottom; outy++, iny++)
	{
		pOut = &(pTarget->bImg.msk[pTarget->bBoxArea.left +
															 pTarget->bSizeImg.cx * outy]);
		pIn = &(pSrc->bImg.msk[pSrc->bBoxArea.left +
													 pSrc->bSizeImg.cx * iny]);
		for(x = pSrc->bBoxArea.left; x < pSrc->bBoxArea.right; x++)
		{
			*pOut = *pOut + *pIn;
			pOut++;
			pIn++;
		}
	}
}

void	G2Dc_merge_heads(const struct tag_ss_bitmap_header *pTarget,
											 const struct tag_ss_bitmap_header *pSrc)
{
	byte *pOut, *pIn;
	int iny, inx, outy;
/*  	uint16 linesize = (pSrc->x2 - pSrc->x1) * sizeof(byte); */

	for (iny = pSrc->y1, outy = pTarget->y1; iny < pSrc->y2; iny++, outy++)
	{
		pOut = &(pTarget->bits[pTarget->x1 + pTarget->width * outy]);
		pIn = &(pSrc->bits[pSrc->x1 + pSrc->width * iny]);
		for (inx = pSrc->x1; inx < pSrc->x2; inx++)
		{
			if (*pIn)
				*pOut = *pIn;
			pIn++;
			pOut++;
		}
	}
}

void	G2Dr_stretch_area(const g2dbase *pTarget, const g2dbase *pSrc)
{
	HSHADE shColor;
	pcolor *pOut, *pInClr;
	byte *pInBit;
	int outx, outy;
	double inx, iny, 
				 offx = (double)(getRectWidth(pTarget->bBoxArea)), 
				 offy = (double)(getRectHeight(pTarget->bBoxArea));

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dr_stretch_area: Src Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2Dr_stretch_area: Target Area out of size");
#endif
	
	if (offx != 0.0)
		offx = (double)(getRectWidth(pSrc->bBoxArea)) / offx;
	if (offy != 0.0)
		offy = (double)(getRectHeight(pSrc->bBoxArea)) / offy;
	switch (pSrc->nType)
	{
		case G2DTrue:
		{
			for (outy = pTarget->bBoxArea.top, iny = (double)(pSrc->bBoxArea.top); 
					 outy < pTarget->bBoxArea.bottom; outy++, iny += offy)
			{
				pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left +
																	 pTarget->bSizeImg.cx * outy]);
				for (outx = pTarget->bBoxArea.left, inx = (double)(pSrc->bBoxArea.left);
						 outx < pTarget->bBoxArea.right; outx++, inx += offx)		
				{
					pInClr = &(pSrc->bImg.clr[(int)(inx) + 
																		pSrc->bSizeImg.cx * (int)(iny)]);
					shColor = shade_get(get_trans(*pInClr));
					*pOut = shade_color(shColor, *pOut, *pInClr);
					pOut++;
				}
			}
			break;
		}
		case G2DPal:
		{
			for (outy = pTarget->bBoxArea.top, iny = (double)(pSrc->bBoxArea.top); 
					 outy < pTarget->bBoxArea.bottom; outy++, iny += offy)
			{
				pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left +
																	 pTarget->bSizeImg.cx * outy]);
				for (outx = pTarget->bBoxArea.left, inx = (double)(pSrc->bBoxArea.left);
						 outx < pTarget->bBoxArea.right; outx++, inx += offx)		
				{
					pInBit = &(pSrc->bImg.bit[(int)(inx) + 
																		pSrc->bSizeImg.cx * (int)(iny)]);
					shColor = shade_get(get_trans(pSrc->pPal[*pInBit]));
					*pOut = shade_color(shColor, *pOut, pSrc->pPal[*pInBit]);
					pOut++;
				}
			}
			break;
		}
		default:	break;
	}
}

void	G2Dc_stretch_area(const g2dbase *pTarget, const g2dbase *pSrc)
{
	pcolor *pOut;
	int outx, outy;
	double inx, iny, 
				 offx = (double)(getRectWidth(pTarget->bBoxArea)), 
				 offy = (double)(getRectHeight(pTarget->bBoxArea));

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dc_stretch_area: Src Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2Dc_stretch_area: Target Area out of size");
#endif

	if (offx != 0.0)
		offx = (double)(getRectWidth(pSrc->bBoxArea)) / offx;
	if (offy != 0.0)
		offy = (double)(getRectHeight(pSrc->bBoxArea)) / offy;
	for (outy = pTarget->bBoxArea.top, iny = (double)(pSrc->bBoxArea.top); 
			 outy < pTarget->bBoxArea.bottom; outy++, iny += offy)
	{
		pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left +
															 pTarget->bSizeImg.cx * outy]);
		for (outx = pTarget->bBoxArea.left, inx = (double)(pSrc->bBoxArea.left);
				 outx < pTarget->bBoxArea.right; outx++, inx += offx)		
		{
			*pOut = (pSrc->bImg.clr[(int)(inx) + pSrc->bSizeImg.cx * (int)(iny)]);
			pOut++;
		}
	}
}

void	G2Dm_stretch_area(const g2dbase *pTarget, const g2dbase *pSrc)
{
	int *pOut;
	int outx, outy;
	double inx, iny, 
				 offx = (double)(getRectWidth(pTarget->bBoxArea)), 
				 offy = (double)(getRectHeight(pTarget->bBoxArea));

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dm_stretch_area: Src Area out of size");
	if (!isBoxInSize(pTarget->bBoxArea, pTarget->bSizeImg))
		Cant_happen("G2Dm_stretch_area: Target Area out of size");
#endif

	if (offx != 0.0)
		offx = (double)(getRectWidth(pSrc->bBoxArea)) / offx;
	if (offy != 0.0)
		offy = (double)(getRectHeight(pSrc->bBoxArea)) / offy;
	for (outy = pTarget->bBoxArea.top, iny = (double)(pSrc->bBoxArea.top); 
			 outy < pTarget->bBoxArea.bottom; outy++, iny += offy)
	{
		pOut = &(pTarget->bImg.msk[pTarget->bBoxArea.left +
															 pTarget->bSizeImg.cx * outy]);
		for (outx = pTarget->bBoxArea.left, inx = (double)(pSrc->bBoxArea.left);
				 outx < pTarget->bBoxArea.right; outx++, inx += offx)		
		{
			*pOut += (pSrc->bImg.msk[(int)(inx) + pSrc->bSizeImg.cx * (int)(iny)]);
			pOut++;
		}
	}
}

void	G2Dr_stretch_head(const g2dbase *pTarget, 
												const ss_bitmap_header *pHead,
												const pcolor *pPal)
{
	HSHADE shColor;
	pcolor *pOut;
	byte *pIn;
	int outx, outy;
	double inx, iny, 
				 offx = (double)(pHead->x2 - pHead->x1) / 
								(double)(getRectWidth(pTarget->bBoxArea)), 
				 offy = (double)(pHead->y2 - pHead->y1) / 
								(double)(getRectHeight(pTarget->bBoxArea));

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pSrc->bBoxArea, pSrc->bSizeImg))
		Cant_happen("G2Dr_stretch_area: Src Area out of size");
#endif

	for (outy = pTarget->bBoxArea.top, iny = (double)(pHead->y1); 
			 outy < pTarget->bBoxArea.bottom; outy++, iny += offy)
	{
		pOut = &(pTarget->bImg.clr[pTarget->bBoxArea.left +
															 pTarget->bSizeImg.cx * outy]);
		for (outx = pTarget->bBoxArea.left, inx = (double)(pHead->x1);
				 outx < pTarget->bBoxArea.right; outx++, inx += offx)		
		{
			pIn = &(pHead->bits[(int)(inx) + pHead->width * (int)(iny)]);
			shColor = shade_get(get_trans(pPal[*pIn]));
			*pOut = shade_color(shColor, *pOut, pPal[*pIn]);
			pOut++;
		}
	}
}

void	 G2Dr_fill_area(const g2dbase *pG2DB)
{
	HSHADE sh = shade_get(get_trans(pG2DB->bOp.clr));
	pcolor *pImg;
	uint16 x, y;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dr_fill_area: Area out of size");
#endif

	for(y = pG2DB->bBoxArea.top; y < pG2DB->bBoxArea.bottom; y++)
	{
		pImg = &pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
														(pG2DB->bSizeImg.cx * y)];
		for(x = pG2DB->bBoxArea.left; x < pG2DB->bBoxArea.right; x++)
		{
			*pImg = shade_color(sh, *pImg, pG2DB->bOp.clr);
			pImg++;
		}
	}
}

void	 G2Dc_fill_area(const g2dbase *pG2DB)
{
	pcolor *pImg;
	uint16 x, y;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dc_fill_area: Area out of size");
#endif

	for(y = pG2DB->bBoxArea.top; y < pG2DB->bBoxArea.bottom; y++)
	{
		pImg = &pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
														(pG2DB->bSizeImg.cx * y)];
		for(x = pG2DB->bBoxArea.left; x < pG2DB->bBoxArea.right; x++)
		{
			*pImg = pG2DB->bOp.clr;
			pImg++;
		}
	}
}

void	G2Dm_fill_area(const g2dbase *pG2DB)
{
	int *pOut;
	uint16 x, y;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dm_fill_rea: Area out of size");
#endif

	for(y = pG2DB->bBoxArea.top; y < pG2DB->bBoxArea.bottom; y++)
	{
		pOut = &pG2DB->bImg.msk[pG2DB->bBoxArea.left + 
														(pG2DB->bSizeImg.cx * y)];
		for(x = pG2DB->bBoxArea.left; x < pG2DB->bBoxArea.right; x++)
		{
			*pOut = *pOut + pG2DB->bOp.trn;
			pOut++;
		}
	}
}

void	 G2Dr_draw_box(const g2dbase *pG2DB, bounding_box b)
{
	HSHADE sh = shade_get(get_trans(pG2DB->bOp.clr));
	pcolor *pImg;
	bounding_box nBox, dBox; 
	uint16 x, y;
/* ff1: i'm playing with the thought to remove b and to 
   honour the area box of g2dbase (would also make the checks
   redundant)
*/
#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dr_draw_box: Area out of size");
#endif

	nBox.left = max(0, pG2DB->bBoxArea.left);
	nBox.top = max(0, pG2DB->bBoxArea.top);
	nBox.right = min(pG2DB->bSizeImg.cx, pG2DB->bBoxArea.right);
	nBox.bottom = min(pG2DB->bSizeImg.cy, pG2DB->bBoxArea.bottom);
	dBox.left = max(nBox.left, b.left);
	dBox.right = min(nBox.right, b.right);	
	dBox.top = max(nBox.top, b.top);
	dBox.bottom = min(nBox.bottom, b.bottom);

	if(b.left == dBox.left)
	{
		pImg = &pG2DB->bImg.clr[b.left + (dBox.top * pG2DB->bSizeImg.cx)];
		for(y = dBox.top; y < dBox.bottom; y++)
		{
			*pImg = shade_color(sh, *pImg, pG2DB->bOp.clr);
			pImg += pG2DB->bSizeImg.cx;
		}
	}
	if(b.right == dBox.right)
	{
		pImg = &pG2DB->bImg.clr[b.right - 1 + (dBox.top * pG2DB->bSizeImg.cx)];
		for(y = dBox.top; y < dBox.bottom; y++)
		{
			*pImg = shade_color(sh, *pImg, pG2DB->bOp.clr);
			pImg += pG2DB->bSizeImg.cx;
		}
	}
	if(b.top == dBox.top)
	{
		pImg = &pG2DB->bImg.clr[dBox.left + (dBox.top * pG2DB->bSizeImg.cx)];
		for(x = dBox.left; x < dBox.right; x++)
		{
			*pImg = shade_color(sh, *pImg, pG2DB->bOp.clr);
			pImg++;
		}
	}
	if(b.bottom == dBox.bottom)
	{
		pImg = &pG2DB->bImg.clr[dBox.left + 
								  ((dBox.bottom - 1) * pG2DB->bSizeImg.cx)];
		for(x = dBox.left; x < dBox.right; x++)
		{
			*pImg = shade_color(sh, *pImg, pG2DB->bOp.clr);
			pImg++;
		}		
	}
}

void	 G2Dc_draw_box(const g2dbase *pG2DB, bounding_box b)
{
	UNUSED(pG2DB);
	UNUSED(b);
}

void	 G2Dr_draw_line(const g2dbase *pG2DB, point p1, point p2)
{
	HSHADE shColor = shade_get(get_trans(pG2DB->bOp.clr));
	pcolor *pOut;
	int i, dx, dy, sdx, sdy, dxabs, dyabs, x, y, scanoff;

	dx		= p2.x - p1.x;                       
	dy		= p2.y - p1.y;                       
	dxabs = abs(dx);                             
	dyabs = abs(dy);
	sdx		= sgn(dx);                               
	sdy		= sgn(dy);
	x			= dyabs >> 1;                                
	y			= dxabs >> 1;
	scanoff = pG2DB->bSizeImg.cx * sdy;

	pOut = &(pG2DB->bImg.clr[p1.x + pG2DB->bSizeImg.cx * p1.y]);
	*pOut = shade_color(shColor, *pOut, pG2DB->bOp.clr);
	if (dxabs >= dyabs)
	{
		for (i = 0; i < dxabs; i++)
		{
			y += dyabs;
			if (y >= dxabs)                     
			{
				y -= dxabs;												
				pOut += scanoff;
			}
			pOut += sdx;
			*pOut = shade_color(shColor, *pOut, pG2DB->bOp.clr);
		}
	}
	else
	{
		for (i = 0; i < dyabs; i++)
		{
			x += dxabs;
			if (x >= dyabs)      
			{
				x -= dyabs;        
				pOut += sdx;
			}
			pOut += scanoff;
			*pOut = shade_color(shColor, *pOut, pG2DB->bOp.clr);
		}
	}
}

void	 G2Dc_draw_line(const g2dbase *pG2DB, point p1, point p2)
{
	pcolor *pOut;
	int i, dx, dy, sdx, sdy, dxabs, dyabs, x, y, scanoff;

	dx		= p2.x - p1.x;                       
	dy		= p2.y - p1.y;                       
	dxabs = abs(dx);                             
	dyabs = abs(dy);
	sdx		= sgn(dx);                               
	sdy		= sgn(dy);
	x			= dyabs >> 1;                                
	y			= dxabs >> 1;
	scanoff = pG2DB->bSizeImg.cx * sdy;

	pOut = &(pG2DB->bImg.clr[p1.x + pG2DB->bSizeImg.cx * p1.y]);
	*pOut = pG2DB->bOp.clr;
	if (dxabs >= dyabs)
	{
		for (i = 0; i < dxabs; i++)
		{
			y += dyabs;
			if (y >= dxabs)                     
			{
				y -= dxabs;												
				pOut += scanoff;
			}
			pOut += sdx;
			*pOut = pG2DB->bOp.clr;
		}
	}
	else
	{
		for (i = 0; i < dyabs; i++)
		{
			x += dxabs;
			if (x >= dyabs)      
			{
				x -= dyabs;        
				pOut += sdx;
			}
			pOut += scanoff;
			*pOut = pG2DB->bOp.clr;
		}
	}
}

void G2Dr_draw_circle(const g2dbase *pG2DB, point center,
											uint16 rad, double start, double end)
{
/*  	HSHADE shClr; */
	pcolor palBuf[0x100];
	byte *pOut = NULL, bVal;
	ss_bitmap_header *pBmp = NULL;
	double curpos = start, toend = end, actx, acty;
	int outx, outy;
	bool crossed;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dr_draw_circle: Area out of size");
#endif
	
/*  	shClr = shade_get(get_trans(pG2DB->bOp.clr)); */
	pBmp = Bmp_create(getRectWidth(pG2DB->bBoxArea),
										getRectHeight(pG2DB->bBoxArea));
	memset(pBmp->bits, 0x00, pBmp->width * pBmp->height);
	Pal_shade_palette(palBuf, 0, 0x100, 
										rgbt(get_red(pG2DB->bOp.clr),
												 get_green(pG2DB->bOp.clr),
												 get_blue(pG2DB->bOp.clr), 0x00), 
										pG2DB->bOp.clr);
	while (curpos >= 360.0)
		curpos -= 360.0;
	while (curpos < 0.0)
		curpos += 360.0;
	while (toend > 360.0)
		toend -= 360.0;
	while (toend < 0.0)
		toend += 360.0;
	crossed = (curpos <= toend);
	while (!crossed || curpos < toend)
	{
		actx = cos(curpos * M_PI / 180.0) * (double)(rad);
		acty = sin(curpos * M_PI / 180.0) * (double)(rad);
		outx = (int)(actx);
		outy = (int)(acty);
		actx -= (double)(outx);
		actx *= 10000.0;
		acty -= (double)(outy);
		acty *= 10000.0;
		if (actx < 0.0)
			actx *= -1.0;
		if (acty < 0.0)
			acty *= -1.0;
		
		if ((center.y + outy) >= 0 && (center.y + outy) < pG2DB->bSizeImg.cy &&
				(center.x + outx) >= 0 && (center.x + outx) < pG2DB->bSizeImg.cx)
		{
			pOut = &(pBmp->bits[center.x + outx - pG2DB->bBoxArea.left +
													pBmp->width * (center.y + outy - pG2DB->bBoxArea.top)]);
			if (abs(outx) > abs(outy))
				bVal = (byte)(((acty)) * 256.0 / 10000.0);
			else
				bVal = (byte)(((actx)) * 256.0 / 10000.0);
			if (*pOut < bVal)
				*pOut = bVal;
		}

		curpos += 0.25;
		if (curpos > 360.0)
		{
			curpos -= 360.0;
			crossed = true;
		}
	}
	G2Dr_head_to_base(pG2DB, pBmp, palBuf);
	Bmp_free_bitmap(pBmp);
}

void G2Dc_draw_circle(const g2dbase *pG2DB, point center,
											uint16 rad, double start, double end)
{
/*  	HSHADE shClr; */
	pcolor palBuf[0x100];
	byte *pOut = NULL, bVal;
	ss_bitmap_header *pBmp = NULL;
	double curpos = start, toend = end, actx, acty;
	int outx, outy;
	bool crossed;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dc_draw_circle: Area out of size");
#endif
	
/*  	shClr = shade_get(get_trans(pG2DB->bOp.clr)); */
	pBmp = Bmp_create(getRectWidth(pG2DB->bBoxArea),
										getRectHeight(pG2DB->bBoxArea));
	memset(pBmp->bits, 0x00, pBmp->width * pBmp->height);
	Pal_shade_palette(palBuf, 0, 0x100, 
										rgbt(get_red(pG2DB->bOp.clr),
												 get_green(pG2DB->bOp.clr),
												 get_blue(pG2DB->bOp.clr), 0x00), 
										pG2DB->bOp.clr);
	while (curpos >= 360.0)
		curpos -= 360.0;
	while (curpos < 0.0)
		curpos += 360.0;
	while (toend > 360.0)
		toend -= 360.0;
	while (toend < 0.0)
		toend += 360.0;
	crossed = (curpos <= toend);
	while (!crossed || curpos < toend)
	{
		actx = cos(curpos * M_PI / 180.0) * (double)(rad);
		acty = sin(curpos * M_PI / 180.0) * (double)(rad);
		outx = (int)(actx);
		outy = (int)(acty);
		actx -= (double)(outx);
		actx *= 10000.0;
		acty -= (double)(outy);
		acty *= 10000.0;
		if (actx < 0.0)
			actx *= -1.0;
		if (acty < 0.0)
			acty *= -1.0;
		
		if ((center.y + outy) >= 0 && (center.y + outy) < pG2DB->bSizeImg.cy &&
				(center.x + outx) >= 0 && (center.x + outx) < pG2DB->bSizeImg.cx)
		{
			pOut = &(pBmp->bits[center.x + outx - pG2DB->bBoxArea.left +
													pBmp->width * (center.y + outy - pG2DB->bBoxArea.top)]);
			if (abs(outx) > abs(outy))
				bVal = (byte)(((acty)) * 256.0 / 10000.0);
			else
				bVal = (byte)(((actx)) * 256.0 / 10000.0);
			if (*pOut < bVal)
				*pOut = bVal;
		}

		curpos += 0.25;
		if (curpos > 360.0)
		{
			curpos -= 360.0;
			crossed = true;
		}
	}
	G2Dc_head_to_base(pG2DB, pBmp, palBuf, true);
	Bmp_free_bitmap(pBmp);
}

void	 G2Dr_do_static(const g2dbase *pG2DB, uint16 center, uint16 height,
											byte dens_x, byte dens_y, byte shades)
{
	HSHADE *shList = NULL, shBuf;
	pcolor *pOut;
	double falldown_y = (100.0 - (double)(dens_y)) / (double)(height / 2),
				 cur_y = 100.0, poss;
	uint16 outx, offy;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dr_do_static: Area out of size");
#endif

	/* make up shade list */
	shList = malloc((shades + 1) * sizeof(HSHADE));
	shList[0] = shade_get(get_trans(pG2DB->bOp.clr));
	for (outx = 1; outx <= shades; outx++)
	{
		if (get_trans(pG2DB->bOp.clr))
			shList[outx] = shade_get((byte)(rand() % get_trans(pG2DB->bOp.clr)));
		else
			shList[outx] = shade_get(0x00);
	}
	/* draw center line */
	if (center >= pG2DB->bBoxArea.top && center < pG2DB->bBoxArea.bottom)
	{
		pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
														 pG2DB->bSizeImg.cx * center]);
		for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
		{
			if((rand() % 100) <= dens_x)
			{
				shBuf = shList[(byte)(rand() % (shades + 1))];
				*pOut = shade_color(shBuf, *pOut, pG2DB->bOp.clr);
			}
			pOut++;
		}
	}
	/* draw other lines */
	for(offy = 1; (offy * 2) <= height; offy++)
	{
		cur_y -= falldown_y;
		poss = (cur_y * (double)(dens_x)) / 100.0;
		if((center - offy) >= pG2DB->bBoxArea.top && 
			 (center - offy) <  pG2DB->bBoxArea.bottom)
		{
			pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
															 pG2DB->bSizeImg.cx * (center - offy)]);
			for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
			{
				if(((double)(rand() % 1000) / 10.0) <= poss)
				{
					shBuf = shList[(byte)(rand() % (shades + 1))];
					*pOut = shade_color(shBuf, *pOut, pG2DB->bOp.clr);
				}
				pOut++;
			}
		}
		if((center + offy) >= pG2DB->bBoxArea.top && 
			 (center + offy) <  pG2DB->bBoxArea.bottom)
		{
			pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
															 pG2DB->bSizeImg.cx * (center + offy)]);
			for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
			{
				if(((double)(rand() % 1000) / 10.0) <= poss)
				{
					shBuf = shList[(byte)(rand() % (shades + 1))];
					*pOut = shade_color(shBuf, *pOut, pG2DB->bOp.clr);
				}
				pOut++;
			}
		}
	}

	free(shList);
}

void	 G2Dc_do_static(const g2dbase *pG2DB, uint16 center, uint16 height,
											byte dens_x, byte dens_y, byte shades)
{
	byte *shList = NULL, t,
			 r = get_red(pG2DB->bOp.clr), 
			 g = get_green(pG2DB->bOp.clr), 
			 b = get_blue(pG2DB->bOp.clr);
	pcolor *pOut;
	double falldown_y = (100.0 - (double)(dens_y)) / (double)(height / 2),
				 cur_y = 100.0, poss;
	uint16 outx, offy;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dc_do_static: Area out of size");
#endif

	/* make up shade list */
	shList = malloc((shades + 1) * sizeof(byte));
	memset(shList, 0x00, (shades + 1) * sizeof(byte));
	shList[0] = get_trans(pG2DB->bOp.clr);
	if (shList[0])
		for (outx = 1; outx <= shades; outx++)
			shList[outx] = (byte)(rand() % shList[0]);
	/* draw center line */
	if (center >= pG2DB->bBoxArea.top && center < pG2DB->bBoxArea.bottom)
	{
		pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
														 pG2DB->bSizeImg.cx * center]);
		for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
		{
			if((rand() % 100) <= dens_x)
			{
				t = shList[(byte)(rand() % (shades + 1))];
				*pOut = rgbt(r, g, b, t);
			}
			pOut++;
		}
	}
	/* draw other lines */
	for(offy = 1; (offy * 2) <= height; offy++)
	{
		cur_y -= falldown_y;
		poss = (cur_y * (double)(dens_x)) / 100.0;
		if((center - offy) >= pG2DB->bBoxArea.top && 
			 (center - offy) <  pG2DB->bBoxArea.bottom)
		{
			pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
															 pG2DB->bSizeImg.cx * (center - offy)]);
			for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
			{
				if(((double)(rand() % 1000) / 10.0) <= poss)
				{
					t = shList[(byte)(rand() % (shades + 1))];
					*pOut = rgbt(r, g, b, t);
				}
				pOut++;
			}
		}
		if((center + offy) >= pG2DB->bBoxArea.top && 
			 (center + offy) <  pG2DB->bBoxArea.bottom)
		{
			pOut = &(pG2DB->bImg.clr[pG2DB->bBoxArea.left + 
															 pG2DB->bSizeImg.cx * (center + offy)]);
			for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
			{
				if(((double)(rand() % 1000) / 10.0) <= poss)
				{
					t = shList[(byte)(rand() % (shades + 1))];
					*pOut = rgbt(r, g, b, t);
				}
				pOut++;
			}
		}
	}

	free(shList);
}

void	 G2Dm_do_static(const g2dbase *pG2DB, uint16 center, uint16 height,
											byte dens_x, byte dens_y, byte shades)
{
	int *shList = NULL, *pOut;
	double falldown_y = (100.0 - (double)(dens_y)) / (double)(height / 2),
				 cur_y = 100.0, poss;
	uint16 outx, offy;

#ifdef G2D_SANITY_CHECKS
	if (!isBoxInSize(pG2DB->bBoxArea, pG2DB->bSizeImg))
		Cant_happen("G2Dm_do_static: Area out of size");
#endif

	/* make up shade list */
	shList = malloc((shades + 1) * sizeof(int));
	shList[0] = pG2DB->bOp.trn;
	for (outx = 1; outx <= shades; outx++)
	{
		if (shList[0])
			shList[outx] = (rand() % shList[0]);
		else
			shList[outx] = 0;
	}
	/* draw center line */
	if (center >= pG2DB->bBoxArea.top && center < pG2DB->bBoxArea.bottom)
	{
		pOut = &(pG2DB->bImg.msk[pG2DB->bBoxArea.left + 
														 pG2DB->bSizeImg.cx * center]);
		for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
		{
			if((rand() % 100) <= dens_x)
			{
				*pOut += shList[(rand() % (shades + 1))];
			}
			pOut++;
		}
	}
	/* draw other lines */
	for(offy = 1; (offy * 2) <= height; offy++)
	{
		cur_y -= falldown_y;
		poss = (cur_y * (double)(dens_x)) / 100.0;
		if((center - offy) >= pG2DB->bBoxArea.top && 
			 (center - offy) <  pG2DB->bBoxArea.bottom)
		{
			pOut = &(pG2DB->bImg.msk[pG2DB->bBoxArea.left + 
															 pG2DB->bSizeImg.cx * (center - offy)]);
			for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
			{
				if(((double)(rand() % 1000) / 10.0) <= poss)
				{
					*pOut += shList[(rand() % (shades + 1))];
				}
				pOut++;
			}
		}
		if((center + offy) >= pG2DB->bBoxArea.top && 
			 (center + offy) <  pG2DB->bBoxArea.bottom)
		{
			pOut = &(pG2DB->bImg.msk[pG2DB->bBoxArea.left + 
															 pG2DB->bSizeImg.cx * (center + offy)]);
			for(outx = pG2DB->bBoxArea.left; outx < pG2DB->bBoxArea.right; outx++)
			{
				if(((double)(rand() % 1000) / 10.0) <= poss)
				{
					*pOut += shList[(rand() % (shades + 1))];
				}
				pOut++;
			}
		}
	}

	free(shList);
}

/* local function; draws a border of color 'bg' around a pixel of 	*/
/* value > 'bg'																											*/
void g2d_draw_border(struct tag_ss_bitmap_header *pHead, byte bg)
{
	int iny, inx,
			bw = pHead->width, bh = pHead->height;
	byte *pIn1, *pIn2;

	/* top and bottom lines */
	pIn1 = pHead->bits;
	pIn2 = &(pHead->bits[pHead->width * (pHead->height - 1)]);
	for (inx = 0; inx < pHead->width; inx++)
	{
		if (!(*pIn1) &&
				((inx < (bw - 1)  && pIn1[+bw + 1]	> bg)	|| /* down right */
				 (									 pIn1[+bw]			> bg)	|| /* down       */
				 (inx > 0					&& pIn1[+bw - 1]	> bg)	)) /* down left  */
			*pIn1 = bg;
		if (!(*pIn2) &&
				((inx > 0					&& pIn2[-bw - 1]	> bg) || /* up left  */
				 (									 pIn2[-bw]			> bg) || /* up       */
				 (inx < (bw - 1)	&& pIn2[-bw + 1]	> bg)	)) /* up right */
			*pIn2 = bg;
		pIn1++;
		pIn2++;
	}
	/* left and right lines */
	pIn1 = pHead->bits;
	pIn2 = &(pHead->bits[pHead->width - 1]);
	for (iny = 0; iny < pHead->height; iny++)
	{
		if (!(*pIn1) &&
				((iny < (bh - 1)  && pIn1[+bw + 1]	> bg)	|| /* right down */
				 (									 pIn1[+1]				> bg)	|| /* right      */
				 (iny > 0					&& pIn1[-bw + 1]	> bg)	)) /* right up   */
			*pIn1 = bg;
		if (!(*pIn2) &&
				((iny > 0					&& pIn2[-bw - 1]	> bg) || /* left up   */
				 (									 pIn2[-1]				> bg) || /* left      */
				 (iny < (bh - 1)	&& pIn2[+bw - 1]	> bg)	)) /* left down */
			*pIn2 = bg;
		pIn1 += bw;
		pIn2 += bw;
	}
	/* area except border */
	for (iny = 1; iny < (pHead->height - 1); iny++)
	{
		pIn1 = &(pHead->bits[bw * iny + 1]);
		for (inx = 1; inx < (pHead->width - 1); inx++)
		{
			if (!(*pIn1) && 
					(pIn1[-1]				> bg || pIn1[+1]			> bg || /* left / right */
					 pIn1[-bw]			> bg || pIn1[+bw]			> bg || /* up / down    */
					 pIn1[-bw - 1]	> bg || pIn1[+bw - 1]	> bg || /* ul, dl       */
					 pIn1[-bw + 1]	> bg || pIn1[+bw + 1]	> bg )) /* ur, dr       */
				*pIn1 = bg;
			pIn1++;
		}
	}
}

struct tag_ss_bitmap_header *G2D_write_string(char * const *pStrings,
																				int nStrings, int nFont, 
																				byte flags,	int vSpace, 
																				int hSpace)
{
	struct tag_ss_bitmap_header *pHead = NULL;
	char **pNewStrings = NULL, **pLineWords, **pWordParts,
			 strSoftHyphen[2] = { 0x02, 0x00 };
	int i, nInLine, nOutLines = 1, nCurLen, nWordLen,
			nCurWord, nCurPart, nNLs,
			nFullLineWidth, nPartLineWidth, nFontWidth = 0, nFontHeight = 0;
	byte *pBuf;
	
	pNewStrings = Str_resize_array(Str_create_array(), 0, 1, 0);
	for (nInLine = 0; nInLine < nStrings; nInLine++)
	{	/* walk through all input lines */
		pLineWords = Str_extract(pStrings[nInLine], " ");
		nCurWord = 0;
		while (pLineWords[nCurWord])
		{ /* now go through all words in this line */
			nCurLen = strlen(pNewStrings[nOutLines - 1]);
			while (pLineWords[nCurWord][0] == 0x0A)
			{
				nOutLines++;
				pNewStrings = Str_resize_array(pNewStrings, 
																			 nOutLines - 1, nOutLines, 
																			 strlen(pLineWords[nCurWord]));
				memmove(pLineWords[nCurWord], &pLineWords[nCurWord][1],
								strlen(pLineWords[nCurWord]) + 1);
			}
			nWordLen = strlen(pLineWords[nCurWord]);
			nNLs = 0;
			while (nWordLen && pLineWords[nCurWord][nWordLen - 1] == 0x0A)
			{ /* search newlines */
				nNLs++;
				pLineWords[nCurWord][nWordLen - 1] = 0x00;
				nWordLen = strlen(pLineWords[nCurWord]);
			}
			pNewStrings[nOutLines - 1] = (char *)realloc(
																			pNewStrings[nOutLines - 1], 
																			nCurLen + nWordLen + 2);
			pWordParts = Str_extract(pLineWords[nCurWord], strSoftHyphen);
			nCurPart = 0;
			while (pWordParts[nCurPart])
			{	/* now go through all word parts... sigh */
				Font_get_size(nFont, pNewStrings[nOutLines - 1], 
											&nFullLineWidth, &nFontHeight);			
				Font_get_size(nFont, pWordParts[nCurPart], 
											&nPartLineWidth, &nFontHeight);			
				if ((nFullLineWidth + nPartLineWidth + 
						 ((flags & F_WSTR_DRAW_BORDER) ? 2 : 0)) >= hSpace &&
            (nCurWord || nCurPart))
				{	/* crossing hSpace. create new line except for first word */
					if (pWordParts[nCurPart + 1])
					{	/* make a break character ('-') if within word */
						nCurLen = strlen(pNewStrings[nOutLines - 1]);	
						pNewStrings[nOutLines - 1][nCurLen + 1] = 0x00;
						pNewStrings[nOutLines - 1][nCurLen] = '-';
					}
					nOutLines++;
					pNewStrings = Str_resize_array(pNewStrings, 
																				 nOutLines - 1, nOutLines,
																				 strlen(pLineWords[nCurWord]));
				}
				strcat(pNewStrings[nOutLines - 1], pWordParts[nCurPart]);
				free(pWordParts[nCurPart++]);
			}
			free(pWordParts);
			nCurLen = strlen(pNewStrings[nOutLines - 1]) + 2;
			pNewStrings[nOutLines - 1] = (char *)realloc(
																		pNewStrings[nOutLines - 1],
																		nCurLen);
			pNewStrings[nOutLines - 1][nCurLen - 1] = 0x00;
			pNewStrings[nOutLines - 1][nCurLen - 2] = ' ';
			free(pLineWords[nCurWord++]);

			while (nNLs--)
			{ /* insert found newlines */
				nOutLines++;
				pNewStrings = Str_resize_array(pNewStrings, 
																			 nOutLines - 1, nOutLines, 0);
			}
		}
		free(pLineWords);
		if (flags & F_WSTR_FORCE_NL)
		{	/* insert forced newline */
			nOutLines++;
			pNewStrings = Str_resize_array(pNewStrings, 
																		 nOutLines - 1, nOutLines, 0);
		}
	}
	if (!strlen(pNewStrings[nOutLines - 1]))
	{	/* remove a void blank line */
		nOutLines--;
		pNewStrings = Str_resize_array(pNewStrings, 
																	 nOutLines + 1, nOutLines, 0);
	}
	
	if (nOutLines)
	{ /* strings reparsed -- now prepare bitmap */
		Font_get_size(nFont, "0", &nFontWidth, &nFontHeight);			
		pHead = Bmp_create(hSpace, ((nFontHeight + vSpace +
																 ((flags & F_WSTR_DRAW_BORDER) ? 1 : 0)) * 
															 nOutLines) + 
															 ((flags & F_WSTR_DRAW_BORDER) ? 1 : 0));		
		memset(pHead->bits, 0x00, pHead->width * pHead->height);
		pBuf = pHead->bits;
		for (i = 0; i < nOutLines; i++)
		{
/*			printf("<%s>\n", pNewStrings[i]); */
			Font_get_size(nFont, pNewStrings[i], 
										&nFullLineWidth, &nFontHeight);
			Font_render_string(nFont, pNewStrings[i], 
												 ((flags & F_WSTR_RIGHT_ALIGN) ? 
														(hSpace - nFullLineWidth -
														 ((flags & F_WSTR_DRAW_BORDER) ? 1 : 0)) : 
													((flags & F_WSTR_DRAW_BORDER) ? 1 : 0)), 
												 (flags & F_WSTR_DRAW_BORDER) ? 1 : 0,
												 hSpace, 
												 nFontHeight + ((flags & F_WSTR_DRAW_BORDER) ? 1 : 0), 
												 pBuf, 0x02);

			free(pNewStrings[i]);
			pBuf += (pHead->width * (nFontHeight + vSpace));
		}
		if (flags & F_WSTR_DRAW_BORDER)
			g2d_draw_border(pHead, 0x01);
	}
	free(pNewStrings);

	return pHead;
}


/*
 * End of src/graph2d.c
 */

/*
 * Emacs magic
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
