/*
 * hud.c -
 *  HUD system
 *
 * This file by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "accel.h"
#include "bitmap.h"
#include "chunks.h"
#include "font.h"
#include "gamemem.h"
#include "graph2d.h"
#include "hud.h"
#include "main.h"
#include "palette.h"
#include "render.h"
#include "shade.h"
#include "timer.h"


/* defines for debugging																*/
/*
#define D_HUD_DEBUG_REGISTER
#define D_HUD_DEBUG_MOUSEEVENT
#define D_HUD_DEBUG_RENDER
*/

#define C_HUD_DESC_MAX		256

/* struct for list																		*/
typedef struct
{
	void				*haiPrev;
	hudArea			*haiThis;
	void				*haiNext;
} hudAreaItem;

typedef struct tag_hudDescription
{
	struct tag_hudDescription	*hdPrev;
	struct tag_hudDescription	*hdNext;

	bounding_box							 hdArea;
	bool											 hdDeactivate;
	uint32										 hdStartLeave;

	char											 hdText[C_HUD_DESC_MAX];

	pcolor										 hdPal[3];
	g2dbase										 hdG2DB;
	hudLayer									 hdLayer;
} hudDescription;

/* local functions 																		*/
bool hud_move_item(hudAreaItem *pHAI, short dir);
bool hud_set_item_last(hudAreaItem *pHAI);
bool hud_set_item_first(hudAreaItem *pHAI);

void hud_free_cursor(void);
void hud_render_cursor(P_HUD_HOOK_RENDER);

hudDescription *hud_desc_find(bounding_box *pBox, char *strDesc);
void hud_desc_detach(hudDescription *pDesc);
void hud_desc_append(hudDescription *pDesc);
void hud_desc_render(P_HUD_HOOK_RENDER);
void hud_desc_timer(uint32 now, uint32 last, void *pData);

bool	hud_mouse_action(byte nAction, bool bStart);

/* hud chain roots 																		*/
hudAreaItem *haiFirst = NULL, *haiLast = NULL;

hudDescription *pDescFirst = NULL, *pDescLast = NULL;

size				 szScreen = { 0, 0 };

/* mouse cursor                                         */
int								 nCurCursor = -1;
point							 ptCurCursor = { 0, 0 };
bounding_box			 boxCursorLock;

pcolor						 palCursor[0x100];
g2dbase						 gbCursor = D_G2D_NULLAREA;
hudLayer					 hlCursor = D_HUD_NULLLAYER(&gbCursor);



void HUD_init(uint16 width, uint16 height)
{
	szScreen.cx = width;
	szScreen.cy = height;

	HUD_unlock_cursor();
	HUD_set_cursor_shade(0xFF);
	ptCurCursor.x = width / 2;
	ptCurCursor.y = height / 2;
	Timer_register_function("HUD.Description", 10, hud_desc_timer, NULL);
}

void HUD_deinit(void)
{
	hudDescription *pDesc;

	Timer_unregister_function("HUD.Description");
	while (haiFirst)
		HUD_unregister_area(haiFirst->haiThis);
	hud_free_cursor();

	pDesc = pDescFirst;
	while (pDesc)
	{
		hud_desc_detach(pDesc);
		free(pDesc);
		pDesc = pDescFirst;
	}
}

void	HUD_reset_layer(hudLayer *pLayer, g2dbase *pG2DB)
{
	memset(pLayer, 0x00, sizeof(hudLayer));
	pLayer->pG2DB = pG2DB;
}
										
hudDescription *hud_desc_find(bounding_box *pBox, char *strDesc)
{
	hudDescription *pDesc = pDescFirst, *pFoundDesc = NULL;

	while (pDesc && !pFoundDesc)
	{
		if (!memcmp(&(pDesc->hdArea), pBox, sizeof(bounding_box)) ||
				!strcmp(pDesc->hdText, strDesc))
			pFoundDesc = pDesc;
		pDesc = pDesc->hdNext;
	}

	return pFoundDesc;
}
							
void hud_desc_detach(hudDescription *pDesc)
{
	if (pDesc == pDescFirst)
		pDescFirst = pDesc->hdNext;
	if (pDesc == pDescLast)
		pDescLast = pDesc->hdPrev;
	if (pDesc->hdPrev)
	{
		pDesc->hdPrev->hdNext = pDesc->hdNext;
		pDesc->hdPrev = NULL;
	}
	if (pDesc->hdNext)
	{
		pDesc->hdNext->hdPrev = pDesc->hdPrev;
		pDesc->hdNext = NULL;
	}
}

void hud_desc_append(hudDescription *pDesc)
{
	pDesc->hdPrev = pDescLast;
	if (pDescLast)
		pDescLast->hdNext = pDesc;
	else
		pDescFirst = pDesc;
	pDescLast = pDesc;
}

void hud_desc_timer(uint32 now, uint32 last, void *pData)
{
	HSHADE hShade = NULL;
	hudDescription *pDesc = NULL, *pDescCur;

	UNUSED(pData);
	UNUSED(last);

	pDesc = pDescFirst;
	while (pDesc)
	{
		pDescCur = pDesc;
		pDesc = pDescCur->hdNext;

		if (pDescCur->hdDeactivate ||
				!isPointinRect(pDescCur->hdArea, ptCurCursor.x, ptCurCursor.y))
		{
			pDescCur->hdDeactivate = true;
			if (!pDescCur->hdStartLeave)
				pDescCur->hdStartLeave = now;
			if ((now - pDescCur->hdStartLeave) < 256)
			{
				hShade = shade_get((byte)(now - pDescCur->hdStartLeave));
				pDescCur->hdPal[2] = rgbt(get_red(pDescCur->hdPal[2]),
																	get_green(pDescCur->hdPal[2]),
																	get_blue(pDescCur->hdPal[2]),
																	shade(hShade, 
																		get_trans(pDescCur->hdPal[2]), 0x00));
				pDescCur->hdPal[1] = rgbt(get_red(pDescCur->hdPal[1]),
																	get_green(pDescCur->hdPal[1]),
																	get_blue(pDescCur->hdPal[1]),
																	shade(hShade, 
																				get_trans(pDescCur->hdPal[1]), 0x00));
			}
			else
			{
				hud_desc_detach(pDescCur);
				G2D_reset_base(&pDescCur->hdG2DB, true);
				free(pDescCur);
			}
		}
	}
}

void hud_desc_render(P_HUD_HOOK_RENDER)
{
	hudDescription *pDesc = pDescFirst;
	int nStartx, nStarty;

	while (pDesc)
	{	
		if (ptCurCursor.x >= (szScreen.cx / 2))
			nStartx = max((int)(ptCurCursor.x) - pDesc->hdG2DB.bSizeImg.cx - 1, 0);
		else
			nStartx = ptCurCursor.x + 1;
		nStarty = max((int)(ptCurCursor.y) - pDesc->hdG2DB.bSizeImg.cy - 1, 0);

		pDesc->hdLayer.ptPos.x = nStartx;
		pDesc->hdLayer.ptPos.y = nStarty;

		pDesc->hdLayer.szSize.cx = min(pDesc->hdG2DB.bSizeImg.cx,
																	 szScreen.cx - nStartx);
		pDesc->hdLayer.szSize.cy = min(pDesc->hdG2DB.bSizeImg.cy,
																	 szScreen.cy - nStarty);

		pDesc->hdG2DB.bBoxArea.right = pDesc->hdLayer.szSize.cx;
		pDesc->hdG2DB.bBoxArea.bottom = pDesc->hdLayer.szSize.cy;

		appendHudLayer(pLayer, pDesc->hdLayer);

		pDesc = pDesc->hdNext;
	}
}

void hud_free_cursor(void)
{
	G2D_reset_base(&gbCursor, true);
	nCurCursor = -1;
}

bool	HUD_show_cursor(int nCursor)
{
	ss_bitmap_header *pCursorBmp = NULL;
	bool rCode = false;

  if (nCursor != nCurCursor)
	{
		pCursorBmp = Bmp_load_bitmap(CHUNK_CURSOR_BASE, nCursor);

		if (pCursorBmp)
		{
			printf("HUD_show_cursor() Showing cursor %d\n", nCursor);
			nCurCursor = nCursor;

			if ((gbCursor.bSizeImg.cx != pCursorBmp->width) ||
					(gbCursor.bSizeImg.cy != pCursorBmp->height))
			{
				G2D_reset_base(&gbCursor, true);
				convertHeadToArea(*pCursorBmp, gbCursor);
				gbCursor.pPal = palCursor;
				gbCursor.nPalSize = 0xFF;

				hlCursor.szSize = gbCursor.bSizeImg;
			}
			free(pCursorBmp);
			rCode = true;
		}
		else
			printf("HUD_show_cursor() Cursor %d failed\n", nCursor);
	}
	else
		rCode = true; /* re-show early-out */

	return rCode;
}

int		HUD_get_cursor(void)
{
	return nCurCursor;
}

void	HUD_move_cursor(int relx, int rely)
{
	ptCurCursor.x = min(max(ptCurCursor.x + relx, boxCursorLock.left), 
											boxCursorLock.right - 1);
	ptCurCursor.y = min(max(ptCurCursor.y + rely, boxCursorLock.top), 
											boxCursorLock.bottom - 1);
	hud_mouse_action(C_HUD_MOUSE_HOVER, true);
}

void	HUD_set_cursor_pos(point ptCursor)
{
	ptCurCursor.x = min(max(ptCursor.x, boxCursorLock.left),
											boxCursorLock.right - 1);
	ptCurCursor.y = min(max(ptCursor.y, boxCursorLock.top),
											boxCursorLock.bottom - 1);
	hud_mouse_action(C_HUD_MOUSE_HOVER, true);
}

void	HUD_get_cursor_pos(point *pptCursor)
{
	memcpy(pptCursor, &ptCurCursor, sizeof(point));
}

void	HUD_get_center_pos(point *pptCenter)
{
	pptCenter->x = (boxCursorLock.left + boxCursorLock.right) / 2;
	pptCenter->y = (boxCursorLock.top + boxCursorLock.bottom) / 2;
}

void	HUD_set_cursor_shade(byte bShade)
{
	Pal_copy_palette(palCursor, RGB_trans, bShade);
	palCursor[0] = rgbt(0x00, 0x00, 0x00, 0x00);
}

void	HUD_lock_cursor(bounding_box *pBox)
{
	memcpy(&boxCursorLock, pBox, sizeof(bounding_box));
}

void	HUD_unlock_cursor()
{
	boxCursorLock.left = boxCursorLock.top = 0;
	boxCursorLock.right = szScreen.cx;
	boxCursorLock.bottom = szScreen.cy;
}

void hud_render_cursor(P_HUD_HOOK_RENDER)
{
	int nHotspotx, nHotspoty;
	
	if (gbCursor.bImg.bit)
	{
		nHotspotx = ptCurCursor.x - (gbCursor.bSizeImg.cx / 2);
		nHotspoty = ptCurCursor.y - (gbCursor.bSizeImg.cy / 2);
		gbCursor.bBoxArea.left = gbCursor.bBoxArea.top = 0;
		if (nHotspotx < 0)
			gbCursor.bBoxArea.left = abs(nHotspotx);
		if (nHotspoty < 0)
			gbCursor.bBoxArea.top = abs(nHotspoty);
		
		hlCursor.ptPos.x = max(nHotspotx, 0);
		hlCursor.ptPos.y = max(nHotspoty, 0);

		gbCursor.bBoxArea.right = min(gbCursor.bSizeImg.cx, 
																	(szScreen.cx - hlCursor.ptPos.x));
		gbCursor.bBoxArea.bottom = min(gbCursor.bSizeImg.cy, 
																	 (szScreen.cy - hlCursor.ptPos.y));

		if (gbCursor.bBoxArea.right <= szScreen.cx &&
				gbCursor.bBoxArea.bottom <= szScreen.cy)
		{	
			appendHudLayer(pLayer, hlCursor);
			hud_desc_render(pLayer);
		}
	}
}

void	HUD_set_description(char *strDesc, pcolor cFG, pcolor cBG,
													bounding_box *pArea)
{
	ss_bitmap_header *pHead = NULL;
	hudDescription *pDesc = NULL;
	int nHeight = 0, nWidth = 0;
	char *pStrings[2] = { NULL, NULL };
								 
	pDesc = hud_desc_find(pArea, strDesc);
	if (pDesc)
	{
		hud_desc_detach(pDesc);
		G2D_reset_base(&pDesc->hdG2DB, true);
	}
	else
	{
		pDesc = (hudDescription *)malloc(sizeof(hudDescription));
		memset(pDesc, 0x00, sizeof(hudDescription));
		G2D_reset_base(&pDesc->hdG2DB, false);
		HUD_reset_layer(&pDesc->hdLayer, &pDesc->hdG2DB);
	}
	pDesc->hdDeactivate = false;
	pDesc->hdStartLeave = 0;
	pDesc->hdPal[1] = cBG;
	pDesc->hdPal[2] = cFG;
	memset(pDesc->hdText, 0x00, C_HUD_DESC_MAX);
	strncpy(pDesc->hdText, strDesc, C_HUD_DESC_MAX - 1);
	memcpy(&(pDesc->hdArea), pArea, sizeof(bounding_box));

	Font_get_size(1, pDesc->hdText, &nWidth, &nHeight);
	pStrings[0] = pDesc->hdText;
	pHead = G2D_write_string(pStrings, 1, 1, F_WSTR_DRAW_BORDER, 0, 
													 nWidth + 3);

	if (pHead)
	{
		convertHeadToArea(*pHead, pDesc->hdG2DB);
		pDesc->hdG2DB.pPal = pDesc->hdPal;
		pDesc->hdG2DB.nPalSize = 0x02;
		free(pHead);
		hud_desc_append(pDesc);
	}
	else
		Cant_happen("HUD_set_description(): Failed to create Image");
}

bool HUD_register_area(hudArea *pHA)
{
	hudAreaItem *pHAI = NULL;
	bool rCode = false;

#ifdef D_HUD_DEBUG_REGISTER
	printf("HUD Area <%s> trying to register...", pHA->haName);
#endif /*D_HUD_DEBUG_REGISTER*/
	if (HUD_find_area(pHA) < 0)
	{
		pHAI = (hudAreaItem *)malloc(sizeof(hudAreaItem));
		if (pHAI != NULL)
		{
			if (haiFirst == NULL)
				haiFirst = pHAI;
			pHAI->haiPrev = haiLast;
			if(pHAI->haiPrev != NULL)
				((hudAreaItem *)pHAI->haiPrev)->haiNext = pHAI;
			pHAI->haiThis = pHA;
			pHAI->haiNext = NULL;
			haiLast = pHAI;
			rCode = true;
#ifdef D_HUD_DEBUG_REGISTER
			printf(" DONE \n", pHA->haName);
#endif /*D_HUD_DEBUG_REGISTER*/
		}
		else
		{
			/* ouch - out of mem?? */
		}
	}
#ifdef D_HUD_DEBUG_REGISTER
	else
	{
		printf("FAILED\n", pHA->haName);
	}
#endif /*D_HUD_DEBUG_REGISTER*/

	return rCode;
}

short HUD_find_area(hudArea *pHA)
{
	hudAreaItem *pHAI = haiFirst;
	short rCode = -1, pos = -1;

	while (rCode == -1 && pHAI != NULL)
	{
		++pos;
		if(pHAI->haiThis != NULL && pHAI->haiThis == pHA)			
			rCode = pos;
		pHAI = pHAI->haiNext;
	}

	return rCode;
}

bool HUD_unregister_area(hudArea *pHA)
{
	hudAreaItem *pHAI = haiFirst;
	bool rCode = false;

	while (!rCode && pHAI != NULL)
	{
		if(pHAI->haiThis != NULL && pHAI->haiThis == pHA)			
		{
			if(pHA->haHooks.hhDestroy != NULL &&
				!((*(pHA->haHooks.hhDestroy))()))
			{
				/*destroy func didn't want to...*/
			}
/*remove from chain*/
			if(pHAI->haiPrev != NULL)					
				((hudAreaItem *)pHAI->haiPrev)->haiNext = pHAI->haiNext;
			else
				haiFirst = pHAI->haiNext;
			if(pHAI->haiNext != NULL)
				((hudAreaItem *)pHAI->haiNext)->haiPrev = pHAI->haiPrev;
			else
				haiLast = pHAI->haiPrev;
/*free mem*/
			free(pHAI);
			rCode = true;
		}
		else
			pHAI = pHAI->haiNext;
	}

	return rCode;
}

short HUD_render(struct tag_camera *pCam)
{
	hudAreaItem *pHAI = haiFirst;
	hudLayer **pLayer = NULL, *pLayerFirst = NULL;
	int nLayer = 0;
	short nAreas = 0;

	pLayer = &pLayerFirst;
	while(pHAI != NULL)
	{
		if(pHAI->haiThis != NULL &&					/*entry there?*/
			pHAI->haiThis->haVisible &&				/*area visible?*/
			pHAI->haiThis->haHooks.hhRender != NULL) /*render func present?*/
		{
#ifdef D_HUD_DEBUG_RENDER
			printf("HUD Area <%s> going to render...",
						 pHAI->haiThis->haName);
#endif /*D_HUD_DEBUG_RENDER*/

			(*pHAI->haiThis->haHooks.hhRender)(pLayer);
			while (pLayer && *pLayer)
			{
				nLayer++;
				pLayer = &((*pLayer)->pNext);
			}
			
#ifdef D_HUD_DEBUG_RENDER
			printf(" DONE \n",
					 pHAI->haiThis->haName);
#endif /*D_HUD_DEBUG_RENDER*/
			++nAreas;
		}
		pHAI = pHAI->haiNext;
	}
	hud_render_cursor(pLayer);
#ifdef D_HUD_DEBUG_RENDER
	printf("HUD_render(): Went through %d Areas, giving %d Layer\n",
				 nAreas, nLayer);
#endif /*D_HUD_DEBUG_RENDER*/
	if (pLayerFirst)
		Acc_draw_overlay(pLayerFirst, pCam);

	return nAreas;
}

void HUD_clear(void)
{
	hudAreaItem *pHAI = haiLast, *pBuf;

	while(pHAI != NULL)
	{
		pBuf = pHAI;
		pHAI = pHAI->haiPrev;
		if(pBuf->haiThis != NULL)
		{
			if(pBuf->haiThis->haHooks.hhDestroy != NULL &&
				!((*(pBuf->haiThis->haHooks.hhDestroy))()))
			{
				/*destroy func didn't want to... but we have to*/
			}
			free(pBuf);
		}
	}
	haiFirst = haiLast = NULL;
}

/*																																	*/
/* input																														*/
/*																																	*/

bool	hud_mouse_action(byte nAction, bool bStart)
{
	hudAreaItem *pHAI = haiLast;
	uint16 offx, offy;
	bool rCode = false;

	while (!rCode && pHAI != NULL)
	{
		if (pHAI->haiThis != NULL &&					/*entry there?*/
				pHAI->haiThis->haVisible &&				/*area visible?*/
				isPointinRect(pHAI->haiThis->haArea, 
											ptCurCursor.x, ptCurCursor.y) &&
				pHAI->haiThis->haHooks.hhMouseEvent != NULL)	/*mouse func present?*/
		{
			offx = ptCurCursor.x - pHAI->haiThis->haArea.left;
			offy = ptCurCursor.y - pHAI->haiThis->haArea.top;
			rCode = (*pHAI->haiThis->haHooks.hhMouseEvent)(nAction, bStart, 
																										 offx, offy);
#ifdef D_HUD_DEBUG_MOUSEEVENT
			if (rCode)
				printf("HUD Area <%s> ate MouseEvent %02X at %d|%d\n",
						 pHAI->haiThis->haName, nAction, ptCurCursor.x, ptCurCursor.y);
#endif /*D_HUD_DEBUG_MOUSEEVENT*/
		}
		pHAI = pHAI->haiPrev;
	}

	return rCode;
}

bool	HUD_select_at_cursor(bool bStart)
{
	return hud_mouse_action(C_HUD_MOUSE_SELECT, bStart);
}

bool	HUD_activate_at_cursor(bool bStart)
{
	return hud_mouse_action(C_HUD_MOUSE_ACTIVATE, bStart);
}

bool	HUD_grab_at_cursor(bool bStart)
{
	return hud_mouse_action(C_HUD_MOUSE_GRAB, bStart);
}

/*																																	*/
/* messages																													*/
/*																																	*/

typedef struct
{
	bool			mmUsed;
	char			mmText[C_MSG_MAXLEN + 1];
	short			mmTime;
	byte			mmFlags;
	pcolor		mmFGColor[2];
	pcolor		mmBGColor[2];
	int				mmFont;
} Msg;								 

/* functions */
void		 msg_timer_messages(uint32 now, uint32 last, void *pData);
void		 msg_render_messages(P_HUD_HOOK_RENDER);
bool		 msg_destroy_messages(P_HUD_HOOK_DEST);

/* hud area struct */
hudArea haMessages =
{ 
	"Messages",				true, { 0, 0, 0, 0 }, 
	{ NULL, msg_render_messages, msg_destroy_messages } 
};

Msg				 Messages[C_MSG_MAXLINE];
pixel			*pixel_Msg = NULL;
g2dbase		 gbMsg = D_G2D_NULLAREA;
hudLayer	 hlMsg = D_HUD_NULLLAYER(&gbMsg);

void MSG_init(bounding_box boxArea)
{
	haMessages.haArea = boxArea;
	memset(Messages, 0x00, sizeof(Msg) * C_MSG_MAXLINE);

	Timer_register_function("HUD.Messages", 5, 
													msg_timer_messages, NULL);
	hlMsg.ptPos.x = boxArea.left;
	hlMsg.ptPos.y = boxArea.top;
	hlMsg.szSize.cx = (uint16)getRectWidth(boxArea);
	hlMsg.szSize.cy = (uint16)getRectHeight(boxArea);
	G2D_init_clrbase(&gbMsg, hlMsg.szSize.cx, hlMsg.szSize.cy);
	G2Dc_fill_area(&gbMsg);

	HUD_register_area(&haMessages);

/*
	MSG_add_message(F_MSG_CENTER, "0", 10000, 6,
									rgbt(0xFF, 0x00, 0x00, 0xFF),
									rgbt(0xFF, 0x00, 0x00, 0x20));
*/
}

void MSG_deinit(void)
{
	Timer_unregister_function("HUD.Messages");
	G2D_reset_base(&gbMsg, true);
}

short MSG_add_message(byte bFlags, const char *pMsg, short nTime,
											int nFont, pcolor clrFg, pcolor clrBg)
{
	short nIndex = -1, i;

	for (i = 0; i < C_MSG_MAXLINE && nIndex == -1; i++)
		if (!Messages[i].mmUsed)
			nIndex = i;
	if (nIndex >= 0)
	{
		Messages[nIndex].mmFlags = bFlags;
		strncpy(Messages[nIndex].mmText, pMsg, C_MSG_MAXLEN + 1);
		Messages[nIndex].mmTime = nTime;
		Messages[nIndex].mmFont = nFont;
		Messages[nIndex].mmFGColor[0] = Messages[nIndex].mmFGColor[1] = clrFg;
		Messages[nIndex].mmBGColor[0] = Messages[nIndex].mmBGColor[1] = clrBg;
		Messages[nIndex].mmUsed = true;
	}

	return nIndex;
}

bool	MSG_edit_message(short nId, byte bMask, byte bFlags, 
											 const char *pMsg, short nTime,
											 int nFont, pcolor clrFg, pcolor clrBg)
{
	bool rCode = false;

	if(Messages[nId].mmUsed)
	{
		if (bMask & F_MSG_MSK_FLAG)
			Messages[nId].mmFlags = bFlags;
		if (bMask & F_MSG_MSK_TEXT)
			strncpy(Messages[nId].mmText, pMsg, C_MSG_MAXLEN + 1);
		if (bMask & F_MSG_MSK_TIME)
			Messages[nId].mmTime = nTime;
		if (bMask & F_MSG_MSK_FONT)
			Messages[nId].mmFont = nFont;
		if (bMask & F_MSG_MSK_COLOR)
		{
			Messages[nId].mmFGColor[0] = clrFg;
			Messages[nId].mmBGColor[0] = clrBg;
		}

		rCode = true;
	}

  return rCode; /* shurely? -jim */
								/* hm? -ff1 */
}

bool	MSG_del_message(short nId)
{
	bool rCode = Messages[nId].mmUsed;

	memset(&Messages[nId], 0x00, sizeof(Msg));

	return rCode;
}

void	msg_timer_messages(uint32 now, uint32 last, void *pData)
{
	HSHADE shFade;
	short i;

	UNUSED(pData);
	for (i = 0; i < C_MSG_MAXLINE; i++)
	{
		if (Messages[i].mmUsed && !(Messages[i].mmFlags & F_MSG_STATIC))
		{
			if (Messages[i].mmTime > 0)
			{
				Messages[i].mmTime -= (short)(now - last);
				if (Messages[i].mmTime < 256)
				{
					/* eat this: shade a trans byte ;)	*/
					shFade = shade_get((byte)(max(Messages[i].mmTime, 0)));
					Messages[i].mmFGColor[1] = 
							rgbt(get_red(Messages[i].mmFGColor[0]),
									 get_green(Messages[i].mmFGColor[0]),
									 get_blue(Messages[i].mmFGColor[0]),
									 shade(shFade, 0x00,
												 get_trans(Messages[i].mmFGColor[0]))
									);
					Messages[i].mmBGColor[1] = 
							rgbt(get_red(Messages[i].mmBGColor[0]),
									 get_green(Messages[i].mmBGColor[0]),
									 get_blue(Messages[i].mmBGColor[0]),
									 shade(shFade, 0x00, 
												 get_trans(Messages[i].mmBGColor[0]))
									);
				}
			}
			if (Messages[i].mmTime <= 0)
				MSG_del_message(i);
		}
	}
}

/*
 * ff1 This render function still can be optimized by
 *		 giving a Message entry a hudLayer & g2dbase;
 *		 BUT: if rendering is done asynchronously, then
 *		 there could be a problem with timer & delete
 */
void msg_render_messages(P_HUD_HOOK_RENDER)
{
	ss_bitmap_header *pHead;
	pcolor palBuf[3];
	char *pString[1];
	int i, nOffset = 0, nStartx, nFontWidth, nFontHeight, nImgWidth;

	G2Dc_fill_area(&gbMsg);
	palBuf[0] = rgbt(0x00, 0x00, 0x00, 0x00);
	for (i = 0; i < C_MSG_MAXLINE; i++)
	{
		if (Messages[i].mmUsed)
		{
			nStartx = 0;
			nImgWidth = gbMsg.bSizeImg.cx;
			if (Messages[i].mmFlags & F_MSG_CENTER ||
					Messages[i].mmFlags & F_MSG_RIGHT)
			{
				Font_get_size(Messages[i].mmFont, Messages[i].mmText,
											&nFontWidth, &nFontHeight);
				nImgWidth = min(nImgWidth, nFontWidth);
			}
			palBuf[1] = Messages[i].mmBGColor[1];
			palBuf[2] = Messages[i].mmFGColor[1];
			pString[0] = Messages[i].mmText;
			pHead = G2D_write_string(&pString[0], 1,
															 Messages[i].mmFont, 
															 F_WSTR_DRAW_BORDER, 0, 
															 nImgWidth);
			gbMsg.bBoxArea.top = nOffset;
			gbMsg.bBoxArea.bottom = gbMsg.bBoxArea.top + pHead->height;
			if (Messages[i].mmFlags & F_MSG_CENTER)
				nStartx = gbMsg.bSizeImg.cx / 2 - pHead->width / 2;
			if (Messages[i].mmFlags & F_MSG_RIGHT)
				nStartx = gbMsg.bSizeImg.cx - pHead->width;
			if (nStartx < 0)
				pHead->y1 -= nStartx;
			nStartx = max(nStartx, 0);
			gbMsg.bBoxArea.left = nStartx;
			gbMsg.bBoxArea.right = min(gbMsg.bBoxArea.left + pHead->width,
																 gbMsg.bSizeImg.cx);
			G2Dc_head_to_base(&gbMsg, pHead, palBuf, true);
			nOffset += pHead->height;
			Bmp_free_bitmap(pHead);
		}
	}
	resetArea(gbMsg);
	appendHudLayer(pLayer, hlMsg);
}

bool	msg_destroy_messages(P_HUD_HOOK_DEST)
{
	short i;

	for (i = 0; i < C_MSG_MAXLINE; i++)
		MSG_del_message(i);

	return true;
}



/*
 * End of src/hud.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
