/*
 * line.c -
 *  3D line drawing
 */

/*
 * File:    src/line.c
 * Purpose: Low-level line drawing functions for the software renderer
 * Prefix:  Line_
 * Note:    
 * Music^H^H^H^H^HRadio show: The Hitch Hiker's Guide to the Galaxy
 * TODO:    
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "gamemem.h"
#include "main.h"
#include "palette.h"
#include "plot.h"
#include "render.h"
#include "vec.h"




/*
 * Imported globals (BAD! NAUGHTY!)
 */
extern int Origin_x;
extern int Origin_y;
extern int Window_width;
extern int Window_height;



/*
 * Used by the cyberspace engine
 * Plot a line (already transformed and clipped)
 * To get some sort of thing up and running quickly it's not shaded or depth
 *  cued yet, just drawn at a single (palette) colour value
 */
void Line_draw (const vec *v1, const vec *v2, pixel *bitmap, uint32 rgb)
{

  /*
   * Get the (sub-pixel) screen coordinates of the endpoints
   */
  double screen_x1 = Origin_x + Camera->screen_z * v1->x / v1->z;
  double screen_y1 = Origin_y - Camera->screen_z * v1->y / v1->z;
  double screen_x2 = Origin_x + Camera->screen_z * v2->x / v2->z;
  double screen_y2 = Origin_y - Camera->screen_z * v2->y / v2->z;

  double delta_x = screen_x2 - screen_x1;
  double delta_y = screen_y2 - screen_y1;

  pixel shade = Pal_rgb_to_pixel (rgb);

  if (fabs (delta_x) > fabs (delta_y))
  {

    /*
     * Further to go x than y
     */
    double gradient = delta_y / delta_x;
    int sx, ex;
    int x;
    long ycnt, ystep;
    if (delta_x < 0)
    {
      /*
       * Make sure we are going from left to right
       */
      double swap_temp;
      swap_temp = screen_x1;
      screen_x1 = screen_x2;
      screen_x2 = swap_temp;
      screen_y1 = screen_y2;
    }

    /*
     * Integer pixel start and end
     * For a "shallow" line a pixel is drawn for each column whose centre is
     *  between the x endpoints
     */
    sx = floor (screen_x1 + 0.5);
    ex = floor (screen_x2 + 0.5);
    assert ((sx >= 0) && (sx <= Window_width));
    assert ((ex >= 0) && (ex <= Window_width));

    /*
     * Starting y is at the midpoint of the first pixel drawn
     */
    screen_y1 += gradient * (sx + 0.5 - screen_x1);
    ycnt  = screen_y1 * 65536;
    ystep = gradient  * 65536;
    for (x = sx; x < ex; ++x)
    {
      int y = ycnt >> 16;
      bitmap [x + Window_width * y] = shade;
      ycnt += ystep;      
    }

  }
  else
  {

    /*
     * Further to go y than x. This is exactly analagous to the above but in
     *  the other direction
     */
    double gradient = delta_x / delta_y;
    int sy, ey;
    int y;
    long xcnt, xstep;
    if (delta_y < 0)
    {
      /*
       * Make sure we are going from top to bottom
       */
      double swap_temp;
      screen_x1 = screen_x2;
      swap_temp = screen_y1;
      screen_y1 = screen_y2;
      screen_y2 = swap_temp;
    }

    /*
     * Integer pixel start and end
     */
    sy = floor (screen_y1 + 0.5);
    ey = floor (screen_y2 + 0.5);
    assert ((sy >= 0) && (sy <= Window_height));
    assert ((ey >= 0) && (ey <= Window_height));

    screen_x1 += gradient * (sy + 0.5 - screen_y1);
    xcnt  = screen_x1 * 65536;
    xstep = gradient  * 65536;
    for (y = sy; y < ey; ++y)
    {
      int x = xcnt >> 16;
      bitmap [x + Window_width * y] = shade;
      xcnt += xstep;      
    }

  }

}




/*
 * As Line_draw() above, but interpolate between 2 rgb values
 * These aren't the fastest line drawing algorithms in the world ...
 * Hmmm, fp exceptions. Better might be to calculate endpoints first off and
 *  check the pathological cases then ...
 */
void Line_draw_inter (const vec *v1, const vec *v2, pixel *bitmap,
		      uint32 rgb1, uint32 rgb2)
{

  /*
   * Get the (sub-pixel) screen coordinates of the endpoints
   */
  double screen_x1 = Origin_x + Camera->screen_z * v1->x / v1->z;
  double screen_y1 = Origin_y - Camera->screen_z * v1->y / v1->z;
  double screen_x2 = Origin_x + Camera->screen_z * v2->x / v2->z;
  double screen_y2 = Origin_y - Camera->screen_z * v2->y / v2->z;

  double delta_x = screen_x2 - screen_x1;
  double delta_y = screen_y2 - screen_y1;
  double fdx = fabs (delta_x);
  double fdy = fabs (delta_y);

  long r, rstep, g, gstep, b, bstep;

  if ((fdx < A_VERY_SMALL_THING) && (fdy < A_VERY_SMALL_THING)) return;

  /*
   * Interpolation in r, g, b
   */
  if (fdx > fdy)
  {

    /*
     * Further to go x than y
     */
    double gradient = delta_y / delta_x;
    int sx, ex;
    int x;
    long ycnt, ystep;
    if (delta_x < 0)
    {
      /*
       * Make sure we are going from left to right
       */
      double swap_temp;
      uint32 trgb;
      swap_temp = screen_x1;
      screen_x1 = screen_x2;
      screen_x2 = swap_temp;
      screen_y1 = screen_y2;
      trgb = rgb1;
      rgb1 = rgb2;
      rgb2 = trgb;
      delta_x = -delta_x;
    }

    /*
     * Integer pixel start and end
     */
    sx = floor (screen_x1 + 0.5);
    ex = floor (screen_x2 + 0.5);
    assert ((sx >= 0) && (sx <= Window_width));
    assert ((ex >= 0) && (ex <= Window_width));

    r = (rgb1 & 0xff) << 16;
    g = (rgb1 & 0xff00) << 8;
    b = (rgb1 & 0xff0000);
    /*
     * The casts here are to force the differences to be signed, or wacky
     *  results ensue
     */
    rstep = (((long) (rgb2 & 0xff) - (long) (rgb1 & 0xff)) << 16) / delta_x;
    gstep = (((long) (rgb2 & 0xff00) - (long) (rgb1 & 0xff00)) << 8) / delta_x;
    bstep = ((long) (rgb2 & 0xff0000) - (long) (rgb1 & 0xff0000)) / delta_x;

    /*
     * Starting y is at the midpoint of the first pixel drawn
     */
    screen_y1 += gradient * (sx + 0.5 - screen_x1);
    ycnt  = screen_y1 * 65536;
    ystep = gradient  * 65536;
    for (x = sx; x < ex; ++x)
    {
      int y = ycnt >> 16;
      bitmap [x + Window_width * y] = Pal_rgb_to_pixel
	((b & 0xff0000) | ((g & 0xff0000) >> 8) | ((r & 0xff0000) >> 16));
      ycnt += ystep;      
      r += rstep;
      g += gstep;
      b += bstep;
    }

  }
  else
  {

    /*
     * Further to go y than x
     */
    double gradient = delta_x / delta_y;
    int sy, ey;
    int y;
    long xcnt, xstep;
    if (delta_y < 0)
    {
      /*
       * Make sure we are going from left to right
       */
      double swap_temp;
      uint32 trgb;
      screen_x1 = screen_x2;
      swap_temp = screen_y1;
      screen_y1 = screen_y2;
      screen_y2 = swap_temp;
      trgb = rgb1;
      rgb1 = rgb2;
      rgb2 = trgb;
      delta_y = -delta_y;
    }

    /*
     * Integer pixel start and end
     */
    sy = floor (screen_y1 + 0.5);
    ey = floor (screen_y2 + 0.5);
    assert ((sy >= 0) && (sy <= Window_height));
    assert ((ey >= 0) && (ey <= Window_height));

    r = (rgb1 & 0xff) << 16;
    g = (rgb1 & 0xff00) << 8;
    b = (rgb1 & 0xff0000);
    rstep = (((long) (rgb2 & 0xff) - (long) (rgb1 & 0xff)) << 16) / delta_y;
    gstep = (((long) (rgb2 & 0xff00) - (long) (rgb1 & 0xff00)) << 8) / delta_y;
    bstep = ((long) (rgb2 & 0xff0000) - (long) (rgb1 & 0xff0000)) / delta_y;

    screen_x1 += gradient * (sy + 0.5 - screen_y1);
    xcnt  = screen_x1 * 65536;
    xstep = gradient  * 65536;
    for (y = sy; y < ey; ++y)
    {
      int x = xcnt >> 16;
      bitmap [x + Window_width * y] = Pal_rgb_to_pixel
	((b & 0xff0000) | ((g & 0xff0000) >> 8) | ((r & 0xff0000) >> 16));
      xcnt += xstep;      
      r += rstep;
      g += gstep;
      b += bstep;
    }

  }

}



