/*
 * map.h -
 *  Map file definitions
 */


#ifndef MAP_H
#define MAP_H


/*
 * #includes
 */
#include "graph2d.h"  /* bounding_box */
#include "matrix.h"
#include "resource.h"





#define MAP_CHUNK_INFO               4
#define MAP_CHUNK_TILE               5
#define MAP_CHUNK_TEXTURE            7
#define MAP_CHUNK_OBJECT             8
#define MAP_CHUNK_OBJ_XREF           9
#define MAP_CHUNK_WEAPON            10
#define MAP_CHUNK_AMMO              11
#define MAP_CHUNK_PROJECTILE        12
#define MAP_CHUNK_EXPLOSIVE         13
#define MAP_CHUNK_PATCH             14
#define MAP_CHUNK_HARDWARE          15
#define MAP_CHUNK_SOFTS             16
#define MAP_CHUNK_FIXTURE           17
#define MAP_CHUNK_MISC              18
#define MAP_CHUNK_SWITCH            19
#define MAP_CHUNK_DOOR              20
#define MAP_CHUNK_ANIMATED          21
#define MAP_CHUNK_TRIGGER           22
#define MAP_CHUNK_CONTAINER         23
#define MAP_CHUNK_CRITTER           24
#define MAP_CHUNK_SURVEILLANCE      43
#define MAP_CHUNK_SURVEILLANCE_CAM  44
#define MAP_CHUNK_ANIM_CONTROL      51
#define MAP_CHUNK_ANIM_SCREEN_COUNT 52


#define UW_MAP_FILE Res_make_path (UW_DATA_PATH, "lev.ark")

/*
 * Tile flags
 */
#define TF_TEXTURE_OFFSET 0x0000001F
#define TF_OWN_TEXTURE    0x00000100
#define TF_WHICH_SLOPE    0x00000C00
#define TF_SLOPE_BOTH     0x00000000
#define TF_SLOPE_OPPOSITE 0x00000400
#define TF_SLOPE_FLOOR    0x00000800
#define TF_SLOPE_CEILING  0x00000C00
#define TF_LOWER_SHADE    0x000F0000
#define TF_UPPER_SHADE    0x0F000000
#define	TF_RENDERED	  0x40000000   /* tile was in view     */
#define	TF_AUTOMAPPED	  0x80000000   /* tile shows up on hud */

/*
 * Extract lighting details from the flag word. The casts are to force the
 *  results to be signed (dark values are signed, but are clamped to
 *  0-(DARK_SHADES - 1) at render time)
 */
#define GET_LOWER_SHADE(t) ((long) ((t).flags & TF_LOWER_SHADE) >> 16)
#define GET_UPPER_SHADE(t) ((long) ((t).flags & TF_UPPER_SHADE) >> 24)

#define GET_WALL_TEXTURE(t) ((t).texture & 0x3f)


#define TILE_SOLID       0x00
#define TILE_OPEN        0x01
#define TILE_DIAGONAL_SE 0x02
#define TILE_DIAGONAL_SW 0x03
#define TILE_DIAGONAL_NW 0x04
#define TILE_DIAGONAL_NE 0x05
#define TILE_SLOPE_N     0x06
#define TILE_SLOPE_E     0x07
#define TILE_SLOPE_S     0x08
#define TILE_SLOPE_W     0x09
#define TILE_VALLEY_NW   0x0A
#define TILE_VALLEY_NE   0x0B
#define TILE_VALLEY_SE   0x0C
#define TILE_VALLEY_SW   0x0D
#define TILE_RIDGE_SE    0x0E
#define TILE_RIDGE_SW    0x0F
#define TILE_RIDGE_NW    0x10
#define TILE_RIDGE_NE    0x11
#define TILE_NUM_TYPES   0x12

#define TILE_DIAGONAL_BASE 0x02
#define TILE_SLOPE_BASE    0x06
#define TILE_DSLOPE_BASE   0x0A




#define RENDER_ACTIVE   0x0001  /* currently visiting this tile        */
#define RENDER_GOT_VIEW 0x0002  /* view coordinates already calculated */
#define RENDER_CEILING  0x0030  /* ceiling orientation                 */
#define RENDER_FLOOR    0x00C0  /* floor orientation                   */

#define TILE_GRID_SIZE 256

/*
 * Map size in tiles
 */
#define MAP_WIDTH  64
#define MAP_HEIGHT 64

/*
 * Wall flags
 */
#define WF_IGNORE     0x0001 /* always reject - 4th wall of diagonal */
#define WF_DRAW       0x0002
#define WF_NO_RECURSE 0x0004 /* tile behind exists but not visible   */



/*
 * Map info chunk. This is chunk (4004 + 100*mapno)
 */
typedef struct tag_map_info
{
  unsigned int xsize;        /* map width in tiles  (always 64) */
  unsigned int ysize;        /* map height in tiles (always 64) */
  unsigned int log_xsize;    /* log_2 map width  (always 6)     */
  unsigned int log_ysize;    /* log_2 map height (always 6)     */
  unsigned int height_shift; /* height scale factor             */
  bool         cyberspace;   /* cyberspace flag                 */
} map_info;

/*
 * Global map info struct (there's only ever one of these, of course)
 */
extern map_info Map_info;


struct tag_map_tile;

typedef struct tag_map_vertex
{
  double x;
  double y;
  double z;
  union
  {
    int dark;   /* for "real-world" */
    uint32 rgb; /* for cyberspace   */
  } u;
} map_vertex;


#define MAX_VERTEX_TILE 24


typedef struct tag_map_tile
{

  short int type;
  short int floor;
  short int ceiling;
  short int slope;
  int index;
  int texture;
  unsigned long flags;

  int x;
  int y;

  byte state [4];

  int num_vertices;
  int vertex [MAX_VERTEX_TILE];

  int first_surface;

  /*
   * Rendering stuff
   */
  unsigned long frame;     /* last frame visited       */
  short_vec world [8];     /* world coords of vertices */
  int rflags;

  /*
   * For the recursive tile search
   */
/*    int render_depth; */

  /*
   * Index into the dark sucker xref table
   */
  int dark_index;

  /*
   * The BSP renderer uses a linked list as its tile queue
   */
  struct tag_map_tile *render_next;

  /*
   * Only the concern of render.c - line queue (cyberspace)
   */
  struct line_entry *line_queue;
/*    bounding_box bbox; */

} map_tile;




#ifdef __cplusplus
extern "C"
{
#endif

struct tag_map_surface;

void Map_init (void);
void Map_deinit(void);
void Map_init_level (int level);

int Map_floor_height (int x, int y);
int Map_ceiling_height (int x, int y);
int Map_tile_height (map_tile *tile, int x, int y, bool ceiling);
void Map_tile_max_height (const map_tile *tile, double *top, double *bottom,
			  bool accumulate);
int Map_terrain_type (const map_tile *tile);

map_tile *Map_get_tile(unsigned short x, unsigned short y);
map_tile *Map_get_tile_v (const Vector *v);
int Map_add_vertex (const map_vertex *v, const map_tile *t);
int Map_add_surface (const struct tag_map_surface *ms);

#ifdef __cplusplus
}
#endif



extern int Height_scale;
extern map_tile   *Tile;
extern map_vertex *Vertex;


#endif /* MAP_H */
