/*
 * model.c -
 *  3D models
 */

/*
 * File:    src/model.c
 * Purpose: 3D model file interpreter and renderer
 * Prefix:  Mod_
 * Note:    
 * Music:   Hardline, "Double Eclipse"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>     /* malloc()      */
#include <memory.h>     /* memset()      */

#include "types.h"
#include "accel.h"
#include "bitmap.h"
#include "chunks.h"
#include "dark.h"
#include "gamemem.h"
#include "container.h"
#include "fixture.h"
#include "graph2d.h"    /* vec2d         */
#include "main.h"       /* Cant_happen() */
#include "map.h"
#include "model.h"
#include "object.h"
#include "option.h"
#include "render.h"
#include "resource.h"
#include "shade.h"      /* DARK_SHADES   */
#include "surface.h"
#include "texture.h"
#include "uwmodel.h"
#include "vec.h"



/*
 * Debugging
 */
#define MODEL_MESSAGES 0


/*
 * Local function prototypes
 */
model *decode_uw_model (int model_no);
static void load_ss_model_node (model *m, model_node *n, byte *rp);



/*
 * #defines and constants
 */

#define MODEL_CHUNK_BASE 2300
#define MAX_MODEL          80

/*
 * For variable-height models
 */
#define VH_THRESHOLD 32768.f     /* max System Shock level height is 16384 */
#define VH_VALUE     40000.f     /* comfortably more than VH_THRESHOLD     */

/*
 * Global variables
 */
model *Model_cache [MAX_MODEL] = {NULL};




static void init_box_cache (void);
static void deinit_box_cache(void);
static model *ensure_box_cache (int width, int height, int depth,
				int top_texture, int side_texture);
static model *ensure_vh_cache  (int model_no, int height);
static void render_model (map_object *obj, const model *mod, byte *colour);
static void render_node (map_object *obj, const model *mod,
			 const model_node *node, Vector *view, byte *colour);



/*
 * model.c is responsible for the Z-buffer as it will normally be the only
 *  thing using it
 */
float *Z_buffer;


extern int Screen_width;
extern int Screen_height;


/*
 * Mod_init() allocates space for the Z-buffer
 */
void Mod_init (int width, int height)
{

  Z_buffer = Alloc_init (width * height * sizeof *Z_buffer, "Mod_init",
			 "Z-buffer");
  /*
   * CHANGEME Should be called from Mod_init_level() except there isn't one
   *  (yet) (ff1: same for deinit below)
   */
  init_box_cache ();

}

void Mod_deinit(void)
{
	int i;

	free(Z_buffer);
	Z_buffer = NULL;
	/* ff1: CHANGEME too -- as above i guess */
	deinit_box_cache();
	for (i = 0; i < MAX_MODEL; i++)
	{
		if (Model_cache[i])
		{
			free(Model_cache[i]);
			Model_cache[i] = NULL;
		}
	}
}


void Mod_init_frame (void)
{
# if 0 /* Just Too Damn Slow ... */
  int i;
  for (i = 0; i < Screen_width * Screen_height; ++i)
  {
    Z_buffer [i] = 0;
  }
# else
  memset (Z_buffer, 0, Screen_width * Screen_height * sizeof *Z_buffer);
# endif
}




/*
 * This should be extern'd in some header file ... but which?
 */
extern Vector Viewpoint;


void Mod_render (map_object *obj, const model *m, byte *colour)
{
  assert (obj != NULL && m != NULL);
  assert (!(m->flags & M3_VAR_HEIGHT));

  render_model (obj, m, colour);
}


void Mod_render_id (map_object *obj, int mno, byte *colour)
{
  model *m = Mod_load_model (mno);
  assert (obj != NULL && m != NULL);
  if (m->flags & M3_VAR_HEIGHT)
  {
    double height = (double)Map_ceiling_height ((int)obj->pos.x, (int)obj->pos.y) - obj->pos.z;
    m = ensure_vh_cache (mno, (int)height);
  }
  render_model (obj, m, colour);
}


void Mod_render_bridge (map_object *obj)
{

  model *m;

  int width, height, depth;
  int top_texture, side_texture;

  /*
   * Bridges are all of class fixture, and store their dimensions and texture
   *  information in the additional fixture data
   */
  fixture *fix = Fix_get_fixture (obj);
  width        =  fix->flags   & 0x000f;
  height       = (fix->flags   & 0x00f0) >> 4;
  depth        = (fix->flags   & 0xff00) >> 8;
  if (obj->object_id >= OBJID_FORCEBRIDGE1)
  {
    int col = (fix->start_frame == 0) ? 0xff : fix->start_frame;
    top_texture  = col + 0x100;
    side_texture = col + 0x100;
  }
  else
  {
    top_texture  = (fix->texture & 0x00ff);
    side_texture = (fix->texture & 0xff00) >> 8;
  }

  /*
   * Get the dimensions in world units (4 per tile size in x and y, 32 in z)
   */
  if (width == 0)
  {
    if (obj->object_id == OBJID_BRIDGE) width = 4;
    else width = 2;
  }
  width *= TILE_GRID_SIZE / 4;

  if (height == 0) height = 4;
  height *= TILE_GRID_SIZE / 4;

  if (depth == 0) depth = 2;
  depth *= TILE_GRID_SIZE / 32;

  m = ensure_box_cache (width, height, depth, top_texture, side_texture);
  render_model (obj, m, NULL);

}





void Mod_render_crate (map_object *obj)
{

  model *m;

  int width = 0, height = 0, depth = 0;
  int top_texture = 12, side_texture = 11;

  /*
   * Crates are all in class container, and store their dimensions and texture
   *  information in the additional texture data
   */
  container *cnr = Cnr_get_container (obj);

  /*
   * Default dimensions for crates
   */
  switch (obj->object_id)
  {
  case OBJID_SMALL_CRATE:
    width = height = depth = 80;
    break;
  case OBJID_LARGE_CRATE:
    width = height = depth = 160;
    break;
  case OBJID_SECURE_CRATE:
    width = height = depth = 240;
    break;
  default:
    Cant_happen ("Mod_render_crate(): don't know the default dimensions for"
		 " this crate");
  }

  /*
   * REMEMBER I suspect that when the height changes, the base remains where
   *  it is and the crate does not remain centred. That's a pain if so, needs
   *  checking
   */
  if (cnr->width)  width  = 10 * cnr->width;
  if (cnr->height) height = 10 * cnr->height;
  if (cnr->depth)  depth  = 10 * cnr->depth;
  if (cnr->top_texture ) top_texture  = cnr->top_texture;
  if (cnr->side_texture) side_texture = cnr->side_texture;

  m = ensure_box_cache (width, height, depth, top_texture, side_texture);
  render_model (obj, m, NULL);

}



/*
 * Used by Mod_load_model ()
 * Convert a set of u,v coordinates to 2 vectors and an anchor
 */
void Mod_uv2anchor_vec (model *m, model_surface *s, double u [], double v [])
{
  Vector side1, t1_v;
  Vector side2, t2_v;
  double u1 = u [s->vertex [0]] - u [s->vertex [1]];
  double v1 = v [s->vertex [0]] - v [s->vertex [1]];
  double u2 = u [s->vertex [2]] - u [s->vertex [1]];
  double v2 = v [s->vertex [2]] - v [s->vertex [1]];

  Vector va = m->vertex [s->vertex [0]];
  Vector vb = m->vertex [s->vertex [1]];
  Vector vc = m->vertex [s->vertex [2]];

  double tp;
  Vector tv;
  double offset;

  /*
   * Correct for variable-height vertices (we don't know where they are yet)
   */
  vec_subtract (side1, va, vb);
  if (side1.z >  VH_THRESHOLD) side1.z =  256.;
  if (side1.z < -VH_THRESHOLD) side1.z = -256.;
  vec_subtract (side2, vc, vb);
  if (side2.z >  VH_THRESHOLD) side2.z =  256.;
  if (side2.z < -VH_THRESHOLD) side2.z = -256.;
  vec_cross (s->normal, side1, side2);
  tp = vec_magnitude (s->normal);
  vec_divide (s->normal, tp);
  vec_cross (t1_v, side1, s->normal);
  vec_cross (t2_v, side2, s->normal);
  tp = vec_dot (t1_v, side2);
  s->u = t1_v; vec_multiply (s->u, u2);
  tv   = t2_v; vec_multiply (tv,   u1);
  vec_subtract (s->u, s->u, tv);
  vec_divide (s->u, tp);

  s->v = t1_v; vec_multiply (s->v, v2);
  tv   = t2_v; vec_multiply (tv,   v1);
  vec_subtract (s->v, s->v, tv);
  vec_divide (s->v, tp);

# if MODEL_MESSAGES
  printf ("(%g, %g, %g)\n", s->u.x, s->u.y, s->u.z);
  printf ("(%g, %g, %g)\n", s->v.x, s->v.y, s->v.z);
  printf ("\n");
# endif
  /*
   * This is a fudgearound ... we ought to be using anchor u,v coords
   */
  tp = vec_magnitude2 (s->u);
  if (fabs (tp) > 1.e-3)
  {
    t1_v = s->u; offset = u [s->vertex [1]] / vec_magnitude2 (s->u);
  }
  else offset = 0.;
  vec_multiply (t1_v, offset);
  tp = vec_magnitude2 (s->v);
  if (fabs (tp) > 1.e-3)
  {
    t2_v = s->v; offset = v [s->vertex [1]] / vec_magnitude2 (s->v);
  }
  else offset = 0.;
  vec_multiply (t2_v, offset);
  vec_subtract (s->anchor, m->vertex [s->vertex [1]], t1_v);
  vec_subtract (s->anchor, s->anchor, t2_v);
  if (s->anchor.z > VH_THRESHOLD) s->anchor.z = 1024;

# if __DEBUG__
  /*
   * Check that we've got a sensible texture mapping
   */
  assert (fabs (vec_dot (s->u, side1)- u1) < A_VERY_SMALL_THING);
  assert (fabs (vec_dot (s->v, side1)- v1) < A_VERY_SMALL_THING);
  assert (fabs (vec_dot (s->u, side2)- u2) < A_VERY_SMALL_THING);
  assert (fabs (vec_dot (s->v, side2)- v2) < A_VERY_SMALL_THING);
# endif /* __DEBUG__ */

}


/*
 * Some useful macros for System Shock model data types
 */
#define READ_M3_FLOAT(f,rp) do { \
  (f) = *rp + (rp [1] << 8) + (rp [2] << 16) + (rp [3] << 24); rp += 4; \
  (f) /= 256; \
} while (0)

#define READ_M3_VECTOR(v,rp) do { \
  READ_M3_FLOAT ((v).x, rp); \
  READ_M3_FLOAT ((v).y, rp); \
  READ_M3_FLOAT ((v).z, rp); \
} while (0)



model *Mod_load_model (int model_no)
{

  block_handle bh;
  size_t size;
  unsigned char *raw;
  unsigned char *rp;
  model *m;

  if ((model_no >= MAX_MODEL) || (model_no < 0))
  {
    Cant_happen ("Mod_load_model(): no such model");
  }

  if (Model_cache [model_no] != NULL)
  {
    return Model_cache [model_no];
  }

# if MODEL_MESSAGES
  printf ("Mod_load_model(): loading model %d\n", model_no);
# endif

  if (Option_underworld)
  {
    Model_cache [model_no] = UW_mod_load (model_no);
    return Model_cache [model_no];
  }

  /*
   * We don't have such a model ... load up its chunk
   */
  bh = Res_load_subchunk (MODEL_CHUNK_BASE + model_no, 0, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Mod_load_model(): couldn't load model chunk");
  }

  raw = *bh;
  rp = raw;

  /*
   * Model header ... I don't know what most of this is
   */
  rp += 0x08;

  m = malloc (sizeof *m);
  m->num_vertices = 0;
  m->flags        = 0;
  load_ss_model_node (m, &(m->root), rp);

/*    Res_free_chunk_by_id (MODEL_CHUNK_BASE + model_no); */

  Model_cache [model_no] = m;
  return m;

}



void load_ss_model_node (model *m, model_node *n, byte *rp)
{

  int i;
  Vector tv;
  Vector normal;
  Vector point;
  double offset;
  bool in_face = false;
  byte *face_end   = NULL;
  byte *left_node  = NULL;
  byte *right_node = NULL;

  n->num_surfaces = 0;
  n->left  = NULL;
  n->right = NULL;

#define CHECK_IN_FACE(s) do { \
  if (!in_face) Cant_happen ("Mod_load_model(): " s); \
} while (0);

  while (true)
  {

    model_surface ts;
    double u [MAX_VERTEX_MODEL];
    double v [MAX_VERTEX_MODEL];

    int command = *rp + (rp [1] << 8); rp += 2;
    if (command == 0) break;
    switch (command)
    {

    case M3_IGNORE:
      break;

    case M3_FACE:               /* face definition           */
      in_face = true;
      face_end = rp + *rp + (rp [1] << 8); rp += 2;
      face_end -= 2; /* already read the command in */
      READ_M3_VECTOR (ts.normal, rp);
      READ_M3_VECTOR (point, rp);
      ts.flags = 0;
      break;

    case M3_VERTICES:           /* define multiple vertices  */
    {
      int nv;
      nv = *rp + (rp [1] << 8); rp += 2;
      rp += 2;
      for (i = 0; i < nv; ++i)
      {
	READ_M3_VECTOR (tv, rp);
	m->vertex [m->num_vertices++] = tv;
      }
      break;
    }

    case M3_FLAT_FACE:          /* define flat-shaded face   */
      CHECK_IN_FACE ("face vertices set outside face definition");
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("Mod_load_model(): too many surfaces");
      }
      ts.flags |= M3F_FLAT;
      ts.num_vertices = *rp + (rp [1] << 8); rp += 2;
      for (i = 0; i < ts.num_vertices; ++i)
      {
	ts.vertex [i] = *rp + (rp [1] << 8); rp += 2;
      }
      n->surface [n->num_surfaces++] = ts;
      break;

    case M3_COLOUR:             /* set face colour           */
      CHECK_IN_FACE ("face shade set outside face definition");
      ts.colour = *rp + (rp [1] << 8); rp += 2;
      break;

    case M3_SORT_PLANE:         /* don't know                */
    {
      byte *planedef = rp - 2;
      int offset;
      READ_M3_VECTOR (normal, rp);
      READ_M3_VECTOR (point, rp);
      offset = *rp + (rp [1] << 8); rp += 2;
      left_node  = planedef + offset;
      offset = *rp + (rp [1] << 8); rp += 2;
      right_node = planedef + offset;
      break;
    }

    case M3_VERTEX_OFFSET_X:    /* vertex offset X direction */
      rp += 2;
      i = *rp + (rp [1] << 8); rp += 2;
      tv = m->vertex [i];
      READ_M3_FLOAT (offset, rp);
      tv.x += offset;
      m->vertex [m->num_vertices++] = tv;
      break;

    case M3_VERTEX_OFFSET_Y:    /* vertex offset Y direction */
      rp += 2;
      i = *rp + (rp [1] << 8); rp += 2;
      tv = m->vertex [i];
      READ_M3_FLOAT (offset, rp);
      tv.y += offset;
      m->vertex [m->num_vertices++] = tv;
      break;

    case M3_VERTEX_OFFSET_Z:    /* vertex offset Z direction */
      rp += 2;
      i = *rp + (rp [1] << 8); rp += 2;
      tv = m->vertex [i];
      READ_M3_FLOAT (offset, rp);
      tv.z += offset;
      m->vertex [m->num_vertices++] = tv;
      break;

    case M3_VERTEX_OFFSET_XY:   /* vertex offset X, Y        */
      rp += 2;
      i = *rp + (rp [1] << 8); rp += 2;
      tv = m->vertex [i];
      READ_M3_FLOAT (offset, rp);
      tv.x += offset;
      READ_M3_FLOAT (offset, rp);
      tv.y += offset;
      m->vertex [m->num_vertices++] = tv;
      break;

    case M3_VERTEX_OFFSET_XZ:   /* vertex offset X, Z        */
      rp += 2;
      i = *rp + (rp [1] << 8); rp += 2;
      tv = m->vertex [i];
      READ_M3_FLOAT (offset, rp);
      tv.x += offset;
      READ_M3_FLOAT (offset, rp);
      tv.z += offset;
      m->vertex [m->num_vertices++] = tv;
      break;

    case M3_VERTEX_OFFSET_YZ:   /* vertex offset Y, Z        */
      rp += 2;
      i = *rp + (rp [1] << 8); rp += 2;
      tv = m->vertex [i];
      READ_M3_FLOAT (offset, rp);
      tv.y += offset;
      READ_M3_FLOAT (offset, rp);
      tv.z += offset;
      m->vertex [m->num_vertices++] = tv;
      break;

    case M3_VERTEX:             /* define single vertex */
      rp += 2;
      READ_M3_VECTOR (tv, rp);
      m->vertex [m->num_vertices++] = tv;
      break;

    case M3_SHADE:              /* separate colour and shade */
    {
      CHECK_IN_FACE ("shade set outside face definition");
      if (ts.flags & M3F_TEXTURED)
      {
	Cant_happen ("Mod_load(): wrong face type for shading");
      }
      ts.flags |= M3F_AUX_SHADE;

      /*
       * Rather sneakily, we put the darkness data in the texture field as we
       *  aren't needing it for anything else
       */
      ts.colour  = *rp + (rp [1] << 8); rp += 2;
      ts.texture = *rp + (rp [1] << 8); rp += 2;
      break;
    }

    case M3_TEXTURE_MAPPING:
      CHECK_IN_FACE ("texture mapping defined outside face definition");
      i = *rp + (rp [1] << 8); rp += 2;
      for (; i > 0; --i)
      {
	int vno = *rp + (rp [1] << 8); rp += 2;
	READ_M3_FLOAT (u [vno], rp);
	READ_M3_FLOAT (v [vno], rp);
      }
      break;

    case M3_TEXTURE:
    {
      CHECK_IN_FACE ("texture set outside face definition");
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("Mod_load_model(): too many surfaces");
      }
      ts.flags |= M3F_TEXTURED;

      /*
       * Set the model texture. Texture 0 is specific to an individual object,
       *  others are taken from the normal model texture block, chunks 475-525
       */
      ts.texture = *rp + (rp [1] << 8); rp += 2;
      if (ts.texture == 0)
      {
	ts.flags |= M3F_OBJ_TEXTURE;
      }
      else ts.texture =
	     Tex_get_cache_id (ts.texture + CHUNK_TEXTURE_MODEL_BASE, 0);

      ts.num_vertices = *rp + (rp [1] << 8); rp += 2;
      for (i = 0; i < ts.num_vertices; ++i)
      {
	int vno = *rp + (rp [1] << 8); rp += 2;
#       if MODEL_MESSAGES
	printf ("%d\t(%g, %g, %g)\t[%g %g]\n", vno,
		m->vertex [vno].x,
		m->vertex [vno].y,
		m->vertex [vno].z, u [vno], v [vno]);
#       endif
	ts.vertex [i] = vno;
      }

      Mod_uv2anchor_vec (m, &ts, u, v);
      n->surface [n->num_surfaces++] = ts;
      break;
    }

    default:
      Cant_happen ("Mod_load_model(): unknown model definition field");
    }

    if (in_face)
    {
      if (rp > face_end)
      {
	Cant_happen ("Mod_load_model(): face length mismatch");
      }
      else if (rp == face_end)
      {
	in_face = false;
      }
    }

  }

  /*
   * Load up the child nodes, if present
   */
  if (left_node != NULL)
  {
    n->left = malloc (sizeof *n->left);
    if (n->left == NULL)
    {
      Cant_happen ("model.c:load_ss_model_node(): can't allocate child node");
    }
    load_ss_model_node (m, n->left, left_node);

    n->right = malloc (sizeof *n->right);
    if (n->right == NULL)
    {
      Cant_happen ("model.c:load_ss_model_node(): can't allocate child node");
    }
    load_ss_model_node (m, n->right, right_node);

  }

}




/*
 * Used by partition_poly()
 * Project a 3D polygon onto an axis-aligned plane orthogonal to the largest
 *  component of the normal vector (to minimise distortion). This will have
 *  the same arrangement of concave and convex vertices as the original and
 *  hence will split in the same way
 */
void project_poly (model *m, model_surface *ms, vec2d *vertex)
{

  /*
   * There are 6 possible cases, each resulting in a different mapping of
   *  axes onto the 2D plane. Now, watch me go wild with cut and paste 8-)
   */
  double ax = fabs (ms->normal.x); /* `absolute X' */
  double ay = fabs (ms->normal.y); /* likewise     */
  double az = fabs (ms->normal.z);
  int i;

  if ((az > ax) && (az > ay))
  {
    /*
     * It's z ... but is it positive or negative?
     */
    if (ms->normal.z > 0)
    {
      for (i = 0; i < ms->num_vertices; ++i)
      {
	vertex [i].x = m->vertex [ms->vertex [i]].x;
	vertex [i].y = m->vertex [ms->vertex [i]].y;
      }
    }
    else
    {
      for (i = 0; i < ms->num_vertices; ++i)
      {
	vertex [i].x = m->vertex [ms->vertex [i]].y;
	vertex [i].y = m->vertex [ms->vertex [i]].x;
      }
    }
  }
  else
  {
    if (ay > ax)
    {
      /*
       * It's y ... but is it positive or negative?
       */
      if (ms->normal.y > 0)
      {
	for (i = 0; i < ms->num_vertices; ++i)
	{
	  vertex [i].x = m->vertex [ms->vertex [i]].z;
	  vertex [i].y = m->vertex [ms->vertex [i]].x;
	}
      }
      else
      {
	for (i = 0; i < ms->num_vertices; ++i)
	{
	  vertex [i].x = m->vertex [ms->vertex [i]].x;
	  vertex [i].y = m->vertex [ms->vertex [i]].z;
	}
      }
    }
    else
    {
      /*
       * It's x ... but is it positive or negative?
       */
      if (ms->normal.x > 0)
      {
	for (i = 0; i < ms->num_vertices; ++i)
	{
	  vertex [i].x = m->vertex [ms->vertex [i]].y;
	  vertex [i].y = m->vertex [ms->vertex [i]].z;
	}
      }
      else
      {
	for (i = 0; i < ms->num_vertices; ++i)
	{
	  vertex [i].x = m->vertex [ms->vertex [i]].z;
	  vertex [i].y = m->vertex [ms->vertex [i]].y;
	}
      }
    }
  }

}



/*
 * Used by Mod_partition_node()
 * Find a convex sub-face of a non-convex poly
 * Returns true if the poly was non-convex, in which case *first contains the
 *  first vertex of the splitting edge and *last the vertex AFTER the last
 *  (no, it's not confusing, I don't know what you mean)
 */
bool partition_poly (model *m, model_surface *ms, int *first, int *last)
{

  /*
   * We do the partitioning in 2D, so we need some 2D vertices
   */
  vec2d vertex [MAX_VERTEX_MS];
  bool is_convex = true;        /* whole poly is convex         */
  bool concave [MAX_VERTEX_MS]; /* concavity tests for vertices */
  vec2d edge [MAX_VERTEX_MS];   /* edge list                    */
  int i;
  int nverts = ms->num_vertices; /* we'll see a lot of this */

  /*
   * First, project the poly onto an axis-aligned plane
   */
  project_poly (m, ms, vertex);

  /*
   * Get an edge list
   */
  vec2_subtract (edge [0], vertex [0], vertex [nverts - 1]);
  for (i = 1; i < ms->num_vertices; ++i)
  {
    vec2_subtract (edge [i], vertex [i], vertex [i - 1]);
  }

  /*
   * For a clockwise-wound poly, vertex ABC is concave if
   *  x_AB*y_BC > x_BC*y_AB (trust me ...)
   */
  for (i = 0; i < ms->num_vertices; ++i)
  {
    int next = (i + 1) % nverts;
    if (edge [i].x * edge [next].y > edge [next].x * edge [i].y)
    {
      is_convex = false;
      concave [i] = true;
    }
    else concave [i] = false;
  }
  if (is_convex) return false; /* whole poly is convex, no need to proceed */

  printf ("Found non-convex poly (%d verts)\n", nverts);

  /*
   * Now comes the fun part
   */
  for (*first = 0; *first < nverts; ++*first)
  {

    /*
     * Search for a concave vertex with a convex one immediately following.
     *  That is where we shall start
     */
#   define next(x) (((x) + 1) % nverts)
#   define prev(x) (((x) + nverts - 1) % nverts)
#   define conc(a,b) ((a).x * (b).y > (b).x * (a).y)
    if (concave [*first] && !concave [next (*first)])
    {
      int second = next (*first);
      vec2d edge2;
      vec2_subtract (edge2, vertex [second], vertex [*first]);

      /*
       * Start with a triangle
       */
      *last = next (second);
      while (1)
      {
	int apex = prev (*last);
	vec2d newedge;
	bool discard = false;

	if (concave [apex]) break; /* new triangle is outside poly */

	/*
	 * Check to see if we have gone past the first vertex: this is the case
	 *  if (last, first, second) is concave
	 */
	vec2_subtract (newedge, vertex [*first], vertex [*last]);
	if (conc (newedge, edge2)) break;

	/*
	 * OK, our candidate poly is still convex and is entirely within the
	 *  main poly. Check if any other vertices are inside the new triangle
	 *  (last, first, apex)
	 */
	for (i = next (*last); i != *first; i = next (i))
	{
	  vec2d ap;
	  vec2d pb;
	  /*
	   * The first edge of the triangle is (last->first)
	   */
	  vec2_subtract (ap, vertex [i], vertex [*last]);
	  vec2_subtract (pb, vertex [*first], vertex [i]);
	  if (!conc (ap, pb)) continue; /* it's outside */

	  /*
	   * The second edge of the triangle is (first->apex)
	   */
	  vec2_subtract (ap, vertex [i], vertex [*first]);
	  vec2_subtract (pb, vertex [apex], vertex [i]);
	  if (!conc (ap, pb)) continue; /* it's outside */

	  /*
	   * The third ... oh, work it out yourself
	   */
	  vec2_subtract (ap, vertex [i], vertex [apex]);
	  vec2_subtract (pb, vertex [*last], vertex [i]);
	  if (conc (ap, pb))
	  {
	    /*
	     * Ah-HA!
	     */
	    discard = true;
	    break;
	  }
	}

	if (discard)
	{
	  /*
	   * Oh, so close. But the new triangle contained another point, and we
	   *  can't have that, or we'll end up including bits that are outside
	   *  the poly
	   */
	  break;
	}

	/*
	 * And try the next vertex
	 */
	*last = next (*last);

      }

      if (*last != next (next (*first)))
      {
	/*
	 * We have at least a triangle to chop off, so return it
	 */
	return true;
      }

    }

#   undef next
#   undef prev
#   undef conc

    /*
     * OK, see if it is immediately clockwise of a convex vertex, and if so,
     *  perform the same search backwards
     * Note that since we are now walking backwards around the poly the sense
     *  of the concavity test is reversed
     * Did I mention going wild with cut and paste?
     */
#   define next(x) (((x) + nverts - 1) % nverts)
#   define prev(x) (((x) + 1) % nverts)
#   define conc(a,b) ((a).x * (b).y < (b).x * (a).y)
    if (concave [*first] && !concave [next (*first)])
    {
      int second = next (*first);
      vec2d edge2;
      vec2_subtract (edge2, vertex [second], vertex [*first]);

      /*
       * Start with a triangle
       */
      *last = next (second);
      while (1)
      {
	int apex = prev (*last);
	vec2d newedge;
	bool discard = false;

	if (concave [apex]) break; /* new triangle is outside poly */

	/*
	 * Check to see if we have gone past the first vertex: this is the case
	 *  if (last, first, second) is concave
	 */
	vec2_subtract (newedge, vertex [*first], vertex [*last]);
	if (conc (newedge, edge2)) break;

	/*
	 * OK, our candidate poly is still convex and is entirely within the
	 *  main poly. Check if any other vertices are inside the new triangle
	 *  (last, first, apex)
	 */
	for (i = next (*last); i != *first; i = next (i))
	{
	  vec2d ap;
	  vec2d pb;
	  /*
	   * The first edge of the triangle is (last->first)
	   */
	  vec2_subtract (ap, vertex [i], vertex [*last]);
	  vec2_subtract (pb, vertex [*first], vertex [i]);
	  if (!conc (ap, pb)) continue; /* it's outside */

	  /*
	   * The second edge of the triangle is (first->apex)
	   */
	  vec2_subtract (ap, vertex [i], vertex [*first]);
	  vec2_subtract (pb, vertex [apex], vertex [i]);
	  if (!conc (ap, pb)) continue; /* it's outside */

	  /*
	   * The third ... oh, work it out yourself
	   */
	  vec2_subtract (ap, vertex [i], vertex [apex]);
	  vec2_subtract (pb, vertex [*last], vertex [i]);
	  if (conc (ap, pb))
	  {
	    /*
	     * Ah-HA!
	     */
	    discard = true;
	    break;
	  }
	}

	if (discard)
	{
	  /*
	   * Oh, so close. But the new triangle contained another point, and we
	   *  can't have that, or we'll end up including bits that are outside
	   *  the poly
	   */
	  break;
	}

	/*
	 * And try the next vertex
	 */
	*last = next (*last);

      }

      if (*last != next (next (*first)))
      {
	/*
	 * We have at least a triangle to chop off, so return it
	 * Invert first and last ...
	 */
	int t = prev (*last);
	*last = prev (*first);
	*first = t;
	return true;
      }

    }

#   undef next
#   undef prev
#   undef conc

    /*
     * Oops, didn't find a poly that time.
     */
  }

  Cant_happen ("model.c:partition_poly(): couldn't split poly");
  return false;

}




/*
 * Ensure a node contains only convex faces, by partitioning any non-convex
 *  ones into convex sub-polys
 */
void Mod_partition_node (model *m, model_node *n)
{

  int nfaces = n->num_surfaces; /* number of possibly convex faces */
  int i;

  for (i = 0; i < nfaces; ++i)
  {
    int first; /* first split vertex */
    int last;  /* can you guess?     */
    model_surface *os = &(n->surface [i]); /* shorthand to original surface */
    while (partition_poly (m, os, &first, &last))
    {
      /*
       * Face was non-convex, and we have a new convex sub-poly to add
       */
      int j;
      int old_nverts = os->num_vertices; /* shorthand */
      int new_nverts;
      model_surface ms = *os;

      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("Mod_partition_node(): too many surfaces");
      }

      /*
       * Actually I was teasing with the "can you guess?" above ... last is
       *  the vertex AFTER the last vertex in the new poly
       */
      if (last > first)
      {
	new_nverts = last - first;
      }
      else
      {
	new_nverts = (last + old_nverts) - first;
      }
      ms.num_vertices = new_nverts;
      for (j = 0; j < new_nverts; ++j)
      {
	ms.vertex [j] = os->vertex [(j + first) % old_nverts];
      }

      /*
       * Add the new surface
       */
/*        ms.colour = UW_debug_shade++; */
      n->surface [n->num_surfaces++] = ms;

      /*
       * Chop the new poly off the old one by deleting vertices
       */
      if (last == 0) last = old_nverts;
      if (last > first)
      {
	for (j = first + 1; j < old_nverts - new_nverts + 2; ++j)
	{
	  os->vertex [j] = os->vertex [j + new_nverts - 2];
	}
      }
      else
      {
	for (j = 0; j < first - last + 2; ++j)
	{
	  os->vertex [j] = os->vertex [j + last - 1];
	}
      }
      os->num_vertices -= new_nverts - 2;

    }
  }

}



static vec rel_pos;
static mat mod2world;
static mat mod2view;
static float model_dark;

void render_model (map_object *obj, const model *mod, byte *colour)
{

  Vector tv;

  Vector view [MAX_VERTEX_MODEL];
  int i;

  /*
   * Get the transformation matrix from model to view coordinates
   */
  if (Option_underworld) Mat_transpose (mod2world, obj->orientation);
  else
  {
    mod2world.x.x =  obj->orientation.x.x;
    mod2world.x.y = -obj->orientation.x.z;
    mod2world.x.z =  obj->orientation.x.y;
    mod2world.y.x =  obj->orientation.y.x;
    mod2world.y.y = -obj->orientation.y.z;
    mod2world.y.z =  obj->orientation.y.y;
    mod2world.z.x =  obj->orientation.z.x;
    mod2world.z.y = -obj->orientation.z.z;
    mod2world.z.z =  obj->orientation.z.y;
  }
  Mat_multiply (mod2view, Camera->orientation, mod2world);

  /*
   * Get the relative object position in model coordinates; add this before
   *  transforming a point
   */
  vec_subtract (tv, obj->pos, Viewpoint);
  mat_vec_multiply_transpose (rel_pos, mod2world, tv);
  if (Option_underworld)
  {
    /*
     * Underworld object positions are at the model "centre" in X and Y and
     *  the bottom in Z
     */
    rel_pos.x -= mod->centre.x;
    rel_pos.y -= mod->centre.y;
  }

  /*
   * Overall dark level at this model. We don't bother with per-surface (nor
   *  yet per-vertex) shading for models at this stage, maybe later
   */
  model_dark = obj->dark + 0.5 -
    DARK_SHADES * Dark_do_point (obj->tile, &(obj->pos));
  if (model_dark < 0.) model_dark = 0.; /* don't allow to saturate */

  /*
   * Transform all the vertices at once ...
   *  consider making a list of faces which pass the backface cull and
   *  transforming only visible vertices, for added efficiency
   */
  for (i = 0; i < mod->num_vertices; ++i)
  {
    vec_add (tv, mod->vertex [i], rel_pos);
    mat_vec_multiply (view [i], mod2view, tv);
  }

  render_node (obj, mod, &(mod->root), view, colour);

}



/*
 * I'm sure this shouldn't be necessary, but cybertoggles are awkward. They
 *  have class 9 (switches), logically enough, and flat shaded faces as usual
 *  for c/space, but there is no colour scheme in objprop.dat for them. The
 *  following is a copy of the cyber items (class 8/5) colour scheme, which
 *  cybertoggles seem to use
 */
static byte Default_cyber_shade [6] = { 0x51, 0x48, 0x45, 0x42, 0x39, 0x36 };


/*
 * Recursively render model nodes
 */
void render_node (map_object *obj, const model *mod, const model_node *node,
		  Vector *view, byte *colour)
{

  int i, j;

  for (i = 0; i < node->num_surfaces; ++i)
  {

    Vector tv;
    surface ts;
    const model_surface *ms = &(node->surface [i]);
    float face_dark;
    extern clipregion_t Object_region_view;

    /*
     * Back-face cull
     */
    vec_add (tv, rel_pos, mod->vertex [ms->vertex [0]]);
    if (vec_dot (ms->normal, tv) > -A_VERY_SMALL_THING)
    {
      continue;
    }

    /*
     * It's drawable ... now make a surface out of it
     */
    ts.num_vertices = ms->num_vertices;
    for (j = 0; j < ts.num_vertices; ++j)
    {
      ts.vertex [j] = view [ms->vertex [j]];
    }
    mat_vec_multiply (ts.normal, mod2view, ms->normal);
    ts.flags = SF_ZBUFFER;
    ts.shade = ms->colour;
    if (ms->flags & M3F_SHADED)
    {
      face_dark = model_dark;
      ts.flags |= SF_SHADED;
    }
    else
    {
      /*
       * Dark_do_plane wants the normal in world coordinates
       */
      vec world_normal;
      mat_vec_multiply (world_normal, mod2world, ms->normal);
      face_dark = obj->dark + 0.5 -
	DARK_SHADES * Dark_do_plane (obj->tile, &(obj->pos), &world_normal);
    }

    /*
     * Set the dark values for the surface vertices
     */
    if (ms->flags & M3F_AUX_SHADE)
    {
      /*
       * Auxiliary palette with extra shading information. This is used for
       *  cyberspace objects in System Shock. The auxpal is passed in in the
       *  colour[] array (6 palette indices), and the face colour field is an
       *  index into this. The face `texture' field contains a shade value to
       *  add to face dark
       */
      if (ms->colour > 0)
      {
	if (ms->colour > 6)
	{
	  printf ("%d\n", ms->colour);
	  Cant_happen ("Mod_render(): bad cyber colour value");
	}
	if (colour == NULL) colour = Default_cyber_shade;
	face_dark += ms->texture * DARK_SHADES / 8;
	ts.shade = colour [ms->colour - 1];
      }
    }

    for (j = 0; j < ts.num_vertices; ++j)
    {
      ts.dark [j] = face_dark;
      if (ms->flags & M3F_SHADED)
      {
	ts.dark [j] += mod->dark [ms->vertex [j]];
      }
      if (ts.dark [j] < 0.) ts.dark [j] = 0.;
      if (ts.dark [j] >= DARK_SHADES) ts.dark [j] = DARK_SHADES - 1;
    }

    /*
     * We ought to clip it against the tile's clip region at this point ...
     *  unfortunately we've long since lost it. Need to sort this out.
     */
    if (Clip_surface_by_region (&ts, &Object_region_view) != CLIP_ACCEPT)
    {
      continue;
    }

    /*
     * Now we draw it!
     */
    if (ms->flags & M3F_TEXTURED)
    {
      int texno = ms->texture;
      if (ms->flags & M3F_OBJ_TEXTURE)
      {
	texno = obj->fetch_texture_hook (obj);
      }
      ts.texture = texno;

      /*
       * For a textured face we must transform the texture map vectors as well
       */
      vec_add (tv, rel_pos, ms->anchor);
      mat_vec_multiply (ts.anchor, mod2view, tv);
      mat_vec_multiply (ts.u,      mod2view, ms->u);
      mat_vec_multiply (ts.v,      mod2view, ms->v);
      ts.flags = SF_TEXTURED | SF_ZBUFFER;
      ts.shade = floor (ts.dark [0]);      /* flat darkness value */
      Acc_draw_surface (&ts);

    }
    else
    {
      /*
       * Check for transparencies (force bridges)
       */
      extern unsigned int First_trans_index;
      if (ts.shade >= First_trans_index) ts.flags = SF_TRANSPARENCY;
      Acc_draw_surface (&ts);
    }
  }

  if (node->left != NULL)
  {
    render_node (obj, mod, node->left, view, colour);
  }
  if (node->right != NULL)
  {
    render_node (obj, mod, node->right, view, colour);
  }

}




/*
 * Box cache
 * A `box' in TSSHP model parlance is a cuboidal 3D object whose dimensions and
 *  texture mapping information are defined at the individual map object rather
 *  than in the model itself (System Shock crates, bridges etc.)
 * Bit 8 in the texture byte indicates that the face is flat-shaded, in which
 *  case the low 8 bits are a palette index. If this is not the case, bit 7
 *  indicates that the texture is taken from the wall textures list rather than
 *  the "special" model textures. The original System Shock datafiles use this
 *  bit for this purpose, so is guaranteed not to need more than 7 bits per
 *  texture. Yes, it's a fudge, but since these fields only exist to test for
 *  a cache hit (the actual rendering uses the model surfaces) it's probably
 *  forgiveable.
 * We could use a hash table for this but really I don't think it's worth the
 *  bother
 */

#define BOX_FLAT_SHADE   0x100
#define BOX_WALL_TEXTURE 0x080

struct box_cache_entry
{
  int width;         /* All of these  */
  int height;        /* must match if */
  int depth;         /* a cache hit   */
  int top_texture;   /* is to be      */
  int side_texture;  /* registered    */
  model *mod;
} *Box_cache;

#define BOX_CACHE_ENTRIES 64
int Num_box_cache_entries;



/*
 * Variable height models (that reach up to the ceiling). Only height and 
 *  model number must match here
 */
struct vh_cache_entry
{
  int height; /* of ceiling above model origin */
  int mod_no;
  model *mod;
} *VH_cache;

#define VH_CACHE_ENTRIES 32
int Num_vh_cache_entries;



/*
 * Called every level
 */
void init_box_cache (void)
{

  if (Box_cache == NULL)
  {
    Box_cache = Alloc_init (BOX_CACHE_ENTRIES * sizeof *Box_cache,
			    "init_box_cache", "box cache");
  }
  Num_box_cache_entries = 0;

  if (VH_cache == NULL)
  {
    VH_cache = Alloc_init (VH_CACHE_ENTRIES * sizeof *VH_cache,
			   "init_box_cache", "VH cache");
  }
  Num_vh_cache_entries = 0;

}


void deinit_box_cache(void)
{
  free (Box_cache);
  Box_cache = NULL;
  free (VH_cache);
  VH_cache = NULL;
}


/*
 * A template for a box object. The dimensions and texture mapping information
 *  will be overridden for each individual box
 */
static const model Box_template =
{
  8, /* num_vertices */
  {
    { -128, -256,  128 }, 
    {  128, -256,  128 },
    {  128, -256, -128 },
    { -128, -256, -128 },
    { -128,    0,  128 }, 
    {  128,    0,  128 },
    {  128,    0, -128 },
    { -128,    0, -128 }
  },
  {0},         /* dark   */
  { 0, 0, 0 }, /* centre */
  { 0, 0, 0 }, /* extent */
  0,           /* flags  */
  {
    6, /* num_surfaces */
    {
      { M3F_TEXTURED, 0, 0, 4, { 0, 1, 2, 3 }, {  0,  0, -1 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_TEXTURED, 0, 0, 4, { 4, 7, 6, 5 }, {  0,  0,  1 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_TEXTURED, 0, 0, 4, { 1, 0, 4, 5 }, {  0, -1,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_TEXTURED, 0, 0, 4, { 2, 1, 5, 6 }, {  1,  0,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_TEXTURED, 0, 0, 4, { 3, 2, 6, 7 }, {  0,  1,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_TEXTURED, 0, 0, 4, { 0, 3, 7, 4 }, { -1,  0,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} }
    },
    { { 0, 0, 0 }, 0 }, /* sort_plane (not used) */
    NULL,
    NULL
  }
};



void set_box_texture_map (model *m, int sno, int texture);

model *ensure_box_cache (int width, int height, int depth,
			 int top_texture, int side_texture)
{

  int i;
  model *m;

  for (i = 0; i < Num_box_cache_entries; ++i)
  {
    if ((Box_cache [i].width        == width) &&
	(Box_cache [i].height       == height) &&
	(Box_cache [i].depth        == depth) &&
	(Box_cache [i].top_texture  == top_texture) &&
	(Box_cache [i].side_texture == side_texture))
    {
#     if MODEL_MESSAGES
      printf ("Cache hit for box %dx%dx%d, texture %02x/%02x\n",
	      width, height, depth, top_texture, side_texture);
#     endif

      return Box_cache [i].mod;
    }
  }

  if (Num_box_cache_entries >= BOX_CACHE_ENTRIES)
  {
    Cant_happen ("model.c:ensure_box_cache(): cache overflow, increase"
		 " BOX_CACHE_ENTRIES");
  }

# if MODEL_MESSAGES
  printf ("Creating new box %dx%dx%d, texture %02x/%02x\n",
	  width, height, depth, top_texture, side_texture);
# endif

  ++Num_box_cache_entries;

  Box_cache [i].width        = width;
  Box_cache [i].height       = height;
  Box_cache [i].depth        = depth;
  Box_cache [i].top_texture  = top_texture;
  Box_cache [i].side_texture = side_texture;

  /*
   * Create a new model from the master box template
   */
  m = malloc (sizeof (model));
  Box_cache [i].mod = m;
  *m = Box_template;

  /*
   * Override the stored vertices with the positions we got from the object
   *  definition. We know which vertex is which in the model, because we put
   *  them there ...
   */
  m->vertex [0].x = m->vertex [3].x = m->vertex [4].x = m->vertex [7].x =
    -(width / 2);
  m->vertex [1].x = m->vertex [2].x = m->vertex [5].x = m->vertex [6].x =
    (width / 2);
  m->vertex [0].y = m->vertex [1].y = m->vertex [2].y = m->vertex [3].y =
    -(depth / 2);
  m->vertex [4].y = m->vertex [5].y = m->vertex [6].y = m->vertex [7].y =
    (depth / 2);
  m->vertex [2].z = m->vertex [3].z = m->vertex [6].z = m->vertex [7].z =
    -(height / 2);
  m->vertex [0].z = m->vertex [1].z = m->vertex [4].z = m->vertex [5].z =
    (height / 2);

  /*
   * Top and bottom surfaces
   */
  set_box_texture_map (m, 0, top_texture);
  set_box_texture_map (m, 1, top_texture);

  /*
   * Side surfaces
   */
  set_box_texture_map (m, 2, side_texture);
  set_box_texture_map (m, 3, side_texture);
  set_box_texture_map (m, 4, side_texture);
  set_box_texture_map (m, 5, side_texture);

  return m;

}



void set_box_texture_map (model *m, int sno, int texture)
{

  double ts;
  model_surface *s = &(m->root.surface [sno]);

  assert ((texture & 0xfffffe00) == 0); /* unused bits */

  /*
   * Find the correct texture chunk or colour. Bit 8 (0x100) indicates a flat
   *  shade. Otherwise, bit 7 (0x80) in the texture number indicates that the
   *  texture is taken from the wall texture list, otherwise it is one of the
   *  "special" model texture chunks
   */
  if (texture & BOX_FLAT_SHADE)
  {
    s->flags  = M3F_FLAT;
    s->colour = texture & 0xff;
  }
  else
  {
    s->flags = M3F_TEXTURED |
      (texture & BOX_WALL_TEXTURE ? M3F_WALL_TEXTURE : 0);
    if (texture & BOX_WALL_TEXTURE) s->texture = texture & 0x7f;
    else
    {
      s->texture = Tex_get_cache_id (CHUNK_TEXTURE_MODEL2_BASE + texture, 0);
    }
  }

  /*
   * Texture mapping. This is straightforward; faces are rectangular and
   *  textures are scaled to fit
   */
  s->anchor = m->vertex [s->vertex [0]];
  vec_subtract (s->u, m->vertex [s->vertex [1]], m->vertex [s->vertex [0]]);
  ts = vec_magnitude2 (s->u) / 256.;
  vec_divide (s->u, ts);
  vec_subtract (s->v, m->vertex [s->vertex [3]], m->vertex [s->vertex [0]]);
  ts = vec_magnitude2 (s->v) / 256.;
  vec_divide (s->v, ts);
  vec_cross (s->normal, s->v, s->u);

}



/*
 * Similar is the handling for the VH cache
 */
void correct_vh_node (model *om, model_node *on, model *m, model_node *n);

model *ensure_vh_cache (int model_no, int height)
{

  int i;
  model *m;
  model *vhm;

  for (i = 0; i < Num_vh_cache_entries; ++i)
  {
    if ((VH_cache [i].height       == height) &&
	(VH_cache [i].mod_no       == model_no))
    {
#     if MODEL_MESSAGES
      printf ("Cache hit for vh %d, model %d\n", height, model_no);
#     endif
      return VH_cache [i].mod;
    }
  }

  /*
   * Not found ... we'll have to load up the template model and correct the
   *  height of any roof vertices
   */
  if (Num_vh_cache_entries >= VH_CACHE_ENTRIES)
  {
    Cant_happen ("model.c:ensure_vh_cache(): cache overflow, increase"
		 " VH_CACHE_ENTRIES");
  }

  /*
   * Load up the original model. The model in the main cache is a placeholder
   *  flagged as variable-height
   */
  m = Mod_load_model (model_no);

  /*
   * Copy the basic information into the new model
   */
  vhm = malloc (sizeof *m);
  *vhm = *m;

  /*
   * Vertex heights
   */
  for (i = 0; i < vhm->num_vertices; ++i)
  {
    if (vhm->vertex [i].z > VH_THRESHOLD)
    {
      vhm->vertex [i].z = height;
    }
  }

  /*
   * Recursively correct model heights for the nodes
   */
  correct_vh_node (m, &(m->root), vhm, &(vhm->root));

  vhm->flags &= ~M3_VAR_HEIGHT;
  VH_cache [Num_vh_cache_entries].height = height;
  VH_cache [Num_vh_cache_entries].mod_no = model_no;
  VH_cache [Num_vh_cache_entries].mod    = vhm;
  ++Num_vh_cache_entries;

  return vhm;

}




void correct_vh_node (model *om, model_node *on, model *m, model_node *n)
{

  /*
   * Recursively correct the children, if any
   */
  if (on->left != NULL)
  {
    n->left = malloc (sizeof *(n->left));
    *(n->left) = *(on->left);
    correct_vh_node (om, on->left, m, n->left);
  }
  else n->left = NULL;

  if (on->right != NULL)
  {
    n->right = malloc (sizeof *(n->right));
    *(n->right) = *(on->right);
    correct_vh_node (om, on->right, m, n->right);
  }
  else n->right = NULL;

}
