/*
 * model.h -
 *  3D models for System Shock
 */


#ifndef MODEL_H
#define MODEL_H



/*
 * #includes
 */
#include "clip.h"      /* plane_t */
#include "types.h"
#include "vec.h"



/*
 * Model definition file entries
 */
#define M3_IGNORE           0x0000         /* don't know what this is for */
#define M3_FACE             0x0001         /* define face                 */
#define M3_VERTICES         0x0003         /* define initial vertices     */
#define M3_FLAT_FACE        0x0004         /* vertex list for flat face   */
#define M3_COLOUR           0x0005         /* colour for flat face        */
#define M3_SORT_PLANE       0x0006         /* sort plane for sub-hulls    */
#define M3_VERTEX_OFFSET_X  0x000A
#define M3_VERTEX_OFFSET_Y  0x000B
#define M3_VERTEX_OFFSET_Z  0x000C
#define M3_VERTEX_OFFSET_XY 0x000D
#define M3_VERTEX_OFFSET_XZ 0x000E
#define M3_VERTEX_OFFSET_YZ 0x000F
#define M3_VERTEX           0x0015
#define M3_SHADE            0x001C         /* colour and shade            */
#define M3_TEXTURE_MAPPING  0x0025
#define M3_TEXTURE          0x0026


/*
 * Underworld model definition commands
 */
#define M3_UW_SORT_PLANE    0x0006
#define M3_UW_SORT_PLANE_ZY 0x000C
#define M3_UW_SORT_PLANE_XY 0x000E
#define M3_UW_SORT_PLANE_XZ 0x0010
#define M3_UW_FACE_PLANE    0x0058         /* define face */
#define M3_UW_FACE_PLANE_ZY 0x005E
#define M3_UW_FACE_PLANE_XY 0x0060
#define M3_UW_FACE_PLANE_XZ 0x0062
#define M3_UW_FACE_PLANE_X  0x0064
#define M3_UW_FACE_PLANE_Z  0x0066
#define M3_UW_FACE_PLANE_Y  0x0068
#define M3_UW_ORIGIN        0x0078
#define M3_UW_VERTEX        0x007A
#define M3_UW_FACE_VERTICES 0x007E
#define M3_UW_VERTICES      0x0082
#define M3_UW_VERTEX_X      0x0086
#define M3_UW_VERTEX_Z      0x0088
#define M3_UW_VERTEX_Y      0x008A
#define M3_UW_VERTEX_ROOF   0x008C
#define M3_UW_VERTEX_XZ     0x0090
#define M3_UW_VERTEX_XY     0x0092
#define M3_UW_VERTEX_YZ     0x0094
#define M3_UW_TEXTURE_FACE  0x00A8
#define M3_UW_TMAP_VERTICES 0x00B4
#define M3_UW_FACE_SHADE    0x00BC
#define M3_UW_VERTEX_DARK   0x00D4
#define M3_UW_FACE_GOURAUD  0x00D6


/*
 * General model flags
 */
#define M3_VAR_HEIGHT 0x0001



/*
 * Model surface flags
 */
#define M3F_FLAT         0x0001
#define M3F_TEXTURED     0x0002
#define M3F_WALL_TEXTURE 0x0004  /* uses wall not model texture            */
#define M3F_OBJ_TEXTURE  0x0008  /* texture specified in object definition */
#define M3F_SHADED       0x0010
#define M3F_AUX_SHADE    0x0020  /* has own dark value + uses aux palette  */


/*
 * Object IDs for `box' objects with draw type 11 (special)
 */
#define OBJID_BRIDGE       174
#define OBJID_CATWALK      175
#define OBJID_FORCEBRIDGE1 181
#define OBJID_FORCEBRIDGE2 183
#define OBJID_SMALL_CRATE  393
#define OBJID_LARGE_CRATE  394
#define OBJID_SECURE_CRATE 395



/*
 * Model texture chunks
 * These are obsolescent - use the defines from chunks.h
 */
#define MOD_TEXTURE_CHUNK_BASE     475
#define MOD_TEXTURE_CHUNK_SPECIAL 2180




#define MAX_VERTEX_MS 32

typedef struct
{
  int flags;                  /* surface type and texture flags */
  int texture;                /* valid if texture-mapped        */
  int colour;                 /* valid if flat-shaded           */
  int num_vertices;           /* should be self-explanatory     */
  int vertex [MAX_VERTEX_MS]; /* index into model's vertex list */
  Vector normal;              /* normal vector in model space   */
  Vector anchor;
  Vector u;
  Vector v;
} model_surface;



#define MAX_VERTEX_MODEL 256
#define MAX_SURFACE_NODE  96


typedef struct tag_model_node
{
  int num_surfaces;
  model_surface surface [MAX_SURFACE_NODE];
  plane_t sort_plane;
  struct tag_model_node *left;
  struct tag_model_node *right;
} model_node;



typedef struct tag_model
{
  int num_vertices;
  Vector vertex [MAX_VERTEX_MODEL]; /* in model space */
  float  dark   [MAX_VERTEX_MODEL];
  Vector extent;
  Vector centre;                    /* object position, in model space */
  int flags;
  model_node root;
} model;



struct tag_map_object;

#ifdef __cplusplus
extern "C"
{
#endif
void Mod_init (int width, int height);
void Mod_deinit(void);
void Mod_init_frame (void);
void Mod_render (struct tag_map_object *obj, const model *m, byte *colour);
void Mod_render_id (struct tag_map_object *obj, int mno, byte *colour);
void Mod_render_bridge (struct tag_map_object *obj);
void Mod_render_crate (struct tag_map_object *obj);
model *Mod_load_model (int model_no);


/*
 * Model helper functions
 */
void Mod_uv2anchor_vec (model *m, model_surface *s, double u [], double v []);
void Mod_partition_node (model *m, model_node *n);


#ifdef __cplusplus
}
#endif


#endif /* sentry */
