/*
 * object.c -
 *  System Shock general object handling
 */

/*
 * File:    src/object.c
 * Purpose: Generic object routines
 * Prefix:  Obj_
 * Note:
 * Music:   Poison, "Flesh & Blood"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>    /* malloc()      */

#if NEED_MEMORY_H
#include <memory.h>		 /* memset()      */
#endif

#include "gamemem.h"
#include "accel.h"
#include "anim.h"
#include "bitmap.h"
#include "chunks.h"
#include "container.h"
#include "critter.h"
#include "dark.h"
#include "door.h"
#include "fixture.h"
#include "gamestring.h"
#include "graph2d.h"
#include "item.h"
#include "main.h"      /* Cant_happen() */
#include "map.h"
#include "model.h"
#include "object.h"
#include "option.h"
#include "physics.h"
#include "render.h"
#include "resource.h"
#include "screen.h"
#include "shade.h"
#include "softs.h"
#include "surface.h"
#include "switch.h"
#include "texture.h"
#include "trigger.h"
#include "weapon.h"

#include "mfd.h"			/* register_target */


/*
 * debug
 */
								/* ff1 - we have to find a way to separate our debug prints... */
                /* jim - yeah. Probably we should have a rule that you aren't
		   allowed to commit with XXX_MESSAGES on, or something. */
#define OBJECT_MESSAGES 0


#define OBJECT_NAME_CHUNK 2157


/*
 * #defines and macros
 */
#define MAP_OBJECT_ARCHIVE_SIZE  0x1B
#define MAP_OBJXREF_ARCHIVE_SIZE 0x0A
#define MAPAIOBJ_UW_ARCHIVE_SIZE 0x1B
#define MAPOBJ_UW_ARCHIVE_SIZE   0x08

#define OBJECT_PROP_ARCHIVE_SIZE 0x1B
#define OBJPROP_UW_ARCHIVE_SIZE  0x0B


#define DEFAULT_BMP_SCALE (Option_underworld ? 4.f : 3.f);



#ifndef M_PI
/*
 * Who ate all the pi?
 */
# define M_PI		3.14159265358979323846	/* pi */
#endif



/*
 * in map.c - should it be declared elsewhere?
 */
extern unsigned long Frame;

extern Vector Viewpoint;
extern Matrix Orientation;

       map_object  *Map_object_list;
static int          Num_map_objects;
static object_xref *Object_xref_table;

extern bool Underworld2;

/*
 * The number of objects of each type. In the original Shock this lives in a
 *  table in the executable. (And indeed in TSSHP, although we probably ought
 *  to put this stuff in a data file of our own somewhere).
 */
int Objects_per_subclass [OBJ_MAXCLASS] [OBJ_MAXSUBCLASS] =
{
  {  5,  2,  2,  2,  3,  2,  0,  0 },  /* 00 weapons               */
  {  2,  2,  3,  2,  2,  2,  2,  0 },  /* 01 ammo                  */
  {  6, 16,  2,  0,  0,  0,  0,  0 },  /* 02 missiles & tracers    */
  {  5,  3,  0,  0,  0,  0,  0,  0 },  /* 03 grenades & explosives */
  {  7,  0,  0,  0,  0,  0,  0,  0 },  /* 04 patches               */
  {  5, 10,  0,  0,  0,  0,  0,  0 },  /* 05 hardware              */
  {  7,  3,  4,  5,  3,  0,  0,  0 },  /* 06 software & logs       */
  {  9, 10, 11,  4,  9,  8, 16, 10 },  /* 07 fixtures              */
  {  8, 10, 15,  6, 12, 12,  9,  8 },  /* 08 */
  {  9,  7,  3, 11,  2,  3,  0,  0 },  /* 09 switches & panels     */
  { 10,  9,  7,  5, 10,  0,  0,  0 },  /* 0A doors & gratings      */
  {  9, 11, 14,  0,  0,  0,  0,  0 },  /* 0B */
  { 13,  1,  5,  0,  0,  0,  0,  0 },  /* 0C traps & triggers      */
  {  3,  3,  4,  8, 13,  7,  8,  0 },  /* 0D containers            */
  {  9, 12,  7,  7,  2,  0,  0,  0 },  /* 0E critters              */
  {  0,  0,  0,  0,  0,  0,  0,  0 }   /* 0F unmapped UW objects   */
};


int Object_index [OBJ_MAXCLASS] [OBJ_MAXSUBCLASS];
int Num_objects;

object_prop *Object_prop_list;

char **Object_name_list;
int Num_object_names;     /* is this strictly necessary? */



/*
 * Local function prototypes
 */
static void render_obj (map_object *obj);
static void try_render_sprite (map_object *obj);
static void render_special (map_object *obj);
static bool object_is_in_tile (const map_object *obj, const map_tile *tile);



/*
 * Local bitmap hook functions
 */
int fetch_texture_default    (map_object *obj);
int fetch_texture_uw         (map_object *obj);
int fetch_texture_uw_tmap    (map_object *obj);
int fetch_texture_uw_crit    (map_object *obj);
int fetch_texture_uw_bridge  (map_object *obj);


void decode_ss_object_prop (int i, const byte *op);
void decode_uw_object_prop (int i, const byte *op);

void Obj_init (void)
{

  int class;
  int subclass;
  int index = 0;
  FILE *propfile;
  unsigned char *objprop;
  unsigned long propsize;
  unsigned char *op;
  int i;

  /*
   * Load up the object name list
   */
  Object_name_list = Str_load_string_chunk
    (Option_underworld ? UW_STRING_OBJECT_NAMES : CHUNK_TEXT_OBJECT_DEBUG,
     &Num_object_names);

  /*
   * Load up the main object properties table
   */
  if (Option_underworld)
  {
    propfile = fopen (Res_make_path (UW_DATA_PATH, "comobj.dat"), "rb");
  }
  else
  {
    propfile = fopen (Res_make_path (CD_DATA_PATH, "objprop.dat"), "rb");
  }

  if (propfile == NULL)
  {
    Cant_happen ("Obj_init(): couldn't open object properties file");
  }

  /*
   * Class-specific bits of the object properties file
   */
  if (!Option_underworld)
  {
    fseek (propfile, 4, SEEK_SET); /* 4 bytes, 0x2d, no idea */
    Wep_load_objprop (propfile);
    fseek (propfile, 0x5ab, SEEK_SET); /* fudge to get to items */
    Item_load_objprop (propfile);
    fseek (propfile, 0x8b9, SEEK_SET); /* and likewise critters */
    Crit_load_objprop (propfile);
  }

  fseek (propfile, 0, SEEK_END);
  propsize = ftell (propfile);
  objprop = malloc (propsize);
  if (objprop == NULL)
  {
    fclose (propfile);
    Cant_happen ("Obj_init(): couldn't allocate scratch property space");
  }
  rewind (propfile);
  if (fread (objprop, 1, propsize, propfile) != propsize)
  {
    Cant_happen ("Obj_init(): error reading object properties");
  }
  fclose (propfile);

  if (Option_underworld)
  {
    /*
     * For Underworld we extrapolate the number of objects from the size of
     *  the common properties list
     */
    Num_objects = propsize / OBJPROP_UW_ARCHIVE_SIZE;
  }
  else
  {
    /*
     * For System Shock it is more or less the other way around
     */
    for (class = 0; class < OBJ_MAXCLASS; ++class)
    {
      for (subclass = 0; subclass < OBJ_MAXSUBCLASS; ++subclass)
      {
	Object_index [class] [subclass] = index;
	index += Objects_per_subclass [class] [subclass];
      }
    }
    Num_objects = index;
  }

# if OBJECT_MESSAGES
  printf ("Obj_init(): %d objects found\n", Num_objects);
# endif

  Object_prop_list = Alloc_init (Num_objects * sizeof(*Object_prop_list),
				 "Obj_init", "object properties");
  if (Option_underworld)
  {
    /*
     * Common-object properties are the whole of comobj.dat
     */
    op = objprop + 2;
  }
  else
  {
    /*
     * The main object property table is at the end of the file
     */
    op = objprop + propsize - Num_objects * OBJECT_PROP_ARCHIVE_SIZE;
  }

  /*
   * And make up the main properties list
   */
  for (i = 0; i < Num_objects; ++i)
  {
    if (Option_underworld)
    {
      decode_uw_object_prop (i, op);
      op += OBJPROP_UW_ARCHIVE_SIZE;
    }
    else
    {
      decode_ss_object_prop (i, op);
      op += OBJECT_PROP_ARCHIVE_SIZE;
    }
  }

  free (objprop);

}

void Obj_deinit(void)
{

  if (!Option_underworld)
  {
    Scr_deinit();
    Cnr_deinit();
    Fix_deinit();
		Trig_deinit();
    Swi_deinit();
    Door_deinit();
  }

  free(Map_object_list);
  Map_object_list = NULL;
  free(Object_prop_list);
  Object_prop_list = NULL;
  free(Object_xref_table);
  Object_xref_table = NULL;
  Str_free_string_chunk(Object_name_list);

}

void decode_ss_object_prop (int i, const byte *op)
{

  static int baseframe = 2;
  object_prop *prop = &(Object_prop_list [i]);

  prop->mass       = *op + 256 * op [1] + 65536 * op [2] +
                                    16777216 * op [3]; op += 4;
  prop->hp         = *op + 256 * op [1]; op += 2;
  prop->armour     = *op++;
  prop->draw_type  = *op++;
  op += 3;
  prop->scale      = *op++;
  op += 2;
  prop->vulns      = *op++;
  prop->special    = *op++;
  op += 2;
  prop->defence    = *op++;
  op++;
  prop->flags      = *op + 256 * op [1]; op += 2;
  prop->info       = *op + 256 * op [1]; op += 2;
  op++;
  prop->extra_bmps = *op >> 4;
  prop->baseframe  = baseframe;
  baseframe += 3 + prop->extra_bmps;

  prop->bmp_chunk  = BITMAP_CHUNK_SPRITE;

  /*
   * Draw type fudge; corpses become orientable bitmaps
   */
  if ((i >= 202) && (i < 211))
  {
    prop->draw_type = OBJDRAW_OSPRITE;
  }
  /* Borland complains ("never used")  op += 2; */

# if OBJECT_MESSAGES
  printf ("Object %d (%s)\t"
	  "armour %d mass %ld.%ldkg flags %02X bmps %d draw %d "
          "vuln %02x/%02x def %d\n",
	  i, Object_name_list [i], prop->armour, prop->mass / 10,
	  prop->mass % 10, prop->flags, prop->extra_bmps, prop->draw_type,
	  prop->vulns, prop->special, prop->defence);
# endif

}



/*
 * Some Underworld objects with "special" draw types
 */
#define UW_OBJID_CAMPFIRE       298
#define UW_OBJID_DOOR_BASE      320
#define UW_OBJID_DOOR_LAST      335
#define UW_OBJID_BENCH          336
#define UW_OBJID_LARGE_BOULDER1 339
#define UW_OBJID_LARGE_BOULDER2 340
#define UW_OBJID_BOULDER        341
#define UW_OBJID_SMALL_BOULDER  342
#define UW_OBJID_SHRINE         343
#define UW_OBJID_TABLE          344
#define UW_OBJID_BEAM           345
#define UW_OBJID_MOONGATE       346
#define UW_OBJID_BARREL         347
#define UW_OBJID_CHAIR          348
#define UW_OBJID_CHEST          349
#define UW_OBJID_NIGHTSTAND     350
#define UW_OBJID_LOTUS          351 /* Lotus Turbo Esprit */
#define UW_OBJID_PILLAR         352
#define UW_OBJID_LEVER_8        353 /* 8-way lever  */
#define UW_OBJID_SWITCH_8       354 /* 8-way switch */
#define UW_OBJID_PAINTING       355
#define UW_OBJID_BRIDGE         356
#define UW_OBJID_GRAVE          357
#define UW_OBJID_WRITING        358
#define UW_OBJID_SHELF          361
#define UW_OBJID_TMAP1          366
#define UW_OBJID_TMAP2          367
#define UW_OBJID_SWITCH_BASE    368
#define UW_OBJID_SWITCH_LAST    383

struct
{
  int id;
  int bmp_chunk;
  int baseframe;
  int draw_type;
  fetch_texture_func hook;
}
UW_special_draw_list [] =
{ 
  { UW_OBJID_TABLE,    BITMAP_CHUNK_UW_DECAL, 32, OBJDRAW_3D,
    fetch_texture_default },
  { UW_OBJID_CHAIR,    BITMAP_CHUNK_UW_DECAL, 38, OBJDRAW_3D,
    fetch_texture_default },
  { UW_OBJID_PILLAR,   BITMAP_CHUNK_UW_DECAL,  0, OBJDRAW_3D,
    fetch_texture_default },
  { UW_OBJID_PAINTING, BITMAP_CHUNK_UW_DECAL, 42, OBJDRAW_3D,
    fetch_texture_default },
  { UW_OBJID_LEVER_8,  BITMAP_CHUNK_UW_DECAL,  4, OBJDRAW_DECAL,
    fetch_texture_default },
  { UW_OBJID_SWITCH_8, BITMAP_CHUNK_UW_DECAL, 12, OBJDRAW_DECAL,
    fetch_texture_default },
  { UW_OBJID_BRIDGE,   BITMAP_CHUNK_UW_DECAL, 30, OBJDRAW_3D,
    fetch_texture_uw_bridge },
  { UW_OBJID_GRAVE,    BITMAP_CHUNK_UW_DECAL, 28, OBJDRAW_3D,
    fetch_texture_default },
  { UW_OBJID_WRITING,  BITMAP_CHUNK_UW_DECAL, 20, OBJDRAW_DECAL,
    fetch_texture_default },
  { UW_OBJID_SHELF,    BITMAP_CHUNK_UW_DECAL, 36, OBJDRAW_3D,
    fetch_texture_default },
  { UW_OBJID_TMAP1,    BITMAP_CHUNK_UW_TMAP,   0, OBJDRAW_DECAL,
    fetch_texture_uw_tmap },
  { UW_OBJID_TMAP2,    BITMAP_CHUNK_UW_TMAP,   0, OBJDRAW_DECAL,
    fetch_texture_uw_tmap },
  { -1,                                    0,  0,             0, NULL }
};


/*
 * Model numbers for objects 336-367, straight out of the exe
 */
int UW_model_list [32] =
{  3,  8,  8,  7,  7,  6,  5, 11, 24,  9, 23, 27, 28, 25, 26,  4,
  10, 16, 17, 13,  2, 19, 18, 29, 30, 31, -1, -1, -1, -1, -1, -1 };


void decode_uw_object_prop (int i, const byte *op)
{

  int sd;
  object_prop *prop = &(Object_prop_list [i]);

  prop->armour    = *op++;
  prop->mass      = *op + 256 * op [1]; op += 2;
  prop->stuff2    = prop->mass & 0x0f;
  prop->mass    >>= 4;
  prop->flags     = *op++;
  prop->value     = *op + 256 * op [1]; op += 2;
  op += 3;
  prop->scale     = *op; /* Borland complains if incremented */
  prop->baseframe = i;

  switch (prop->flags & 0x0f)
  {
  case 0:
  case 2:
    if ((i >= 336) && (i < 368) && (UW_model_list [i - 336] >= 0))
    { 
      prop->draw_type = OBJDRAW_3D;
      prop->info = UW_model_list [i - 336];
    }
    else prop->draw_type = OBJDRAW_SPRITE;
    break;
  case 4:
    prop->draw_type = OBJDRAW_DECAL;
    break;
  default:
    prop->draw_type = OBJDRAW_SPRITE;
  }

  if ((i >= UW_OBJID_DOOR_BASE) && (i <= UW_OBJID_DOOR_LAST))
  {
    prop->draw_type = OBJDRAW_SPECIAL;
  }
  else if ((i >= UW_OBJID_ANIM_BASE) && (i <= UW_OBJID_ANIM_LAST))
  {
    uw_animobj *uwanim = Anim_get_uw_info (i);
    prop->bmp_chunk  = 1; /* animo.gr */
    prop->extra_bmps = uwanim->num_frames - 1;
    prop->baseframe  = uwanim->base_frame;
  }

  /*
   * Critter draw type fudge
   */
  if ((i >= CRIT_UW_BASE_OBJID) &&
      (i < CRIT_UW_BASE_OBJID + CRIT_UW_NUM_CRITS))
  {
    prop->draw_type = OBJDRAW_CRITTER;
  }

  /*
   * Switch hook fudge
   */
  if ((i >= UW_OBJID_SWITCH_BASE) && (i <= UW_OBJID_SWITCH_LAST))
  {
    prop->bmp_chunk = BITMAP_CHUNK_UW_SWITCH;
    prop->baseframe = i - UW_OBJID_SWITCH_BASE;
  }

  /*
   * Any objects which UW considers "special" but which we treat as having one
   *  of the predefined draw types
   */
  for (sd = 0; UW_special_draw_list [sd].hook != NULL; ++sd)
  {
    if (i == UW_special_draw_list [sd].id)
    {
      prop->bmp_chunk = UW_special_draw_list [sd].bmp_chunk;
      prop->baseframe = UW_special_draw_list [sd].baseframe;
      prop->draw_type = UW_special_draw_list [sd].draw_type;
    }
  }

# if OBJECT_MESSAGES
  printf ("Object %d (%s)\t%x armour %d mass %ld.%ldkg flags %02X value %d\n",
	  i, Object_name_list [i], prop->stuff2, prop->armour, prop->mass / 10,
	  prop->mass % 10, prop->flags, prop->value);
# endif
	  
}




void decode_uw_map_object (int i, const byte *op);
void decode_ss_map_object (int i, const byte *op);
void init_uw_fetch_hook   (map_object *obj);


void Obj_init_level (int level)
{

  block_handle bh;
  size_t size;
  int num_objs;
  int i;
  unsigned char *op;
  Matrix tm;
  Matrix rot;
  byte *uw2raw;
  byte uwraw [0x3300];

  /*
   * Initialise the classes
   */
  if (!Option_underworld)
  {
    Fix_init_level (level);
    Swi_init_level (level);
		Trig_init_level (level);
    Door_init_level (level);
    Cnr_init_level (level);
    Scr_init_level (level);
  }

  /*
   * Initialise the animation table for this level
   */
  Anim_init_level (level);

  /*
   * Load up the map object list
   * This is really horrible - a cut-and-paste from map.c
   * Ought to call from Map_init_level() and leave the raw data lying about
   *  for Obj_init_level() to use
   */
  if (Option_underworld)
  {
    FILE *levfile = fopen (UW_MAP_FILE, "rb");
    if (levfile == NULL)
    {
      Cant_happen ("Map_init_level(): failed to open UW level file");
    }
    if (Underworld2)
    {
      Underworld2 = true;
      uw2raw = Res_load_uw2_map (levfile, level, 0, &size);
      op = uw2raw + 0x4000;
    }
    else
    {
      /*
       * Underworld 1 detected
       */
      unsigned long offs;
      fseek (levfile, 0x02 + 4 * level, SEEK_SET);
      READ_LE_LONG (levfile, offs);
      fseek (levfile, offs + 0x4000, SEEK_SET);
      fread (uwraw, 1, 0x3300, levfile);
      op = uwraw;
    }
    fclose (levfile);
    Num_map_objects = 1024;
  }
  else
  {
    bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_OBJECT, &size);
    if ((bh == NULL) || (*bh == NULL))
    {
      Cant_happen ("Obj_init_level(): couldn't load map object chunk");
    }
    if (size % MAP_OBJECT_ARCHIVE_SIZE != 0)
    {
      Cant_happen ("Obj_init_level(): not exact number of map objects");
    }
    Num_map_objects = size / MAP_OBJECT_ARCHIVE_SIZE;
    op = (unsigned char *) *bh;
  }

  Map_object_list = Alloc_init (Num_map_objects * sizeof(*Map_object_list),
				"Obj_init_level", "object list");
  for (i = 0; i < Num_map_objects; ++i)
  {

    map_object *obj   = &(Map_object_list [i]);
    object_prop *prop;
    if (Option_underworld)
    {
      decode_uw_map_object (i, op);
      op += (i < 0x100) ? MAPAIOBJ_UW_ARCHIVE_SIZE : MAPOBJ_UW_ARCHIVE_SIZE;
    }
    else
    {
      decode_ss_map_object (i, op);
      op += MAP_OBJECT_ARCHIVE_SIZE;
    }

    obj->frame      = 0;
    obj->ms         = NULL;
    obj->base_texid = -1;
    obj->fetch_texture_hook = fetch_texture_default;
/*      obj->bmp_scale = DEFAULT_BMP_SCALE; */

#   if OBJECT_MESSAGES
    if (obj->used)
    {
      Obj_dump_object_info (obj);
    }
#   endif
/*      if (obj->used && (obj->ai != 255)) printf ("%02x\n", obj->ai); */

    /*
     * Perform some post-processing : get the orientation matrices for each
     *  object
     */

    /*
     * The orientation matrix is object-to-world i.e. we pre-multiply an object
     *  coordinate with it to get a world coordinate.
     * Start with the pitch matrix
     */
    tm.x.x  =  1.;
    tm.x.y  =  0.;
    tm.x.z  =  0.;
    tm.y.x  =  0.;
    tm.y.y  =  cos (obj->pitch   * M_PI / 128.);
    tm.y.z  = -sin (obj->pitch   * M_PI / 128.);
    tm.z.x  =  0.;
    tm.z.y  =  sin (obj->pitch   * M_PI / 128.);
    tm.z.z  =  cos (obj->pitch   * M_PI / 128.);

    /*
     * And add bank (incidentally, these are the angle names used by DromEd /
     *  ShockEd; they're good enough)
     */
    rot.x.x =  cos (obj->bank    * M_PI / 128.);
    rot.x.y =  0.;
    rot.x.z =  sin (obj->bank    * M_PI / 128.);
    rot.y.x =  0.;
    rot.y.y =  1.;
    rot.y.z =  0.;
    rot.z.x = -sin (obj->bank    * M_PI / 128.);
    rot.z.y =  0.;
    rot.z.z =  cos (obj->bank    * M_PI / 128.);

/*      Mat_multiply (obj->orientation, rot, tm); */
    Mat_multiply (obj->orientation, tm, rot);
    tm = obj->orientation;

    /*
     * And finally multiply with heading
     */
    rot.x.x =  cos (obj->heading * M_PI / 128.);
    rot.x.y =  sin (obj->heading * M_PI / 128.);
    rot.x.z =  0.;
    rot.y.x = -sin (obj->heading * M_PI / 128.);
    rot.y.y =  cos (obj->heading * M_PI / 128.);
    rot.y.z =  0.;
    rot.z.x =  0.;
    rot.z.y =  0.;
    rot.z.z =  1.;

    Mat_multiply (obj->orientation, rot, tm);

    /*
     * Initialise the rflags
     */
    obj->rflags = 0;

    prop = &(Object_prop_list [obj->object_id]);

  }

  Res_free_chunk_by_id (4000 + 100*level + MAP_CHUNK_OBJECT);

  /*
   * The object cross-reference table
   */
  if (!Option_underworld)
  {
    /*
     * System Shock has an explicit object cross-reference table stored with
     *  the map
     */
    bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_OBJ_XREF, &size);
    if ((bh == NULL) || (*bh == NULL))
    {
      Cant_happen ("Obj_init_level(): couldn't load object xref chunk");
    }
    if (size % MAP_OBJXREF_ARCHIVE_SIZE != 0)
    {
      Cant_happen ("Obj_init_level(): not exact number of map objects");
    }
    num_objs = size / MAP_OBJXREF_ARCHIVE_SIZE;
    op = (unsigned char *) *bh;
  }
  else num_objs = Num_map_objects;

  Object_xref_table = Alloc_init (num_objs * sizeof *Object_xref_table,
				  "Obj_init_level", "xref table");
  for (i = 0; i < num_objs; ++i)
  {
/*      map_object *obj = &(Map_object_list [i]); */
    if (Option_underworld)
    {
      /*
       * Underworld doesn't (as far as I know) have an explicit object cross-
       *  reference table, so I'm going to make one up: one entry per object
       */
      Object_xref_table [i].tile       = NULL;
      Object_xref_table [i].index      = i;
      Object_xref_table [i].obj_chain  = Map_object_list [i].next;
      Object_xref_table [i].tile_chain = i;
    }
    else
    {
      int tile_x, tile_y;
      tile_x = *op + 256 * op [1]; op += 2;
      tile_y = *op + 256 * op [1]; op += 2;
      Object_xref_table [i].tile       = &(Tile [tile_x + 64 * tile_y]);
      Object_xref_table [i].index      = *op + 256 * op [1]; op += 2;
      Object_xref_table [i].obj_chain  = *op + 256 * op [1]; op += 2;
      Object_xref_table [i].tile_chain = *op + 256 * op [1]; op += 2;
    }
/*      if (obj->object_id == 366) Obj_dump_object_info (obj); */
  }

  if (!Option_underworld)
  {
    Res_free_chunk_by_id (4000 + 100*level + MAP_CHUNK_OBJ_XREF);
  }

  for (i = 0; i < Num_map_objects; ++i)
  {

    map_object *obj = &(Map_object_list [i]);
    object_prop *prop = &(Object_prop_list [obj->object_id]);

    /*
     * Make sure we have the right hook(s)
     * Why don't we just have an array of 16 init functions and call on object
     *  class?
     */
    switch (obj->class)
    {
    case OBJCLASS_FIXTURE:
      Fix_init_object (obj);
      break;
    case OBJCLASS_DOOR:
      Door_init_object (obj);
      break;
    case OBJCLASS_SWITCH:
      Swi_init_object (obj);
      break;
    case OBJCLASS_ANIMATED:
      if (prop->extra_bmps > 0)
      {
        Anim_add (0.25, obj, prop->extra_bmps + 1, 0);
      }
      break;
    case OBJCLASS_CRITTER:
      if (prop->draw_type == OBJDRAW_CRITTER)
      {
        obj->fetch_texture_hook = Crit_fetch_texture;
      }
      break;
    case OBJCLASS_UNDERWORLD:
      init_uw_fetch_hook (obj);
      break;
    default:
      break;
    }

    /*
     * Give any forcedoors their permanent map_surface
     */
    if (prop->draw_type == OBJDRAW_FORCEDOOR)
    {
      Obj_make_decal_surface (obj, TILE_GRID_SIZE, TILE_GRID_SIZE,
			      SF_TRANSPARENCY | MSF_2SIDED);
    }

  }

}




object_xref *Obj_get_reference(int index)
{
	object_xref *pReference = NULL;

  /*
   * Object 0 means "no object"!
   */
	if (index > 0 /* && index < num_obj */)
		pReference = &(Object_xref_table[index]);

	return pReference;
}

map_object *Obj_get_object(int index)
{
	map_object *pObject = NULL;

	if (index >= 0 /* && index < num_obj */)
		pObject = &(Map_object_list[index]);

	return pObject;
}

object_prop *Obj_get_property(int id)
{
	object_prop *pProp = NULL;

	if (id >= 0 /* && index < num_obj */)
		pProp = &(Object_prop_list[id]);

	return pProp;
}

void decode_uw_map_object (int i, const byte *op)
{
  uint32 obword;
  map_object *mo = &(Map_object_list [i]);

  /*
   * First object-definition word
   */
  obword = *op + 256 * op [1] + 65536 * op [2] + 16777216 * op [3]; op += 4;
  mo->used      = 1;
  mo->object_id = obword & 0x1ff; obword >>= 9;
  mo->stuff1    = obword & 0x07f; obword >>= 7;
  mo->stuff3    = obword;
  mo->pos.z     = obword & 0x07f; obword >>= 7;
  mo->pos.z    *= Height_scale / 4;
  mo->pitch     = 0;
  mo->heading   = obword & 0x007; obword >>= 3;
  mo->bank      = 0;
  mo->pos.y     = obword & 0x007; obword >>= 3;
  mo->pos.x     = obword & 0x007; /* obword >>= 3; */

  /*
   * Second object-definition word
   */
  obword = *op + 256 * op [1] + 65536 * op [2] + 16777216 * op [3];
  /* op += 4; */
  mo->quality   = obword & 0x03f; obword >>= 6;
  mo->next      = obword & 0x3ff; obword >>= 10;
  mo->stuff2    = obword & 0x03f; obword >>= 6;
  mo->link2     = obword & 0x3ff; /* obword >>= 10; */
  mo->dark      = INKY_BLACKNESS;
  mo->state     = mo->stuff1 & 7;

  /*
   * Position correction dependent on heading (tentative solution)
   */
  switch (mo->heading)
  {
  case 0:
  case 1:
    mo->pos.x++;
    mo->pos.y++;
    break;
  case 2:
  case 3:
    mo->pos.x++;
    break;
  case 4:
  case 5:
    break;
  case 6:
  case 7:
    mo->pos.y++;
    break;
  }
  mo->pos.x *= 32;
  mo->pos.y *= 32;

  /*
   * Angle correction: System Shock angles are in slyces (1/256 circle),
   *  Underworld half-quadrants (1/8 circle)
   */
  mo->heading *= 32;

  /*
   * C/S/T mapping
   */
  if ((mo->object_id >= UW_OBJID_DOOR_BASE) &&
      (mo->object_id <= UW_OBJID_DOOR_LAST))
  {
    mo->class    = OBJCLASS_DOOR;
    mo->subclass = DOOR_SUBCLASS_3D;
    mo->type     = mo->object_id - UW_OBJID_DOOR_BASE;
  }
  else if ((mo->object_id >= UW_OBJID_ANIM_BASE) &&
           (mo->object_id <= UW_OBJID_ANIM_LAST))
  {
    mo->class    = OBJCLASS_ANIMATED;
    mo->subclass = 0;
    mo->type     = mo->object_id - UW_OBJID_ANIM_BASE;
  }
  else
  {
    mo->class    = OBJCLASS_UNDERWORLD;
    mo->subclass = mo->object_id >> 8;
    mo->type     = mo->object_id & 0xff;
  }

}





void decode_ss_map_object (int i, const byte *op)
{

  int objid;
  map_object *mo = &(Map_object_list [i]);

  mo->used        = *op++;
  mo->class       = *op++;
  mo->subclass    = *op++;
  mo->class_index = *op + 256 * op [1]; op += 2;
  mo->tile_xref   = *op + 256 * op [1]; op += 2;
  mo->prev        = *op + 256 * op [1]; op += 2;
  mo->next        = *op + 256 * op [1]; op += 2;
  mo->pos.x       = *op + 256 * op [1]; op += 2;
  mo->pos.y       = *op + 256 * op [1]; op += 2;
  mo->pos.z       = *op++ * Height_scale / 8;
  mo->pitch       = *op++;
  mo->heading     = *op++;
  mo->bank        = *op++;
  mo->ai          = *op++;
  mo->type        = *op++;
  mo->stuff1      = *op + 256 * op [1]; op += 2;
  mo->stuff2      = *op++;
  mo->state       = *op++;
  mo->stuff3      = *op++;
  mo->flags       = *op; /* Borland complains if incremented */

  /*
   * System Shock doesn't store the object ID directly, so we must work it out
   *  from the class, subclass and type
   */
  objid = Object_index [mo->class] [mo->subclass] + mo->type;
  assert (objid < Num_objects);
  mo->object_id = objid;

  mo->dark = Shd_get_static_dark (mo->pos.x, mo->pos.y, mo->pos.z);

}



void init_uw_fetch_hook (map_object *obj)
{

  int sd;
  int id = obj->object_id;

  obj->fetch_texture_hook = fetch_texture_uw; /* default */

  if ((id >= CRIT_UW_BASE_OBJID) &&
      (id < CRIT_UW_BASE_OBJID + CRIT_UW_NUM_CRITS))
  {
    obj->fetch_texture_hook = fetch_texture_uw_crit;
  }
  else if ((id >= UW_OBJID_SWITCH_BASE) && (id <= UW_OBJID_SWITCH_LAST))
  {
    obj->fetch_texture_hook = fetch_texture_default;
  }
  else for (sd = 0; UW_special_draw_list [sd].hook != NULL; ++sd)
  {
    if (obj->object_id == UW_special_draw_list [sd].id)
    {
      obj->fetch_texture_hook = UW_special_draw_list [sd].hook;
    }
  }

}



/*
 * Prep the object renderer for a scene
 */
void Obj_init_scene (void)
{

  /*
   * Clear all ORF_DRAWABLE flags so that the objects will be drawn in the
   *  right place and time
   */
  int i;
  for (i = 0; i < Num_map_objects; ++i)
  {
    Map_object_list [i].rflags &= ~ORF_DRAWABLE;
  }

}




/*
 * This is called when a tile is being queued, to set the depth and rendering
 *  tile for the objects
 */
void Obj_queue_tile (map_tile *tile)
{

  int index;
  map_object *obj;

  for (index = tile->index; index != 0;
       index = Object_xref_table [index].obj_chain)
  {
    obj = &(Map_object_list  [Object_xref_table [index].index]);

    /*
     * An object is drawn `in' the first tile in which it is encountered
     */
    if (!(obj->rflags & ORF_DRAWABLE))
    {
      obj->rflags |= ORF_DRAWABLE;
      obj->tile  = tile;
    }

  }

}



#define MAX_DEPTH_SORT_TILE 32

/*
 * The current object clip region, in view space. Used by models (which pre-
 *  transform to view space to save time) and sprites (which are inherently
 *  view-space objects)
 */
clipregion_t Object_region_view;




/*
 * Render all the objects in a given inclusion volume
 * Objects are clipped to the given inclusion volume unless they exist in the
 *  `from' tile, in which case they will be drawn later (this is to prevent
 *  objects from being drawn twice if they are in two tiles, one of which can
 *  only be seen through the other)
 */
void Obj_render_tile (map_tile *tile, map_tile *from,
                      const clipregion_t *frustrum)
{

  map_object *obj;
  int i, j, index, num_objs = 0;
  struct
  {
    Vector view;
    int obj_index;
  } depth_list [MAX_DEPTH_SORT_TILE];
  double depth;

  /*
   * Set up the object clip region
   */
  if (frustrum == NULL)
  {
    /*
     * Null is allowed, just clip to the view frustrum in this case
     */
    Clip_copy_region (&Object_region_view, &(Camera->view_frustrum));
  }
  else Clip_region_world_to_view (&Object_region_view, frustrum);

  for (i = 0, index = tile->index; i < MAX_DEPTH_SORT_TILE; ++i)
  {
    Vector tv;
    object_prop *prop;
    if (index == 0) break;
    obj = &(Map_object_list  [Object_xref_table [index].index]);
    if (!obj->used) break;
    prop = &(Object_prop_list [obj->object_id]);

    /*
     * Reject if we are viewing the object through another tile in which it is
     *  elegible to be drawn later
     */
    if (!object_is_in_tile (obj, from))
    {

      /*
       * Decals and screens have already been done
       */
      if ((prop->draw_type != OBJDRAW_DECAL) &&
          (prop->draw_type != OBJDRAW_SCREEN))
      {
        /*
         * Get the position of the object in view coordinates
         */
        vec_subtract (tv, obj->pos, Viewpoint);
        mat_vec_multiply (depth_list [num_objs].view, Orientation, tv);
        obj->view = depth_list [num_objs].view;
        depth_list [num_objs].obj_index = index;
        ++num_objs;
      }
    }

    index = Object_xref_table [index].obj_chain;
  }


  /*
   * Sort the objects in the tile by view depth.
   * This is O(n^2) with the number of objects in the tile. Deal with it.
   * A more serious problem is that it won't actually work for 3D objects
   *  versus sprites. Perhaps we ought just to bung ALL objects at the
   *  Z-buffer and wash our hands of the whole sorry affair.
   */
  for (i = 0; i < num_objs; ++i)
  {
    depth = -1.e10;
    index = -1;
    for (j = 0; j < num_objs; ++j)
    {
      if (depth_list [j].obj_index == 0) continue; /* already done this one */
      if (depth_list [j].view.z > depth)
      {
        depth = depth_list [j].view.z;
        index = j;
      }
    }
    if (index < 0) break;   /* no objects found */
    obj = &(Map_object_list
	    [Object_xref_table [depth_list [index].obj_index].index]);

    /*
     * If an object exists in more than one tile in a sector (which have the
     *  same inclusion volumes) only draw it in the frontmost drawable. This
     *  will be the case if the tile render field in the object is set
     *  appropriately.
     * As the renderer stands, this is only applicable to cyberspace, in which
     *  case the sector contains the entire level (and its BSP tree). For real
     *  space the sector contains only one tile, so it will always be the
     *  frontmost.
     */
    if ((obj->tile == tile) || !Map_info.cyberspace)
    {
      render_obj (obj);
    }
    depth_list [index].obj_index = 0;
  }

  Clip_free_region (&Object_region_view);

}




/*
 * Decals are drawn separately for all inclusion volumes in a tile, before the
 *  other objects are rendered
 */
void Obj_render_tile_decals (map_tile *tile, const clipregion_t *frustrum)
{

  map_object *obj;
  int i = tile->index; /* index into object xref table */
  surface ts;

  while (i != 0)
  {
    object_prop *prop;
    int texid;

    obj = &(Map_object_list  [Object_xref_table [i].index]);
    if (Object_xref_table [i].tile == NULL)
    {
      /*
       * Not encountered this (Underworld) object yet. Set its position
       */
      Object_xref_table [i].tile = tile;
      obj->pos.x += tile->x * TILE_GRID_SIZE;
      obj->pos.y += tile->y * TILE_GRID_SIZE;

      /*
       * A teaser for Underworld ... if it's a campfire, add a light for it
       */
      if (obj->object_id == UW_OBJID_CAMPFIRE)
      {
        Dark_add_static_sucker (obj, 20000., 512.);
      }

    }

    /*
     * Set up for next object
     */
    i = Object_xref_table [i].obj_chain;

    /*
     * Check that it is a decal
     */
    prop = &(Object_prop_list [obj->object_id]);
    if ((prop->draw_type != OBJDRAW_DECAL) &&
        (prop->draw_type != OBJDRAW_SCREEN)) continue;

    texid = obj->fetch_texture_hook (obj);

    /*
     * Try to draw it. Decals are unordered as they cannot interfere with one
     *  another
     */
    if (obj->ms == NULL)
    {
      /*
       * No cached map surface ... if we've got this far, just use the
       *  defaults i.e. the size of the bitmap, masked, not reversible
       */
      ss_bitmap_header *bmp = Tex_get_texture (texid);
      Obj_make_decal_surface (obj, 4 * bmp->width, 4 * bmp->height,
                              SF_TEXTURED | SF_MASKED | SF_SHADED);
    }

    Sur_map_to_surface (&ts, obj->ms);
    if (ts.texture < 0)
    {
      ts.texture = texid;
    }


    /*
     * NOTE while I'm at it: light values for decal surfaces not
     *  coinciding with map vertices are even wronger because they use
     *  60 (15*4) as inky blackness rather than 63 which they should
     * Need correction for this
     */

    /*
     * Light sourcing
     */
    Dark_do_gouraud (tile, &ts);

    if (Clip_surface_by_region (&ts, frustrum))
    {
      Ren_transform_surface (&ts); /* should be done later */
      Acc_draw_surface (&ts);

    }

  }

}





/*
 * Calculate the view octant, from 0 to 7, of object obj when viewed from the
 *  current renderer camera. This is used to determine which bitmap should be
 *  displayed for an orientable sprite
 * Octant 6 represents an object facing directly towards the viewer. Octant
 *  0 represents an object facing towards the left. (Note that these are
 *  System Shock octants; Underworld uses 0 towards, 2 left. This is corrected
 *  for elsewhere)
 */
#define ARCTAN_22_5 0.41421356
#define ARCTAN_67_5 2.41421356
int Obj_get_octant (map_object *obj)
{

  Vector rel_pos;
  double x;
  double y;
  double atn; /* arctangent of the angle (y / x) */
  int octant;

  /*
   * Get the relative vector from the object to the view position
   */
  vec_subtract (rel_pos, Camera->position, obj->pos);

  /*
   * We don't care about the orientation Z axis (straight up), only X and Y
   * Dot product with the inverse (=transpose) gives object coords
   */
  x = obj->orientation.x.x * rel_pos.x + obj->orientation.y.x * rel_pos.y;
  y = obj->orientation.x.y * rel_pos.x + obj->orientation.y.y * rel_pos.y;

  /*
   * Short-circuit in case we are directly to one side
   */
  if (x == 0.)
  {
    if (y > 0) return 6; else return 2;
  }
  if (y == 0.)
  {
    if (x > 0) return 4; else return 0;
  }

  /*
   * First figure out which quadrant we are in
   */
  atn = y / x;
  if (atn < 0)
  {
    /*
     * -/+ or +/-
     */
    if (atn < -ARCTAN_67_5)      octant = 6;
    else if (atn < -ARCTAN_22_5) octant = 7;
    else                         octant = 0;
    if (y < 0) octant = (octant + 4) % 8;
  }
  else
  {
    /*
     * +/+ or -/-
     */
    if (atn < ARCTAN_22_5)       octant = 0;
    else if (atn < ARCTAN_67_5)  octant = 1;
    else                         octant = 2;
    if (y > 0) octant += 4;
  }

  return octant;

}




/*
 * For informational / debugging purposes
 */
void Obj_dump_object_info (map_object *obj)
{

  printf ("Object %d %d/%d/%d %s @(%g, %g, %g)\tclassidx %d AI %d HP %d"
	  " ?? %02x state %d ?? %02x flags %02x q=%d link=%d\n",
	  obj - Map_object_list,
	  obj->class, obj->subclass, obj->type,
	  Object_name_list [obj->object_id],
	  obj->pos.x, obj->pos.y, obj->pos.z,
	  obj->class_index, obj->ai, obj->stuff1, obj->stuff2,
	  obj->state, obj->stuff3, obj->flags, obj->quality, obj->link2);
  if (Option_underworld && obj->link2)
  {
/*      printf ("-->"); */
/*      Obj_dump_object_info (&(Map_object_list [obj->link2])); */
  }

}




/*
 * Set up a map surface for a decal object. These all have the same
 *  orientation / texture mapping
 */
void Obj_make_decal_surface (map_object *obj, int width, int height, int flags)
{

  map_surface tms;
  Vector u;
  Vector v;
  map_vertex tv = {0, 0, 0, {0} };
  double ts;
  extern map_surface *Surface; /* from map.c */

  /*
   * Decal u vector is always in the direction of object x. Decal v vector is
   *  always in the direction of negative object z.
   */
  u.x =  obj->orientation.x.x;
  u.y =  obj->orientation.y.x;
  u.z =  obj->orientation.z.x;
  vec_multiply (u, width);

  v.x = -obj->orientation.x.z;
  v.y = -obj->orientation.y.z;
  v.z = -obj->orientation.z.z;
  vec_multiply (v, height);

  /*
   * No prizes for guessing which orientation vector gives us the surface
   *  normal, then ...
   */
  tms.normal.x = -obj->orientation.x.y;
  tms.normal.y = -obj->orientation.y.y;
  tms.normal.z = -obj->orientation.z.y;

  /*
   * Now to populate the surface
   */
  tms.texture = obj->base_texid;

  /*
   * The surface vertices (in reverse order, to leave the anchor in tv)
   */
  tms.num_vertices = 4;
  if (Option_underworld)
  {
    /*
     * Underworld decals are aligned with the bottom of the bitmap ...
     */
    tv.x = u.x; tv.y = u.y; tv.z = u.z;
  }
  else
  {
    /*
     * ... whereas System Shock decals use the middle
     */
    vec_subtract (tv, u, v);
  }
  assert (!Map_info.cyberspace); /* using dark, must be real world */
  vec_divide (tv, 2); vec_subtract (tv, obj->pos, tv);
  tv.u.dark = Shd_get_static_dark (tv.x, tv.y, tv.z);
  tms.vertex [3] = Map_add_vertex (&tv, obj->tile);

  vec_add (tv, tv, u);
  tv.u.dark = Shd_get_static_dark (tv.x, tv.y, tv.z);
  tms.vertex [2] = Map_add_vertex (&tv, obj->tile);

  vec_subtract (tv, tv, v);
  tv.u.dark = Shd_get_static_dark (tv.x, tv.y, tv.z);
  tms.vertex [1] = Map_add_vertex (&tv, obj->tile);

  vec_subtract (tv, tv, u);
  tv.u.dark = Shd_get_static_dark (tv.x, tv.y, tv.z);
  tms.vertex [0] = Map_add_vertex (&tv, obj->tile);

  tms.tile_next = -1; /* not linked into tile surfaces */
  tms.flags     = flags;

  /*
   * Now the texture mapping
   */
  tms.anchor.x = tv.x; tms.anchor.y = tv.y; tms.anchor.z = tv.z;
  ts = width * width / 256.;  vec_divide (u, ts);
  tms.u = u;
  ts = height * height / 256.; vec_divide (v, ts);
  tms.v = v;

  tms.behind = NULL;

  obj->ms = &(Surface [Map_add_surface (&tms)]);

}




void render_obj (map_object *obj)
{

  object_prop *prop = &(Object_prop_list [obj->object_id]);

  if (Map_info.cyberspace && 
      (obj->class == OBJCLASS_SOFTS || obj->class == OBJCLASS_FIXTURE))
  {
    Sof_render_cyber (obj);
    return;
  }

  switch (prop->draw_type)
  {

  case OBJDRAW_3D:
  {
    /*
     * If it's a cyber object, we'll need a colour scheme
     * This is pretty nasty -- why not just store the colour scheme with the
     *  object?
     */
    byte *colour;
    switch (obj->class)
    {
    case OBJCLASS_PROJECTILE:
      colour = Wep_get_colour (obj);
      break;
    case OBJCLASS_ITEM:
      colour = Item_get_colour (obj);
      break;
    case OBJCLASS_CRITTER:
      colour = Crit_get_colour (obj);
      break;
    default:
      colour = NULL;
      break;
    }
    Mod_render_id (obj, prop->info, colour);
    break;
  }

  case OBJDRAW_CRITTER:
    /*
     * Register the critter as a target only in the main view (if I've
     *  correctly interpreted firefreak's intentions)
     */
    if (Camera->flags & CF_MAIN) MFD_register_target (obj);

	/* remember to draw attention to this deliberate ... */
	/* FALL THROUGH! */

  case OBJDRAW_SPRITE:
  case OBJDRAW_OSPRITE:
  {
    try_render_sprite (obj);
    break;
  }

  /*
   * It seems to me that a fragmented object just uses its normal sprite
   *  but expanded
   */
  case OBJDRAW_FRAGMENTS: /* c/space fragmented thing */
  {
    ss_bitmap_header *bmp, *zbmp;
    assert (Map_info.cyberspace);
    bmp  = Tex_get_texture (Tex_get_cache_id (prop->bmp_chunk,
                                              prop->baseframe));
    zbmp = Tex_get_texture (Tex_get_cache_id (prop->bmp_chunk,
                                              prop->baseframe + 1));
    assert (zbmp->width == bmp->width);
    assert (zbmp->height == bmp->height);
    Acc_draw_fragmented_sprite (&(obj->pos), &(obj->orientation), bmp,
				zbmp->bits, 8);
    break;
  }

  case OBJDRAW_INVISIBLE:
    break;

  case OBJDRAW_SCREEN:
  case OBJDRAW_DECAL:
    /*
     * These have already been drawn
     */
    break;

  case OBJDRAW_SPECIAL:
    render_special (obj);
    break;

  case OBJDRAW_FORCEDOOR:
  {
    door *pDoor = Door_get_door(obj->class_index);
    surface ts;

    Sur_map_to_surface (&ts, obj->ms);
    Ren_transform_surface (&ts); /* should be done later */
    if (pDoor->colour == 0) ts.shade = 0xff; /* default transparent red */
    else ts.shade = pDoor->colour;

    /*
     * Should be object's inclusion volume ...
     */
    if (Clip_surface_by_view_frustrum (&ts))
    {
      Acc_draw_surface (&ts);
    }
    break;
  }

  default:
    Cant_happen ("Obj_render_tile(): unknown object draw type");
    break;

  }

}



#define SPRITE_ALBEDO 0.7 /* sprite "reflectiveness" (all right, fudge) */

/*
 * Render a sprite.
 * This makes a surface for the sprite in view space, clips it to the current
 *  object inclusion volume and passes to Acc_draw_sprite() (which assumes a
 *  constant z value for its surface)
 */
void try_render_sprite (map_object *obj)
{

  ss_bitmap_header *bmp = Tex_get_texture (obj->fetch_texture_hook (obj));
  int dark;
  surface ts; /* temp surface, as per */
  vec tv;
  float scale = DEFAULT_BMP_SCALE;

  if (bmp == NULL) /* hmmm ... */
  {
    return;
  }

  /*
   * Start populating the surface vertices, at the top left corner which is
   *  also the texture anchor point
   */
  tv.x = obj->view.x - 0.5 * scale * bmp->width;
  if (Option_underworld) tv.y = obj->view.y + scale * bmp->height;
  else tv.y = obj->view.y + 0.5 * scale * bmp->height;
  tv.z = obj->view.z;

  ts.anchor = tv;
  ts.vertex [0] = tv;
  tv.x += scale * bmp->width;
  ts.vertex [1] = tv;
  tv.y -= scale * bmp->height;
  ts.vertex [2] = tv;
  tv.x -= scale * bmp->width;
  ts.vertex [3] = tv;

  /*
   * u and v vectors. Trivial, for a sprite
   */
  ts.u.x =  scale; ts.u.y =  0;      ts.u.z =  0;
  ts.v.x =  0;     ts.v.y = -scale;  ts.v.z =  0;

  ts.normal.x = 0; ts.normal.y = 0; ts.normal.z = -1;

  ts.num_vertices = 4;
  assert (obj->tile != NULL);
  dark = obj->dark - SPRITE_ALBEDO * DARK_SHADES *
    Dark_do_point (obj->tile, &(obj->pos));
  if (dark < 0) dark = 0;
  ts.dark [0] = ts.dark [1] = ts.dark [2] = ts.dark [3] = dark;

  ts.flags = SF_TEXTURED | SF_MASKED;

  if (Clip_surface_by_region (&ts, &Object_region_view))
  {
    Acc_draw_sprite (&ts, bmp);
  }

}


void render_special (map_object *obj)
{

  if ((obj->class == OBJCLASS_DOOR) && (obj->subclass == DOOR_SUBCLASS_3D))
  {
    Door_render_3d (obj);
    return;
  }

  /*
   * More convenient to switch on object ID methinks
   */
  switch (obj->object_id)
  {

  case OBJID_BRIDGE:
  case OBJID_CATWALK:
  case OBJID_FORCEBRIDGE1:
  case OBJID_FORCEBRIDGE2:
  {
    Mod_render_bridge (obj);
    break;
  }

  case OBJID_SMALL_CRATE:
  case OBJID_LARGE_CRATE:
  case OBJID_SECURE_CRATE:
  {
    Mod_render_crate (obj);
    break;
  }

  default:
    break;
  }
}




/*
 * Returns true if a xref exists between tile tile and object obj
 */
bool object_is_in_tile (const map_object *obj, const map_tile *tile)
{
  int refno = obj->tile_xref;
  while (refno != 0)
  {
    object_xref *xref = &(Object_xref_table [refno]);
    if (xref->tile == tile) return true;
    refno = xref->tile_chain;
    if (refno == obj->tile_xref) break; /* tile chains are circular! */
  }
  return false;
}





/*
 * Bitmap fetch functions for decals and sprites
 */


/*
 * This is for objects with only one texture, whose id is stored in the
 *  base_texid field
 */
int Obj_fetch_texture_single (map_object *obj)
{
  return obj->base_texid;
}


/*
 * This is for animated objects, which have several textures depending on the
 *  state field
 */
int Obj_fetch_texture_anim (map_object *obj)
{
  return obj->base_texid + obj->state;
}


/*
 * This is for oriented sprites, which have 8 textures depending on their
 *  relative orientation
 */
int Obj_fetch_texture_osprite (map_object *obj)
{
  return obj->base_texid + Obj_get_octant (obj);
}


int fetch_texture_default (map_object *obj)
{
  object_prop *prop = &(Object_prop_list [obj->object_id]);

  if (obj->class == OBJCLASS_ANIMATED)
  {
    obj->base_texid =
      Tex_get_cache_block (prop->bmp_chunk, prop->baseframe,
                           prop->extra_bmps + 1, TEXTURE_BLOCK_SUBCHUNK);
    obj->fetch_texture_hook = Obj_fetch_texture_anim;
  }
  else if (prop->draw_type == OBJDRAW_OSPRITE) /* corpses */
  {
    assert (prop->extra_bmps == 7);
    obj->base_texid = Tex_get_cache_block (prop->bmp_chunk, prop->baseframe,
                                           8, TEXTURE_BLOCK_SUBCHUNK);
    obj->fetch_texture_hook = Obj_fetch_texture_osprite;
  }
  else
  {
    obj->base_texid = Tex_get_cache_id (prop->bmp_chunk,
                                        prop->baseframe + obj->state);
    obj->fetch_texture_hook = Obj_fetch_texture_single;
  }
  return obj->fetch_texture_hook (obj);
}



int fetch_texture_uw (map_object *obj)
{
  /*
   * Nasty fudge for switch objects
   */
  int chunk = BITMAP_CHUNK_UW_SPRITE;
  int subchunk = obj->object_id;
  if ((obj->object_id >= 368) && (obj->object_id < 384))
  {
    chunk = BITMAP_CHUNK_UW_SWITCH;
    subchunk -= 368;
  }
  obj->base_texid = Tex_get_cache_id (chunk, subchunk);
  obj->fetch_texture_hook = Obj_fetch_texture_single;       /* snap pointer */
  return obj->base_texid;
}



int fetch_texture_uw_tmap (map_object *obj)
{
  obj->base_texid = obj->stuff2;
  obj->fetch_texture_hook = Obj_fetch_texture_single;
  return obj->base_texid;
}


int fetch_texture_uw_crit (map_object *obj)
{
  int variant = 0;
  int frame;
  int segment;
  int bmpno = (6 + Obj_get_octant (obj)) % 8;

  if (Underworld2)
  {
    segment = bmpno;
  }
  else
  {
    segment = 0x80 + bmpno;
  }
  frame = Crit_get_uw_segment_frame (obj->object_id, segment, 0, &variant);
  return Tex_get_cache_id (FAKE_CRIT_FRAME_CHUNK + obj->object_id -
                           CRIT_UW_BASE_OBJID, frame);
}


int fetch_texture_uw_bridge (map_object *obj)
{
  int bmpno = obj->stuff1 & 15;
  if (bmpno < 2)
  {
    return fetch_texture_default (obj);
  }
  else
  {
    bmpno -= 2;
    if (Underworld2) obj->base_texid = bmpno;
    else obj->base_texid = 48 + bmpno;
  }
  obj->fetch_texture_hook = Obj_fetch_texture_single;
  return obj->base_texid;
}








/*
 * End of src/object.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
