/*
 * object.h -
 *  Misc. object definitions
 */



/*
 * sentry
 */
#ifndef OBJECT_H
#define OBJECT_H



/*
 * #includes 
 */
#include "graph2d.h"
#include "matrix.h"




#define OBJ_MAXCLASS    16
#define OBJ_MAXSUBCLASS  8



/*
 * Object classes
 */
#define OBJCLASS_WEAPON      0
#define OBJCLASS_AMMO        1
#define OBJCLASS_PROJECTILE  2
#define OBJCLASS_EXPLOSIVE   3
#define OBJCLASS_PATCH       4
#define OBJCLASS_HARDWARE    5
#define OBJCLASS_SOFTS       6
#define OBJCLASS_FIXTURE     7   /* bad name, also contains bridges */
#define OBJCLASS_ITEM        8   /* mostly pickable */
#define OBJCLASS_SWITCH      9
#define OBJCLASS_DOOR       10   /* and gratings */
#define OBJCLASS_ANIMATED   11   /* are these the only ones? */
#define OBJCLASS_TRIGGER    12
#define OBJCLASS_CONTAINER  13
#define OBJCLASS_CRITTER    14
#define OBJCLASS_UNDERWORLD 15   /* catchall class for UW objects */



#define OBJDRAW_3D         1
#define OBJDRAW_SPRITE     2
#define OBJDRAW_SCREEN     3
#define OBJDRAW_CRITTER    4
#define OBJDRAW_OSPRITE    5  /* orientable sprite (TSSHP not Shock type) */
#define OBJDRAW_FRAGMENTS  6  /* c-space, made out of bits                */
#define OBJDRAW_INVISIBLE  7  /* don't draw this at all                   */
#define OBJDRAW_DECAL      8  /* oriented bitmap                          */
#define OBJDRAW_SPECIAL   11
#define OBJDRAW_FORCEDOOR 12  /* force door                               */



/*
 * Object class ranges for Underworld
 */
#define UW_OBJID_ANIM_BASE  448
#define UW_OBJID_ANIM_LAST  459


/*
 * Object flags
 */
#define OF_INVENTORY  0x0001      /* inventory item (main or access)       */
#define OF_TOUCHABLE  0x0002      /* can be touched                        */
#define OF_CONSUMABLE 0x0010      /* inv. item is consumed when used       */
#define OF_BLOCK3D    0x0020      /* can block 3D map search               */
#define OF_OPENS      0x0100      /* blocks passage when shut (door)       */
#define OF_SOLID      0x0200      /* can't walk or fall through it         */
#define OF_MISSILE    0x0800      /* explodes on hit (missile or live)     */


/*
 * Object rflags
 */
#define ORF_DRAWABLE 0x0001



/*
 * Octants for orientable sprites
 */
#define OCTANT_LEFT        0
#define OCTANT_LEFT_BACK   1
#define OCTANT_BACK        2
#define OCTANT_RIGHT_BACK  3
#define OCTANT_RIGHT       4
#define OCTANT_RIGHT_FRONT 5
#define OCTANT_FRONT       6
#define OCTANT_LEFT_FRONT  7


/*
 * Maximum number of cached bitmaps per object
 */
#define MAX_CACHED_BITMAPS 8


/*
 * A single value made from class/subclass/type, for use in switches etc.
 */
#define CST(o) (((o)->class << 16) + ((o)->subclass << 8) + (o)->type)



/*
 * Generic object properties. I still don't know what most of this does
 */
typedef struct
{

  long  mass;       /* in units of 100g                       */
  int   hp;         /* starting hit points                    */
  short armour;     /* what you need AP rounds to get through */
  short draw_type;  /* OBJDRAW_ constant above                */
  short scale;
  short vulns;      /* weapon vulnerabilities mask            */
  short special;    /* weapon special vulnerability mods      */
  short defence;    /* defence modifier value                 */
  int   flags;
  int   info;       /* 3D model chunk for drawtype 1          */
  int   extra_bmps; /* no. extra bitmaps (more than 3)        */
  int   baseframe;  /* in its bitmap chunk bmp_chunk          */
  int   value;      /* how much does it cost? (UW only)       */
  int   stuff2;     /* UW field which I don't know what it is */

  int   bmp_chunk;  /* chunk where its bitmaps live (TSSHP)   */

} object_prop;



struct tag_ss_bitmap_header;
struct tag_map_object;
struct tag_map_surface;
struct tag_map_tile;


/*
 * This is a hook off which to hang functions to fetch textures for bitmap-
 *  oriented objects (sprites and decals)
 */
typedef int (*fetch_texture_func) (struct tag_map_object *);



/*
 * Map object definition
 */
typedef struct tag_map_object
{

  short used;       /* object slot is in use  */
  short class;
  unsigned short subclass;
  int class_index;
  int tile_xref;
  int prev;         /* linked list of slots   */
  int next;         /* linked list of slots   */
  vec   pos;        /* world position         */
  short pitch;
  short heading;
  short bank;       /* 3 angles               */
  short ai;
  unsigned short type;
  int stuff1;       /* ??? decimal ... HP?    */
  int stuff2;       /* ??? don't know         */
  int state;
  int stuff3;       /* ??? don't know         */
  int flags;
  int quality;      /* for Underworld objects */
  int link2;        /* Underworld, quantity   */

  /*
   * Flattened object id ... index into the common object properties table
   */
  int object_id;

  /*
   * Renderer stuff
   */
  Matrix orientation;
  unsigned long frame;
  int rflags;
  int depth;
  struct tag_map_tile *tile;
  bounding_box bbox;
  struct tag_map_surface *ms;  /* map surface (decals)             */
  int dark;                    /* base static dark value (sprites) */
  Vector view;                 /* hopefully won't need this long   */

  /*
   * Textures. Most objects have a single texture in the main texture cache.
   *  Orientable sprites have 8. Animation objects may have more than one.
   *  Critters appear as orientable sprites to the renderer but their base
   *  texture ID is set by the AI code depending on AI state and frame.
   */
  int base_texid;
  fetch_texture_func fetch_texture_hook;

  /*
   * In theory the bitmap scale. Actually I have no clue how this is determined
   */
/*    float bmp_scale; */

} map_object;




/*
 * Cross-reference object <-> tile
 */
typedef struct
{

  struct tag_map_tile *tile;
  short index;      /* into the map_object list */
  short obj_chain;  /* next object this tile    */
  short tile_chain; /* next tile this object    */

} object_xref;




#ifdef __cplusplus
extern "C"
{
#endif




void Obj_init (void);
void Obj_deinit(void);
void Obj_init_level (int level);
int  Obj_flatten (int class, int subclass, int type);

/*
 * Render functions
 */
struct tag_clip_region;

void Obj_init_scene (void);
void Obj_queue_tile (struct tag_map_tile *tile);
void Obj_render_tile (struct tag_map_tile *tile, struct tag_map_tile *from,
		      const struct tag_clip_region *frustrum);
void Obj_render_tile_decals (struct tag_map_tile *tile,
			     const struct tag_clip_region *frustrum);
int  Obj_get_octant  (map_object *obj);
void Obj_dump_object_info (map_object *obj);
void Obj_make_decal_surface (map_object *obj, int width, int height,
			     int flags);

object_xref *Obj_get_reference(int index);
map_object *Obj_get_object(int index);
object_prop *Obj_get_property(int id);

/*
 * Standard texture fetch functions
 */
int Obj_fetch_texture_single  (map_object *obj);
int Obj_fetch_texture_anim    (map_object *obj);
int Obj_fetch_texture_osprite (map_object *obj);

#ifdef __cplusplus
}
#endif




/*
 * Global variables
 */
extern object_prop *Object_prop_list;




#endif /* sentry */

