/*
 * option.c -
 *  Command-line options
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gamemem.h"
#include "main.h"




/*
 * Local typedefs and structs
 */


/*
 * Options. Options may be abbreviated, first match in the list gets it
 * The `variable' field is cast to its proper type when the option is parsed.
 *  Need I go into the consequences of putting a variable of the wrong type
 *  here? I thought not.
 */
typedef struct
{
  const char *name;         /* Option string (full)   */
  int         type;         /* int, string, bool etc. */
  void       *variable;     /* Where to store it      */
  int         min;          /* Minimum for ints       */
  int         max;          /* Maximum for ints       */
} option_t;




/*
 * Types of option
 */
#define OPTION_INVALID 0
#define OPTION_INT     1
#define OPTION_BOOL    2
#define OPTION_STRING  3
#define OPTION_FLOAT   4



/*
 * Option public global variables and their defaults
 */
int Option_start_level   = 1;
int Option_frame_times   = 0;
int Option_print_map     = 0;
int Option_start_x       = -1;
int Option_start_y       = -1;
int Option_screen_width  = 320;
int Option_screen_height = 240;
int Option_underworld    = 0;
int Option_extract       = -1;
int Option_extract_sub   = 0;
int Option_extract_all   = 0;
int Option_fov           = 90;
int Option_dump_chunk    = -1;
int Option_dump_strings  = 0;
int Option_fullscreen    = 0;
int Option_no_alogs      = 0;
int Option_no_vmail      = 0;
int Option_no_hud        = 0;

char Option_hd_dir [256] = "";
char Option_cd_dir [256] = "";
char Option_uw_dir [256] = "";
char Option_mfd_cfg[256] = "";
char Option_con_cfg[256] = "control.cfg";

float Option_dark_gamma  = -1.f;

/*
 * The list of options
 */
static const option_t Option [] =
{
  { "start-level", OPTION_INT,     &Option_start_level,   0,    79 },
  { "frame-times", OPTION_BOOL,    &Option_frame_times,   0,     0 },
  { "print-map",   OPTION_BOOL,    &Option_print_map,     0,     0 },
  { "xpos",        OPTION_INT,     &Option_start_x,       0, 16383 },
  { "ypos",        OPTION_INT,     &Option_start_y,       0, 16383 },
  { "width",       OPTION_INT,     &Option_screen_width,  0,  1600 },
  { "height",      OPTION_INT,     &Option_screen_height, 0,  1200 },
  { "uw",          OPTION_BOOL,    &Option_underworld,    0,     0 },
  { "extract",     OPTION_INT,     &Option_extract,       0, 16383 },
  { "extract-sub", OPTION_INT,     &Option_extract_sub,   0, 16383 },
  { "extract-all", OPTION_BOOL,    &Option_extract_all,   0,     0 },
  { "fov",         OPTION_INT,     &Option_fov,           1,   179 },
  { "dump",        OPTION_INT,     &Option_dump_chunk,    0, 32768 },
  { "dump-strings",OPTION_BOOL,    &Option_dump_strings,  0,     0 },
  { "hddata",      OPTION_STRING,   Option_hd_dir,        0,   256 },
  { "cddata",      OPTION_STRING,   Option_cd_dir,        0,   256 },
  { "uwdata",      OPTION_STRING,   Option_uw_dir,        0,   256 },
  { "mfdcfg",	   OPTION_STRING,   Option_mfd_cfg,       0,   256 },
  { "concfg",      OPTION_STRING,   Option_con_cfg,       0,   256 },
  { "dark-gamma",  OPTION_FLOAT,   &Option_dark_gamma,    0, 99999 },
  { "fullscreen",  OPTION_BOOL,	   &Option_fullscreen,    0,     0 },
  { "no-alogs",    OPTION_BOOL,    &Option_no_alogs,      0,     0 },
  { "no_vmail",    OPTION_BOOL,    &Option_no_vmail,      0,     0 },
  { "no-hud",      OPTION_BOOL,    &Option_no_hud,        0,     0 },
  {  NULL,         OPTION_INVALID,  NULL,                 0,     0 }
};





static int process_option (const char *name, const char *parm, int on);
static void parse_config_file (void);



/*
 * The function to parse the command line
 * Takes the parameters as passed to main() plus a starting index. Should
 *  first be called with `start' set to 1 to skip the command name
 * Returns the index of the first non-option parameter (not beginning with `-'
 *  or an argument to a parameter)
 */
int Opt_parse (int argc, char *argv [], int start)
{

  int i;
  int offs;       /* into argument string */
  int skip;       /* argument found by process_option, skip */
  char *arg;

  /*
   * Paranoia
   */
  if (argv == NULL)
  {
    Cant_happen ("Opt_parse(): passed bogus argument list");
  }

  /*
   * Parse any configuration options from the config file. These can be
   *  overridden from the command line
   */
  parse_config_file ();

  for (i = start; i < argc; ++i)
  {

    /*
     * Paranoia
     */
    if (argv [i] == NULL)
    {
      Cant_happen ("Opt_parse(): passed bogus argument count");
    }

    arg = argv [i];
    if ((arg [0] != '-') && (arg [0] != '+'))
    {
      /*
       * Not an option
       */
      break;
    }

    offs = 1;
    if (arg [offs] == '-')
    {
      /*
       * Double minus sign. A grotty GNU-ism, but we'll allow it cos I'm
       *  feeling lenient
       */
      ++offs;

      /*
       * Allow an option `--' on its own to explicitly terminate the options
       *  list. Yes, I know that was a split infinitive.
       */
      if (arg [offs] == '\0')
      {
	++i;
	break;
      }
    }
    else if (arg [offs] == '\0')
    {
      /*
       * An option `-' on its own, however, is a syntax error.
       */
      fprintf (stderr, "Warning: bad option `-'\n");
      continue;
    }

    /*
     * We've (apparently) found an option. Search for it in the options list
     * As a special case we allow a single-letter abbreviation followed
     *  immediately (in the same argument) by a numeric parameter. Wrong
     *  option type will be caught later
     */
    if (i < argc - 1)
    {
      skip = process_option (arg + offs, argv [i + 1], (arg [0] == '-'));
    }
    else
    {
      /*
       * Last option ... better not need a parameter
       */
      skip = process_option (arg + offs, NULL, (arg [0] == '-'));
    }

    if (skip) ++i;

  }

  return i;

}








int process_option (const char *name, const char *parm, int on)
{

  int skip = 0;
  int single = 0; /* single-letter abbreviation with numeric after */
  int val;        /* for ints */
  char *end;      /* for strtol() */
  float fval;     /* for floats */
  int opt;

  /*
   * Check for a single-character abbreviation with numeric parameter
   */
  if (isdigit (name [1]))
  {
    single = 1;
    parm = name + 1;
  }

  for (opt = 0; Option [opt].name != NULL; ++opt)
  {
    if ((single && (Option [opt].name [0] == name [0])) ||
	(strncmp (Option [opt].name, name, strlen (name)) == 0))
    {
      break;
    }
  }

  if (Option [opt].name == NULL)
  {
    fprintf (stderr, "Warning: unrecognised option `%s'\n", name);
    return 0;
  }

  /*
   * Check for the existence of the parameter, if needed
   */
  if ((Option [opt].type != OPTION_BOOL) && (parm == NULL))
  {
    fprintf (stderr, "Warning: missing parameter to option `%s'\n", name);
    return 0;
  }

  switch (Option [opt].type)
  {
  case OPTION_INT:
    val = strtol (parm, &end, 0);
    if (*end != '\0')
    {
      fprintf (stderr, "Warning: trailing garbage in parameter to option"
	       " `%s'\n", name);
    }
    if ((val >= Option [opt].min) && (val <= Option [opt].max))
    {
      *((int *) (Option [opt].variable)) = val;
    }
    else
    {
      fprintf (stderr, "Warning: parameter to option `%s' out of range\n",
	       name);
    }
    skip = !single;
    break;

  case OPTION_BOOL:
    if (single)
    {
      /*
       * Single-letter boolean option followed by numeric argument --
       *  syntax error
       */
      fprintf (stderr, "Warning: parameter to boolean argument `%s'\n", name);
      return 0;
    }
    *((int *) (Option [opt].variable)) = on;
    break;

  case OPTION_STRING:
  {
    char *svar = (char *) Option [opt].variable;
    strncpy (svar, parm, Option [opt].max);
    svar [Option [opt].max - 1] = '\0';
    skip = 1;
    break;
  }

  case OPTION_FLOAT:
  {
    fval = (float)(strtod (parm, &end));
    if (*end != '\0')
    {
      fprintf (stderr, "Warning: trailing garbage in parameter to option"
	       " `%s'\n", name);
    }
    if ((fval >= Option [opt].min) && (fval <= Option [opt].max))
    {
      *((float *) (Option [opt].variable)) = fval;
    }
    else
    {
      fprintf (stderr, "Warning: parameter to option `%s' out of range\n",
	       name);
    }
    skip = !single;
    break;
  }

  case OPTION_INVALID:
    Cant_happen ("Opt_parse(): invalid option encountered"
		 " (off end of options list?)");

  default:
    Cant_happen ("Opt_parse(): error found in options list, fix option.c");

  }

  return skip;

}




#define CONFIG_FILE "cyb.cfg"
#define MAX_CONFIG_LINE 1024

void parse_config_file (void)
{

  char line [MAX_CONFIG_LINE];
  char *lp;
  char *name;
  char *parm;

  FILE *config = fopen (CONFIG_FILE, "rt");
  if (config == NULL)
  {
    fprintf (stderr, "Warning: couldn't open configuration file\n");
    return;
  }

  while (fgets (line, MAX_CONFIG_LINE, config) != NULL)
  {

    if (line [strlen (line) - 1] == '\n')
    {
      line [strlen (line) - 1] = '\0';
    }

    lp = line;
    while (isspace (*lp)) ++lp;

    if ((*lp == '\0') || (*lp == '#'))
    {
      /*
       * Blank line or comment
       */
      continue;
    }

    name = lp;
    while (!isspace (*lp) && (*lp != '\0')) ++lp;
    if (*lp != '\0')
    {
      /*
       * Terminate name, look for parameter
       */
      *lp = '\0';
      ++lp;
    }

    while (isspace (*lp)) ++lp;
    if (*lp == '\0')
    {
      parm = NULL;
    }
    else
    {
      parm = lp;
    }

    /*
     * Got the option ... now process it
     */
    process_option (name, parm, 1);

  }

}

