/*
 * palette.c -
 *  System Shock palette stuff
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>		/*abs*/
#include <memory.h>

#include "types.h"
#include "anim.h"
#include "game.h"
#include "gamemem.h"
#include "main.h"
#include "option.h"
#include "palette.h"
#include "resource.h"
#include "shade.h"

/*
 * #defines and constants
 */

/*  #define PALETTE_CHUNK_BASE 700 */
#define PALETTE_CHUNK_BASE 0



/*
 * Global variables
 */
int Palette [256] [3];
uint32 RGB_trans [256];

/*
 * First transparency value in the indexed palette
 */
unsigned int First_trans_index;


/*
 * The colour cube, for back-conversions from 24 to 8 bits
 */
#define COLOUR_CUBE_SIDE 32
byte *Colour_cube;


/*
 * Palette-based animations. Each game has a different set of these.
 * Since palette index 0 is special, and never appears in a palette anim
 *  sequence, we can use it as a terminator
 */
struct palanim_sequence { const byte start; const byte nframes; byte frame; };

struct palanim_sequence SS_palanim [] =
{ { 0x04, 4, 0 },
  { 0x08, 4, 0 },
  { 0x0c, 4, 0 },
  { 0x10, 4, 0 },
  { 0x14, 4, 0 },
  { 0x18, 4, 0 },
  { 0x00, 0, 0 } };

struct palanim_sequence UW_palanim [] =
{ { 0x10, 5, 0 },
  { 0x15, 3, 0 },  /* lava */
  { 0x30, 4, 0 },
  { 0x34, 4, 0 },
  { 0x38, 4, 0 },
  { 0x3c, 4, 0 },  /* water */
  { 0x00, 0, 0 } };

struct palanim_sequence UW2_palanim [] =
{ { 0x03, 5, 0 },
  { 0x08, 3, 0 },  /* lava */
  { 0xe0, 4, 0 },
  { 0xe4, 4, 0 },
  { 0xe8, 4, 0 },
  { 0xec, 4, 0 },  /* water */
  { 0x00, 0, 0 } };

struct palanim_sequence *Pal_anim;



/*
 * Palette index flags
 */
#define PF_TRANSPARENT 0x01
#define PF_ANIMATED    0x02

static byte *Pal_flags;


void Pal_init (void)
{

  if (Option_underworld)
  {
    Pal_load_palette (0);
    if (Underworld2)
    {
      extern byte UW2_palflags [];
      Pal_anim  = UW2_palanim;
      Pal_flags = UW2_palflags;
    }
    else
    {
      extern byte UW_palflags [];
      Pal_anim  = UW_palanim;
      Pal_flags = UW_palflags;
    }
  }
  else
  {
    extern byte SS_palflags [];
    Res_add_resfile (Res_make_path (CD_DATA_PATH, "cutspal.res"));
    Res_add_resfile (Res_make_path (CD_DATA_PATH, "splshpal.res"));
    Res_add_resfile (Res_make_path (CD_DATA_PATH, "gamepal.res"));
    Pal_load_palette (700);
    Pal_anim  = SS_palanim;
    Pal_flags = SS_palflags;
  }

  /*
   * Load up the colour cube
   */
  {
    size_t cubesize = COLOUR_CUBE_SIDE * COLOUR_CUBE_SIDE * COLOUR_CUBE_SIDE;
    FILE *ccfile;
    Colour_cube = Alloc_init (cubesize, "Pal_init", "colour cube");

    if (Option_underworld)
    {
      int r, g, b;
      for (r = 0; r < COLOUR_CUBE_SIDE; ++r)
      {
	for (g = 0; g < COLOUR_CUBE_SIDE; ++g)
	{
	  for (b = 0; b < COLOUR_CUBE_SIDE; ++b)
	  {
	    Colour_cube [r + COLOUR_CUBE_SIDE * g +
			COLOUR_CUBE_SIDE * COLOUR_CUBE_SIDE * b]
	      = (byte)Pal_approx_color (rgb (8 * r, 8 * g, 8 * b), RGB_trans);
	  }
	}
      }
    }
    else
    {
      ccfile = fopen (Res_make_path (CD_DATA_PATH, "ipal.dat"), "rb");
      if (ccfile == NULL)
      {
	Cant_happen ("Pal_init(): failed to load colour cube");
      }
      if (fread (Colour_cube, 1, cubesize, ccfile) != cubesize)
      {
	Cant_happen ("Pal_init(): failed to load colour cube");
      }
      Colour_cube [0] = 1; /* for some reason black maps to transparent */
      fclose (ccfile);
    }
  }

}

void Pal_deinit(void)
{
	free(Colour_cube);
	Colour_cube = NULL;
}




static void advance_palanim (anim_control *ac)
{
  int i;
  UNUSED (ac); /* morally justified, since we're a callback */

  for (i = 0; Pal_anim [i].start > 0; ++i)
  {
    int col, dark;
    int start = Pal_anim [i].start;
    uint32 rgb_save;
    Pal_anim [i].frame++;
    if (Pal_anim [i].frame >= Pal_anim [i].nframes)
    {
      Pal_anim [i].frame = 0;
    }

    /*
     * Master RGB table
     */
    rgb_save = RGB_trans [start];
    for (col = 0; col < Pal_anim [i].nframes - 1; ++col)
    {
      RGB_trans [col + start] = RGB_trans [1 + col + start];
    }
    RGB_trans [col + start] = rgb_save;

    /*
     * Dark table
     */
    for (dark = 0; dark < DARK_SHADES; ++dark)
    {
      pixel dark_save = Dark_table [dark] [start];
      for (col = 0; col < Pal_anim [i].nframes - 1; ++col)
      {
        Dark_table [dark] [col + start] = Dark_table [dark] [1 + col + start];
      }
      Dark_table [dark] [col + start] = dark_save;
    }
  }      

}



/*
 * Initialise the palette for a level. This just registers the palette anims
 *  with the animation subsystem, since they are clobbered when not in a level
 */
void Pal_init_level (void)
{

  anim_control *ac = Anim_add (0.25, NULL, 0, AF_PAUSED);
  ac->advance_hook = advance_palanim;

}


void Pal_exit_level (void)
{
}



void Pal_load_palette (int offset)
{

  int i;
  int c;
  const unsigned char *ppal;

  if (Option_underworld)
  {
    FILE *palfile = fopen (Res_make_path (UW_DATA_PATH, "pals.dat"), "rb");
    unsigned char rawpal [0x300];
    if (palfile == NULL)
    {
      Cant_happen ("Pal_load_palette(): failed to open UW palette file");
    }
    fseek (palfile, 0x300 * offset, SEEK_SET);
    if (fread (rawpal, 1, 0x300, palfile) != 0x300)
    {
      fprintf (stderr, "Pal_load_palette(): failed to load palette %d\n",
	       offset);
      Cant_happen ("Pal_load_palette(): couldn't load UW palette");
    }
    ppal = rawpal;
    for (i = 0; i < 256; ++i)
    {
      for (c = 0; c < 3; ++c)
      {
        Palette [i] [c] = 4 * *ppal++;
      }
    }
    fclose (palfile);
    First_trans_index = 0xf0;
  }
  else
  {
    size_t size;
    block_handle hpal
      = Res_load_chunk_by_id (PALETTE_CHUNK_BASE + offset, &size);

    if ((hpal == NULL) || (*hpal == NULL))
    {
      fprintf (stderr, "Pal_load_palette(): failed to load palette %d\n",
	       offset);
      return;
    }

    ppal = *hpal;
    for (i = 0; i < 256; ++i)
    {
      for (c = 0; c < 3; ++c)
      {
        Palette [i] [c] = *ppal++;
      }
    }

    Res_free_chunk_by_id (PALETTE_CHUNK_BASE + offset);
    First_trans_index = 0xe0;

  }

  for (i = 0; i < 256; ++i)
  {
    RGB_trans [i] = rgbt (Palette [i] [0], Palette [i] [1], Palette [i] [2], 0xFF);
  }
	/* apply transparency values */
	RGB_trans[0x00] = rgbt(0x00, 0x00, 0x00, 0x00);
	if (Option_underworld)
	{ /* UW uses transparency for shadows */
		RGB_trans[0xFB] = rgbt(0x20, 0x20, 0x20, 0x20);	/* UW1 (?) */
		RGB_trans[0xFC] = rgbt(0x00, 0x00, 0x00, 0x20);	/* UW1 (?) */
		RGB_trans[0xFD] = rgbt(0x20, 0x20, 0x20, 0x20); /* UW2 (?) */
	}
	else
	{	/* SSs special vals - set by guessing */
		RGB_trans[0xF7] = rgbt(0x00, 0x00, 0x00, 0x20);
		RGB_trans[0xF8] = rgbt(0x00, 0x00, 0x00, 0x10);
		RGB_trans[0xFC] = rgbt(0x20, 0x20, 0x20, 0x20);
    RGB_trans[0xFD] = rgbt(0x00, 0x00, 0xff, 0x30); /* fd-ff are force */
    RGB_trans[0xFE] = rgbt(0x00, 0xff, 0x00, 0x30); /* doors / bridges */
    RGB_trans[0xFF] = rgbt(0xff, 0x00, 0x00, 0x30);
	}
}

void Pal_shade_palette(pcolor *pPal, int iFrom, int iTo,
											 pcolor cFrom, pcolor cTo)
{
	HSHADE shClr;
	int i, iAmount;
	byte fRed, fGreen, fBlue, fTrans, tRed, tGreen, tBlue, tTrans;
	double curShade, shStep;

	iAmount = iTo - iFrom;
	shStep = (double)(0x100) / (double)(iAmount);
	curShade = 0.0;
	fRed = get_red(cFrom);
	fGreen = get_green(cFrom);
	fBlue = get_blue(cFrom);
	fTrans = get_trans(cFrom);
	tRed = get_red(cTo);
	tGreen = get_green(cTo);
	tBlue = get_blue(cTo);
	tTrans = get_trans(cTo);
	for (i = iFrom; i < iTo; i++, curShade += shStep)
	{
		curShade = min(255.0, curShade);
		shClr = shade_get((byte)(curShade));
		pPal[i] = rgbt(shade(shClr, fRed, tRed),
									 shade(shClr, fGreen, tGreen),
									 shade(shClr, fBlue, tBlue),
									 shade(shClr, fTrans, tTrans)
									);
	}
	pPal[iFrom] = cFrom;
	pPal[iTo - 1] = cTo;
}

void Pal_filter_palette(pcolor *pPalOut, const pcolor *pPalIn,
												pcolor cFilter, int nPalSize, int nGamma)
{
	HSHADE hShade;
	int i, nSquare, nRed, nGreen, nBlue, nIntensity;

	for (i = 0; i < nPalSize; i++)
	{
		nRed = get_red(pPalIn[i]);
		nGreen = get_green(pPalIn[i]);
		nBlue = get_blue(pPalIn[i]);
		nSquare = nRed * nRed +	nGreen * nGreen + nBlue * nBlue;
		/* Maximum Intensity: 0x2FA03 (0xFF * 0xFF * 3) */
		nIntensity = ((nSquare * 0xFF) / 0x2FA03);
		hShade = shade_get((byte)min(0xFF, max(0, nIntensity + nGamma)));
		pPalOut[i] = rgbt(shade(hShade, 0x00, get_red(cFilter)),
											shade(hShade, 0x00, get_green(cFilter)),
											shade(hShade, 0x00, get_blue(cFilter)),
											get_trans(cFilter));
	}
}

void Pal_copy_palette(pcolor *pPalOut, const pcolor *pPalIn, byte nShade)
{
	int i;
	byte *pTrans;

	memcpy(pPalOut, pPalIn, sizeof(pcolor) * 0x100);
	pTrans = &(((byte *)(pPalOut))[3]);
	for (i = 0; i < 0x100; i++, pTrans += 4)
		*pTrans = nShade;
}

byte Pal_approx_color(pcolor truecolor, pcolor *pPal)
{
	pcolor *pCur;
	int approx = 0;
	uint32 curdist = 0xFFFFFFFFUL, buf;
	unsigned int i;

  /*
   * Use First_trans_index here to make sure transparent colours aren't
   *  considered for nearest-approximations
   */
	pCur = pPal;
	for(i = 0; i < First_trans_index && curdist != 0; i++)
	{
    int delta_r, delta_g, delta_b;
    if (Pal_flags [i] != 0) continue;

	  /*
	   * Mean-square difference seems to work better here -jim
	   */
	  delta_r = get_red   (*pCur) - get_red   (truecolor);
	  delta_g = get_green (*pCur) - get_green (truecolor);
	  delta_b = get_blue  (*pCur) - get_blue  (truecolor);
	  ++pCur;
	  buf = (delta_r * delta_r + delta_g * delta_g + delta_b * delta_b);
		if(buf < curdist)
		{
			curdist = buf;
			approx = i;
		}
	}

	return (byte) approx;
}

byte Pal_rgb_to_index (uint32 tc)
{
/*
  return Colour_cube [(get_red (tc) >> 3) + 32 * (get_green (tc) >> 3) +
		     32*32 * (get_blue (tc) >> 3)];
*/

/* ff1: i like to optimize such things ;) */
  /* jim - yes, but this isn't really an optimisation; a compiler which can't
     figure out to replace a multiply by an integer power-of-2 constant with
     a shift isn't really one we should be using. Besides, we should be
     multiplying with COLOUR_CUBE_SIDE, not 32 */

/*      (shouldn't this return byte, not pixel?) */
  /* jim - yes it should. Now we've thrashed out the difference between
     rgb->index and rgb->pixel it does 8-) */
	return Colour_cube [ (get_red		(tc)	>> 3)				|
											((get_green (tc)	>> 3) << 5) |
											((get_blue	(tc)	>> 3) << 10)];
}








/*
 * End of src/palette.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
