/*
 * physics.c -
 *  Game physics
 */

/*
 * File:    src/physics.c
 * Purpose: Game physics
 * Prefix:  Phy_
 * Note:    
 * Music:   Bon Jovi, "Bon Jovi"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <math.h>
#include <stdlib.h>
#include <memory.h>

#include "types.h"
#include "graph2d.h" /* control.h needs point struct (move to types.h?) */

#include "control.h"
#include "gamemem.h"
#include "map.h"
#include "matrix.h"
#include "object.h"
#include "render.h"


/*
 * We want pi!
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif



/*
 * Local function prototypes
 */
static void left_func  (double elapsed);
static void right_func (double elapsed);
static void stepleft_func  (double elapsed);
static void stepright_func (double elapsed);
static void near_func  (double elapsed);
static void far_func   (double elapsed);

static void height_func (double rel);

/*
 * Global variables
 */

/*
 * Avatar orientation (not necessarily the same as camera orientation ... does
 *  not take into account look up/down
 */
double Avatar_angle = M_PI;      /* anticlockwise from east */
Vector Avatar_pos;
Matrix Avatar_orn = {{0., 1., 0.}, {0., 0., 1.}, {-1., 0., 0.}};


/*
 * A map object for the avatar
 */
map_object Avatar_obj;



/*
 * In game units per sec
 */
static const int Run_speed = 500;

static const double Turn_angle = M_PI / 2.;

static ss_controlset Control;




/*
 * Call this every physics frame
 */
void Phy_frame (double elapsed)
{

  /*
   * Physics is pretty trivial at the moment - just gather controls and move
   *  the viewpoint accordingly
   */
  Con_snapshot (&Control, true);

	if (Control.bControl.run)
	{
		near_func (elapsed * Control.bControl.run);
	}
	else
	{
		if (Control.bControl.forwards)
		{
			near_func (elapsed);
		}
	}
	if (Control.rControl.forwards)
	{
		near_func (elapsed * (Control.rControl.forwards / M_PI));
	}
  if (Control.rControl.back)
  {
    far_func (elapsed * (Control.rControl.back / M_PI));
  }
  if (Control.bControl.back)
  {
    far_func (elapsed);
  }
  if (Control.rControl.right)
  {
    right_func (elapsed * (Control.rControl.right / M_PI));
  }
  if (Control.bControl.right)
  {
    right_func (elapsed);
  }
  if (Control.rControl.left)
  {
    left_func (elapsed * (Control.rControl.left / M_PI));
  }
  if (Control.bControl.left)
  {
    left_func (elapsed);
  }
  if (Control.rControl.stepright)
  {
    stepright_func (elapsed * (Control.rControl.stepright / M_PI));
  }
  if (Control.bControl.stepright)
  {
    stepright_func (elapsed);
  }
  if (Control.rControl.stepleft)
  {
    stepleft_func (elapsed * (Control.rControl.stepleft / M_PI));
  }
  if (Control.bControl.stepleft)
  {
    stepleft_func (elapsed);
  }
  if (Control.rControl.height)
  {
    height_func (Control.rControl.height);
  }

  /*
   * Not too sure about this either -jim
   */
  memset(&Control.rControl, 0x00, sizeof(RelativeControl));

  if (!Map_info.cyberspace)
  {
    /*
     * Nasty fudge for viewpoint height
     */
    Avatar_pos.z = Map_floor_height ((int)Avatar_pos.x, (int)Avatar_pos.y);
  }

  /*
   * Update the avatar's map object to reflect the control actions
   */
  Avatar_obj.pos         = Avatar_pos;
  Avatar_obj.orientation = Avatar_orn;

}




/*
 * Get the world position of a point relative to a map_object
 * I'm not entirely sure this should be the province of the physics engine, but
 *  it'll do ok here for now
 */
void Phy_get_rel_pos (const map_object *obj, const vec *rel_pos, vec *ret_pos)
{
  /*
   * Relative position in world coordinates
   */
  mat_vec_multiply_transpose (*ret_pos, obj->orientation, *rel_pos);

  /*
   * Absolute position in world coordinates
   */
  vec_add (*ret_pos, *ret_pos, obj->pos);

}




/*
 * Turn
 */
static void left_func (double elapsed)
{

  double ct = cos (elapsed * Turn_angle);
  double st = sin (elapsed * Turn_angle);
  Matrix tm = Avatar_orn;

  Avatar_angle += elapsed * Turn_angle;
  while (Avatar_angle >= 2. * M_PI) Avatar_angle -= 2. * M_PI;
  Avatar_orn.x.x = ct * tm.x.x - st * tm.x.y;
  Avatar_orn.y.x = ct * tm.y.x - st * tm.y.y;
  Avatar_orn.z.x = ct * tm.z.x - st * tm.z.y;
  Avatar_orn.x.y = ct * tm.x.y + st * tm.x.x;
  Avatar_orn.y.y = ct * tm.y.y + st * tm.y.x;
  Avatar_orn.z.y = ct * tm.z.y + st * tm.z.x;

}

static void right_func (double elapsed)
{

  double ct = cos (elapsed * Turn_angle);
  double st = sin (elapsed * Turn_angle);
  Matrix tm = Avatar_orn;

  Avatar_angle -= elapsed * Turn_angle;
  while (Avatar_angle < 0.) Avatar_angle += 2. * M_PI;
  Avatar_orn.x.x = ct * tm.x.x + st * tm.x.y;
  Avatar_orn.y.x = ct * tm.y.x + st * tm.y.y;
  Avatar_orn.z.x = ct * tm.z.x + st * tm.z.y;
  Avatar_orn.x.y = ct * tm.x.y - st * tm.x.x;
  Avatar_orn.y.y = ct * tm.y.y - st * tm.y.x;
  Avatar_orn.z.y = ct * tm.z.y - st * tm.z.x;

}



/*
 * Strafe
 */
static void stepleft_func  (double elapsed)
{
  Avatar_pos.x += elapsed * Run_speed * Avatar_orn.x.x;
  Avatar_pos.y += elapsed * Run_speed * Avatar_orn.x.y;
  Avatar_pos.z += elapsed * Run_speed * Avatar_orn.x.z;
}

static void stepright_func (double elapsed)
{
  Avatar_pos.x += elapsed * Run_speed * -Avatar_orn.x.x;
  Avatar_pos.y += elapsed * Run_speed * -Avatar_orn.x.y;
  Avatar_pos.z += elapsed * Run_speed * -Avatar_orn.x.z;
}



/*
 * Move forward/backward
 */
static void near_func (double elapsed)
{
  Avatar_pos.x += elapsed * Run_speed * Avatar_orn.z.x;
  Avatar_pos.y += elapsed * Run_speed * Avatar_orn.z.y;
  Avatar_pos.z += elapsed * Run_speed * Avatar_orn.z.z;
}

static void far_func (double elapsed)
{
  Avatar_pos.x -= elapsed * Run_speed * Avatar_orn.z.x;
  Avatar_pos.y -= elapsed * Run_speed * Avatar_orn.z.y;
  Avatar_pos.z -= elapsed * Run_speed * Avatar_orn.z.z;
}

static void height_func (double rel)
{
	extern int Screen_height;
  Matrix rot;
  Matrix tm;

  double ct;
  double st;

  /*
   * Calculate x-based rotation
   */
/*
  if (rel)
  {
    ct = cos (rel * M_PI / (Screen_width));
    st = sin (rel * M_PI / (Screen_width));
    rot.x.x = ct;
    rot.x.y = -st;
    rot.x.z = 0;
    rot.y.x = st;
    rot.y.y = ct;
    rot.y.z = 0;
    rot.z.x = 0;
    rot.z.y = 0;
    rot.z.z = 1;
    Mat_multiply (tm, Camera->orientation, rot);
    Camera->orientation = tm;
  }
*/
  /*
   * Calculate y-based rotation
   */

  if (rel)
  {
    ct = cos (rel * M_PI / (Screen_height));
    st = sin (rel * M_PI / (Screen_height));
    rot.x.x = 1;
    rot.x.y = 0;
    rot.x.z = 0;
    rot.y.x = 0;
    rot.y.y = ct;
    rot.y.z = st;
    rot.z.x = 0;
    rot.z.y = -st;
    rot.z.z = ct;
    Mat_multiply (tm, rot, Avatar_orn);
    Avatar_orn = tm;
  }

}
