/*
 * plot.c -
 *  I'm hatching a plot ...
 */

/*
 * File:    src/plot.c
 * Purpose: Low-level plot and texture-map functions for the software renderer
 * Prefix:  Plot_
 * Note:    Things will migrate here from texture.c over time
 * Music:   Little Angels, "Young Gods"
 * TODO:    Reunify the scan convert functions
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "accel.h"
#include "bitmap.h"
#include "gamemem.h"
#include "graph2d.h"
#include "hud.h"
#include "main.h"
#include "palette.h"
#include "plot.h"
#include "render.h"
#include "shade.h"
#include "surface.h"
#include "texture.h"
#include "vec.h"


/*
 * Debug
 */
#define PLOT_MESSAGES 0



/*
 * Imported globals (BAD! NAUGHTY!)
 */
extern int Origin_x;
extern int Origin_y;
extern int Window_width;
extern int Window_height;

extern float *Z_buffer;

extern raster_surface Scratch_raster;




/*
 * Mapping for a surface. This contains all the information an inner-loop
 *  function could possibly need for performing the plot
 */
struct surface_map
{
  raster_surface *rs;

  /*
   * Texture mapping stuff (valid if surface is texture-mapped, obviously)
   */
  double duz_dx;
  double duz_dy;
  double dvz_dx;
  double dvz_dy;

  double dzinv_dx;
  double dzinv_dy;

  double ax;
  double ay;

  /*
   * Initial values for u/z and v/z at top scanline
   */
  double uz_y;
  double vz_y;

  ss_bitmap_header *bmp;

  /*
   * flat shade ... is a RGBA value if surface is transparent, else palette
   *  index (for now)
   */
  pcolor shade;

};


/*
 * Local function prototypes
 */
bool scan_convert_surface (const surface *s, struct surface_map *sm);
bool scan_convert_gouraud (surface *s, struct surface_map *sm);



/*
 * Inner loop plot functions
 */
typedef void (*plot_func) (struct surface_map *sm, pixel *bitmap);

void plot_inner_textured     (struct surface_map *sm, pixel *bitmap);
void plot_inner_zbuffer      (struct surface_map *sm, pixel *bitmap);
void plot_inner_txt_zbf      (struct surface_map *sm, pixel *bitmap);
void plot_inner_txt_shd      (struct surface_map *sm, pixel *bitmap);
void plot_inner_txt_msk_shd  (struct surface_map *sm, pixel *bitmap);
void plot_inner_zbf_shd      (struct surface_map *sm, pixel *bitmap);
void plot_inner_transparency (struct surface_map *sm, pixel *bitmap);
void plot_inner_notimp       (struct surface_map *sm, pixel *bitmap);

/*
 * Inner loop table. 32 of these for different combinations of the 5 surface
 *  flags (not all are allowed)
 */
static plot_func Inner_loop [32] =
{
                             /* txt | msk | zbf | shd | trn */
  plot_inner_notimp,         /*     |     |     |     |     */
  plot_inner_textured,       /*  x  |     |     |     |     */
  plot_inner_notimp,         /*     |  x  |     |     |     */
  plot_inner_notimp,         /*  x  |  x  |     |     |     */
  plot_inner_zbuffer,        /*     |     |  x  |     |     */
  plot_inner_txt_zbf,        /*  x  |     |  x  |     |     */
  plot_inner_notimp,         /*     |  x  |  x  |     |     */
  plot_inner_notimp,         /*  x  |  x  |  x  |     |     */
  plot_inner_notimp,         /*     |     |     |  x  |     */
  plot_inner_txt_shd,        /*  x  |     |     |  x  |     */
  plot_inner_notimp,         /*     |  x  |     |  x  |     */
  plot_inner_txt_msk_shd,    /*  x  |  x  |     |  x  |     */
  plot_inner_zbf_shd,        /*     |     |  x  |  x  |     */
  plot_inner_notimp,         /*  x  |     |  x  |  x  |     */
  plot_inner_notimp,         /*     |  x  |  x  |  x  |     */
  plot_inner_notimp,         /*  x  |  x  |  x  |  x  |     */
  plot_inner_transparency,   /*     |     |     |     |  x  */
  plot_inner_notimp,         /*  x  |     |     |     |  x  */
  plot_inner_notimp,         /*     |  x  |     |     |  x  */
  plot_inner_notimp,         /*  x  |  x  |     |     |  x  */
  plot_inner_notimp,         /*     |     |  x  |     |  x  */
  plot_inner_notimp,         /*  x  |     |  x  |     |  x  */
  plot_inner_notimp,         /*     |  x  |  x  |     |  x  */
  plot_inner_notimp,         /*  x  |  x  |  x  |     |  x  */
  plot_inner_notimp,         /*     |     |     |  x  |  x  */
  plot_inner_notimp,         /*  x  |     |     |  x  |  x  */
  plot_inner_notimp,         /*     |  x  |     |  x  |  x  */
  plot_inner_notimp,         /*  x  |  x  |     |  x  |  x  */
  plot_inner_notimp,         /*     |     |  x  |  x  |  x  */
  plot_inner_notimp,         /*  x  |     |  x  |  x  |  x  */
  plot_inner_notimp,         /*     |  x  |  x  |  x  |  x  */
  plot_inner_notimp,         /*  x  |  x  |  x  |  x  |  x  */
};



/*
 * Distance between 2 consecutive fragments
 */
#define FRAG_SEP 16.f;

void Plot_fragmented_sprite (vec *pos, mat *orn, ss_bitmap_header *bmp, 
			     byte *zarray, int frag_size)
{

  vec u, v, normal;
  vec tv;
  vec frag_pos;  /* position of whatever fragment */
  vec frag_line; /* "left" of fragment scanline   */
  int x, y;
  byte *bp = bmp->bits;

  /*
   * What passes for texture mapping
   */
  u.x =  orn->x.x * FRAG_SEP;
  u.y =  orn->y.x * FRAG_SEP;
  u.z =  orn->z.x * FRAG_SEP;

  v.x = -orn->x.z * FRAG_SEP;
  v.y = -orn->y.z * FRAG_SEP;
  v.z = -orn->z.z * FRAG_SEP;

  normal.x = -orn->x.y * FRAG_SEP;
  normal.y = -orn->y.y * FRAG_SEP;
  normal.z = -orn->z.y * FRAG_SEP;

  /*
   * Top left corner of the bitmap
   */
  tv = u;
  vec_multiply (tv, 0.5 * bmp->width);
  vec_subtract (frag_pos, *pos, tv);
  tv = v;
  vec_multiply (tv, 0.5 * bmp->height);
  vec_subtract (frag_pos, frag_pos, tv);

  for (y = 0; y < bmp->height; ++y)
  {
    frag_line = frag_pos;
    for (x = 0; x < bmp->width; ++x)
    {
      int index = *bp++;
      int z = *zarray++;
      if (index != 0)
      {
	vec tv = normal;
        vec_multiply (tv, 0xd0 - z);
	vec_add (tv, tv, frag_pos);
	if (tv.z > 0.)
	{
	  Acc_draw_fragment (&tv, Pal_index_to_rgb (index), frag_size);
	}
      }
      vec_add (frag_pos, frag_pos, u);
    }
    vec_add (frag_pos, frag_line, v);
  }
}



void Plot_fragment (vec *pos, uint32 rgb, int frag_size)
{

  vec tv;
  vec view;
  int start_x, end_x;
  int start_y, end_y;
  int x, y;
  pixel colour = Pal_rgb_to_pixel (rgb);

  extern float *Z_buffer;

  /*
   * Transform the position to view coordinates
   */
  vec_subtract (tv, *pos, Camera->position);
  mat_vec_multiply (view, Camera->orientation, tv);

  /*
   * Screen coordinates of the fragment corners
   */
  start_x = floor ((view.x - frag_size) * Camera->screen_z / view.z + 0.5);
  end_x   = floor ((view.x + frag_size) * Camera->screen_z / view.z + 0.5);
  start_y = floor ((view.y + frag_size) * Camera->screen_z / view.z + 0.5);
  end_y   = floor ((view.y - frag_size) * Camera->screen_z / view.z + 0.5);

  /*
   * Clipping. Where OUGHT we to do this?
   */
  start_x += Origin_x; end_x += Origin_x;
  start_y = Origin_y - start_y; end_y = Origin_y - end_y;
  start_x = (start_x > 0) ? start_x : 0;
  end_x = (end_x < Window_width) ? end_x : Window_width;
  start_y = (start_y > 0) ? start_y : 0;
  end_y = (end_y < Window_height) ? end_y : Window_height;

  if (Camera->flags & CF_ZBUFFER)
  {
    /*
     * The Z-buffer is set up for use by the texture mapper, which uses a
     *  different transform, so some conversion is needed here
     */
    float zval = Camera->screen_z / view.z;
    for (y = start_y; y < end_y; ++y)
    {
      pixel *bits = Camera->framebuffer + y * Window_width;
      float *zbits = Z_buffer + y * Window_width + start_x;
      for (x = start_x; x < end_x; ++x)
      {
	if (zval > *zbits)
	{
	  bits [x] = colour;
	  *zbits = zval;
	}
	++zbits;
      }
    }
  }
  else
  {
    for (y = start_y; y < end_y; ++y)
    {
      pixel *bits = Camera->framebuffer + y * Window_width;
      for (x = start_x; x < end_x; ++x)
      {
        bits [x] = colour;
      }
    }
  }

}





#define SPRITE_MESSAGES 0

#define Z_DIVIDE(v,a) (Camera->screen_z * (v).a / (v).z);

struct raster_edge
{
  int start_vert;
  int end_vert;
  int count;      /* no. scanlines */
  double dx_dy;
};


/*
 * This is an ordinary sprite, scaled but not rotated, with an option to use
 *  the Z-buffer so that it sorts correctly against 3D objects
 */
void Plot_sprite (surface *s, ss_bitmap_header *bmp)
{

  double screen_x [MAX_VERTEX_SURFACE], screen_y [MAX_VERTEX_SURFACE];

  /*
   * Scale the sprite according to the surface u and v vectors
   * Since it is a sprite, we can assume that u points right and v down, in
   *  screen space
   */
  double du_dx =  s->anchor.z / (Camera->screen_z * s->u.x);
  double dv_dy = -s->anchor.z / (Camera->screen_z * s->v.y);

  int top_vertex = 0;
  double top_y = DBL_MAX;

  /*
   * Edges
   */
  struct raster_edge left, right;

  /*
   * "Anchor" point for the bitmap
   */
  double ax, ay;

  /*
   * u, v coords
   */
  double u;
  double v;
  int iv;

# if Z_BUFFER_SPRITES
  float zinv;
# endif

  int x, y;
  pixel *bits;

	/* for transparent critter */
	HSHADE hShade = NULL;	
	pcolor clrCritter;

  int i;
  int dark = (int) s->dark [0];          /* dark value is constant */
  pixel *dt = &(Dark_table [dark] [0]);  /* shade table for this dark value */

  bool end = false;

  /*
   * Minimum depth for drawing a sprite
   */
  if (s->anchor.z < 1) return;

# if Z_BUFFER_SPRITES
  zinv = 1. / s->anchor.z;
# endif

  du_dx *= 1 - A_VERY_SMALL_THING;
  dv_dy *= 1 - A_VERY_SMALL_THING;

  /*
   * Convert the surface vertices to screen space and sort by y
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    screen_x [i] = Origin_x + Z_DIVIDE (s->vertex [i], x);
    screen_y [i] = Origin_y - Z_DIVIDE (s->vertex [i], y);
    if (screen_y [i] < top_y)
    {
      top_y = screen_y [i];
      top_vertex = i;
    }
  }

  ax = Origin_x + Z_DIVIDE (s->anchor, x);
  ay = Origin_y - Z_DIVIDE (s->anchor, y);
  y = (int) POS2PIXEL (top_y);
  left.start_vert  = left.end_vert  = top_vertex;
  right.start_vert = right.end_vert = top_vertex;

  v = dv_dy * (top_y - ay);

  do
  {

    int ycount;
    double sx, ex;
    left.count  = (int) POS2PIXEL (screen_y [left.end_vert ]) - y;
    right.count = (int) POS2PIXEL (screen_y [right.end_vert]) - y;
    ycount = min (left.count, right.count);

    left.count  -= ycount;
    right.count -= ycount;

    sx = 0.5 + screen_x [left.start_vert] +
      left.dx_dy * (y - screen_y [left.start_vert]);
    ex = 0.5 + screen_x [right.start_vert] +
      right.dx_dy * (y - screen_y [right.start_vert]);

    for (; ycount > 0; ++y, --ycount)
    {

#     if Z_BUFFER_SPRITES
      float *zbits = Z_buffer + sbox.left + Window_width * y;
#     endif
      byte *texbits;
      int lx = floor (ex); /* "last x" */

      u = (sx - ax) * du_dx - 0.5;
      iv = (int)(v);
      x = floor (sx);
      texbits = &(bmp->bits [bmp->width * iv]);
      bits = &(Camera->framebuffer [x + Window_width * y]);

      for (; x < lx; ++x)
      {

        /*
         * All sprites are masked
         */
        int iu = (int) u;
        byte index = texbits [iu];
#       if Z_BUFFER_SPRITES
        if (index && (*zbits <= zinv))
#       else
        if (index)
#       endif
        {
          /* ff1: adding transparency */
          if (get_trans (RGB_trans [index]) == 0xFF)
          {	/* full color */
            *bits = dt [index];
          }
          else
          {	/* those do not apply to darkness */
            clrCritter = RGB_trans [index];
            hShade = shade_get(get_trans(clrCritter));
            *bits = Pal_rgb_to_pixel(shade_color(hShade, 
                                                 Pal_pixel_to_rgb(*bits),
                                                 clrCritter));
          }

#         if Z_BUFFER_SPRITES
          *zbits = (float)(zinv);
#         endif
        }
        ++bits;
        u += du_dx;

      }

      v += dv_dy;
      sx += left.dx_dy;
      ex += right.dx_dy;

    }

    if (left.count == 0)
    {
      /*
       * Need to recalculate for the left-hand edge
       */
      double dy;
      left.start_vert = left.end_vert;
      left.end_vert   = left.end_vert - 1;
      if (left.end_vert < 0) left.end_vert = s->num_vertices - 1;
      dy = screen_y [left.end_vert] - screen_y [left.start_vert];
      if (dy < 0.) end = true;
      if (dy > A_VERY_SMALL_THING)
      {
        left.dx_dy = (screen_x [left.end_vert] - screen_x [left.start_vert]) /
          dy;
      }
    }
  
    if (right.count == 0)
    {
      /*
       * Ditto right-hand edge
       */
      double dy;
      right.start_vert = right.end_vert;
      right.end_vert   = right.end_vert + 1;
      if (right.end_vert >= s->num_vertices) right.end_vert = 0;
      dy = screen_y [right.end_vert] - screen_y [right.start_vert];
      if (dy < 0.) end = true;
      if (dy > A_VERY_SMALL_THING)
      {
        right.dx_dy =
          (screen_x [right.end_vert] - screen_x [right.start_vert]) / dy;
      }
    }
  
  } while (!end);

}







void Plot_surface (surface *s)
{

  double plane_d;

  struct surface_map sm;

  double ts;

  /*
   * This bit inherited from Tex_plot_obm() : only decals are ever 2-sided
   * It's not necessarily the case that all backface culling has been done at
   *  this stage, we might want to let the hardware handle it
   */
  if (vec_dot (s->normal, s->vertex [0]) > 0.)
  {

    if (s->flags & MSF_2SIDED)
    {
      /*
       * It's facing the wrong way ... reverse it
       */
      vec rev_v [MAX_VERTEX_SURFACE];
      float rev_dark [MAX_VERTEX_SURFACE];
      int i;

      for (i = 0; i < s->num_vertices; ++i)
      {
        rev_v [i] = s->vertex [s->num_vertices - 1 - i];
        rev_dark [i] = s->dark [s->num_vertices - 1 - i];
      }
      for (i = 0; i < s->num_vertices; ++i)
      {
        s->vertex [i] = rev_v    [i];
        s->dark   [i] = rev_dark [i];
      }
    }
    else return; /* backface cull */
  }

  /*
   * Calculate the z gradients (constant over our surface) and rasterise
   */
  plane_d = vec_dot (s->normal, s->vertex [0]);
  if (plane_d == 0.) return;
  sm.dzinv_dy = -s->normal.y / plane_d;
  sm.dzinv_dx =  s->normal.x / plane_d;

  sm.rs  = &Scratch_raster;
  if (s->flags & SF_SHADED)
  {
    if (!scan_convert_gouraud (s, &sm)) return;
  }
  else
  {
    if (!scan_convert_surface (s, &sm)) return;
  }

  if (s->flags & SF_TEXTURED)
  {

    /*
     * Need texture mapping information
     */
    int y = Scratch_raster.line [0].y;
    int tex_width, tex_height;
    ss_bitmap_header *bmp = Tex_get_texture (s->texture);

    assert (bmp != NULL);
    tex_width   = bmp->width;
    tex_height  = bmp->height;

    sm.duz_dy = -(s->u.y + sm.dzinv_dy * vec_dot (s->anchor, s->u));
    sm.duz_dx =  (s->u.x - sm.dzinv_dx * vec_dot (s->anchor, s->u));
    sm.dvz_dy = -(s->v.y + sm.dzinv_dy * vec_dot (s->anchor, s->v));
    sm.dvz_dx =  (s->v.x - sm.dzinv_dx * vec_dot (s->anchor, s->v));
    ts = (256. / tex_width) + 1.e-8;
    sm.duz_dx /= ts;
    sm.duz_dy /= ts;
    ts = (256. / tex_height) + 1.e-8;
    sm.dvz_dx /= ts;
    sm.dvz_dy /= ts;

    sm.ax = (Origin_x + Camera->screen_z * s->anchor.x / s->anchor.z) - 0.5;
    sm.ay =  Origin_y - Camera->screen_z * s->anchor.y / s->anchor.z;

    sm.uz_y = (0.5 + y - sm.ay) * sm.duz_dy;
    sm.vz_y = (0.5 + y - sm.ay) * sm.dvz_dy;

    if (bmp == NULL)
    {
      Cant_happen ("texture.c:plot_surface(): NULL bitmap for texture-map");
    }
    sm.bmp = bmp;

  }
  if (!(s->flags & SF_TEXTURED) && !(s->flags & SF_SHADED) &&
      !(s->flags & SF_TRANSPARENCY))
  {
    /*
     * Colour. For a flat-shaded face the colour passed to the inner function
     *  is the actual pixel value to write to the framebuffer, whereas for a
     *  Gouraud shaded face it is the palette index. Yes, it's nasty.
     */
    int dark = (int) s->dark [0];
    assert (dark >= 0 && dark < DARK_SHADES);
    sm.shade = Dark_table [dark] [s->shade];
  }
  else sm.shade = s->shade;

  /*
   * We've got the gradients, now where do we start?
   * Projecting the anchor point onto the screen is dangerous, and risks a
   *  divide by zero. We can't clip it, as that would defeat the purpose.
   * Project the first point we are about to draw back onto the texture plane
   *  and calculate the starting u, v from there
   */
# if 0
  l = Scratch_raster.line;

  tv.x = l->start_x + 0.5 - Origin_x;
  tv.y = -(y + 0.5)       - Origin_y;
  tv.z = Camera->screen_z;
  vec_cross (tex_normal, s->u, s->v);
  plane_d = vec_dot (s->anchor, tex_normal);
  ts = plane_d / vec_dot (tv, tex_normal);
  vec_multiply (tv, ts);
/*    printf ("%g %g\n", plane_d, vec_dot (tv, tex_normal)); */
  vec_subtract (tv, tv, s->anchor);
  sm.uz_y = vec_dot (tv, s->u);
  sm.vz_y = vec_dot (tv, s->v);
# endif

  Inner_loop [s->flags & 0x1f] (&sm, Camera->framebuffer);

}


g2dbase gbTrueScreen = D_G2D_NULLAREA;
g2dbase *pgbTrueScreen = NULL;

void Plot_overlay(struct tag_hudLayer *pLayer,
									struct tag_camera *pCam)
{
	g2dbase gbBuffer = D_G2D_NULLAREA, 
					gbScreen = D_G2D_NULLAREA;

	gbScreen.bImg.pxl = pCam->framebuffer;
	gbScreen.bBoxArea.left = gbScreen.bBoxArea.top = 0;
	gbScreen.bSizeImg.cx = gbScreen.bBoxArea.right = pCam->window_width;
	gbScreen.bSizeImg.cy = gbScreen.bBoxArea.bottom = pCam->window_height;
	if (BITS_PER_PIXEL < 24)
	{
		gbScreen.nType = G2DPixel;
		/* who's going to free that one? */
		if (!gbTrueScreen.bImg.clr)
			G2D_init_clrbase(&gbTrueScreen, gbScreen.bSizeImg.cx,
																			gbScreen.bSizeImg.cy);
		pgbTrueScreen = &gbTrueScreen;
		G2D_pixel_to_color(pgbTrueScreen, &gbScreen, RGB_trans);
	}
	else
	{
		gbScreen.nType = G2DTrue;
		pgbTrueScreen = &gbScreen;
	}
	while (pLayer)
	{
		if (pLayer->pG2DB)
		{
			pgbTrueScreen->bBoxArea.left = pLayer->ptPos.x;
			pgbTrueScreen->bBoxArea.top = pLayer->ptPos.y;
			pgbTrueScreen->bBoxArea.right = pgbTrueScreen->bBoxArea.left +
																		  pLayer->szSize.cx;
			pgbTrueScreen->bBoxArea.bottom = pgbTrueScreen->bBoxArea.top +
																		   pLayer->szSize.cy;
			if (getRectWidth(pLayer->pG2DB->bBoxArea) == pLayer->szSize.cx &&
					getRectHeight(pLayer->pG2DB->bBoxArea) == pLayer->szSize.cy)
			{
				switch (pLayer->pG2DB->nType)
				{
					case G2DTrue:
					case G2DPal:
					{
						G2Dr_merge_areas(pgbTrueScreen, pLayer->pG2DB);
						break;
					}
					case G2DPixel:
					{
						G2D_init_clrbase(&gbBuffer, 
														 (uint16)getRectWidth(pLayer->pG2DB->bBoxArea),
														 (uint16)getRectHeight(pLayer->pG2DB->bBoxArea));
						G2D_pixel_to_color(&gbBuffer, pLayer->pG2DB, RGB_trans);
						G2D_set_shades(&gbBuffer, get_trans(pLayer->pG2DB->bOp.clr));
						G2Dr_merge_areas(pgbTrueScreen, &gbBuffer);
						G2D_reset_base(&gbBuffer, true);
						break;
					}
					case G2DMask:
					default:
					{
						Cant_happen("Plot_overlay(): Unrenderable Layer");
						break;
					}
				}
			}
			else
			{
				switch (pLayer->pG2DB->nType)
				{
					case G2DPal:
					case G2DTrue:
					{
						G2Dr_stretch_area(pgbTrueScreen, pLayer->pG2DB);
						break;
					}
					case G2DPixel:
					{
						G2D_init_clrbase(&gbBuffer, 
														 (uint16)getRectWidth(pLayer->pG2DB->bBoxArea),
														 (uint16)getRectHeight(pLayer->pG2DB->bBoxArea));
						G2D_pixel_to_color(&gbBuffer, pLayer->pG2DB, RGB_trans);
						G2D_set_shades(&gbBuffer, get_trans(pLayer->pG2DB->bOp.clr));
						G2Dr_stretch_area(pgbTrueScreen, &gbBuffer);
						G2D_reset_base(&gbBuffer, true);
						break;
					}
					case G2DMask:
					default:
					{
						Cant_happen("Plot_overlay(): Unrenderable Layer");
						break;
					}
				}
			}
		}
		resetArea(*pgbTrueScreen);
		pLayer = pLayer->pNext;
	}
	if (BITS_PER_PIXEL < 24)
		G2D_color_to_pixel(&gbScreen, pgbTrueScreen, RGB_trans);
}


/*
 * Rasterise! Let no scanlines on screen get pasteurise!
 *  [ OK, that's enough Tom Lehrer parody for one function ... ed ]
 * Convert a surface s, whose view/screen coordinates we have prepared earlier
 *  [ That's enough Blue Peter parody for one function as well ]
 *  and whose (constant) z gradients are given in dzinv_dx and dzinv_dy, to a 
 *  raster_surface raster
 * This function assumes that the surface is convex and vertices are in
 *  clockwise order (which should always be the case) and that the passed-in
 *  raster_surface has enough free scanline slots to accommodate the surface
 *
 * At some point in the next version or so, the 2 scan_convert functions will
 *  be unified, handling Gouraud shading or not depending on the surface flags.
 *  The inner loop functions should properly handle a raster surface without
 *  contiguous y, so blank lines and (possibly) boundaries between two copies
 *  of the texture can be dealt with.
 */
bool scan_convert_surface (const surface *s, struct surface_map *sm)
{

  raster_surface *rs = sm->rs;
  double dzinv_dx = sm->dzinv_dx;
  double dzinv_dy = sm->dzinv_dy;

  int top_vertex = 0;
  double top_y =  1.e10;
  int top_scanline;
  int bot_vertex = 0;
  double bot_y = -1.e10;
  int bot_scanline;
  int next_vertex;
  int next_y;

  int i;
  int y;
  double x;
  double dx_dy;
  double delta_y;

  double screen_x [MAX_VERTEX_SURFACE];
  double screen_y [MAX_VERTEX_SURFACE];
  int scanline [MAX_VERTEX_SURFACE];

  /*
   * 1/Z coordinate. Due to the perspective correction this is directly
   *  proportional to the screen X and Y coordinates. The proof is left as an
   *  exercise for the reader 8-) actually there's a very good explanation of
   *  it at some website I can't remember the URL of but I'll put in soon
   */
  double zinv;

  int line;
  int oline;

  /*
   * Firstly, we need screen coordinates for each vertex of the surface
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    screen_x [i] = Origin_x +
      Camera->screen_z * s->vertex [i].x / s->vertex [i].z;
    screen_y [i] = Origin_y - 
      Camera->screen_z * s->vertex [i].y / s->vertex [i].z;
    scanline [i] = (int)(POS2PIXEL (screen_y [i]));
  }


  /*
   * Now sort the vertices by height
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    if (screen_y [i] < top_y)
    {
      top_y = screen_y [i];
      top_vertex = i;
    }
    if (screen_y [i] > bot_y)
    {
      bot_y = screen_y [i];
      bot_vertex = i;
    }
  }
  top_scanline = scanline [top_vertex];
  bot_scanline = scanline [bot_vertex];

  /*
   * Because the vertices are specified clockwise and the surface is convex
   *  in 2D, we can divide the edges into left- (bot_vertex -> top_vertex)
   *  and right-hand (top_vertex -> bot_vertex) and handle them separately
   */
  y = top_scanline;

  /*
   * Left-hand vertices
   */
  line = 0;
  for (i = top_vertex; i != bot_vertex; i = next_vertex)
  {

    /*
     * Stuff done every left-hand edge
     */
    next_vertex = (i + s->num_vertices - 1) % s->num_vertices;
    if (screen_y [next_vertex] == screen_y [i]) continue;

    next_y = scanline [next_vertex];
    dx_dy   = (screen_x [next_vertex] - screen_x [i]) /
              (screen_y [next_vertex] - screen_y [i]);
    delta_y = 0.5 + y - screen_y [i];
    x       = 0.5 + screen_x [i] + delta_y * dx_dy;

    for (; y < next_y; ++y)
    {
      if (line >= Window_height)
      {
/*  	Cant_happen ("scan_convert(): too many scanlines"); */
        /*
         * An infinitely thin surface might just have slipped through the
         *  clipping
         */
        return false;
      }
      rs->line [line  ].y       = y;
      rs->line [line++].start_x = x;
      x += dx_dy;
    }

  }

  /*
   * Right-hand vertices. This is exactly the same except we step round the
   *  surface in the opposite direction
   */
  line = 0;
  y = top_scanline;
  for (i = top_vertex; i != bot_vertex; i = next_vertex)
  {

    /*
     * Stuff done every left-hand edge
     */
    next_vertex = (i + 1) % s->num_vertices;
    if (screen_y [next_vertex] == screen_y [i]) continue;

    next_y = scanline [next_vertex];
    dx_dy   = (screen_x [next_vertex] - screen_x [i]) /
              (screen_y [next_vertex] - screen_y [i]);
    delta_y = 0.5 + y - screen_y [i];
    x       = 0.5 + screen_x [i] + delta_y * dx_dy;

    for (; y < next_y; ++y)
    {
      if (line >= Window_height)
      {
/*  	Cant_happen ("scan_convert(): too many scanlines"); */
        /*
         * An infinitely thin surface might just have slipped through the
         *  clipping
         */
        return false;
      }
      rs->line [line++].end_x = x;
      x += dx_dy;
    }

  }

  /*
   * Cleanup and inverse-z values.
   * First we find the real top scanline to draw; this is the first line for
   *  which end_x is greater than start_x
   */
  for (oline = 0; oline < (bot_scanline - top_scanline); ++oline)
  {
    if (rs->line [oline].end_x > rs->line [oline].start_x)
    {
      break;
    }
  }

  top_scanline += oline;

  delta_y = 0.5 + top_scanline - screen_y [top_vertex];
  zinv    = (Camera->screen_z / s->vertex [top_vertex].z) + delta_y * dzinv_dy;
  for (line = 0, y = top_scanline; y < bot_scanline; ++y)
  {
    /*
     * Unfortunately we can't do this; a surface with a very sharp top corner
     *  might have one or more isolated pixels separated by blank lines, which
     *  would cause the rest of the surface to be rejected. That's pixels for
     *  you ...
     * FIXME Now that we aren't using contiguous y any more, we can ...
     */
/*      if (rs->line [oline].end_x <= rs->line [oline].start_x) */
/*      { */
/*        break; */
/*      } */
    rs->line [line].y       = rs->line [oline].y;
    rs->line [line].start_x = rs->line [oline].start_x;
    rs->line [line].end_x   = rs->line [oline].end_x;  
    rs->line [line].zinv    = zinv +
      (0.5 + rs->line [line].start_x - screen_x [top_vertex]) * dzinv_dx;
    zinv += dzinv_dy;
    ++line;
    ++oline;
  }

  assert (line == y - top_scanline); /* temporary */
  rs->num_lines = line;
  return true;

}





#if 0
bool scan_convert_gouraud (surface *s, struct surface_map *sm)
{

  raster_surface *rs = sm->rs;
  double dzinv_dx = sm->dzinv_dx;
  double dzinv_dy = sm->dzinv_dy;

  int top_vertex = 0;
  double top_y =  1.e10;
  int top_scanline;
  int bot_vertex = 0;
  double bot_y = -1.e10;
  int bot_scanline;

  int i;
  int y, next_y;
  double delta_y, deltar_y, deltal_y;
  float start_dark, end_dark;
  float deltal_dark, deltar_dark;

  double screen_x [MAX_VERTEX_SURFACE];
  double screen_y [MAX_VERTEX_SURFACE];
  int scanline [MAX_VERTEX_SURFACE];

  /*
   * 1/Z coordinate. Due to the perspective correction this is directly
   *  proportional to the screen X and Y coordinates. The proof is left as an
   *  exercise for the reader 8-) actually there's a very good explanation of
   *  it at some website I can't remember the URL of but I'll put in soon
   */
  double zinv;
  double dlx_dy, drx_dy;
  double start_x, end_x;

  int line;
  int left, right, next_left, next_right;

  /*
   * Firstly, we need screen coordinates for each vertex of the surface
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    screen_x [i] = Origin_x +
      Camera->screen_z * s->vertex [i].x / s->vertex [i].z;
    screen_y [i] = Origin_y -
      Camera->screen_z * s->vertex [i].y / s->vertex [i].z;
    scanline [i] = (int)(POS2PIXEL (screen_y [i]));

    /*
     * Lighting paranoia
     */
    assert (s->dark [i] >= 0 && s->dark [i] < DARK_SHADES);
  }


  /*
   * Now sort the vertices by height
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    if (screen_y [i] < top_y)
    {
      top_y = screen_y [i];
      top_vertex = i;
    }
    if (screen_y [i] > bot_y)
    {
      bot_y = screen_y [i];
      bot_vertex = i;
    }
  }
  if (top_vertex == bot_vertex) return false; /* nothing to do! */
  top_scanline = scanline [top_vertex];
  bot_scanline = scanline [bot_vertex];

  y = top_scanline;

  line = 0;
  delta_y = 0.5 + top_scanline - screen_y [top_vertex];
  zinv    = (Camera->screen_z / s->vertex [top_vertex].z) + delta_y * dzinv_dy;
  left = top_vertex;
  right = top_vertex;

  next_y = y;
  next_left = left;
  next_right = right;
  do
  {

    /*
     * Sort out the next vertices for each side
     */
    while (next_y > y)
    {

      if (line >= Window_height)
      {
        /*  	Cant_happen ("scan_convert(): too many scanlines"); */
        /*
         * An infinitely thin surface might just have slipped through the
         *  clipping
         * 02-Apr-2002 But it shouldn't have, surely? Earlier handling of
         *  special cases wanted! -jim
         */
        return false;
      }
      rs->line [line].y          = y;
      rs->line [line].start_dark = start_dark;
      rs->line [line].end_dark   = end_dark;
      rs->line [line].start_x    = floor (start_x);
      rs->line [line].end_x      = floor (end_x);
      rs->line [line].zinv       = zinv +
        (0.5 + rs->line [line].start_x - screen_x [top_vertex]) * dzinv_dx;
      ++line;
      start_x += dlx_dy;
      end_x   += drx_dy;
      start_dark += deltal_dark;
      end_dark   += deltar_dark;
      zinv += dzinv_dy;
      ++y;

    }
    
    if (y == scanline [next_left ])
    {
      left  = next_left;
      next_left  = (left < 1) ? s->num_vertices - 1 : left - 1;
      dlx_dy   = (screen_x [next_left ] - screen_x [left ]) /
                 (screen_y [next_left ] - screen_y [left ]);
      deltal_y = 0.5 + y - screen_y [left ];
      start_x  = 0.5 + screen_x [left ] + deltal_y * dlx_dy;
      deltal_dark = (s->dark [next_left] - s->dark [left]) /
        (screen_y [next_left] - screen_y [left]);
      start_dark = s->dark [left];
    }

    if (y == scanline [next_right])
    {
      right = next_right;
      next_right = (right >= s->num_vertices - 1) ? 0 : right + 1;
      drx_dy   = (screen_x [next_right] - screen_x [right]) /
                 (screen_y [next_right] - screen_y [right]);
      deltar_y = 0.5 + y - screen_y [right];
      end_x    = 0.5 + screen_x [right] + deltar_y * drx_dy;
      deltar_dark = (s->dark [next_right] - s->dark [right]) /
        (screen_y [next_right] - screen_y [right]);
      end_dark = s->dark [right];
    }

    next_y = min (scanline [next_left], scanline [next_right]);

  } while (left != bot_vertex);
  assert (scanline [right] == scanline [bot_vertex]);

  assert (line == bot_scanline - top_scanline); /* temporary */
  rs->num_lines = line;
  return true;

}
#else
bool scan_convert_gouraud (surface *s, struct surface_map *sm)
{

  raster_surface *rs = sm->rs;
  double dzinv_dx = sm->dzinv_dx;
  double dzinv_dy = sm->dzinv_dy;

  int top_vertex = 0;
  double top_y =  1.e10;
  int top_scanline;
  int bot_vertex = 0;
  double bot_y = -1.e10;
  int bot_scanline;
  int next_vertex;
  int next_y;

  int i;
  int y;
  double x;
  double dx_dy;
  double delta_y;
  float dark;
  float delta_dark;

  double screen_x [MAX_VERTEX_SURFACE];
  double screen_y [MAX_VERTEX_SURFACE];
  int scanline [MAX_VERTEX_SURFACE];

  /*
   * 1/Z coordinate. Due to the perspective correction this is directly
   *  proportional to the screen X and Y coordinates. The proof is left as an
   *  exercise for the reader 8-) actually there's a very good explanation of
   *  it at some website I can't remember the URL of but I'll put in soon
   */
  double zinv;

  int line;

  /*
   * Firstly, we need screen coordinates for each vertex of the surface
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    screen_x [i] = Origin_x +
      Camera->screen_z * s->vertex [i].x / s->vertex [i].z;
    screen_y [i] = Origin_y -
      Camera->screen_z * s->vertex [i].y / s->vertex [i].z;
    scanline [i] = (int)(POS2PIXEL (screen_y [i]));

    /*
     * Lighting paranoia
     */
    assert (s->dark [i] >= 0 && s->dark [i] < DARK_SHADES);
  }


  /*
   * Now sort the vertices by height
   */
  for (i = 0; i < s->num_vertices; ++i)
  {
    if (screen_y [i] < top_y)
    {
      top_y = screen_y [i];
      top_vertex = i;
    }
    if (screen_y [i] > bot_y)
    {
      bot_y = screen_y [i];
      bot_vertex = i;
    }
  }
  top_scanline = scanline [top_vertex];
  bot_scanline = scanline [bot_vertex];

  /*
   * Because the vertices are specified clockwise and the surface is convex
   *  in 2D, we can divide the edges into left- (bot_vertex -> top_vertex)
   *  and right-hand (top_vertex -> bot_vertex) and handle them separately
   */
  y = top_scanline;

  /*
   * Left-hand vertices
   */
  line = 0;
  delta_y = 0.5 + top_scanline - screen_y [top_vertex];
  zinv    = (Camera->screen_z / s->vertex [top_vertex].z) + delta_y * dzinv_dy;
  for (i = top_vertex; i != bot_vertex; i = next_vertex)
  {

    /*
     * Stuff done every left-hand edge
     */
/*      next_vertex = (i + s->num_vertices - 1) % s->num_vertices; */
    if (i < 1) next_vertex = s->num_vertices - 1; else next_vertex = i - 1;
    if (screen_y [next_vertex] == screen_y [i]) continue;

    next_y = scanline [next_vertex];
    if (next_y == y) continue;
    dx_dy   = (screen_x [next_vertex] - screen_x [i]) /
              (screen_y [next_vertex] - screen_y [i]);
    delta_y = 0.5 + y - screen_y [i];
    x       = 0.5 + screen_x [i] + delta_y * dx_dy;
    delta_dark = (s->dark [next_vertex] - s->dark [i]) / (next_y - y);
    dark = s->dark [i];

    for (; y < next_y; ++y)
    {
      if (line >= Window_height)
      {
/*  	Cant_happen ("scan_convert(): too many scanlines"); */
        /*
         * An infinitely thin surface might just have slipped through the
         *  clipping
         * 02-Apr-2002 But it shouldn't have, surely? Earlier handling of
         *  special cases wanted! -jim
         */
        return false;
      }
      rs->line [line].y          = y;
      rs->line [line].start_dark = dark;
      rs->line [line].start_x    = (int)(x);
      rs->line [line].zinv       = zinv +
        (0.5 + rs->line [line].start_x - screen_x [top_vertex]) * dzinv_dx;
      ++line;
      x += dx_dy;
      dark += delta_dark;
      zinv += dzinv_dy;
    }

  }

  /*
   * Right-hand vertices. This is exactly the same except we step round the
   *  surface in the opposite direction
   */
  line = 0;
  y = top_scanline;
  for (i = top_vertex; i != bot_vertex; i = next_vertex)
  {

    /*
     * Stuff done every left-hand edge
     */
    next_vertex = (i + 1) % s->num_vertices;
    if (screen_y [next_vertex] == screen_y [i]) continue;

    next_y = scanline [next_vertex];
    if (next_y == y) continue;
    dx_dy   = (screen_x [next_vertex] - screen_x [i]) /
              (screen_y [next_vertex] - screen_y [i]);
    delta_y = 0.5 + y - screen_y [i];
    x       = 0.5 + screen_x [i] + delta_y * dx_dy;
    delta_dark = (s->dark [next_vertex] - s->dark [i]) / (next_y - y);
    dark = s->dark [i];

    for (; y < next_y; ++y)
    {
      if (line >= Window_height)
      {
/*  	Cant_happen ("scan_convert(): too many scanlines"); */
        /*
         * An infinitely thin surface might just have slipped through the
         *  clipping
         */
        return false;
      }
      rs->line [line].end_dark = dark;
      rs->line [line++].end_x = (int)(x);
      x += dx_dy;
      dark += delta_dark;
    }

  }

  assert (line == bot_scanline - top_scanline); /* temporary */
  rs->num_lines = line;
  return true;

}
#endif



void plot_inner_textured (struct surface_map *sm, pixel *bitmap)
{

  raster_surface *rs = sm->rs;
  int ycount = rs->num_lines;
  struct raster_line *l = rs->line;
  double uz_y = sm->uz_y;
  double vz_y = sm->vz_y;
  byte *tex_bits   = sm->bmp->bits;
  int   tex_width  = sm->bmp->width;
  int   tex_height = sm->bmp->height;

  for (; ycount > 0; --ycount, ++l)
  {

    double zinv = l->zinv;
    int x = l->start_x;
    double uz = uz_y + (0.5 + x - sm->ax) * sm->duz_dx;
    double vz = vz_y + (0.5 + x - sm->ax) * sm->dvz_dx;

    pixel *bmp = &(bitmap[x + Window_width * l->y]);

#   if TEXTURE_PARANOID
    if ((l->start_x < 0) || (l->end_x > Window_width) ||
        (l->y < 0) || (l->y > Window_height))
    {
      Cant_happen ("Tex_plot_projected(): trying to draw off screen");
    }
#   endif

    for (; x < l->end_x; ++x)
    {

      double z = 1. / zinv;
      int u = (int)(uz * z);
      int v = (int)(vz * z);

      v %= tex_height;

      *bmp++ = Pal_index_to_pixel (tex_bits [u + tex_width * v]);
      uz += sm->duz_dx;
      vz += sm->dvz_dx;
      zinv += sm->dzinv_dx;

    }

    uz_y += sm->duz_dy;
    vz_y += sm->dvz_dy;

  }

}


void plot_inner_txt_shd (struct surface_map *sm, pixel *bitmap)
{

  raster_surface *rs = sm->rs;
  int ycount = rs->num_lines;
  struct raster_line *l = rs->line;
  double uz_y = sm->uz_y;
  double vz_y = sm->vz_y;
  byte *tex_bits   = sm->bmp->bits;
  int   tex_width  = sm->bmp->width;
  int   tex_height = sm->bmp->height;

  for (; ycount > 0; --ycount, ++l)
  {

    double zinv = l->zinv;
    int x = l->start_x;
    int shade = 0;
    int dark, delta_dark;
    double uz, vz;
    pixel *bmp;
    pixel *dt = &(Dark_table [0] [0]);

    if (x != l->end_x)
    {
      dark = (int)(65536 * l->start_dark);
      uz = uz_y + (x - sm->ax) * sm->duz_dx;
      vz = vz_y + (x - sm->ax) * sm->dvz_dx;
      delta_dark = (int)(65536 * (l->end_dark - l->start_dark) /
			 (l->end_x - x));
      bmp = &(bitmap[x + Window_width * l->y]);
    }

#   if TEXTURE_PARANOID
    if ((l->start_x < 0) || (l->end_x > Window_width) ||
        (l->y < 0) || (l->y > Window_height))
    {
      Cant_happen ("Tex_plot_projected(): trying to draw off screen");
    }
#   endif

    for (; x < l->end_x; ++x)
    {

      float z = (float)(1. / zinv);
      int u = (int)(uz * z);
      int v = (int)(vz * z);

      v %= tex_height;

/*        if (shade != (dark >> 16)) */
/*        {  */
/*          shade = dark >> 16; */
/*          dt = &(Dark_table [shade] [0]); */
/*        } */
      *bmp++ = Dark_table [dark >> 16] [tex_bits [u + tex_width * v]];
      dark += delta_dark;

      uz += sm->duz_dx;
      vz += sm->dvz_dx;
      zinv += sm->dzinv_dx;

    }

    uz_y += sm->duz_dy;
    vz_y += sm->dvz_dy;

  }

}


void plot_inner_txt_msk_shd (struct surface_map *sm, pixel *bitmap)
{

  raster_surface *rs = sm->rs;
  int ycount = rs->num_lines;
  struct raster_line *l = rs->line;
  double uz_y = sm->uz_y;
  double vz_y = sm->vz_y;
  byte *tex_bits   = sm->bmp->bits;
  int   tex_width  = sm->bmp->width;
  int   tex_height = sm->bmp->height;

  for (; ycount > 0; --ycount, ++l)
  {

    double zinv = l->zinv;
    int x = l->start_x;
    int shade = 0;
    int dark, delta_dark;
    double uz, vz;
    pixel *bmp;
    pixel *dt = &(Dark_table [0] [0]);

    if (x != l->end_x)
    {
      dark = (int)(65536 * l->start_dark);
      uz = uz_y + (x - sm->ax) * sm->duz_dx;
      vz = vz_y + (x - sm->ax) * sm->dvz_dx;
      delta_dark = (int)(65536 * (l->end_dark - l->start_dark) /
			 (l->end_x - x));
      bmp = &(bitmap[x + Window_width * l->y]);
    }

#   if TEXTURE_PARANOID
    if ((l->start_x < 0) || (l->end_x > Window_width) ||
        (l->y < 0) || (l->y > Window_height))
    {
      Cant_happen ("Tex_plot_projected(): trying to draw off screen");
    }
#   endif

    for (; x < l->end_x; ++x)
    {

      float z = (float)(1. / zinv);
      int u = (int)(uz * z);
      int v = (int)(vz * z);

      v %= tex_height;

      if (tex_bits [u + tex_width * v])
      {
        if (shade != (dark >> 16))
        {
          shade = dark >> 16;
          dt = &(Dark_table [shade] [0]);
        }
        *bmp = dt [tex_bits [u + tex_width * v]];
      }
      ++bmp;
      dark += delta_dark;

      uz += sm->duz_dx;
      vz += sm->dvz_dx;
      zinv += sm->dzinv_dx;

    }

    uz_y += sm->duz_dy;
    vz_y += sm->dvz_dy;

  }

}


void plot_inner_zbuffer (struct surface_map *sm, pixel *bitmap)
{

  raster_surface *rs = sm->rs;
  int ycount = rs->num_lines;
  struct raster_line *l = rs->line;
  double dzinv_dx = sm->dzinv_dx;
  pixel col = sm->shade;

  for (; ycount > 0; --ycount, ++l)
  {

    int x = l->start_x;
    double zinv = l->zinv;

    pixel *bits  = &(bitmap[x + Window_width * l->y]);
    float *zbits = Z_buffer + x + Window_width * l->y;

#   if TEXTURE_PARANOID
    if ((l->start_x < 0) || (l->end_x > Window_width) ||
        (l->y < 0) || (l->y > Window_height))
    {
      Cant_happen ("Tex_plot_projected(): trying to draw off screen");
    }
#   endif

    for (; x < l->end_x; ++x)
    {

      if (*zbits <= zinv)
      {
        *zbits = (float)(zinv);
        *bits  = col;
      }
      ++bits;
      ++zbits;
      zinv += dzinv_dx;

    }

  }

}


void plot_inner_txt_zbf (struct surface_map *sm, pixel *bitmap)
{

  raster_surface *rs = sm->rs;
  int ycount = rs->num_lines;
  struct raster_line *l = rs->line;
  double uz_y = sm->uz_y;
  double vz_y = sm->vz_y;
  byte *tex_bits   = sm->bmp->bits;
  int   tex_width  = sm->bmp->width;
  int   tex_height = sm->bmp->height;
  pixel *dt = &(Dark_table [sm->shade] [0]);  /* shade for this dark value */

  for (; ycount > 0; --ycount, ++l)
  {

    double zinv = l->zinv;
    int x = l->start_x;
    double uz = uz_y + (0.5 + x - sm->ax) * sm->duz_dx;
    double vz = vz_y + (0.5 + x - sm->ax) * sm->dvz_dx;
    pixel *bits  = &(bitmap[x + Window_width * l->y]);
    float *zbits = Z_buffer + x + Window_width * l->y;

#   if TEXTURE_PARANOID
    if ((l->start_x < 0) || (l->end_x > Window_width) ||
        (l->y < 0) || (l->y > Window_height))
    {
      Cant_happen ("Tex_plot_projected(): trying to draw off screen");
    }
#   endif

    /*
     * FIXME Should clean-up zero length scanlines in scan_convert_gouraud()
     */
    if (l->end_x > l->start_x)
    {
      assert (zinv > A_VERY_SMALL_THING);
/*      if (uz < -1000.) */
/*      { */
/*        printf ("%g %g\n", uz, vz); */
/*      } */
/*      for (; x < l->end_x; ++x) */
      while (1)
      {

        double z = 1. / zinv;
        int u = (int)(uz * z);
        int v = (int)(vz * z);

        /*
         * Temporary (I hope)
         */
        u %= tex_width;
        v %= tex_height;
/*        u &= tex_width - 1; */
/*        v &= tex_height - 1; */

        if (*zbits <= zinv)
        {
          *bits = dt [tex_bits [u + tex_width * v]];
          *zbits = (float)(zinv);
        }

        if (++x >= l->end_x) break;
        ++bits;
        ++zbits;
        uz += sm->duz_dx;
        vz += sm->dvz_dx;
        zinv += sm->dzinv_dx;

      }

      assert (zinv > A_VERY_SMALL_THING);

    }
    uz_y += sm->duz_dy;
    vz_y += sm->dvz_dy;

  }

}



void plot_inner_zbf_shd (struct surface_map *sm, pixel *bitmap)
{

  raster_surface *rs = sm->rs;
  int ycount = rs->num_lines;
  struct raster_line *l = rs->line;
  double dzinv_dx = sm->dzinv_dx;
  int col = sm->shade;

  for (; ycount > 0; --ycount, ++l)
  {

    double zinv = l->zinv;
    int x = l->start_x;
    int shade = 0;
    int dark, delta_dark;
    pixel *bits;
    float *zbits;
    pixel *dt = &(Dark_table [0] [0]);

    if (x != l->end_x)
    {
      dark = (int)(65536 * l->start_dark);
      delta_dark = (int)(65536 * (l->end_dark - l->start_dark) /
			 (l->end_x - x));
      bits = &(bitmap [x + Window_width * l->y]);
      zbits = Z_buffer + x + Window_width * l->y;
    }

#   if TEXTURE_PARANOID
    if ((l->start_x < 0) || (l->end_x > Window_width) ||
        (l->y < 0) || (l->y > Window_height))
    {
      Cant_happen ("Tex_plot_projected(): trying to draw off screen");
    }
#   endif

    for (; x < l->end_x; ++x)
    {

      if (shade != (dark >> 16))
      {
        shade = dark >> 16;
        dt = &(Dark_table [shade] [0]);
      }

      if (*zbits <= zinv)
      {
        *zbits = (float)(zinv);
        *bits  = dt [col];
      }
      ++bits;
      ++zbits;
      dark += delta_dark;
      zinv += dzinv_dx;

    }

  }

}




void plot_inner_transparency (struct surface_map *sm, pixel *bitmap)
{

  raster_surface *rs = sm->rs;
  int ycount = rs->num_lines;
  struct raster_line *l = rs->line;
  pcolor clr = RGB_trans [sm->shade]; /* assumes it IS a transparency */
  HSHADE sh = shade_get(get_trans(clr));

  for (; ycount > 0; --ycount, ++l)
  {

    int x = l->start_x;
    pixel *bits = &(bitmap [x + Window_width * l->y]);

#   if TEXTURE_PARANOID
    if ((l->start_x < 0) || (l->end_x > Window_width) ||
        (l->y < 0) || (l->y > Window_height))
    {
      Cant_happen ("Tex_plot_projected(): trying to draw off screen");
    }
#   endif

    for (; x < l->end_x; ++x)
    {

#if BITS_PER_PIXEL < 32
      *bits = Pal_rgb_to_pixel (shade_color (sh, Pal_pixel_to_rgb (*bits),
					     clr));
#else
      *bits = shade_color (sh, *bits, clr);
#endif
      ++bits;

    }

  }

}

void plot_inner_notimp (struct surface_map *sm, pixel *bitmap)
{
  /* unused */  sm = sm, bitmap = bitmap;
  Cant_happen ("texture.c:plot_inner(): flag combination not supported");
}




/*
 * End of src/plot.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
