/*
 * quadtree.c -
 *  (2D) map partitioning by means of a quadtree
 */

/*
 * File:    src/quadtree.c
 * Purpose: Quadtrees, building and navigating
 * Prefix:  Quad_
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2002 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>

#include "clip.h"
#include "main.h"
#include "map.h"
#include "quadtree.h"


/*
 * Debug
 */
#ifdef __DEBUG__
#define QUADTREE_MESSAGES 0
#endif


/*
 * Global variables
 */


/*
 * The root of the level quadtree
 */
quad_node *Quad_root;



static double Max_quad_search_distance = 4096.;



/*
 * Local function prototypes
 */
static void get_node_bounds (quad_node *node, double *top, double *bottom);
static void partition_node (quad_node *node);
static bool is_node_empty (quad_node *node);
static quad_node *new_node (int xmin, int ymin, int xmax, int ymax);
static void free_node (quad_node *node);




void Quad_init (void)
{
}



void Quad_exit (void)
{
  free_node (Quad_root);
}






/*
 * Should be called after the map is set up (it uses Map_info), to initialise
 *  the level quadtree
 */
void Quad_init_level (void)
{

  double zmax = -DBL_MAX, zmin = DBL_MAX;

# if QUADTREE_MESSAGES
  printf ("Quad_init_level(): creating level quadtree\n");
# endif

  /*
   * The root node for the level contains the entire level; tree bounding
   *  volumes haven't been set up at this stage but we know where the split
   *  planes should be
   */
  Quad_root = new_node (0, 0, Map_info.xsize, Map_info.ysize);

  /*
   * Partition the root node and its children
   */
  partition_node (Quad_root);

  /*
   * Get the bounding volumes for the root node and its children
   */
  get_node_bounds (Quad_root, &zmax, &zmin);

}



/*
 * Navigate a quadtree in front-to-back order, calling the callback function
 *  for each non-empty node within the clipping region
 */
void Quad_navigate (quad_node *node, const vec *refpoint, 
		    const clipregion_t *region, quad_callback callback,
		    bool noclip)
{

  vec tv;

  assert (node != NULL);

  if ((node->child [0] == NULL) && (node->tile == NULL))
  {
    /*
     * Node is empty, reject
     */
    return;
  }

  vec_subtract (tv, node->centre, *refpoint);
  if (vec_magnitude2 (tv) > (node->radius + Max_quad_search_distance) *
      (node->radius + Max_quad_search_distance)) return;

  /*
   * This is taken directly from the BSP code
   */
  if (!noclip)
  {
    switch (Clip_ellipsoid_z_by_region (&(node->centre), node->radius,
					node->z_scale, region))
    {
    case CLIP_WITHOUT:
      return;           /* outside clipping volume */
    case CLIP_WITHIN:
      noclip = true;    /* aha!                    */
      break;
    case CLIP_INTERSECT:
      break;            /* need further clipping   */
    default:
      Cant_happen ("Quad_navigate(): illegal clip status");
    }
  }

  if (node->child [0] == NULL)
  {
    /*
     * Leaf node
     */
    callback (node);
    return;
  }

  /*
   * Recurse down the tree depending on what quadrant the reference point is
   *  in
   */
  if (refpoint->x < node->centre.x)
  {
    if (refpoint->y < node->centre.y)
    {
      /*
       * Top-left, quadrant 0
       */
      Quad_navigate (node->child [0], refpoint, region, callback, noclip);
      Quad_navigate (node->child [1], refpoint, region, callback, noclip);
      Quad_navigate (node->child [3], refpoint, region, callback, noclip);
      Quad_navigate (node->child [2], refpoint, region, callback, noclip);
    }
    else
    {
      /*
       * Bottom left, quadrant 3
       */
      Quad_navigate (node->child [3], refpoint, region, callback, noclip);
      Quad_navigate (node->child [0], refpoint, region, callback, noclip);
      Quad_navigate (node->child [2], refpoint, region, callback, noclip);
      Quad_navigate (node->child [1], refpoint, region, callback, noclip);
    }
  }
  else
  {
    if (refpoint->y < node->centre.y)
    {
      /*
       * Top right, quadrant 1
       */
      Quad_navigate (node->child [1], refpoint, region, callback, noclip);
      Quad_navigate (node->child [0], refpoint, region, callback, noclip);
      Quad_navigate (node->child [2], refpoint, region, callback, noclip);
      Quad_navigate (node->child [3], refpoint, region, callback, noclip);
    }
    else
    {
      /*
       * Bottom right, quadrant 2
       */
      Quad_navigate (node->child [2], refpoint, region, callback, noclip);
      Quad_navigate (node->child [1], refpoint, region, callback, noclip);
      Quad_navigate (node->child [3], refpoint, region, callback, noclip);
      Quad_navigate (node->child [0], refpoint, region, callback, noclip);
    }
  }

}



void get_node_bounds (quad_node *node, double *top, double *bottom)
{

  double zmax = -DBL_MAX, zmin = DBL_MAX;
  double node_width;

  assert (node != NULL);

  if (node->child [0])
  {
    /*
     * Node has children: recurse
     */
    get_node_bounds (node->child [0], &zmax, &zmin);
    get_node_bounds (node->child [1], &zmax, &zmin);
    get_node_bounds (node->child [2], &zmax, &zmin);
    get_node_bounds (node->child [3], &zmax, &zmin);
  }
  else if (node->tile)
  {
    /*
     * Node contains a tile: get its extent from the map
     */
    Map_tile_max_height (node->tile, &zmax, &zmin, false);
  }
  else
  {
    /*
     * Node is empty: nothing is ever going to be drawn in it, so set its
     *  extent to zero and return
     */
    node->centre.z = 0.;
    node->radius   = 0.;
    node->z_scale  = 0.;
    return;
  }

  /*
   * Set the extents from those we have collected
   */
  node_width = TILE_GRID_SIZE * (node->xmax - node->xmin);
  node->centre.z = zmin + (zmax - zmin) / 2.;
  node->radius   = 0.5 * sqrt (3 * node_width * node_width);
  node->z_scale  = (zmax - zmin) / node_width;

# if QUADTREE_MESSAGES
  printf ("Quadnode [%d %d %d %d] (%g, %g, %g) rad %g scale %g\n",
	  node->xmin, node->ymin, node->xmax, node->ymax,
	  node->centre.x, node->centre.y, node->centre.z,
	  node->radius, node->z_scale);
# endif

  /*
   * Propagate the maximum extents up to the next level
   */
  if (zmax > *top)    *top    = zmax;
  if (zmin < *bottom) *bottom = zmin;

}



/*
 * Split a node into 4 children, if it is not empty and it is larger than a
 *  tile, otherwise make it into a leaf
 */
void partition_node (quad_node *node)
{

  int node_width  = node->xmax - node->xmin;
  int node_height = node->ymax - node->ymin;

  assert (node_width == node_height); /* not supporting non-square levels */
  assert (node_width > 0);

# if QUADTREE_MESSAGES
  printf ("partition_node(): [%d %d %d %d] ", node->xmin, node->xmax,
	  node->ymin, node->ymax);
# endif

  if (is_node_empty (node))
  {
    /*
     * The node contains no drawable tiles, set it to be a leaf but without an
     *  associated tile
     */
#   if QUADTREE_MESSAGES
    printf ("empty\n");
#   endif
    node->tile = NULL;
  }
  else if (node_width == 1)
  {
    /*
     * We've got down to the level of a single tile, so set leaf parameters
     */
#   if QUADTREE_MESSAGES
    printf ("leaf\n");
#   endif
    node->tile = &(Tile [node->ymin * Map_info.xsize + node->xmin]);
    node->centre.x = TILE_GRID_SIZE * node->xmin + TILE_GRID_SIZE / 2;
    node->centre.y = TILE_GRID_SIZE * node->ymin + TILE_GRID_SIZE / 2;
  }
  else
  {
    /*
     * The node needs to be partitioned into 4 children
     */
    int centre_x = node->xmin + node_width  / 2;
    int centre_y = node->ymin + node_height / 2;
#   if QUADTREE_MESSAGES
    printf ("splitting ...\n");
#   endif
    node->child [0] = new_node (node->xmin, node->ymin, centre_x, centre_y);
    node->child [1] = new_node (centre_x, node->ymin, node->xmax, centre_y);
    node->child [2] = new_node (centre_x, centre_y, node->xmax, node->ymax);
    node->child [3] = new_node (node->xmin, centre_y, centre_x, node->ymax);
    node->centre.x = centre_x * TILE_GRID_SIZE;
    node->centre.y = centre_y * TILE_GRID_SIZE;

    /*
     * And ... recurse
     */
    partition_node (node->child [0]);
    partition_node (node->child [1]);
    partition_node (node->child [2]);
    partition_node (node->child [3]);
  }

}




/*
 * Find out if a node is empty i.e. contains no tiles that are not SOLID
 */
bool is_node_empty (quad_node *node)
{

  int x, y;
  for (x = node->xmin; x < node->xmax; ++x)
  {
    for (y = node->ymin; y < node->ymax; ++y)
    {
      if (Tile [x + Map_info.xsize * y].type != TILE_SOLID) return false;
    }
  }
  return true;

}




/*
 * Allocate and initialise a node with the given dimensions
 */
quad_node *new_node (int xmin, int ymin, int xmax, int ymax)
{

  quad_node *node = malloc (sizeof *node);
  if (node == NULL)
  {
    Cant_happen ("quadtree.c:new_node(): couldn't allocate space for node");
  }
  node->xmin = xmin;
  node->ymin = ymin;
  node->xmax = xmax;
  node->ymax = ymax;
  node->child [0] = node->child [1] = node->child [2] = node->child [3] = NULL;
  node->tile = NULL;

  return node;

}



/*
 * Free a quadtree node, recursively freeing its children
 */
void free_node (quad_node *node)
{
  int i;

  if (node == NULL) return;

  /*
   * Strictly speaking, a quadtree is malformed if a node has any number of
   *  children other than 0 and 4, but it's probably too late to check it now
   */
  for (i = 0; i < 4; ++i)
  {
    free_node (node->child [i]);
  }

}
