/*
 * quadtree.h -
 *  Partitioning the (2D) map by means of a quadtree
 */

/*
 * File:    src/quadtree.h
 * Purpose: Defines and types for quadtree partitioning
 * Prefix:  Quad_
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2002 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


#ifndef QUADTREE_H
#define QUADTREE_H


/*
 * #includes
 */
#include "vec.h"



/*
 * A node in a quadtree
 */
typedef struct tag_quad_node
{

  /*
   * Dimensions of the node. Convenient to have while compiling and searching
   *  the quadtree, although it can of course be deduced
   */
  int xmin, ymin, xmax, ymax;

  /*
   * A quadtree, of course, has 4 children, unless it is a leaf node
   * These are arranged in the usual order, i.e. clockwise from top left
   */
  struct tag_quad_node *child [4];

  /*
   * Tile pointer, valid for a leaf node
   * This may be NULL even if the node is a leaf, indicating that the node
   *  contains no drawable tiles. This may be the case for any quadtree, since
   *  the root is always the entire map
   */
  struct tag_map_tile *tile;

  /*
   * The bounding spheroid. This is equivalent to that for a BSP node; as a
   *  useful shortcut, the centre vector also gives us the 2 split planes
   */
  vec centre;
  double radius;
  double z_scale;

} quad_node;



/*
 * A callback function for use when navigating a quadtree
 */
typedef void (*quad_callback) (quad_node *);



#ifdef __cplusplus
extern "C"
{
#endif

struct tag_clip_region;

/*
 * Function prototypes
 */
void Quad_init (void);
void Quad_exit (void);
void Quad_init_level (void);
void Quad_navigate (quad_node *node, const vec *refpoint,
		    const struct tag_clip_region *region,
		    quad_callback callback, bool noclip);

#ifdef __cplusplus
}
#endif



/*
 * Global variables
 */
extern quad_node *Quad_root;


#endif /* QUADTREE_H */
