/*
 * render.c -
 *  Rendering frame
 */

/*
 * File:    src/render.c
 * Purpose: Render outer frame function
 * Prefix:  Ren_
 * Note:    No, there isn't a file with prefix "Stimpy_"
 * Music:   Van Halen, "1984"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */




/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>

#include "types.h"
#include "accel.h"
#include "anim.h"
#include "bsp.h"
#include "clip.h"
#include "dark.h"
#include "gamemem.h"
#include "graph2d.h"
#include "hud.h"
#include "main.h"    /* Cant_happen() */
#include "map.h"
#include "matrix.h"
#include "model.h"
#include "object.h"
#include "option.h"
#include "plot.h"
#include "quadtree.h"
#include "render.h"
#include "shade.h"
#include "surface.h"
#include "texture.h"



/*
 * Got to get this sorted somewhere
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif



/*
 * Debugging
 */
#if __DEBUG__
static bool Render_debug = false;
static FILE *Render_debug_file;   /* logfile for massive spew */
#endif


/*
 * Global variables
 */
extern int Screen_width;
extern int Screen_height;

/*  double Screen_z; */
double Screen_scale;      /* Screen_z is Screen_scale * screen width */


camera Avatar_camera;
camera *Camera = &Avatar_camera;


/*
 * Elapsed (renderer) time since game start
 */
double Render_frame_time = 0.;
unsigned long Frame;




#define MAX_QUEUED_CAMERAS 12    /* main, 8 screen, 3 sensaround */

static struct
{
  int next_frame;   /* index of start of next frame             */
  int insert_pos;   /* next queue position to insert new camera */
  int render_pos;   /* next queue position of camera to render  */
  camera *cam_list [MAX_QUEUED_CAMERAS];
} Camera_queue;




/*
 * Inclusion volume (linked list of frustra)
 */
#define MAX_INC_VOL 512
static struct inc_vol
{
  clipregion_t frustrum;
  struct pvt     *from;   /* visible from this PVT */
  struct inc_vol *next;
} *Inc_vol_list;
int Inc_vol_ptr;

struct portal;

/*
 * A potentially visible tile (catchy name eh?)
 */
#define MAX_PVT 256
static struct pvt
{
  map_tile *tile;
  int visited; /* visit count (number of tiles with portals into it) */
  struct portal  *portal;
  struct inc_vol *included; /* included via these volumes */
  struct pvt *next;         /* doubly linked list         */
  struct pvt *prev;
} *PVT_list;
int PVT_ptr;

/*
 * The list of potential portals
 */
#define MAX_PORTAL 512
static struct portal
{
  surface *s;
  struct pvt *behind;
  struct portal *next; /* in list of portals for this PVT  */
} *Portal_list;
int Portal_ptr;

#define MAX_SEARCH_DISTANCE 2048




/*
 * The line queue (for cyberspace)
 */
#define LINE_QUEUE_SIZE 4096
struct line_entry
{
  int from;                     /* index into the map vertex list        */
  int to;                       /* ------------- " --------------        */
  struct line_entry *hash_next; /* hash table for quick duplicate lookup */
  struct line_entry *tile_next; /* tile's line queue is a linked list    */
  bool backface;                /* belongs to a backface (for sorting)   */
  map_tile *tile;               /* frontmost tile line belongs to        */
} *Line_queue;
int Line_queue_ptr;

struct line_entry *Line_hash_table [256];



/*
 * Local function prototypes
 */
static void render_portal_map (const clipregion_t *frustrum);
static void recurse_pvt (struct pvt *t, const vec *start,
			 const clipregion_t *frustrum);
static void queue_pvt   (struct pvt *t);
static void render_pvt  (struct pvt *t);
static struct pvt *sort_pvt (struct pvt *t);
static void render_tree (const clipregion_t *frustrum);
static void try_queue_tile (map_tile *tile);



void Ren_init (void)
{

  /*
   * Set the screen scale from the FOV angle
   */
  Screen_scale = 1. / (2 * tan (Option_fov * M_PI / 360.));

  /*
   * Initialise the avatar camera
   */
  Avatar_camera.window_width  = Screen_width;
  Avatar_camera.window_height = Screen_height;
  /*
   * Do we or don't we have access to the framebuffer here?
   */
  Avatar_camera.screen_z      = Screen_width * Screen_scale;
  Avatar_camera.flags         = CF_MAIN | CF_ZBUFFER;
# if HAVE_OPENGL
  Avatar_camera.flags |= CF_3D_API;
# endif
  Clip_init_camera_frustrum (&Avatar_camera);

  /*
   * Initialise the camera queue
   */
  Camera_queue.next_frame = 0;
  Camera_queue.insert_pos = 0;
  Camera_queue.render_pos = -1;  /* to prevent "no room" error first time */

  /*
   * Initialise the PVT list
   */
  PVT_list = Alloc_init (MAX_PVT * sizeof *PVT_list, "Ren_init", "PVT list");
  Portal_list = Alloc_init (MAX_PORTAL * sizeof *Portal_list, "Ren_init",
			    "portal list");
  Inc_vol_list = Alloc_init (MAX_INC_VOL * sizeof *Inc_vol_list, "Ren_init",
			     "inclusion volumes");

  /*
   * Initialise the line queue for cyberspace
   */
  Line_queue = Alloc_init (LINE_QUEUE_SIZE * sizeof *Line_queue, "Ren_init",
			   "line queue");

# if __DEBUG__
  if (Render_debug)
  {
    Render_debug_file = fopen ("render.log", "wt");
  }
# endif

}


extern Matrix Avatar_orn;
extern Vector Avatar_pos;

void Ren_frame (double elapsed, pixel *framebuf)
{

  Render_frame_time += elapsed;

  /*
   * Call Anim_frame() to advance any active animations
   */
  Anim_frame (elapsed);

  /*
   * This is a horrible fudge. The avatar will have a physics body to which the
   *  main camera is attached ... when we get some physics
   */
  Avatar_camera.orientation = Avatar_orn;
  Avatar_camera.position    = Avatar_pos;
  Avatar_camera.position.z += 160.;
  Avatar_camera.framebuffer = framebuf;

  /*
   * Queue the main camera to be drawn ... at the end of the queue for this
   *  frame
   */
  Ren_queue_camera (&Avatar_camera);

  /*
   * Set up the lighting for this frame
   */
  Dark_frame ();

  /*
   * Work through the camera queue for this frame
   */
  Camera_queue.render_pos = Camera_queue.next_frame;
  Camera_queue.next_frame = Camera_queue.insert_pos;
  while (Camera_queue.render_pos != Camera_queue.next_frame)
  {
    Ren_scene (Camera_queue.cam_list [Camera_queue.render_pos]);
    Camera_queue.cam_list [Camera_queue.render_pos] = NULL;
    Camera_queue.render_pos = (Camera_queue.render_pos + 1) %
      MAX_QUEUED_CAMERAS;
  }

  Camera_queue.render_pos = -1;  /* not rendering now */

  if (!Option_no_hud) HUD_render(&Avatar_camera);
}



void Ren_queue_camera (camera *c)
{
  int i;
  for (i = Camera_queue.next_frame; i != Camera_queue.insert_pos;
       i = (i + 1) % MAX_QUEUED_CAMERAS);
  {
    if (Camera_queue.cam_list [i] == c)
    {
      /*
       * Already queued ... can't queue the same camera more than once per
       *  frame
       */
      return;
    }
  }
  if (Camera_queue.insert_pos == Camera_queue.render_pos)
  {
    /*
     * Frames have collided ... no room
     */
    Cant_happen ("Ren_queue_camera(): no room in queue");
  }
  Camera_queue.cam_list [Camera_queue.insert_pos] = c;
  Camera_queue.insert_pos = (Camera_queue.insert_pos + 1) % MAX_QUEUED_CAMERAS;
}




/*
 * Renderer globals: are these all really necessary?
 */
Vector Viewpoint;
Matrix Orientation;
int Origin_x;
int Origin_y;
int Window_width;
int Window_height;


/*
 * The map itself, defined in map.c (no surprise)
 */
extern map_vertex  *Vertex;
extern map_surface *Surface;
extern map_tile    *Tile;


void Ren_scene (camera *c)
{

  clipregion_t view_frustrum;

  Camera = c;

  /*
   * Set up the accelerated (or not) renderer for this scene
   */
  Ren_spew ("\n\nRen_scene(): starting scene");
  Acc_setup_scene (c);

  ++Frame;
  Mod_init_frame ();
  Obj_init_scene ();

  /*
   * Transform the camera frustrum to get the view frustrum in world space
   */
  Clip_transform_camera_frustrum (&view_frustrum, Camera);

  /*
   * Reset the surface stack
   */
  Sur_reset ();

  if (Map_info.cyberspace)
  {
    /*
     * Clear out the line hash table ready for the c/space engine
     */
    int i;
    for (i = 0; i < 256; ++i)
    {
      Line_hash_table [i] = NULL;
    }
    Line_queue_ptr = 0;
    render_tree (&view_frustrum);
  }
  else
  {
    Ren_spew ("Ren_scene(): map is a portal map");
    render_portal_map (&view_frustrum);
  }

  /*
   * Clean up
   */
  Clip_free_region (&view_frustrum);

  Acc_end_scene ();

}


void render_portal_map (const clipregion_t *frustrum)
{

  int tile_x, tile_y;
  map_tile *tile;
  struct pvt *curpvt;
  struct pvt *last_pvt;
  vec tv;

  vec_add (tv, Camera->orientation.z, Camera->position);
  tile_x = (int)(tv.x / TILE_GRID_SIZE);
  tile_y = (int)(tv.y / TILE_GRID_SIZE);

  if ((tile_x < 0) || (tile_x > 63) || (tile_y < 0) || (tile_y > 63))
  {
    Cant_happen ("Map_render_scene(): viewpoint off map");
  }
  tile = &(Tile [tile_x + 64 * tile_y]);
  tile->frame = Frame;

  Ren_spew ("render_portal_map(): tile is (%d, %d), %p", tile_x, tile_y, tile);

  /*
   * Tile list pass 1:
   * Get a list of potentially visible tiles. These are tiles which intersect
   *  the view frustrum and which might be visible from the viewpoint tile.
   * Not guaranteed to be properly depth sorted at this point
   */
  Portal_ptr = 0;
  PVT_list [0].tile    = tile;
  PVT_list [0].visited = 0;
  PVT_list [0].portal  = NULL;
  PVT_list [0].next    = NULL;
  PVT_list [0].prev    = NULL;

  PVT_ptr = 1;
  Ren_spew ("render_portal_map(): pass 1, generating PVT list");
  recurse_pvt (&(PVT_list [0]), &(Camera->position), frustrum);

  /*
   * Tile list pass 2:
   * Partially sort the tile list so every tile is positioned after all tiles
   *  it can be seen through
   */
  last_pvt = sort_pvt (&(PVT_list [0]));

  /*
   * Tile list pass 3:
   * Now we go through the tiles with a proper clip volume, queueing surfaces
   *  and objects that are drawable
   * Kick off the inclusion lists with the first PVT
   */
  Clip_copy_region (&(Inc_vol_list [0].frustrum), frustrum);
  Inc_vol_list [0].from = NULL;
  Inc_vol_list [0].next = NULL;
  Inc_vol_ptr = 1;
  curpvt = &(PVT_list [0]);
  curpvt->included = &(Inc_vol_list [0]);
  while (curpvt != NULL)
  {
    queue_pvt (curpvt);
    curpvt = curpvt->next;
  }

  /*
   * Tile list pass 4:
   * Render all the tiles in the PVT list in back-to-front order
   */
  for (curpvt = last_pvt; curpvt != NULL; curpvt = curpvt->prev)
  {
    render_pvt (curpvt);
  }

}





void recurse_pvt (struct pvt *t, const vec *start,
		  const clipregion_t *frustrum)
{

  map_tile *tile = t->tile;
  int sno = tile->first_surface;
  struct portal *cur_portal = NULL;

  tile->rflags |= RENDER_ACTIVE;

  /*
   * The PVT scan is very quick and dirty; a portal (and the tile beyond it) is
   *  considered potentially visible if it passes the backface test and if any
   *  vertex of it is within the view frustrum
   */
  while (sno != -1)
  {

    map_surface *ms = &(Surface [sno]);
    Vector tv;
    double dist;
    surface *s = Sur_next ();

    sno = ms->tile_next;
    if (!(ms->flags & MSF_PORTAL)) continue; /* only interested in portals */

    Ren_spew ("recurse_pvt(): found portal (map_surface %p)", ms);
    vec_subtract (tv, ms->anchor, *start);
    dist = vec_dot (tv, ms->normal);

    Ren_spew ("recurse_pvt(): portal normal (%g, %g, %g), dist %g",
	      ms->normal.x, ms->normal.y, ms->normal.z, dist);
    if (dist > A_VERY_SMALL_THING) continue;
    if (dist < -MAX_SEARCH_DISTANCE) continue;

    Ren_spew ("recurse_pvt(): portal %p passed back-face cull", ms);
    Sur_map_to_surface (s, ms);
    if (Clip_surface_by_region (s, frustrum))
    {

      /*
       * Portal is potentially visible: add the tile behind it as a PVT if it
       *  isn't one already
       */
      map_tile *behind = ms->behind;
      struct pvt *newpvt;
      struct portal *newport;

      if (behind->rflags & RENDER_ACTIVE)
      {
	Ren_spew ("recurse_pvt(): loop detected, rejecting this portal");
	continue;
      }

      newport = &(Portal_list [Portal_ptr++]);
      Ren_spew ("recurse_pvt(): portal %p passed clipping", ms);

      assert (dist <= A_VERY_SMALL_THING); /* special case removed */

      /*
       * This is a stop-gap to stop a surface used for an inclusion volume from
       *  being clipped twice, which it doesn't seem to like
       */
      Sur_map_to_surface (s, ms);
      Sur_push ();

      if (behind->frame == Frame)
      {
	int i;
	for (i = 0; i < PVT_ptr; ++i)
	{
	  if (PVT_list [i].tile == behind) break;
	}
	if (i >= PVT_ptr)
	{
	  Cant_happen ("render.c:recurse_pvt():"
		       " tile marked potentially visible but not in PVT list");
	}
	PVT_list [i].visited++;
	newpvt = &(PVT_list [i]);
      }
      else
      {
	behind->frame = Frame;
	if (PVT_ptr >= MAX_PVT)
	{
	  Cant_happen ("render.c:recurse_pvt():"
		       " too many PVTs, increase MAX_PVT");
	}
	newpvt = &(PVT_list [PVT_ptr++]);
	newpvt->tile     = behind;
	newpvt->visited  = 1;
	newpvt->portal   = NULL;
	newpvt->included = NULL;
	newpvt->next     = NULL;
	recurse_pvt (newpvt, start, frustrum);
      }

      /*
       * Add a portal for it
       */
      if (Portal_ptr >= MAX_PORTAL)
      {
	Cant_happen ("render.c:recurse_pvt():"
		     " too many portals, increase MAX_PORTAL");
      }
      newport->s      = s;
      newport->behind = newpvt;
      newport->next   = cur_portal;
      cur_portal = newport;

    } /* if (v >= ms->num_vertices) */

  }

  t->portal = cur_portal;
  tile->rflags &= ~RENDER_ACTIVE;

}




struct pvt *sort_pvt (struct pvt *t)
{

  struct pvt *last = t;
  struct portal *portal = t->portal;

  while (portal != NULL)
  {
    struct pvt *behind = portal->behind;
    behind->visited--;
    if (behind->visited == 0)
    {
      last->next   = behind;
      behind->prev = last;
      last = sort_pvt (behind);
    }
    portal = portal->next;
  }
  return last;

}




void queue_pvt (struct pvt *t)
{

  map_tile *tile = t->tile;
  struct inc_vol *iv = t->included;

  /*
   * Quick reject for PVTs with no inclusion volumes
   */
  if (iv == NULL) return;

  while (iv != NULL)
  {

    struct portal *portal = t->portal;
    clipregion_t *frustrum = &(iv->frustrum);

    /*
     * Queue any objects visible in this tile
     */
    Obj_queue_tile (tile);

    /*
     * Don't queue surfaces yet ... they're clipped and drawn on the fly when
     *  the time comes to do the actual rendering
     */

    /*
     * Now go through all the portals, adding to the inclusion list of any
     *  tiles visible through this one
     */
    while (portal != NULL)
    {
      /*
       * Try to clip the portal against the current frustrum
       */
      surface ts;
      Sur_copy (&ts, portal->s);
      if (Clip_surface_by_region (&ts, frustrum))
      {

	struct pvt *behind = portal->behind;
	struct inc_vol *newiv;

	/*
	 * The portal is visible, and has been clipped. Add it to the surface
	 *  queue and to the inclusion list for the target PVT
	 */
	if (Inc_vol_ptr >= MAX_INC_VOL)
	{
	  Cant_happen ("render.c:queue_pvt(): too many inclusion volumes");
	}
	newiv = &(Inc_vol_list [Inc_vol_ptr++]);
	if (ts.flags & SF_NOCLIP)
	{
	  Clip_copy_region (&(newiv->frustrum), frustrum);
	}
	else
	{
	  Clip_init_region (&(newiv->frustrum));
	  if (!Clip_region_from_surface_convex (&(newiv->frustrum), &ts,
						&(Camera->position)))
	  {
	    /*
	     * Can't make a sensible clipregion, copy the including region
	     */
	    Clip_copy_region (&(newiv->frustrum), frustrum);
	  }
	}
	newiv->from = t;
	newiv->next = behind->included;
	behind->included = newiv;
      }

      portal = portal->next;

    }
    iv = iv->next;
  }

}


void render_pvt (struct pvt *t)
{

  map_tile *tile = t->tile;
  struct inc_vol *iv = t->included;

  /*
   * Quick reject for PVTs with no inclusion volumes
   */
  if (iv == NULL) return;

  while (iv != NULL)
  {

    clipregion_t *frustrum = &(iv->frustrum);
    int sno = tile->first_surface;
    map_tile *from = (iv->from == NULL) ? NULL : iv->from->tile;

    /*
     * Here we are interested only in textured surfaces, not in portals
     */
    while (sno != -1)
    {

      map_surface *ms = &(Surface [sno]);
      surface ts;
      Vector tv;

      sno = ms->tile_next;
      if (!(ms->flags & SF_TEXTURED)) continue;

      /*
       * Backface cull
       */
      vec_subtract (tv, ms->anchor, Camera->position);
      if (vec_dot (tv, ms->normal) > A_VERY_SMALL_THING) continue;

      /*
       * If rendering to a 3D API, we throw it the surface at this point
       */
      Sur_map_to_surface (&ts, ms);
      ts.flags = SF_TEXTURED | SF_SHADED;

      /*
       * Lighting
       * FIXME Apply lighting only once to a surface, not for every IV!
       */
      Dark_do_gouraud (tile, &ts);

      if (Clip_surface_by_region (&ts, frustrum))
      {
	/*
	 * Got at least part of it, time to render
	 */
	Ren_transform_surface (&ts);
	Acc_draw_surface (&ts);

      }

    }

    /*
     * Render the tile decals for this IV
     * This is a fudge pending proper decals-as-map-surfaces handling
     */
    Obj_render_tile_decals (tile, frustrum);

    /*
     * Do the rest of the objects in this IV
     * This takes the tiles on both sides of the portal: if an object exists
     *  in the nearer tile it is not rendered (it will be dealt with later)
     */
    Obj_render_tile (tile, from, frustrum);

    /*
     * Retire the frustrum, we won't need it again
     */
    Clip_free_region (frustrum);
    iv = iv->next;
  }

  /*
   * Mark it as rendered for the automap
   */
  tile->flags |= TF_RENDERED;

}



void Ren_transform_surface (surface *s)
{

  int v;
  vec tv;

  for (v = 0; v < s->num_vertices; ++v)
  {
    vec_subtract (tv, s->vertex [v], Camera->position);
    mat_vec_multiply (s->vertex [v], Orientation, tv);
  }

  /*
   * And on to the texture mapping
   */
  vec_subtract (tv, s->anchor, Camera->position);
  mat_vec_multiply (s->anchor, Orientation, tv);

  mat_vec_multiply (tv, Orientation, s->u);
  s->u = tv;
  mat_vec_multiply (tv, Orientation, s->v);
  s->v = tv;

  /*
   * Normal
   */
  mat_vec_multiply (tv, Orientation, s->normal);
  s->normal = tv;

}




/*
 * Here be the spacetree renderer stuff
 */
static void recurse_bsp_node (const bsp_node *node,
			      const clipregion_t *frustrum, bool noclip);

/*
 * The tile queue (linked list of drawable tiles for the current tree)
 */
map_tile *tile_queue;



void try_render_line (int from, int to, const clipregion_t *frustrum)
{

  /*
   * Make up a line_seg structure
   */
  line_seg tl; /* "temp line" */

  assert (Map_info.cyberspace);
  tl.v1.x = Vertex [from].x;
  tl.v1.y = Vertex [from].y;
  tl.v1.z = Vertex [from].z;
  tl.rgb1 = Vertex [from].u.rgb;

  tl.v2.x = Vertex [to  ].x;
  tl.v2.y = Vertex [to  ].y;
  tl.v2.z = Vertex [to  ].z;
  tl.rgb2 = Vertex [to  ].u.rgb;

  /*
   * And pass it to the low-level renderer
   */
  Acc_draw_line (&tl, frustrum);

}



/*
 * The quadtree callback for cyberspace is very simple, it just queues the
 *  tile and its objects
 */
void quadtree_cyber_callback (quad_node *node)
{
  try_queue_tile (node->tile);
  Obj_queue_tile (node->tile);
}



void render_tree (const clipregion_t *frustrum)
{

  map_tile *tile;

  tile_queue = NULL;

  /*
   * Recurse the tree to pick up tiles
   */
# if QUADTREES
  Quad_navigate (Quad_root, &(Camera->position), frustrum,
		 quadtree_cyber_callback, false);
# else
  recurse_bsp_node (BSP_root, frustrum, false);
# endif

  /*
   * Testing our hash function
   */
# if 0
  {
    int i, util = 0;
    for (i = 0; i < 255; ++i)
    {
      if (Line_hash_table [i]) ++util;
    }
    printf ("Hash table is %d%% utilised with %d lines\n", 100 * util / 256,
	    Line_queue_ptr);
  }
# endif

  tile = tile_queue;
  while (tile != NULL)
  {
    /*
     * Render the tile's lines
     */
    struct line_entry *line = tile->line_queue;
    while (line != NULL)
    {
      if (!line->backface) try_render_line (line->from, line->to, frustrum);
      line = line->tile_next;
    }

    /*
     * Render the tile's objects
     */
    Obj_render_tile (tile, NULL, NULL);

    line = tile->line_queue;
    while (line != NULL)
    {
      if (line->backface) try_render_line (line->from, line->to, frustrum);
      line = line->tile_next;
    }

    /*
     * This is for debugging only - cyberspace doesn't have an automap
     */
    tile->flags |= TF_RENDERED;
    tile = tile->render_next;

  }

}


void recurse_bsp_node (const bsp_node *node, const clipregion_t *frustrum,
		       bool noclip)
{

  if (!noclip)
  {
    /*
     * The node needs clipping against the view frustrum. There are 3 possible
     *  outcomes: The node is rejected (CLIP_WITHOUT) - just return. The node
     *  is entirely within (CLIP_WITHIN) the view frustrum, in which case all
     *  its children must also and we can set the noclip flag to save time.
     *  Else the node intersects one or more planes of the frustrum
     *  (CLIP_INTERSECT) in which case its children must also be clipped, so
     *  don't set noclip
     */
    switch (Clip_ellipsoid_z_by_region (&(node->centre), node->radius,
					node->z_scale, frustrum))
    {
    case CLIP_WITHOUT:
      return;          /* unceremonious early-out */
    case CLIP_WITHIN:
      noclip = true;   /* aha!                    */
      break;
    case CLIP_INTERSECT:
      break;
    default:
      Cant_happen ("render.c:recurse_bsp_node(): illegal clip status");
    }
  }

  if (node->front != NULL)
  {
    /*
     * The node has children (always 2). Find out on which side of the split
     *  plane lies the viewpoint
     */
    bsp_node *near_child; /* child on side of viewpoint */
    bsp_node *far_child;  /* child on other side        */
    double dist = vec_dot (Camera->position, node->split_plane.normal) -
      node->split_plane.np;
    if (dist > 0)
    {
      /*
       * On the normal side: the BSP tree counts this as "front"
       */
      near_child = node->front;
      far_child  = node->back;
    }
    else
    {
      dist = -dist;
      near_child = node->back;
      far_child  = node->front;
    }

    /*
     * Traverse the tree from front to back to get the tile list
     */
    recurse_bsp_node (near_child, frustrum, noclip);
    if (dist < MAX_SEARCH_DISTANCE)
    {
      recurse_bsp_node (far_child, frustrum, noclip);
    }

  }

  if (node->tile != NULL)
  {
    /*
     * The node has a tile in it. Queue it to be drawn
     */
    try_queue_tile (node->tile);
    Obj_queue_tile (node->tile);
  }

}




/*
 * Cyberspace only
 */
struct line_entry *new_line (void)
{
  if (Line_queue_ptr >= LINE_QUEUE_SIZE)
  {
    Cant_happen ("render.c:new_line(): too many lines,"
		 " increase LINE_QUEUE_SIZE");
  }
  return &(Line_queue [Line_queue_ptr++]);
}


void try_queue_line (map_tile *tile, int from, int to, bool backface)
{
  /*
   * Only queue a line once, so look for duplicates in the hash table
   * To get the hash bucket, multiply the from & to vertex indices and take the
   *  low byte of the result. Should be a nice quick test but at the same time
   *  minimise collisions for nearby lines
   */
  int bucket = (from * to * 3) & 0xff;
  struct line_entry *line = Line_hash_table [bucket];
  while (line != NULL)
  {
    if (((from == line->from) && (to == line->to)) ||
	((to == line->from) && (from == line->to)))
    {

      /*
       * Draw lines as late as possible, so backface takes priority within a
       *  given tile
       */
      if ((line->tile == tile) && backface) line->backface = true;

      /*
       * Already queued this line, reject it
       */
      return;
    }
    line = line->hash_next;
  }

  /*
   * It's the first time this line has been seen, so add it to the hash table
   *  and queue it in the tile
   */
  line = new_line ();
  line->from = from;
  line->to = to;
  line->hash_next = Line_hash_table [bucket];
  Line_hash_table [bucket] = line;
  line->tile_next = tile->line_queue; 
  line->backface  = backface;
  line->tile      = tile;
  tile->line_queue = line;

}


void try_queue_surface (map_tile *tile, map_surface *ms)
{
  int i, prev = ms->num_vertices - 1;
  bool backface;
  vec tv;
  assert (Map_info.cyberspace);         /* only c/space uses this queue type */

  /*
   * If line sorting according to backfaces, do the backface check here and
   *  pass a flag into try_queue_line
   */
  vec_subtract (tv, Camera->position, ms->anchor);
  backface = (vec_dot (tv, ms->normal) < 0.);
  for (i = 0; i < ms->num_vertices; ++i)
  {
    try_queue_line (tile, ms->vertex [i], ms->vertex [prev], backface);
    prev = i;
  }
}


/*
 * Try to queue a tile in the BSP tile queue
 */
void try_queue_tile (map_tile *tile)
{

  /*
   * Add the tile to the BSP tile queue
   */
  tile->render_next = tile_queue;
  tile_queue = tile;

  /*
   * Cyberspace only: queue any lines belonging to the tile
   */
  if (Map_info.cyberspace)
  {
    int sno = tile->first_surface;
    tile->line_queue = NULL;        /* initialise the tile's line queue */
    while (sno != -1)
    {
      map_surface *ms = &(Surface [sno]);
      if (!(ms->flags & MSF_PORTAL)) try_queue_surface (tile, ms);
      sno = ms->tile_next;
    }
  }

  /*
   * The BSP tree is only (currently) used in cyberspace
   */

}



/*
 * Debug output function
 */
void Ren_spew (const char *format, ...)
{
# if __DEBUG__
  if (Render_debug)
  {
    va_list ap;
    va_start (ap, format);
    vfprintf (Render_debug_file, format, ap);
    fprintf (Render_debug_file, "\n");
    fflush (Render_debug_file);
  }
# endif
}
