/*
 * resource.c -
 *  stuff for dealing with Looking Glass resfiles
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>		/* memset() */
#include <string.h>		/* strcpy(), strcmp() */

#include "gamemem.h"
#include "main.h"
#include "option.h"
#include "resource.h"


/*
 * Debugging
 */
#if __DEBUG__
#define RESOURCE_MESSAGES 0
#endif


/*
 * Local function prototypes
 */
/*  static int alloc_cache_block (long size); */
/*  static void free_cache_block (int block); */
/*  void dump_cache_list (void); */

static int read_resfile_header (resfile_header *rh);
static void read_resfile_chunkdir (resfile_header *rh);
static void free_chunk (resfile_dirent *rd);
static resfile_dirent *find_chunk_by_id (int id, resfile_header **prh);
static resfile_dirent *load_chunk (resfile_header *rh, int cno);
/*  static void dump_chunk (resfile_header *rh, int cno); */
static void unpack_data (unsigned char *pack,    unsigned char *unpack,
			 unsigned long packsize, unsigned long unpacksize);

unsigned char *unpack_uw2 (FILE *file, size_t *s);


/*
 * #defines and constants
 */
#define CACHE_NONE   0
#define CACHE_BLOCK  1
#define CACHE_FREE   2
#define CACHE_ANCHOR 3



/*
 * Local typedefs and structs
 */
typedef struct
{

  int type;                /* none, block, free, anchor      */
  resfile_dirent *chunk;   /* in resfile directory           */
  int subchunk;            /* -1 refers to entire chunk      */
  unsigned long pos;       /* offset from beginning of cache */
  unsigned long length;    /* in bytes                       */
  int next_free;           /* index, for free block chain    */
  int prev_free;           /* index, for free block chain    */
  int next_used;           /* index, for used block chain    */
  int prev_used;           /* index, for used block chain    */

} cache_block;





/*
 * Global variables
 */
resfile_header *Resfile_list;

/*
 * 2 MB chunk cache
 */
static const int Chunk_cache_size = 2*1024*1024;
/*  static const int Max_blocks       = 256; */

static unsigned char *Chunk_cache;
/*  static cache_block   *Block_list; */

/*  static int Next_block = 2; */
/*  static int Mopper_upper = 0; */
/*  static int Allocator    = 1; */



#ifndef MAX_PATH
#define MAX_PATH 512
#endif

/*
 * Resource file directories
 */
char CD_data_path [MAX_PATH] = "/cdrom/cdrom/data";
char HD_data_path [MAX_PATH] = "/cdrom/hd/data";
char UW_data_path [MAX_PATH] = "/windows/games/uw";






void Res_init (void)
{

  if (Option_underworld)
  {
    /*
     * Underworld doesn't use resource files as such
     */
/*      return; */
 }

  /*
   * Initialise the block list and the chunk cache
   */
  Chunk_cache = Alloc_init (Chunk_cache_size, "Res_init", "chunk cache");

# if 0
  Block_list = malloc (Max_blocks * sizeof (cache_block));
  if (Block_list == NULL)
  {
    Cant_happen ("Res_init(): failed to allocate block list");
  }
	memst(Block_list, 0x00, Max_blocks * sizeof(cache_block));
  /*
   * The initial chunk cache consists of the anchor block and a single free
   *  block which contains the entire chunk cache
   * The allocator's length is always zero so that it does not get chosen as
   *  a cache block in case there are no free blocks
   */
  Block_list [0].type = CACHE_ANCHOR;
  Block_list [0].length    = 0;
  Block_list [0].next_free = 1;
  Block_list [0].prev_free = 1;
  Block_list [0].next_used = 0;
  Block_list [0].prev_used = 0;

  Block_list [1].type = CACHE_FREE;
  Block_list [1].pos       = 0;
  Block_list [1].length    = Chunk_cache_size;
  Block_list [1].next_free = 0;
  Block_list [1].prev_free = 0;
  Block_list [1].next_used = 0;
  Block_list [1].prev_used = 0;

  for (i = 2; i < Max_blocks; ++i)
  {
    Block_list [i].type = CACHE_NONE;
  }
# endif

# if RESOURCE_MESSAGES
  printf ("Res_init(): block list initialised\n");
# endif

  /*
   * Option - directories where data is to be found
   */
  if (Option_hd_dir [0] != '\0') strcpy (HD_data_path, Option_hd_dir);
  if (Option_cd_dir [0] != '\0') strcpy (CD_data_path, Option_cd_dir);
  if (Option_uw_dir [0] != '\0') strcpy (UW_data_path, Option_uw_dir);

}




int Res_add_resfile (const char *fname)
{

  resfile_header **rp = &Resfile_list;
  resfile_header *rh = Resfile_list;
  FILE *resfile;

  while (rh != NULL)
  {
    if (strcmp (fname, rh->fname) == 0)
    {
      break;
    }
    rp = &(rh->next);
    rh = rh->next;
  }

  if (rh == NULL)
  {
    resfile = fopen (fname, "rb");
    if (resfile == NULL)
    {
      return 0;
    }
    rh = malloc (sizeof (resfile_header));
    if (rh == NULL)
    {
      Cant_happen ("Res_add_resfile(): couldn't allocate space for header");
    }
		memset(rh, 0x00, sizeof(resfile_header));
    *rp = rh;
    rh->resfile = resfile;
    rh->next    = NULL;
    read_resfile_header (rh);
    read_resfile_chunkdir (rh);
    strcpy (rh->fname, fname);
    
  }

  return 1;

}



void Res_free_all (void)
{

  resfile_header *rh = Resfile_list;
  resfile_header *rn;

  while (rh != NULL)
  {
    Res_free_resfile (rh);
    fclose (rh->resfile);
    rn = rh->next;
    free (rh);
    rh = rn;
  }

  Resfile_list = NULL;
	free(Chunk_cache);
	Chunk_cache = NULL;
}



void Res_print_resfile_info (void)
{

  resfile_header *rh = Resfile_list;
  int i;
  FILE *resfile;

  while (rh != NULL)
  {

    printf ("Resource file %s\tchunks %04X -> %04X\n", rh->fname,
	    rh->lowest_chunkid, rh->highest_chunkid);
    resfile = rh->resfile;
    if (resfile == NULL)
    {
      return;
    }
    printf ("%s\tchunk dir at %08lX\n", rh->id, rh->offs_dir_file);
    printf ("%d chunks, first chunk at 0x%08lX\n", rh->dir->nchunks,
	    rh->dir->offs_first_chunk_file);

    for (i = 0; i < rh->dir->nchunks; ++i)
    {
      printf ("chunk %d (%04X) size %06lX type %02X filesize %06lX (%02X)\t"
	      "@%08lX\n", i, rh->dir->entry [i].chunkid,
	      rh->dir->entry [i].size, rh->dir->entry [i].type,
	      rh->dir->entry [i].filesize, rh->dir->entry [i].what,
	      rh->dir->entry [i].offs_chunk_file);
    }

/*    dump_chunk (resfile, &head, 161); */

    rh = rh->next;

  }

/*    free (head.dir->entry); */
/*    free (head.dir); */

/*    Res_free_resfile (&head); */

}




block_handle Res_load_chunk_by_id (int id, size_t *size)
{

  resfile_header *rh = Resfile_list;
  resfile_dirent *rd;
  int i;

# if RESOURCE_MESSAGES
  printf ("Res_load_chunk_by_id(): loading chunk %d\n", id);
# endif
  while (rh != NULL)
  {

    if ((rh->lowest_chunkid <= id) && (rh->highest_chunkid >= id))
    {
      for (i = 0; i < rh->dir->nchunks; ++i)
      {
	if (rh->dir->entry [i].chunkid == id)
	{
/*  	  dump_chunk (rh, i); */
	  rd = load_chunk (rh, i);
	  if (rd == NULL)
	  {
#           if RESOURCE_MESSAGES
	    printf ("Res_load_chunk_by_id(): warning: load_chunk() failed\n");
#           endif
	    return NULL;
	  }
#         if RESOURCE_MESSAGES
	  printf ("Res_load_chunk_by_id(): found chunk %d ok\n", id);
#         endif
	  *size = rd->size;
	  return &(rd->chunk);
	}
      }
    }

    rh = rh->next;

  }

  return NULL;

}


static char Pathname [MAX_PATH];

const char *Res_make_path (int base, const char *filename)
{

  switch (base)
  {
  case CD_DATA_PATH:
    sprintf (Pathname, "%s/%s", CD_data_path, filename);
    break;
  case HD_DATA_PATH:
    sprintf (Pathname, "%s/%s", HD_data_path, filename);
    break;
  case UW_DATA_PATH:
    sprintf (Pathname, "%s/data/%s", UW_data_path, filename);
    break;
  case UW_BASE_PATH:
    sprintf (Pathname, "%s/%s", UW_data_path, filename);
    break;
  default:
    Cant_happen ("Res_make_path(): unknown resource location");
  }

  return Pathname;

}





resfile_dirent *load_chunk (resfile_header *rh, int cno)
{

  unsigned char *chunk;
  unsigned char tc [4];
  resfile_dirent *rd = &(rh->dir->entry [cno]);
  int i;
  FILE *resfile = rh->resfile;

  if (rd->chunk == NULL)
  {

    switch (rd->type)
    {

    case CTYPE_RAW:
#     if RESOURCE_MESSAGES
      printf ("load_chunk(): got uncompressed chunk\n");
#     endif
      rd->chunk = malloc (rd->filesize);
      fseek (resfile, rd->offs_chunk_file, SEEK_SET);
      fread (rd->chunk, 1, rd->filesize, resfile);
      break;

    case CTYPE_COMPRESSED:
#     if RESOURCE_MESSAGES
      printf ("load_chunk(): got compressed chunk\n");
#     endif
      chunk = malloc (rd->filesize);
      rd->chunk = malloc (rd->size);
      fseek (resfile, rd->offs_chunk_file, SEEK_SET);
      fread (chunk, 1, rd->filesize, resfile);
      unpack_data (chunk, rd->chunk, rd->filesize, rd->size);
			free(chunk);
      break;

    case CTYPE_SUBDIR:
      /*
       * Read in the subchunk directory only
       */
			if (!rd->offs_subchunk_file)
			{	/* ff1: only once please... */
#if RESOURCE_MESSAGES
		    printf ("load_chunk(): got subdir chunk\n");
#endif
				fseek (resfile, rd->offs_chunk_file, SEEK_SET);
				fread (tc, 2, 1, resfile);
				rd->nsubchunks = tc [0] + (tc [1] << 8);
				rd->offs_subchunk_file = malloc ((1 + rd->nsubchunks) * sizeof (long));
				rd->subchunk = malloc ((1 + rd->nsubchunks) * sizeof (void *));
				for (i = 0; i <= rd->nsubchunks; ++i)
				{
					fread (tc, 4, 1, resfile);
					rd->offs_subchunk_file [i] = rd->offs_chunk_file
						+ tc [0] + (tc [1] << 8) + (tc [2] << 16) + (tc [3] << 24);
#if RESOURCE_MESSAGES
					printf ("subchunk %d at %08lx\n", i, rd->offs_subchunk_file [i]);
#endif
					rd->subchunk [i] = NULL;
				}
			}
      break;

    case CTYPE_SUBDIR_COMP:
#     if RESOURCE_MESSAGES
      printf ("load_chunk(): got compressed subdir chunk\n");
#     endif

      /*
       * Read the subchunk directory and uncompress the chunk proper
       */
      fseek (resfile, rd->offs_chunk_file, SEEK_SET);
      fread (tc, 2, 1, resfile);
      rd->nsubchunks = tc [0] + (tc [1] << 8);
      rd->offs_subchunk_file = malloc ((rd->nsubchunks + 1) * sizeof (long));
      rd->subchunk = malloc ((rd->nsubchunks + 1) * sizeof (void *));
      for (i = 0; i <= rd->nsubchunks; ++i)
      {
	fread (tc, 4, 1, resfile);
	rd->offs_subchunk_file [i] =
	  tc [0] + (tc [1] << 8) + (tc [2] << 16) + (tc [3] << 24)
	  - (2 + 4 + 4 * rd->nsubchunks);
	rd->subchunk [i] = NULL;
      }
      chunk = malloc (rd->filesize - (2 + 4 + 4 * rd->nsubchunks));
      rd->chunk = malloc (rd->size - (2 + 4 + 4 * rd->nsubchunks));
      fread (chunk, 1, rd->filesize - (2 + 4 + 4 * rd->nsubchunks), resfile);
      unpack_data (chunk, rd->chunk,
		   rd->filesize - (2 + 4 + 4 * rd->nsubchunks),
		   rd->size - (2 + 4 + 4 * rd->nsubchunks));
      free (chunk);
			chunk = NULL;

    default:
      break;

    }

  }

  return rd;

}




block_handle Res_load_subchunk (int id, int sub, size_t *size)
{

  resfile_header *rh;
  resfile_dirent *rd = find_chunk_by_id (id, &rh);
  if (rd == NULL)
  {
    return NULL;
  }

  Res_load_chunk_by_id (id, size);
  if (sub >= rd->nsubchunks)
  {
    return NULL;
  }

/*    printf ("hello %d\n", rd->type); */

  *size = rd->offs_subchunk_file [sub + 1] - rd->offs_subchunk_file [sub];

  switch (rd->type)
  {

  case CTYPE_SUBDIR:
		if (!rd->subchunk [sub])
		{	/* ff1: only once please... */
			rd->subchunk [sub] = malloc (*size);
			fseek (rh->resfile, rd->offs_subchunk_file [sub], SEEK_SET);
			fread (rd->subchunk [sub], 1, *size, rh->resfile);
		}
    return &(rd->subchunk [sub]);

  case CTYPE_SUBDIR_COMP:
    rd->subchunk [sub] = ((char *) rd->chunk) + rd->offs_subchunk_file [sub];
    return &(rd->subchunk [sub]);

  default:
    break;

  }

  return NULL;

}




void Res_free_chunk_by_id (int id)
{
  free_chunk (find_chunk_by_id (id, NULL));
}




void Res_free_resfile (resfile_header *rh)
{

  int i;

  for (i = 0; i < rh->dir->nchunks; ++i)
  {
    free_chunk (&(rh->dir->entry [i]));
  }
  free (rh->dir->entry);
  free (rh->dir);
	rh->dir = NULL;
}




#if 0
int alloc_cache_block (long size)
{

  int eater;
  int finder;
  int block;

  /*
   * Sanity check first ... if the size we want is bigger than the cache size
   *  we are clearly never going to succeed
   */
  if (size > Chunk_cache_size)
  {
#   if RESOURCE_MESSAGES
    printf ("alloc_cache_block(): block size too big, increase"
	    " Chunk_cache_size\n");
#   endif
    return 0;
  }
  if (size == 0)
  {
#   if RESOURCE_MESSAGES
    printf ("alloc_cache_block(): zero length block requested\n");
#   endif
    return 0;
  }

  /*
   * Look for a block which is big enough
   */
  finder = Allocator;
  do
  {

    if (Block_list [finder].type == CACHE_FREE)
    {
      if (Block_list [finder].length >= size)
      {
	/*
	 * Found one!
	 */
	break;
      }
    }
    else
    {
      /*
       * Sanity check here
       */
      if (Block_list [finder].type != CACHE_ANCHOR)
      {
	Cant_happen ("alloc_cache_block(): free list contains non-free block");
      }
    }
    finder = Block_list [finder].next_free;

  } while (finder != Allocator);

  if (Block_list [finder].length < size)
  {

    /*
     * No single block exists which is large enough. Enter the mopper-upper
     */
    finder = Block_list [Mopper_upper].next_free;
    eater  = Block_list [Mopper_upper].next_used;
    if ((Block_list [finder].pos > Chunk_cache_size - size) &&
	(Block_list [eater].pos  > Chunk_cache_size - size))
    {
      /*
       * No room at the end; we must start again at the beginning
       */
      finder = Block_list [0].next_free;
      eater  = Block_list [0].next_used;
    }

    if (Block_list [eater].pos < Block_list [finder].pos)
    {
      /*
       * Must insert a zero-length free block
       */
      finder = Next_block;
      Block_list [finder].type = CACHE_FREE;
      Block_list [finder].pos       = Block_list [eater].pos;
      Block_list [finder].length    = 0;
      Block_list [finder].next_free = Block_list [eater].next_free;
      Block_list [finder].prev_free = Block_list [eater].prev_free;
      Block_list [Block_list [finder].next_free].prev_free = finder;
      Block_list [Block_list [finder].prev_free].next_free = finder;
    }

    while (Block_list [finder].length < size)
    {
#     if RESOURCE_MESSAGES
      printf ("alloc_cache_block(): freeing block %d\n", eater);
#     endif
      free_cache_block (eater);
      eater = Block_list [finder].next_used;
    }

  }

  /*
   * Insert a used block
   */
  if (Block_list [finder].length == size)
  {
    block = finder;
  }
  else
  {

    /*
     * Get a new block and set up its links
     */
    block = Next_block;
    while (Block_list [block].type != CACHE_NONE)
    {
      ++block;
      if (block == Max_blocks)
      {
	block = 1;
      }
      if (block == Next_block)
      {
	Cant_happen ("alloc_cache_block(): block list full");
      }
    }
    Block_list [block].next_free = finder;
    Block_list [block].prev_free = Block_list [finder].prev_free;
    Block_list [block].next_used = Block_list [finder].next_used;
    Block_list [block].prev_used = Block_list [finder].prev_used;
    Block_list [block].pos       = Block_list [finder].pos;
    Block_list [block].length    = size;
    Block_list [finder].pos += size;
    Block_list [finder].length -= size;
  }

  Block_list [block].type = CACHE_BLOCK;

  Block_list [Block_list [block].next_free].prev_used = block;
  Block_list [Block_list [block].prev_free].next_used = block;
  Block_list [Block_list [block].next_used].prev_used = block;
  Block_list [Block_list [block].prev_used].next_used = block;

  Block_list [Block_list [block].next_free].prev_free =
    Block_list [block].prev_free;
  Block_list [Block_list [block].prev_free].next_free =
    Block_list [block].next_free;
  Block_list [Block_list [block].next_used].prev_free =
    Block_list [block].prev_free;
  Block_list [Block_list [block].prev_used].next_free =
    Block_list [block].next_free;

  return block;

}




void free_cache_block (int block)
{

  int link_free;
/*    int link_used; */

  if ((block >= Max_blocks) || (block <= 0))
  {
    Cant_happen ("free_cache_block(): block id out of range");
  }
  if (Block_list [block].type != CACHE_BLOCK)
  {
#   if RESOURCE_MESSAGES
    printf ("free_cache_block(): tried to free unused block\n");
#   endif
    return;
  }

  /*
   * Remove the block from the used list and put it on the free list
   */
  Block_list [block].type = CACHE_FREE;
  Block_list [Block_list [block].prev_used].next_used
    = Block_list [block].next_used;
  Block_list [Block_list [block].next_used].prev_used
    = Block_list [block].prev_used;
  Block_list [Block_list [block].next_free].prev_free = block;
  Block_list [Block_list [block].prev_free].next_free = block;

  /*
   * Amalgamate the blocks, if needed
   */
  link_free = Block_list [block].prev_free;
  if (block == Block_list [link_free].next_used)
  {

    /*
     * Amalgamate the previous block and this one
     */
    if (Block_list [link_free].pos + Block_list [link_free].length !=
	Block_list [block].pos)
    {
      Cant_happen ("free_cache_block(): blocks not contiguous, should be");
    }
    Block_list [block].length += Block_list [link_free].length;
    Block_list [block].pos = Block_list [link_free].pos;
    Block_list [block].prev_free = Block_list [link_free].prev_free;
    Block_list [link_free].type = CACHE_NONE;
  }

  link_free = Block_list [block].next_free;
  if (block == Block_list [link_free].prev_used)
  {

    /*
     * Amalgamate this block and the next one
     */
    if (Block_list [block].pos + Block_list [block].length !=
	Block_list [link_free].pos)
    {
      Cant_happen ("free_cache_block(): blocks not contiguous, should be");
    }
    Block_list [block].length += Block_list [link_free].length;
    Block_list [block].next_free = Block_list [link_free].next_free;
    Block_list [link_free].type = CACHE_NONE;
  }

}




void dump_cache_list (void)
{

  int i;

  for (i = 0; i < Max_blocks; ++i)
  {

    if (Block_list [i].type == CACHE_NONE)
    {
      continue;
    }

    printf ("Block %d type ", i);
    switch (Block_list [i].type)
    {
    case CACHE_BLOCK:
      printf ("block ");
      break;
    case CACHE_FREE:
      printf ("free  ");
      break;
    case CACHE_ANCHOR:
      printf ("anchor");
      break;
    default:
      printf ("??????");
      break;
    }

    printf (" start %lu length %lu\n",
	    Block_list [i].pos, Block_list [i].length);
    printf ("\tnext free %d prev free %d next used %d prev used %d\n",
	    Block_list [i].next_free, Block_list [i].prev_free,
	    Block_list [i].next_used, Block_list [i].prev_used);

  }

}
#endif



int read_resfile_header (resfile_header *rh)
{

  unsigned char hdr [0x80];

  fseek (rh->resfile, 0, SEEK_SET);
  if (fread (hdr, 1, 0x80, rh->resfile) != 0x80)
  {
    return 0;
  }

  memcpy (rh->id, hdr, 0x7c);
  rh->offs_dir_file = hdr [0x7c] + (hdr [0x7d] << 8) +
                     (hdr [0x7e] << 16) + (hdr [0x7f] << 24);
  rh->dir = NULL;
  return 1;

}



void read_resfile_chunkdir (resfile_header *rh)
{

  resfile_chunkdir *rc;
  unsigned char raw [10];
  long offs;
  int i;
  FILE *resfile = rh->resfile;

  if ((fseek (resfile, rh->offs_dir_file, SEEK_SET) < 0) ||
      (fread (raw, 1, 6, resfile) != 6))
  {
    return;
  }

  rc = malloc (sizeof (resfile_chunkdir));
  if (rc == NULL)
  {
    return;
  }
	memset(rc, 0x00, sizeof(resfile_chunkdir));
  rh->dir = rc;
  rh->highest_chunkid = 0;
  rh->lowest_chunkid = 0xFFFF;

  rc->nchunks = raw [0] + (raw [1] << 8);
  rc->offs_first_chunk_file = raw [2] + (raw [3] << 8) +
                             (raw [4] << 16) + (raw [5] << 24);

  /*
   * Now to read in the directory proper
   */
  rc->entry = malloc (rc->nchunks * sizeof (resfile_dirent));
  if (rc->entry == NULL)
  {
    free (rc);
    return;
  }
	memset(rc->entry, 0x00, rc->nchunks * sizeof(resfile_dirent));
  offs = rc->offs_first_chunk_file;
  for (i = 0; i < rc->nchunks; ++i)
  {
    if (fread (raw, 1, 10, resfile) != 10)
    {
      free (rc->entry);
      free (rc);
      return;
    }
    rc->entry [i].chunkid  = raw [0] + (raw [1] << 8);
    rc->entry [i].size     = raw [2] + (raw [3] << 8) + (raw [4] << 16);
    rc->entry [i].type     = raw [5];
    rc->entry [i].filesize = raw [6] + (raw [7] << 8) + (raw [8] << 16);
    rc->entry [i].what     = raw [9];
    rc->entry [i].offs_chunk_file = offs;
    rc->entry [i].chunk    = NULL;
    rc->entry [i].offs_subchunk_file = NULL;

    if (rc->entry [i].chunkid < rh->lowest_chunkid)
    {
      rh->lowest_chunkid = rc->entry [i].chunkid;
    }
    if (rc->entry [i].chunkid > rh->highest_chunkid)
    {
      rh->highest_chunkid = rc->entry [i].chunkid;
    }

    offs += (rc->entry [i].filesize + 3) & ~3;
  }

  return;

}



void free_chunk (resfile_dirent *rd)
{

  int i;

  if (rd != NULL)
  {
    if (rd->chunk != NULL)
    {
      free (rd->chunk);
      rd->chunk = NULL;
    }
    if (rd->offs_subchunk_file != NULL)
    {
      for (i = 0; i < rd->nsubchunks; ++i)
      {
				/* ff1: don't free them all - could get too chaotic ;)	*/
				/*			the comp subdirs only have re-pointers in				*/
				/*			rd->subchunk[] to the content of rd->chunk...		*/
				if (rd->subchunk [i] != NULL && rd->type != CTYPE_SUBDIR_COMP)
				{
					free (rd->subchunk [i]);
					rd->subchunk[i] = NULL;
				}
      }
			rd->nsubchunks = 0;
      free (rd->subchunk);
			rd->subchunk = NULL;
      free (rd->offs_subchunk_file);
			rd->offs_subchunk_file = NULL;
    }
  }

}




static resfile_dirent *find_chunk_by_id (int id, resfile_header **prh)
{

  resfile_header *rh = Resfile_list;
  int i;

  while (rh != NULL)
  {

    if ((rh->lowest_chunkid <= id) && (rh->highest_chunkid >= id))
    {
      for (i = 0; i < rh->dir->nchunks; ++i)
      {
	if (rh->dir->entry [i].chunkid == id)
	{
	  if (prh != NULL)
	  {
	    *prh = rh;
	  }
	  return &(rh->dir->entry [i]);
	}
      }
    }

    rh = rh->next;

  }

  return NULL;

}



void Res_dump_chunk (int id, unsigned int width)
{ 

  unsigned char *chunk;
  resfile_header *rh;
  resfile_dirent *rd = find_chunk_by_id (id, &rh);
  size_t size;
  unsigned int i;

  if (rd == NULL) Cant_happen ("Res_dump_chunk(): couldn't find chunk dirent");
  printf ("%d\n", rd->type);

# if 0
  if ((rd->type == CTYPE_SUBDIR) || (rd->type == CTYPE_SUBDIR_COMP))
  {
    for (i = 0; i < rd->nsubchunks; ++i)
    {
      printf ("%d\n");
      Res_dump_subchunk (id, i, width);
    }
    return;
  }
# endif

  Res_load_chunk_by_id (id, &size);
  chunk = rd->chunk;

  if (chunk == NULL) Cant_happen ("Res_dump_chunk(): couldn't load chunk");

  for (i = 0; i < rd->size; ++i)
  {
    printf ("%02X ", chunk [i]);
    if ((i % width) == width - 1)
    {
      printf ("\n");
    }
  }
  if ((i % width) != width) printf ("\n");

}



void Res_dump_subchunk (int id, int subchunk, unsigned int width)
{

  block_handle bh;
  size_t size;
  unsigned char *chunk;
  unsigned int i;

  bh = Res_load_subchunk (id, subchunk, &size);
  chunk = *bh;
  for (i = 0; i < size; ++i)
  {
    printf ("%02X ", chunk [i]);
    if ((i % width) == width - 1)
    {
      printf ("\n");
    }
  }

}




/*
 * Load up an Underworld II map resource
 */
unsigned char *Res_load_uw2_map (FILE *f, int map, int chunk, size_t *size)
{

  size_t dirsize;
/*    size_t mapoffs; */
  size_t ckoffs;
  size_t offset;
  unsigned int flags;
  size_t filesize;
  size_t packsize;
  unsigned char *raw;

  fseek (f, 0, SEEK_SET);
  READ_LE_LONG (f, dirsize);

  ckoffs = map * 4;
  if (ckoffs >= dirsize)
  {
    Cant_happen ("Res_load_uw2_map(): invalid map number");
  }

  ckoffs += chunk * dirsize; /* offset within directory block */
  fseek (f, 6 + ckoffs, SEEK_SET);
  READ_LE_LONG (f, offset);
  fseek (f, 6 + 4 * dirsize + ckoffs, SEEK_SET);
  READ_LE_LONG (f, flags);
  fseek (f, 6 + 8 * dirsize + ckoffs, SEEK_SET);
  READ_LE_LONG (f, packsize);

  fseek (f, 6 + 12 * dirsize + ckoffs, SEEK_SET);
  READ_LE_LONG (f, filesize);

  /*
   * Check that the resource actually exists in the file. This seems to be
   *  determined by bit 1 of the flags word except for texture map resources
   *  which are always present if their map is present ...
   */
  if ((chunk != 1) && !(flags & 2))
  {
    return NULL;
  }

  /*
   * Texture map resources might not be compressed ... this looks to be bit 1
   *  again
   */
  fseek (f, offset, SEEK_SET);
  if ((chunk == 1) && !(flags & 2))
  {
    *size = packsize;
    raw = malloc (packsize);
    if (raw == NULL)
    {
      Cant_happen ("Res_load_uw2_map(): failed to allocate space");
    }
    if (fread (raw, 1, packsize, f) != packsize)
    {
      Cant_happen ("Res_load_uw2_map(): couldn't load map chunk");
    }
  }
  else
  {
    raw = unpack_uw2 (f, size);
  }

  return raw;

}



void unpack_data (unsigned char *pack,    unsigned char *unpack,
		  unsigned long packsize, unsigned long unpacksize)
{

  unsigned char *byteptr;
  unsigned char *exptr;
  unsigned long word = 0;  /* initialise to stop "might be used before set" */
  unsigned long bytesdone;
  int nbits;
/*    int type; */
  int val;

  int ntokens = 0;
  static int offs_token [16384];
  static int len_token  [16384];
  static int org_token  [16384];

  int i;

  for (i = 0; i < 16384; ++i)
  {
    len_token [i] = 1;
    org_token [i] = -1;
  }
  memset (unpack, 0, unpacksize);

  byteptr = pack;
  exptr   = unpack;
  nbits = 0;

  bytesdone = 0;
  while (bytesdone < unpacksize)
  {

    while (nbits < 14)
    {
      word = (word << 8) + *byteptr++;
      nbits += 8;
    }
    if (byteptr > pack + packsize)
    {
      Cant_happen ("unpack_data(): read off end of compressed data");
    }

    nbits -= 14;
    val = (word >> nbits) & 0x3FFF;
    if (val == 0x3FFF)
    {
      break;
    }

    if (val == 0x3FFE)
    {
      for (i = 0; i < 16384; ++i)
      {
	len_token [i] = 1;
	org_token [i] = -1;
      }
      ntokens = 0;
      continue;
    }

    if (ntokens < 16384)
    {
      offs_token [ntokens] = bytesdone;
      if (val >= 0x100)
      {
	org_token [ntokens] = val - 0x100;
      }
    }
    ++ntokens;

    if (val < 0x100)
    {
      *exptr++ = val;
    }
    else
    {
      val -= 0x100;

      if (len_token [val] == 1)
      {	  
	if (org_token [val] != -1)
	{
	  len_token [val] += len_token [org_token [val]];
	}
	else
	{
	  len_token [val] += 1;
	}
      }
      for (i = 0; i < len_token [val]; ++i)
      {
	*exptr++ = unpack [i + offs_token [val]];
      }

    }
    bytesdone = exptr - unpack;
    
  }

}



/*
 * Underworld II compression
 */
unsigned char *unpack_uw2 (FILE *file, size_t *s)
{

  unsigned char *unpack;
  unsigned char *upp;
  unsigned char *rp;

  int which;
  int record;
  int offset;
  int count;

  READ_LE_LONG (file, *s);
  unpack = malloc (*s + 100);
  if (unpack == NULL)
  {
    Cant_happen ("unpack_uw2(): couldn't allocate space for unpacked data");
  }

  upp = unpack;
  rp = unpack;
  while (upp < unpack + *s)
  {

    /*
     * A compressed block consists of a flag byte followed by 8 records.
     * If a bit in the flag byte is 1, its record is a literal byte to be
     *  written to the output
     * If it is a 0, we have a repeat record.
     */
    which = fgetc (file);
    for (record = 0; record < 8 && (upp < unpack + *s); ++record)
    {
      if (which & 1)
      {
	*upp++ = fgetc (file);  /* easy */
      }
      else
      {
	/*
	 * A repeat record consists of 2 bytes containing a 12-bit offset and
	 *  a 4-bit count. The first byte contains the low 8 bits of the count.
	 *  The second contains the high 4 bits of the offset (high nybble)
	 *  and the count (low nybble)
	 */
	offset = fgetc (file);
	count  = fgetc (file);
	offset += (count & 0xF0) << 4;
	count  = (count & 0x0F) + 3;
	offset = (offset + 18);
	if (offset > upp - unpack) offset -= 0x1000;
	while (offset < upp - unpack - 0x1000) offset += 0x1000;
	for (; count > 0; --count)
	{
	  *upp++ = rp [offset++];
	}
	assert (upp <= unpack + *s);
      }
      which >>= 1;
    }

  }

  return unpack;

}

/*
 * Method: int Res_get_numSubchunks_by_id(int id, size_t *size)
 * Author: BryanN
 * Returns: "-1" on error, otherwise....
 * Description: Its overly large and messy *grin*, 
 * but it returns the number of Subchunks in a chunk. 
 * Very handy for logs!
 * 
 * I wrote it to remove the need for magic numbers in log loops (see viewlog.c)
 */

int Res_get_numSubchunks_by_id(int id, size_t *size)
{
  resfile_header *rh = Resfile_list;
  resfile_dirent *rd = find_chunk_by_id (id, &rh);

  if (rd == NULL)
  {
	  return -1;
  }

  Res_load_chunk_by_id (id, size);
  return (int)rd->nsubchunks;
}
