/*
 * resource.h -
 *  definitions for Looking Glass resource files
 */


#ifndef RESOURCE_H
#define RESOURCE_H



/*
 * #includes
 */
#include <stdio.h>

#include "types.h"




/*
 * Chunk types
 */
#define CTYPE_RAW         0x00
#define CTYPE_COMPRESSED  0x01
#define CTYPE_SUBDIR      0x02
#define CTYPE_SUBDIR_COMP 0x03


/*
 * What is in a chunk
 */
#define CHUNK_PALETTE 0x00
#define CHUNK_STRING  0x01
#define CHUNK_BITMAP  0x02
#define CHUNK_FONT    0x03
#define CHUNK_VIDEO   0x04
#define CHUNK_FX      0x07
#define CHUNK_OBJ3D   0x0F
#define CHUNK_AUDIO   0x11
#define CHUNK_MAP     0x30



#define RESFILE_NAME_MAX 1024


/*
 * Where data files live (configurable)
 */
#define CD_DATA_PATH 1
#define HD_DATA_PATH 2
#define UW_DATA_PATH 3
#define UW_BASE_PATH 4


typedef struct
{
  unsigned short chunkid;
  unsigned long  size;
           int   type;
  unsigned long  filesize;
           int   what;
           long  offs_chunk_file;

  /*
   * The whole chunk, and nothing but the chunk
   */
           void *chunk;

  /*
   * Subdirectory for type 2/3 chunks
   */
  int   nsubchunks;
  long *offs_subchunk_file;
  void **subchunk;

} resfile_dirent;


typedef struct
{
  unsigned short  nchunks;
  long            offs_first_chunk_file;
  resfile_dirent *entry;
} resfile_chunkdir;


typedef struct resfile_header
{

  char   id [0x7c];
  long   offs_dir_file;
  FILE  *resfile;
  resfile_chunkdir *dir;

  /*
   * For optimisation purposes
   */
  unsigned short highest_chunkid;
  unsigned short lowest_chunkid;

  char fname [RESFILE_NAME_MAX];
  struct resfile_header *next;

} resfile_header;



/*
 * Macros to read little-endian values from resource files (used by Underworld
 *  resource handlers)
 */
#define READ_LE_SHORT(f,v) do { \
  v = fgetc (f);                \
  v += (fgetc (f) << 8);        \
} while (0)

#define READ_LE_LONG(f,v) do {  \
  v = fgetc (f);                \
  v += fgetc (f) << 8;          \
  v += fgetc (f) << 16;         \
  v += fgetc (f) << 24;         \
} while (0)



#ifdef __cplusplus
extern "C"
{
#endif



void Res_init (void);
void Res_free_resfile (resfile_header *rh);
int  Res_add_resfile (const char *fname);
void Res_free_all (void);
block_handle Res_load_chunk_by_id (int id, size_t *size);
block_handle Res_load_subchunk (int id, int sub, size_t *size);
void Res_free_chunk_by_id (int id);
const char *Res_make_path (int base, const char *filename);

int Res_get_numSubchunks_by_id(int id, size_t *size);

/*
 * Debugging
 */
void Res_print_resfile_info (void);
void Res_dump_chunk (int id, unsigned int width);
void Res_dump_subchunk (int id, int subchunk, unsigned int width);

/*
 * Ultima Underworld resources
 */
unsigned char *Res_load_uw2_map (FILE *f, int map, int chunk, size_t *size);



#ifdef __cplusplus
}
#endif




/*
 * Global variables
 */
extern char CD_data_path [];
extern char HD_data_path [];
extern char UW_data_path [];



#endif /* RESOURCE_H */
