/*
 * screen.c -
 *  In-game display screens
 */

/*
 * File:    src/screen.c
 * Purpose: Screen decal objects, animations and cameras
 * Prefix:  Scr_
 * Note:    
 * Music:   Terrorvision, "How To Make Friends And Influence People"
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "anim.h"
#include "bitmap.h"
#include "chunks.h"
#include "clip.h"
#include "fixture.h"
#include "font.h"
#include "gamemem.h"
#include "graph2d.h"
#include "main.h"
#include "map.h"
#include "object.h"
#include "option.h"
#include "palette.h"
#include "render.h"
#include "resource.h"
#include "surface.h"
#include "texture.h"



/*
 * Got to get this sorted somewhere
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif


/*
 * Special values of the "start frame" field for animated screens
 */
#define SCREEN_SPECIAL_START         0x00F6
#define SCREEN_SPECIAL_STATIC_SHODAN 0x00F6
#define SCREEN_SPECIAL_STATIC        0x00F7
#define SCREEN_SPECIAL_SURVEILLANCE  0x00F8  /* F8-FF */
#define SCREEN_SPECIAL_TEXT          0x0100
#define SCREEN_SPECIAL_SCROLL        0x0080  /* text scroll flag        */
#define SCREEN_SPECIAL_RANDOM_DIGIT  0x007f  /* random 0-9 in CPU rooms */

#define MAX_SURVEILLANCE_CAMERA 8

#define SCREEN_ADVANCE_DELAY    0.5  /* in seconds               */

#define NUM_SCREEN_FRAMES \
  (1+CHUNK_TEXTURE_SCREEN_LAST-CHUNK_TEXTURE_SCREEN_BASE)

#define NUM_SHODAN_FRAMES       6    /* +1 for the static itself */
#define BASE_FRAME_SHODAN      64
#define SHODAN_EMERGE_CHANCE    4    /* each advance, out of 256 */
#define CODE_SWITCH_CHANCE     16    /* each advance, out of 256 */


/*
 * Local typedefs and structs
 */





/*
 * Local function prototypes
 */
static void init_anim   (map_object *mo);
static void choose_hook (map_object *mo, fixture *fix);

static int fetch_texture_anim (map_object *obj);
static int fetch_texture_surveillance (map_object *obj);
static int fetch_texture_static (map_object *obj);
static int fetch_texture_shodan (map_object *obj);
static int fetch_texture_digit  (map_object *obj);
static int fetch_texture_text   (map_object *obj);

static void advance_shodan      (anim_control *anim);
/*  static void advance_text_scroll (anim_control *anim); */
static void advance_digit       (anim_control *anim);
static void advance_code        (anim_control *anim);


/*
 * Global variables
 */
static struct
{
  int viewpoint_id; /* ID of a dummy object to provide a transform */
  camera cam;
  ss_bitmap_header *bmp;
} Surveillance_camera [MAX_SURVEILLANCE_CAMERA];

/*
 * Size of the bitmap into which we render for a surveillance screen. Must be
 *  a power of 2
 */
static int Surveillance_bitmap_size = 64;

/*
 * Matrix to transform between object and camera coordinates, for surveillance
 */
Matrix Cam_correction = { { 0, 0, 1 }, {-1, 0, 0 }, { 0, 1, 0 } };



/*
 * Have we got an anim_control for the random digit screens? Nasty fudge to
 *  make sure we only get one even if there are several on the level
 */
static bool Digit_anim_control;
static int  Digit;

static int Digit_texture [10] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

/*
 * Base texid for the animated screens block
 */
static int Base_screen_texture = -1;



/*
 * Called from Obj_init_level()
 */
void Scr_init_level (int level)
{

  block_handle bh;
  size_t size;
  int i;
  unsigned char *sp;

  /*
   * Load up the surveillance camera list
   */
  bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_SURVEILLANCE, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Scr_init_level(): couldn't load surveillance info");
  }
  if (size != 2 * MAX_SURVEILLANCE_CAMERA)
  {
    Cant_happen ("Scr_init_level(): surveillance info chunk size mismatch");
  }

  sp = (unsigned char *) *bh;
  for (i = 0; i < MAX_SURVEILLANCE_CAMERA; ++i)
  {
    camera *cam = &(Surveillance_camera [i].cam);
    int view = *sp + 256 * sp [1]; sp += 2;
    Surveillance_camera [i].viewpoint_id = view;
    if (view != 0)
    {
      cam->window_width  = Surveillance_bitmap_size;
      cam->window_height = Surveillance_bitmap_size;
      cam->framebuffer   =
	malloc (Surveillance_bitmap_size * Surveillance_bitmap_size *
		sizeof *(cam->framebuffer));
      if (cam->framebuffer == NULL)
      {
	Cant_happen ("Scr_init_level(): failed to allocate camera buffer");
      }
      cam->screen_z =
	Surveillance_bitmap_size / (2 * tan (Option_fov * M_PI / 360.));
      cam->flags    = 0;
      Clip_init_camera_frustrum (cam);

      Surveillance_camera [i].bmp =
	Bmp_create (Surveillance_bitmap_size, Surveillance_bitmap_size);

      printf ("Scr_init_level(): camera %d points to object %d\n", i, view);
    }

  }

  Res_free_chunk_by_id (4000 + 100*level + MAP_CHUNK_SURVEILLANCE);

  /*
   * In case of random-digit screens
   */
  Digit_anim_control = false;  

  /*
   * Grab a block of texids for the animated screens (102 of them! but the
   *  actual bitmaps will only be loaded if necessary, fortunately)
   */
  Base_screen_texture = Tex_get_cache_block (CHUNK_TEXTURE_SCREEN_BASE,
                                             0, NUM_SCREEN_FRAMES,
                                             TEXTURE_BLOCK_CHUNK);
# if TEXTURE_MESSAGES
  printf ("Scr_init_level(): got %d texids from %d\n", NUM_SCREEN_FRAMES,
          Base_screen_texture);
# endif

}

void Scr_deinit(void)
{
	int i;

  for (i = 0; i < MAX_SURVEILLANCE_CAMERA; ++i)
  {
		if (Surveillance_camera [i].cam.framebuffer)
		{
      free(Surveillance_camera [i].cam.framebuffer);
			Surveillance_camera [i].cam.framebuffer = NULL;
		}
		if (Surveillance_camera[i].bmp)
		{
			if (Surveillance_camera [i].bmp->bits)
				free(Surveillance_camera [i].bmp->bits);
			free(Surveillance_camera[i].bmp);
		}
	}
}

extern fixture *Fixture_list;
extern map_object *Map_object_list;






#define OBJID_SCREEN      132
#define OBJID_TMAP        133
#define OBJID_SUPERSCREEN 134
#define OBJID_BIGSCREEN   135


/*
 * Initialise screen fetch hooks
 */
void Scr_init_object (map_object *mo)
{

  static const int Screen_size [] =
  {
    128, /* screen      */
    256, /* tmap        */
    512, /* superscreen */
    256  /* bigscreen   */
  };

  int fi = Fix_check_index (mo);
  choose_hook (mo, &(Fixture_list [fi]));

  init_anim (mo);

  if ((mo->object_id >= OBJID_SCREEN) && (mo->object_id <= OBJID_BIGSCREEN))
  {
    /*
     * Set it up a map_surface
     */
    Obj_make_decal_surface (mo, Screen_size [mo->type - 6],
			    Screen_size [mo->type - 6], SF_TEXTURED);
  }

}



/*
 * Choose a bitmap fetch hook appropriate to the screen type
 * Assumes that fix is a valid fixture and is actually associated with mo
 */
void choose_hook (map_object *mo, fixture *fix)
{
  if (fix->start_frame >= SCREEN_SPECIAL_TEXT)
  {
    int txtno = fix->start_frame & 0xff;
    if (txtno == SCREEN_SPECIAL_RANDOM_DIGIT)
    {
      mo->fetch_texture_hook = fetch_texture_digit;
    }
    else
    {
      mo->fetch_texture_hook = fetch_texture_text;
    }
  }
  else if (fix->start_frame >= SCREEN_SPECIAL_SURVEILLANCE)
  {
    mo->fetch_texture_hook = fetch_texture_surveillance;
  }
  else if (fix->start_frame == SCREEN_SPECIAL_STATIC)
  {
    mo->fetch_texture_hook = fetch_texture_static;
  }
  else if (fix->start_frame == SCREEN_SPECIAL_STATIC_SHODAN)
  {
    mo->fetch_texture_hook = fetch_texture_shodan;
  }
  else
  {
    mo->fetch_texture_hook = fetch_texture_anim;
  }

}




/*
 * Called from Scr_init_hook() to set up the animation
 */
void init_anim (map_object *mo)
{
  int fi = Fix_check_index (mo);

# if 0 /* diagnostic */
  extern map_object *Map_object_list;
  printf ("%04x\t%d %d %d %02x\t%04x %04x\n", mo - Map_object_list,
	  Fixture_list [fi].u.num_frames,
	  mo->state, Fixture_list [fi].start_frame,
	  Fixture_list [fi].flags, (int)mo->pos.x, (int)mo->pos.y);
# endif

  if (Fixture_list [fi].start_frame == SCREEN_SPECIAL_STATIC_SHODAN)
  {
    /*
     * SHODAN screens are handled in a slightly wacky way
     */
    Fixture_list [fi].anim = Anim_add (SCREEN_ADVANCE_DELAY, mo,
				       NUM_SHODAN_FRAMES,
				       AF_BIDIRECTIONAL | AF_PAUSED);
    Fixture_list [fi].anim->advance_hook = advance_shodan;
  }
  else if ((Fixture_list [fi].start_frame == 
	    SCREEN_SPECIAL_TEXT + SCREEN_SPECIAL_RANDOM_DIGIT) &&
	   !Digit_anim_control)
  {
    Fixture_list [fi].anim = Anim_add (SCREEN_ADVANCE_DELAY, mo, 1,
				       AF_PAUSED);
    Fixture_list [fi].anim->advance_hook = advance_digit;
    Digit_anim_control = true;
  }
  else if (Fixture_list [fi].u.num_frames > 1)
  {
    /*
     * It's got more than one frame, so it's animated. Add it to the
     *  animations list
     */
    Fixture_list [fi].anim = Anim_add (SCREEN_ADVANCE_DELAY, mo, 
				       Fixture_list [fi].u.num_frames,
				       Fixture_list [fi].flags & 1);
  }

  if (Fixture_list [fi].texture && Fixture_list [fi].anim)
  {
    /*
     * The "texture" field is set for code screens indicating that the screen
     *  will switch randomly between displaying an animation and a code
     */
    Fixture_list [fi].anim->advance_hook = advance_code;
  }

}



/*
 * Bitmap fetch hooks
 */

/*
 * An ordinary animated screen. Frame determination is handled by anim.c
 */
int fetch_texture_anim (map_object *obj)
{

  /*
   * For an animated screen bmpno is the frame, which is set in object.c
   *  according to the frame time
   */
  int fi = Fix_check_index (obj);

  obj->base_texid = Base_screen_texture + Fixture_list [fi].start_frame;
  obj->fetch_texture_hook = Obj_fetch_texture_anim;
  return obj->base_texid + obj->state;
}



/*
 * A surveillance camera. This is cross-referenced to the surveillance control
 *  chunk which points to a dummy object giving the transformation
 * We return the previous contents of the framebuffer and queue the camera to
 *  be rendered again next frame
 */
int fetch_texture_surveillance (map_object *obj)
{

  int fi = Fix_check_index (obj);
  int special = Fixture_list [fi].start_frame;
  ss_bitmap_header *bmp;

  /*
   * Surveillance camera. Get its bitmap from the previous framebuffer
   */
  int scam = special - SCREEN_SPECIAL_SURVEILLANCE;
  int i;
  if (Surveillance_camera [scam].viewpoint_id == 0)
  {
    /*
     * Camera without a viewpoint ... set it to produce static
     */
    obj->fetch_texture_hook = fetch_texture_static;
    return fetch_texture_static (obj);
  }

  bmp = Surveillance_camera [scam].bmp;
  if (bmp == NULL)
  {
    Cant_happen ("screen.c: fetch_texture_special(): NULL cache bitmap");
  }
  if (obj->base_texid < 0)
  {
    obj->base_texid = Tex_add_cache (bmp);
  }

  for (i = 0; i < Surveillance_bitmap_size * Surveillance_bitmap_size; ++i)
  {
    bmp->bits [i] =
      Pal_pixel_to_index (Surveillance_camera [scam].cam.framebuffer [i]);
  }

  /*
   * Queue the camera to be rendered into next frame
   */
  i = Surveillance_camera [scam].viewpoint_id;
  Surveillance_camera [scam].cam.position = Map_object_list [i].pos;
  Surveillance_camera [scam].cam.orientation.x.x =
    Map_object_list [i].orientation.x.x;
  Surveillance_camera [scam].cam.orientation.x.y =
    Map_object_list [i].orientation.y.x;
  Surveillance_camera [scam].cam.orientation.x.z =
    Map_object_list [i].orientation.z.x;
  Surveillance_camera [scam].cam.orientation.y.x =
    Map_object_list [i].orientation.x.z;
  Surveillance_camera [scam].cam.orientation.y.y =
    Map_object_list [i].orientation.y.z;
  Surveillance_camera [scam].cam.orientation.y.z =
    Map_object_list [i].orientation.z.z;
  Surveillance_camera [scam].cam.orientation.z.x =
    Map_object_list [i].orientation.x.y;
  Surveillance_camera [scam].cam.orientation.z.y =
    Map_object_list [i].orientation.y.y;
  Surveillance_camera [scam].cam.orientation.z.z =
    Map_object_list [i].orientation.z.y;
  Ren_queue_camera (&Surveillance_camera [scam].cam);

  return obj->base_texid;

}



#if 0
/*
 * Scratch space to render static into
 */
static /* no pun intended */ ss_bitmap_header *Static_buffer;
#endif

static const int Static_bitmap_size = 32;

int fetch_texture_shodan (map_object *obj)
{

  if (obj->state == 0)
  {
    /*
     * State 0 is static
     */
    return fetch_texture_static (obj);
  }
  return Base_screen_texture + BASE_FRAME_SHODAN + obj->state - 1;

}



void advance_texture_static (ss_bitmap_header *bmp)
{
  int i;

  /*
   * Bitmap paranoia
   */
  assert (bmp != NULL);
  assert (bmp->width == Static_bitmap_size);
  assert (bmp->height == Static_bitmap_size);
  assert (bmp->bits != NULL);
  for (i = 0; i < Static_bitmap_size * Static_bitmap_size; ++i)
  {
    uint32 shd = (rand () & 0xff) * 0x010101;
    bmp->bits [i] = Pal_rgb_to_index (shd);
  }
}


/*
 * Just plain static
 */
int fetch_texture_static (map_object *obj)
{

  if (obj->base_texid < 0)
  {
    ss_bitmap_header *bmp =
      Bmp_create (Static_bitmap_size, Static_bitmap_size);
    obj->base_texid = Tex_add_cache (bmp);
    Tex_set_refresh (obj->base_texid, advance_texture_static);
  }
  /*
   * We remain as the object texture fetch hook because we are also used by
   *  fetch_texture_shodan()
   */
  return obj->base_texid;

}




#define TEXT_SCREEN_BMP_SIZE 64
#define TEXT_SCREEN_FONT      3

/*
 * For the random digits in CPU rooms. These use the global Digit from 0-9
 *  set by the digit anim_control
 * We should probably have a block of 10 TEXTURE_BLOCK_UNIQUE textures and set
 *  them all up in one go, but it is hardly desperately important
 */
int fetch_texture_digit (map_object *obj)
{
  UNUSED (obj);
  if (Digit_texture [Digit] < 0)
  {
    char digit [2];
    ss_bitmap_header *bmp = Bmp_create (TEXT_SCREEN_BMP_SIZE,
                                        TEXT_SCREEN_BMP_SIZE);
    int w, h;
    digit [0] = '0' + Digit;
    digit [1] = '\0';
    memset (bmp->bits, 1, TEXT_SCREEN_BMP_SIZE * TEXT_SCREEN_BMP_SIZE);
    Font_get_size (TEXT_SCREEN_FONT, digit, &w, &h);
    Font_render_string (TEXT_SCREEN_FONT, digit,
                        (TEXT_SCREEN_BMP_SIZE - w) / 2,
                        (TEXT_SCREEN_BMP_SIZE - h) / 2,
                        TEXT_SCREEN_BMP_SIZE, TEXT_SCREEN_BMP_SIZE,
                        bmp->bits, 0);
    Digit_texture [Digit] = Tex_add_cache (bmp);
  }
  return Digit_texture [Digit];
}


/*
 * For scrolling text screens, render a screenful of text with its starting
 *  line dependent on object state
 */
ss_bitmap_header *render_scroll_text (int txtno, int state)
{

  ss_bitmap_header *bmp = Bmp_create (TEXT_SCREEN_BMP_SIZE,
                                      TEXT_SCREEN_BMP_SIZE);
  int w, h;
  int y;
  size_t size;

  txtno = (txtno & 0x7f) + state;
  Font_get_size (TEXT_SCREEN_FONT, "", &w, &h); /* get height */
  for (y = 0; y < TEXT_SCREEN_BMP_SIZE - h; y += h)
  {
    block_handle bh = Res_load_subchunk (CHUNK_TEXT_SCREEN, txtno++, &size);
    if ((bh == NULL) || (*bh == NULL))
    {
      /*
       * No string!
       */
      Cant_happen ("screen.c:fetch_texture_text(): can't load screen text");
    }
    Font_render_string (TEXT_SCREEN_FONT, *bh, 1, y, TEXT_SCREEN_BMP_SIZE,
                        TEXT_SCREEN_BMP_SIZE, bmp->bits, 0);
  }

  return bmp;

}


int fetch_texture_text (map_object *obj)
{

  ss_bitmap_header *bmp;
  int fi = Fix_check_index (obj);
  int txtno = Fixture_list [fi].start_frame & 0xff;
  int w, h; /* text string dimensions */

  /*
   * Text subchunk in the screen text chunk
   */
  int nlines = 1;
  int widest;
  int x, y;
  size_t size;
  block_handle bh;
  char *screentext;
  char *tp;
  int i;

  if (obj->base_texid >= 0) return obj->base_texid;     /* already got it */

# if SCREEN_MESSAGES
  printf ("Text screen %04x %04x %04x %04x %04x\n",
    	  Fixture_list [fi].u.num_frames, Fixture_list [fi].flags,
	  Fixture_list [fi].texture, Fixture_list [fi].start_frame,
	  Fixture_list [fi].junk);
# endif

  if (txtno & SCREEN_SPECIAL_SCROLL)
  {
    /*
     * Scrolling text ... this starts off at text string
     *  (start_frame + state) and carries on until it runs out of bitmap
     */
    obj->base_texid = Tex_get_cache_block (-1, -1,
                                           Fixture_list [fi].u.num_frames,
                                           TEXTURE_BLOCK_UNIQUE);
    for (i = 0; i < Fixture_list [fi].u.num_frames; ++i)
    {
      Tex_set_bitmap (obj->base_texid + i, render_scroll_text (txtno, i));
    }
    obj->fetch_texture_hook = Obj_fetch_texture_anim;
    return obj->base_texid + obj->state;
  }

  /*
   * Create us a bitmap, and initialise it to black
   */
  bmp = Bmp_create (TEXT_SCREEN_BMP_SIZE, TEXT_SCREEN_BMP_SIZE);
  memset (bmp->bits, 1, TEXT_SCREEN_BMP_SIZE * TEXT_SCREEN_BMP_SIZE);
  obj->base_texid = Tex_add_cache (bmp);

  bh = Res_load_subchunk (CHUNK_TEXT_SCREEN, txtno, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    printf ("%04x %04x\n", Fixture_list [fi].start_frame, Fixture_list [fi].u.num_frames);
    return -1;
  }
/*        Cant_happen ("screen.c:fetch_texture_text(): couldn't load screen text"); */
  screentext = (char *) *bh;

  for (tp = screentext; *tp != '\0'; ++tp)
  {
    if (*tp == '\n')
    {
      *tp = '\0';
      ++nlines;
    }
  }

  widest = 0;
  tp = screentext;
  for (i = 0; i < nlines; ++i)
  {
    Font_get_size (TEXT_SCREEN_FONT, tp, &w, &h);
    if (w > widest) widest = w;
    tp += strlen (tp) + 1;
  }

  x = (TEXT_SCREEN_BMP_SIZE - widest) / 2;
  y = (TEXT_SCREEN_BMP_SIZE - h * nlines) / 2;
  tp = screentext;
  for (i = 0; i < nlines; ++i)
  {
    Font_render_string (TEXT_SCREEN_FONT, tp, x, y, TEXT_SCREEN_BMP_SIZE,
                        TEXT_SCREEN_BMP_SIZE, bmp->bits, 0);
    y += h;
    tp += strlen (tp) + 1;
  }

  return obj->base_texid;

}





/*
 * Advance hook function for SHODAN screens
 */
void advance_shodan (anim_control *anim)
{

  map_object *mo = anim->obj;

  if (mo->state == 0)
  {
    /*
     * For our purposes, state 0 represents static ...
     * We might want to randomly switch to SHODAN animation, though
     */
    if ((rand () & 0xFF) >= SHODAN_EMERGE_CHANCE)
    {
      /*
       * No SHODAN this time
       */
      anim->flags |= AF_PAUSED;
    }
    else
    {
      mo->state = 1;
      anim->flags &= ~(AF_PAUSED | AF_BACKWARDS);
    }
  }

}




#if 0 /* probably put this back with improvements later */
/*
 * Advance hook function for text scroll screens
 */
void advance_text_scroll (anim_control *anim)
{

  map_object *mo = anim->obj;
  int fi = Fix_check_index (mo);
  int txtno = (Fixture_list [fi].start_frame & 0x7f) + mo->state;
  size_t size;
  block_handle bh = Res_load_subchunk (CHUNK_TEXT_SCREEN, txtno, &size);
  char *screentext;
  int w, h;
  ss_bitmap_header *bmp = mo->bmp_cache [0];

  if (bmp == NULL)
  {
    /*
     * The screen should always have had a bitmap set up before it gets an
     *  anim_control, so a NULL here is a bug
     */
    Cant_happen ("screen.c:advance_text_scroll(): no bitmap for scroll text");
  }

  if ((bh == NULL) || (*bh == NULL))
  {
    /*
     * No string!
     */
    Cant_happen ("screen.c:advance_text_scroll(): can't load screen text");
  }
  screentext = (char *) *bh;

  /*
   * Scroll up the existing text in the bitmap
   */
  Font_get_size (TEXT_SCREEN_FONT, screentext, &w, &h);
  memmove (bmp->bits, bmp->bits + h * TEXT_SCREEN_BMP_SIZE, 
	   (TEXT_SCREEN_BMP_SIZE - h) * TEXT_SCREEN_BMP_SIZE);
  memset (bmp->bits + (TEXT_SCREEN_BMP_SIZE - h) * TEXT_SCREEN_BMP_SIZE,
	  1, h * TEXT_SCREEN_BMP_SIZE);

  Font_render_string (TEXT_SCREEN_FONT, screentext,
		      1, TEXT_SCREEN_BMP_SIZE - h,
		      TEXT_SCREEN_BMP_SIZE, TEXT_SCREEN_BMP_SIZE,
		      bmp->bits, 0);

}
#endif



/*
 * Advance hook function for random-digit screens
 */
void advance_digit (anim_control *anim)
{

  /*
   * This is very simple ... all we need to do is select a new random digit
   *  for the screen(s) to display
   */
  Digit = (int) ((((double) rand()) / RAND_MAX) * 10.);

}



/*
 * Advance hook function for code screens. We might randomly want to switch
 *  between displaying an animated bitmap and displaying a code
 */
void advance_code (anim_control *anim)
{
  map_object *mo = anim->obj;

  if ((rand () & 0xff) < CODE_SWITCH_CHANCE)
  {
    int fi = Fix_check_index (mo);
    int old_frame = Fixture_list [fi].start_frame;

    /*
     * Clear out the bitmap cache for this object
     */
/*      clear_cache (mo); */

    /*
     * Swap the aux and start_frame fields to switch display type
     */
    Fixture_list [fi].start_frame = Fixture_list [fi].aux;
    Fixture_list [fi].aux = old_frame;

    /*
     * Set up the new bitmap fetch hook appropriately
     */
    choose_hook (mo, &(Fixture_list [fi]));
  }

}




/*
 * End of src/screen.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
