/*
 * shade.c -
 *  transparency at it's glory - 256 shades per colorpart
 *  plus darkness
 *
 * This file (Transparency) by Christian Haas 2001 christian.haas@eunet.at
 * Darkness by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "gamemem.h"
#include "main.h"     /* Cant_happen() */
#include "option.h"
#include "palette.h"
#include "resource.h"
#include "types.h"
#include "shade.h"

#define	TABLESIZE			131072L			  /* 512 * 256 */
#define	GET_ADDR(s)		(&(shadetable[((uint32)s << 9L) | 0x100]))

static short shadetable[TABLESIZE];

HCSHADE	shade_transparent = GET_ADDR(0x00);
HCSHADE	shade_opage				= GET_ADDR(0xFF);


/*
 * jim - this is the darkness table for lighting
 */
pixel Dark_table [DARK_SHADES] [256];

/*
 * This is "relative" gamma, in that it modifies the rate of brightness
 *  drop-off with decreasing light levels, rather than "absolute" gamma which
 *  brightens the entire palette. It can be overridden with the option
 *  `-dark-gamma' (Option_dark_gamma)
 */
float Dark_gamma = 1.2f;


/* inits the shadetable - the REAL heart of the transparency */
void shade_init(void)
{
	int sh, diff;
	HSHADE cur;
  FILE *shadefile;
  int i; /* palette index */

	memset(shadetable, 0x00, TABLESIZE * sizeof(short));

	for(sh = 0; sh < 256; sh++)
	{
		cur = GET_ADDR(sh);
		for(diff = -255; diff < 256; diff++)
		{
			cur[diff] = ((sh < abs(diff)) ? sh : abs(diff));
			if(diff < 0)
				cur[diff] *= -1;	
		}
	}

  /*
   * jim - set up the darkness table
   */
  if (Option_underworld)
  {
    shadefile = fopen (Res_make_path (UW_DATA_PATH, "light.dat"), "rb");
  }
  else
  {
    shadefile = fopen (Res_make_path (CD_DATA_PATH, "shadtabl.dat"), "rb");
  }
  if (shadefile == NULL)
  {
    Cant_happen ("shade_init(): couldn't open shade table file");
  }

  if (Option_dark_gamma > 0)
  {
    Dark_gamma = Option_dark_gamma;
  }

  /*
   * Read in the first (full bright) and last (inky blackness) shade table
   *  entries and interpolate
   */
  for (i = 0; i < 256; ++i)
  {
    Dark_table [0] [i] = Pal_index_to_pixel (fgetc (shadefile));
  }
  fseek (shadefile, -256, SEEK_END);
  for (i = 0; i < 256; ++i)
  {
    Dark_table [DARK_SHADES - 1] [i] = Pal_index_to_pixel (fgetc (shadefile));
  }

  fclose (shadefile);
  for (sh = 1; sh < DARK_SHADES - 1; ++sh)
  {
    float ll = ((float) DARK_SHADES - 1 - sh) / (DARK_SHADES - 1);
    ll = pow (ll, 1. / Dark_gamma);
    for (i = 0; i < 256; ++i)
    {
      uint32 lt = Pal_pixel_to_rgb (Dark_table [0] [i]);
      uint32 dk = Pal_pixel_to_rgb (Dark_table [DARK_SHADES - 1] [i]);
      float r = get_red   (dk) + ll * (get_red   (lt) - get_red   (dk));
      float g = get_green (dk) + ll * (get_green (lt) - get_green (dk));
      float b = get_blue  (dk) + ll * (get_blue  (lt) - get_blue  (dk));
      Dark_table [sh] [i] = Pal_rgb_to_pixel (rgb ((int) r, (int) g, (int) b));
    }
  }

}

/* get a shade                                               */
HSHADE shade_get(byte s)
{
	return GET_ADDR(s);
}




#include "map.h"        /* needed by Shd_build_static_darkmap() */

/*
 * Upper and lower dark values for each grid point
 */
static float Upper_darkmap [MAP_WIDTH + 1] [MAP_HEIGHT + 1] = {{0.0f}};
static float Lower_darkmap [MAP_WIDTH + 1] [MAP_HEIGHT + 1] = {{0.0f}};

/*
 * defined in map.c (probably should be extern'd in map.h)
 */
extern int Level_height;

/*
 * jim again - build the static dark map for a System Shock level
 */
void Shd_build_static_darkmap (map_vertex *vertex, int num_vertices)
{

  int x;
  int y;
  int v; /* vertex index */

  extern map_tile *Tile;   /* from map.c */

  /*
   * Compile the upper and lower darkmaps by averaging the upper and lower
   *  dark values respectively for each tile meeting at a grid point
   */
  for (x = 0; x < MAP_WIDTH; ++x)
  {
    for (y = 0; y < MAP_HEIGHT; ++y)
    {

      float delta_dark;
      float upper;
      float lower;
      map_tile *tile = &(Tile [x + MAP_WIDTH * y]);

      upper = (float)(GET_UPPER_SHADE (*tile) - (tile->state [3] >> 4));
      lower = (float)(GET_LOWER_SHADE (*tile) - (tile->state [3] & 0xf));
      if (tile->ceiling == tile->floor) delta_dark = 0.0f;
      else delta_dark = (upper - lower) / (tile->ceiling - tile->floor);
      lower -= delta_dark * tile->floor;
      upper += delta_dark * (Level_height - tile->ceiling);
/*        printf ("%g %g %g\n", delta_dark, upper, lower); */

      /*
       * Add to the upper darkmap
       */
/*        Upper_darkmap [x    ] [y    ] += upper; */
/*        Upper_darkmap [x + 1] [y    ] += upper; */
/*        Upper_darkmap [x    ] [y + 1] += upper; */
/*        Upper_darkmap [x + 1] [y + 1] += upper; */

      /*
       * Add to the lower darkmap
       */
/*        Lower_darkmap [x    ] [y    ] += lower; */
/*        Lower_darkmap [x + 1] [y    ] += lower; */
/*        Lower_darkmap [x    ] [y + 1] += lower; */
/*        Lower_darkmap [x + 1] [y + 1] += lower; */
      Upper_darkmap [x] [y] = upper * 4;
      Lower_darkmap [x] [y] = lower * 4;
    }
  }

  /*
   * Correct the edge and corner values
   */
  for (x = 0; x <= MAP_WIDTH; ++x)
  {
    Upper_darkmap [x] [0] *= 2.0f;
    Lower_darkmap [x] [0] *= 2.0f;
    Upper_darkmap [x] [MAP_HEIGHT] *= 2.0f;
    Lower_darkmap [y] [MAP_HEIGHT] *= 2.0f;
  }
  for (y = 0; y <= MAP_HEIGHT; ++y)
  {
    Upper_darkmap [0] [y] *= 2.0f;
    Lower_darkmap [0] [y] *= 2.0f;
    Upper_darkmap [MAP_WIDTH] [y] *= 2.0f;
    Lower_darkmap [MAP_WIDTH] [y] *= 2.0f;
  }

  /*
   * Now to set the light values for the vertices themselves
   * This is a straightforward linear interpolation between the nearest 2 or
   *  4 darkmap entries in the tables we have just calculated
   */

# define ON_GRID(x) ( \
  ((x)/TILE_GRID_SIZE - floor ((x)/TILE_GRID_SIZE) < A_VERY_SMALL_THING) || \
  (ceil ((x)/TILE_GRID_SIZE) - (x)/TILE_GRID_SIZE  < A_VERY_SMALL_THING)    \
  )

  assert (!Map_info.cyberspace); /* using vertex dark values */

  for (v = 0; v < num_vertices; ++v)
  {
    float dark;
    if (ON_GRID (vertex [v].x))
    {
      x = (int)(floor ((vertex [v].x + 1.) / TILE_GRID_SIZE));
      if (ON_GRID (vertex [v].y))
      {
	/*
	 * We're on a grid post, so a single interpolation between floor and
	 *  ceiling values is ok
	 */
	y = (int)(floor ((vertex [v].y + 1.) / TILE_GRID_SIZE));
	dark = (float)(Lower_darkmap [x] [y] +
	  (vertex [v].z / Level_height) *
	  (Upper_darkmap [x] [y] - Lower_darkmap [x] [y]));
	vertex [v].u.dark = (int)((dark / 60.) * DARK_SHADES);
      }
      else
      {
	/*
	 * Need a bilinear interpolation in y and z
	 */
	float dark1, dark2;
	y = (int)(floor (vertex [v].y / TILE_GRID_SIZE));
	dark1 = (float)(Lower_darkmap [x] [y] +
	  (vertex [v].z / Level_height) *
	  (Upper_darkmap [x] [y] - Lower_darkmap [x] [y]));
	dark2 = (float)(Lower_darkmap [x] [y + 1] +
	  (vertex [v].z / Level_height) *
	  (Upper_darkmap [x] [y + 1] - Lower_darkmap [x] [y + 1]));
	dark = (float)(dark1 + (vertex [v].y / TILE_GRID_SIZE - y) * (dark2 - dark1));
	vertex [v].u.dark = (int)((dark / 60.) * DARK_SHADES);
      }
    }
    else if (ON_GRID (vertex [v].y))
    {
      /*
       * Need a bilinear interpolation in x and z
       */
      float dark1, dark2;
      x = (int)(floor (vertex [v].x / TILE_GRID_SIZE));
      dark1 = (float)(Lower_darkmap [x] [y] +
	(vertex [v].z / Level_height) *
	(Upper_darkmap [x] [y] - Lower_darkmap [x] [y]));
      dark2 = (float)(Lower_darkmap [x + 1] [y] +
	  (vertex [v].z / Level_height) *
	  (Upper_darkmap [x + 1] [y] - Lower_darkmap [x + 1] [y]));
      dark = dark1 + (float)((vertex [v].x / TILE_GRID_SIZE - x) * (dark2 - dark1));
      vertex [v].u.dark = (int)((dark / 60.) * DARK_SHADES);
    }
    else
    {
      Cant_happen
	("Shd_build_static_darkmap(): vertex in illegal grid position");
    }
    if (vertex [v].u.dark >= DARK_SHADES) vertex [v].u.dark = INKY_BLACKNESS;
    if (vertex [v].u.dark <  0          ) vertex [v].u.dark = FULL_BRIGHT;
  }

}




/*
 * Get the dark value for a point on a post (a tile corner in x and y)
 *  x and y parameters are in whole tiles
 */
float Shd_get_dark_post (int x, int y, double z)
{
  if ((x < 0) || (x > MAP_WIDTH) || (y < 0) || (y > MAP_HEIGHT))
    Cant_happen ("Shd_get_dark_post(): post position off map");

  return (float) (Lower_darkmap [x] [y] + (z / Level_height) *
		  (Upper_darkmap [x] [y] - Lower_darkmap [x] [y]));
}



/*
 * Get the static darkmap value for an arbitrary point, using an inverse-square
 *  weighted average of the 4 dark posts for that tile
 * Probably very slow, but we shouldn't be calling it often, if at all, after
 *  the map has been set up
 */
float Shd_get_static_dark (double x, double y, double z)
{

  int post_x = (int) (x / TILE_GRID_SIZE);
  int post_y = (int) (y / TILE_GRID_SIZE);
  double rx = x - TILE_GRID_SIZE * post_x;
  double ry = y - TILE_GRID_SIZE * post_y;
  double ir2 [4];  /* inverse r^2 */
  double sum = 0.;
  float dark [4];
  float result = 0.;
  int i;

  if ((post_x < 0) || (post_x >= MAP_WIDTH) ||
      (post_y < 0) || (post_y >= MAP_HEIGHT))
  {
/*      Cant_happen ("Shd_get_static_dark(): post position off map"); */
    return 0;
  }

  /*
   * Collect the 4 dark post values at this z for our weighted average
   */
  dark [0] = Shd_get_dark_post (post_x,     post_y,     z);
  dark [1] = Shd_get_dark_post (post_x + 1, post_y,     z);
  dark [2] = Shd_get_dark_post (post_x + 1, post_y + 1, z);
  dark [3] = Shd_get_dark_post (post_x,     post_y + 1, z);

  /*
   * Inverse r values, first x^2
   */
  ir2 [0] = ir2 [3] = rx * rx;
  rx = TILE_GRID_SIZE - rx;
  ir2 [1] = ir2 [2] = rx * rx;

  /*
   * Now add in y^2 (and an extra 1 to avoid divides-by-zero)
   */
  ir2 [0] += 1 + ry * ry;
  ir2 [1] += 1 + ry * ry;
  ry = TILE_GRID_SIZE - ry;
  ir2 [2] += 1 + ry * ry;
  ir2 [3] += 1 + ry * ry;

  /*
   * Invert and collect the sum
   */
  for (i = 0; i < 4; ++i)
  {
    ir2 [i] = 1. / ir2 [i];
    sum += ir2 [i];
  }

  /*
   * And perform the weighted average proper
   */
  for (i = 0; i < 4; ++i)
  {
    result += dark [i] * ir2 [i] / sum;
  }

  /*
   * FIXME This is a fudge. Ought to investigate why bogus light values are
   *  possible at all and fix where appropriate
   */
  if (result < 0) result = 0;
  if (result >= DARK_SHADES - 1) result = DARK_SHADES - 1;
  return result;

}
