/*
 * surface.c -
 *  The surface cache
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "gamemem.h"
#include "graph2d.h"
#include "main.h"
#include "map.h"
#include "surface.h"





/*
 * #defines and constants
 */
#define SURFACE_CACHE_SIZE 1024




/*
 * Global variables
 */
surface *Surface_cache;
int Next_surface;




void Sur_init (void)
{
  Surface_cache = Alloc_init (SURFACE_CACHE_SIZE * sizeof (surface),
			      "Sur_init", "surface cache");
  Next_surface = 0;
}

void Sur_deinit(void)
{
	free(Surface_cache);
	Surface_cache = NULL;
}



/*
 * Reset the surface stack
 */
void Sur_reset (void)
{
  Next_surface = 0;
}



surface *Sur_next (void)
{

  return &(Surface_cache [Next_surface]);

}



void Sur_push (void)
{

  if (Next_surface >= SURFACE_CACHE_SIZE)
  {
    Cant_happen ("Sur_push(): too many surfaces, increase SURFACE_CACHE_SIZE");
  }

  ++Next_surface;

}


void Sur_pop (void)
{

  if (Next_surface == 0)
  {
    Cant_happen ("Sur_pop(): surface cache underflow");
  }

  --Next_surface;

}



surface *Sur_peek (void)
{

  if (Next_surface == 0)
  {
    return NULL;
  }
  return &(Surface_cache [Next_surface - 1]);

}




/*
 * Check that a surface is flat i.e. that all its vertices are coplanar.
 * Returns true if this is the case
 */
bool Sur_check_flatness (const surface *s)
{

  Vector tv;
  int i;

  if (s->num_vertices < 4)
  {
    /*
     * A surface with fewer than 4 vertices cannot be other than flat
     */
    return true;
  }

  for (i = 0; i < s->num_vertices; ++i)
  {
    int next = (i + 1) % s->num_vertices;
    vec_subtract (tv, s->vertex [next], s->vertex [i]);
    if (fabs (vec_dot (tv, s->normal)) > A_VERY_SMALL_THING)
    {
      return false;
    }
  }

  return true;

}




void Sur_map_to_surface (surface *s, const map_surface *ms)
{
  int v;
  extern map_vertex *Vertex;

  assert (s != NULL);
  assert (ms != NULL);

  s->tile = -1; /* no valid tile yet */

  s->texture = ms->texture; /* might not be valid */
  s->num_vertices = ms->num_vertices;

  assert (!Map_info.cyberspace); /* using vertex dark value */
  for (v = 0; v < s->num_vertices; ++v)
  {
    map_vertex *mv = &(Vertex [ms->vertex [v]]);
    s->vertex [v].x = mv->x;
    s->vertex [v].y = mv->y;
    s->vertex [v].z = mv->z;
    s->dark [v] = 0.5 + mv->u.dark; /* 1/2 to DARK_SHADES - 1/2 */
  }
  s->normal = ms->normal;
  s->anchor = ms->anchor;
  s->u      = ms->u;
  s->v      = ms->v;

  s->flags = ms->flags;

}




void Sur_copy (surface *dst, const surface *src)
{
  int i;

  dst->tile         = src->tile;
  dst->texture      = src->texture;
  dst->shade        = src->shade;
  dst->num_vertices = src->num_vertices;

  for (i = 0; i < dst->num_vertices; ++i)
  {
    dst->vertex [i] = src->vertex [i];
    dst->dark   [i] = src->dark   [i];
  }

  dst->anchor       = src->anchor;
  dst->normal       = src->normal;
  dst->u            = src->u;
  dst->v            = src->v;
  dst->flags        = src->flags;
}
