/*
 * surface.h -
 *  Definitions for the surface cache
 */

/*
 * File:    src/surface.h
 * Purpose: Definitions for 3D surfaces (including map_surface)
 * Prefix:  Sur_
 * Note:    
 * Music:   Fairport Convention, "Red & Gold"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


#ifndef SURFACE_H
#define SURFACE_H




/*
 * #includes
 */
#include "graph2d.h"  /* point */
#include "types.h"
#include "vec.h"



#define MAX_VERTEX_SURFACE 32
#define MAX_VERTEX_MAP      6



/*
 * Surface flags
 */
#define SF_TEXTURED     0x01  /* is texture-mapped               */
#define SF_MASKED       0x02  /* colour 0 is transparent         */
#define SF_ZBUFFER      0x04  /* uses the Z-buffer               */
#define SF_SHADED       0x08  /* Gouraud shaded                  */
#define SF_TRANSPARENCY 0x10  /* firefreak's cute transparencies */

#define SF_NOCLIP       0x20  /* for portals, NO clipping volume */


typedef struct tag_surface
{

  /*
   * The tile to which the surface belongs
   */
  int tile;

  /*
   * The texture and/or shade
   */
  int texture;
  pcolor shade;

  /*
   * How many vertices do we have
   */
  int num_vertices;

  /*
   * The coordinates of the corners
   */
  vec vertex [MAX_VERTEX_SURFACE];

  /*
   * Lighting values for each corner vertex
   */
  float dark [MAX_VERTEX_SURFACE];

  /*
   * Anchor point for the texture in 3D space. This is a point somewhere on
   *  the plane of the surface which is at u,v coordinates (0,0) (not
   *  necessarily within the surface)
   */
  vec anchor;

  /*
   * The normal vector
   */
  vec normal;

  /*
   * The vectors u and v in view coordinates
   */
  vec u;
  vec v;

  /*
   * Rendering flags
   */
  int flags;

} surface;


/*
 * map_surface flags
 * Mustn't overlap with the SF_ flags above ... best to use a different byte
 */
#define MSF_PORTAL       0x0100  /* can see through to next tile */
#define MSF_2SIDED       0x0200  /* drawn from both sides (door) */


struct tag_map_tile;

typedef struct tag_map_surface
{

  int texture;                 /* valid if texture-mapped                  */
  int num_vertices;            /* should be self-explanatory               */
  int vertex [MAX_VERTEX_MAP]; /* index into main vertex list              */
  int tile_next;               /* in linked list of surfaces for this tile */
  int flags;                   /* MSF_ flags above                         */
  vec normal;                  /* normalised to magnitude 1                */
  vec anchor;                  /* a point in worldspace with (u,v) = (0,0) */
  vec u;                       /* the remaining 2 vectors which define the */
  vec v;                       /*  texture mapping                         */
  struct tag_map_tile *behind; /* valid if it's a portal                   */

} map_surface;




/*
 * A 3D line segment, used for cyberspace (and map notes, when they exist)
 */
typedef struct tag_line_seg
{
  vec v1;
  vec v2;
  uint32 rgb1; /* RGB colour value for v1 */
  uint32 rgb2; /* RGB colour value for v2 */
} line_seg;




/*
 * A rasterised surface for the use of the renderer ... this consists of start
 *  and end x, start 1/z for a number of scanlines
 */
typedef struct
{

  int num_lines;
  struct raster_line
  {
    int y;            /* screen y coord in pixels from top */
    int start_x;
    int end_x;
    float start_dark;
    float end_dark;
    double zinv;
  } *line;

} raster_surface;




#ifdef __cplusplus
extern "C"
{
#endif



void Sur_init (void);
void Sur_deinit(void);
void Sur_reset (void);
surface *Sur_next (void);
void Sur_push (void);
void Sur_pop (void);
surface *Sur_peek (void);
bool Sur_check_flatness (const surface *s);
void Sur_map_to_surface (surface *s, const map_surface *ms);
void Sur_copy (surface *dest, const surface *source);

#ifdef __cplusplus
}
#endif




/*
 * External global variables defined in surface.c
 */



#endif /* SURFACE_H */

