/*
 * texture.c -
 *  Textures
 */

/*
 * File:    src/texture.c
 * Purpose: Texture management
 * Prefix:  Tex_
 * Note:    
 * Music:   Toto, "Toto IV"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 */



/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "types.h"
#include "accel.h"
#include "anim.h"
#include "bitmap.h"
#include "chunks.h"
#include "game.h"
#include "gamemem.h"
#include "main.h"
#include "map.h"
#include "option.h"
#include "palette.h"
#include "plot.h"
#include "render.h"
#include "resource.h"
#include "shade.h"
#include "surface.h"
#include "texture.h"


/*
 * Debug
 */
#define TEXTURE_MESSAGES 0



/*
 * #defines and macros
 */
#define MAX_TEXTURE 272

#define UW_WALL_TEXTURE_FILE  Res_make_path (UW_DATA_PATH, "w64.tr")
#define UW_FLOOR_TEXTURE_FILE Res_make_path (UW_DATA_PATH, "f32.tr")
#define UW2_TEXTURE_FILE      Res_make_path (UW_DATA_PATH, "t64.tr")





/*
 * The texture cache. Each level may have up to a maximum of 64 textures which
 *  are pre-cached on level load (under 1.5MB)
 */
#define TCF_ANIMATED  0x01 /* is self-animating (wall texture)    */
#define TCF_DIRECTION 0x02 /* is currently in reverse             */
#define TCF_REVERSE   0x04 /* is reversible (back and forth)      */
#define TCF_REFRESH   0x08 /* call its refresh hook when accessed */

typedef struct
{
  int chunk;     /* bitmap chunk (wall texs have hi-res base chunkno) */
  int subchunk;  /* subchunk or offset (chunk/subchunk unique entry)  */
  int texanim;   /* index into the texanims list, if animated         */
  int flags;     /* for animations                                    */
  int bmpno;     /* index into the internal bitmaps list              */
  tex_refresh_func refresh_hook;
} texture_cache_entry;

texture_cache_entry *Texture_cache;
static const int Max_tex_cache = 512;
int Num_tex_cache; /* how many cache slots are actually occupied   */


#define MAX_TEX_BITMAP 512
static ss_bitmap_header **Texture_bitmap;
int Num_tex_bitmaps;


static int Num_textures;
static texture_prop *Texture_prop_list;



/*
 * Animated textures
 */
struct texanim
{
  int texid;      /* Texture ID in texture cache */
  int start_bmp;  /* base bitmap in internal bitmaps list */
  int num_frames;
  int frame;      /* current frame 0 to num_frames - 1    */
};

#define MAX_TEXANIM 8
static struct texanim *Texture_anim;
static int Num_texanims;

#define MAX_TEXANIM_SEQ 8
static int Texanim_last [MAX_TEXANIM_SEQ];


raster_surface Scratch_raster;



/*
 * Local function prototypes
 */
static int find_cache_texture (int chunk, int subchunk, int start);
static int new_tex_bitmap (void);


void load_ss_texture_prop (void)
{

  FILE *texpropfile = fopen (Res_make_path (CD_DATA_PATH, "textprop.dat"),
                             "rb");
  int i;
  long len;
  fseek (texpropfile, 0, SEEK_END);
  len = ftell (texpropfile);

  /*
   * The texture properties file looks to be truncated ... I THINK I've got the
   *  right properties applied to the right textures
   */
  if (len % TEXPROP_FILE_SIZE != 0)
  {
    Cant_happen ("Tex_init(): not whole number of texture properties");
  }
  Num_textures = len / TEXPROP_FILE_SIZE - 1;
  Texture_prop_list = Alloc_init (Num_textures * sizeof *Texture_prop_list,
                                  "Tex_init", "texture properties");

  fseek (texpropfile, 4, SEEK_SET);
  for (i = 0; i < Num_textures; ++i)
  {

    texture_prop *tp = &(Texture_prop_list [i]);

    fseek (texpropfile, 2, SEEK_CUR);
    READ_LE_LONG (texpropfile, tp->stuff);
    tp->climbable  = fgetc (texpropfile);
    fgetc (texpropfile);
    tp->starfield  = fgetc (texpropfile);
    tp->anim_group = fgetc (texpropfile);
    tp->anim_no    = fgetc (texpropfile);
    tp->type       = TERRAIN_SOLID;
#   if TEXTURE_MESSAGES
    printf ("%04x %08lx ", i, tp->stuff);
    if (tp->climbable) printf ("c "); else printf ("  ");
    printf ("%02x %02x %02x\n", tp->starfield, tp->anim_group, tp->anim_no);
#   endif

    assert (tp->anim_group < MAX_TEXANIM_SEQ); 
    if (tp->anim_no > Texanim_last [tp->anim_group])
    {
      Texanim_last [tp->anim_group] = tp->anim_no;
    }
  }

  fclose (texpropfile);

}



void load_uw_texture_prop (void)
{

  FILE *file = fopen (Res_make_path (UW_DATA_PATH, "terrain.dat"), "rb");
  int i;

  if (file == NULL)
  {
    Cant_happen ("texture.c:load_uw_texture_prop():"
                 " couldn't open terrain file");
  }    

  /*
   * We allow 512 texture properties for Underworld, though there aren't really
   *  that many textures
   */
  Num_textures = 512;
  Texture_prop_list = Alloc_init (Num_textures * sizeof *Texture_prop_list,
                                  "Tex_init", "texture properties");

  for (i = 0; i < Num_textures; ++i)
  {
    Texture_prop_list [i].stuff      = 10; /* same as Shock */
    Texture_prop_list [i].climbable  = false;
    Texture_prop_list [i].anim_group = 0;
    Texture_prop_list [i].anim_no    = 0;
    READ_LE_SHORT (file, Texture_prop_list [i].type);
  }

}


void Tex_init (void)
{

  extern int Screen_height;

  /*
   * Texture properties
   */
  if (Option_underworld) load_uw_texture_prop ();
  else load_ss_texture_prop ();

  /*
   * Allocate space for the texture cache and initialise the pointers to
   *  within it
   */
  Texture_cache = Alloc_init (Max_tex_cache * sizeof *Texture_cache,
                              "Tex_init", "texture cache");

  Texture_bitmap = Alloc_init (MAX_TEX_BITMAP * sizeof *Texture_bitmap,
                               "Tex_init", "texture bitmaps");

  Texture_anim = Alloc_init (MAX_TEXANIM * sizeof *Texture_anim,
                             "Tex_init", "texture animations");

  /*
   * Scratch space for the rasteriser
   * Should be in Plot_init() except that there isn't one, of course
   */
  Scratch_raster.line = Alloc_init
    (Screen_height * sizeof *(Scratch_raster.line), "Tex_init", "rasteriser");

}





void Tex_exit (void)
{

  int i;

	free(Scratch_raster.line);
	Scratch_raster.line = NULL;

  for (i = 0; i < Num_tex_bitmaps; ++i)
  {
    Bmp_free_bitmap (Texture_bitmap [i]);
  }
  free (Texture_bitmap);
	Texture_bitmap = NULL;
  free (Texture_cache);
  Texture_cache = NULL;
}




/*
 * Used by load_uw_texture_cache()
 */
void load_uw_texture (int texid, FILE *file, int texno, int size)
{
  unsigned long offs;

  fseek (file, 4 + texno * 4, SEEK_SET);
  READ_LE_LONG (file, offs);
  Texture_bitmap [Num_tex_bitmaps] = Bmp_create (size, size);
  fseek (file, offs, SEEK_SET);
  fread (Texture_bitmap [Num_tex_bitmaps]->bits, 1, size*size, file);

  Texture_cache [texid].chunk    = CHUNK_TEXTURE_128_BASE;
  Texture_cache [texid].subchunk = texno + (size == 32) * 256;
  Texture_cache [texid].texanim  = -1;
  Texture_cache [texid].flags    = 0;
  Texture_cache [texid].bmpno    = Num_tex_bitmaps++;
  Texture_cache [texid].refresh_hook = NULL;
}



void load_uw_texture_cache (int level_no)
{

  FILE *levfile   = fopen (UW_MAP_FILE, "rb");
  FILE *wallfile  = fopen (UW_WALL_TEXTURE_FILE, "rb");
  FILE *floorfile = fopen (UW_FLOOR_TEXTURE_FILE, "rb");
  size_t texsize;
  int ntex = 64;
  int i;
  byte *chunk = NULL;

  if (wallfile == NULL)
  {
    /*
     * Couldn't open Underworld texture files, try for Underworld II
     */
    wallfile = fopen (UW2_TEXTURE_FILE, "rb");
    if (wallfile == NULL)
    {
      Cant_happen ("Tex_init_level(): can't open UW2 file");
    }
    chunk = Res_load_uw2_map (levfile, level_no, 1, &texsize);
  }
  else if ((levfile == NULL) || (floorfile == NULL))
  {
    Cant_happen ("Tex_init_level(): can't open UW file");
  }
  else
  {
    /*
     * We're in the Stygian Abyss, so find the level archive
     * Surely there should be a Res_ function for this?
     */
    long offs;
    fseek (levfile, 0x4A + 4 * level_no, SEEK_SET);
    ntex = 48; /* 48 wall textures */
    READ_LE_LONG (levfile, offs);
    fseek (levfile, offs, SEEK_SET);
    chunk = malloc (0x86);
    fread (chunk, 1, 0x86, levfile);
  }

  for (i = 0; i < ntex; ++i)
  {
    load_uw_texture (i, wallfile, chunk [2*i] + (chunk [2*i + 1] << 8), 64);
  }
  for (i = ntex; i < 58; ++i)
  {
    load_uw_texture (i, floorfile, chunk [2*i] + (chunk [2*i + 1] << 8), 32);
  }
  for (; i < 64; ++i)
  {
    int doorno = chunk [116 + i - 58];
    Texture_bitmap [Num_tex_bitmaps] =
      Bmp_load_bitmap (12, doorno);
    Texture_cache [i].chunk    = 12;
    Texture_cache [i].subchunk = doorno;
    Texture_cache [i].texanim  = -1;
    Texture_cache [i].flags    = 0;
    Texture_cache [i].bmpno    = Num_tex_bitmaps++;
    Texture_cache [i].refresh_hook = NULL;
  }
  Num_tex_cache = 64;

  if (floorfile != NULL) fclose (floorfile);
  fclose (wallfile);
  fclose (levfile);
  free (chunk);

}


/*
 * Used by load_ss_texture_cache()
 */
void load_ss_cache_texture (int cno, int texno)
{

  assert (cno < Max_tex_cache);
  if (texno > MAX_TEXTURE)
  {
    Cant_happen ("Tex_init_level(): bad texture");
  }

  Texture_bitmap [Num_tex_bitmaps] =
    Bmp_load_bitmap (CHUNK_TEXTURE_128_BASE + texno, 0);

  Texture_cache [cno].chunk    = CHUNK_TEXTURE_128_BASE;
  Texture_cache [cno].subchunk = texno;
  Texture_cache [cno].texanim  = -1;
  Texture_cache [cno].flags    = 0;
  Texture_cache [cno].bmpno    = Num_tex_bitmaps++;
  Texture_cache [cno].refresh_hook = NULL;

}


void load_ss_texture_cache (int level_no)
{

  int ntex;
  int i;
  byte *chunk = NULL;
/*    unsigned long anim_present [MAX_TEXANIM_SEQ] = {0}; */

  size_t size;
  block_handle bh = Res_load_chunk_by_id
    (4000 + 100*level_no + MAP_CHUNK_TEXTURE, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Tex_init_level(): failed to load texture list");
  }

  ntex = size / 2;
  if (ntex > 64)
  {
    Cant_happen ("Tex_init_level(): too many textures for level");
  }

  /*
   * Load and cache each texture that the level refers to
   */
  chunk = *bh;
  for (i = 0; i < ntex; ++i)
  {
    int texno = chunk [2*i] + (chunk [2*i + 1] << 8);
    load_ss_cache_texture (i, texno);
#   if TEXTURE_MESSAGES
    printf ("Tex_init_level(): got texture %d\n", texno);
#   endif
  }
  Num_tex_cache = ntex;

  printf ("Tex_init_level(): got %d textures\n", Num_tex_cache);

}



/*
 * Find a global texture reference in the texture cache
 */
int find_cache_texture (int chunk, int subchunk, int start)
{
  int i;
  for (i = start; i < Num_tex_cache; ++i)
  {
    if ((Texture_cache [i].chunk == chunk) &&
        (Texture_cache [i].subchunk == subchunk)) return i;
  }
  return -1;
}



/*
 * This is trivial at the moment, but might become less so if we get purgeable
 *  textures
 */
int new_tex_bitmap (void)
{
  if (Num_tex_bitmaps >= MAX_TEX_BITMAP)
  {
    Cant_happen ("texture.c:new_tex_bitmap(): too many bitmaps");
  }
  Texture_bitmap [Num_tex_bitmaps] = NULL;
  return Num_tex_bitmaps++;
}




/*
 * Animation advance hook for animated textures
 */
static void advance_texanim (anim_control *ac)
{
  int i;
  UNUSED (ac); /* morally justified, since we're a callback */

  for (i = 0; i < Num_texanims; ++i)
  {

    int texid = Texture_anim [i].texid;
    int flags = Texture_cache [texid].flags;

    if (flags & TCF_REVERSE)
    {
      if (flags & TCF_DIRECTION)
      {
        if (Texture_anim [i].frame == 0)
        {
          Texture_cache [texid].flags &= ~TCF_DIRECTION;
        }
        else --Texture_anim [i].frame;
      }
      else
      {
        if (Texture_anim [i].frame == Texture_anim [i].num_frames - 1)
        {
          Texture_cache [texid].flags |= TCF_DIRECTION;
        }
        else ++Texture_anim [i].frame;
      }
    }
    else
    {
      ++Texture_anim [i].frame;
      if (Texture_anim [i].frame == Texture_anim [i].num_frames)
      {
        Texture_anim [i].frame = 0;
      }
    }
    Texture_cache [texid].bmpno =
      Texture_anim [i].start_bmp + Texture_anim [i].frame;
  }

}



/*
 * Initialise the texture cache for a level. This assumes that no textures have
 *  yet been loaded into the texture cache
 */
void Tex_init_level (int level_no)
{

  int i;        /* misc. loop index */
  int base_bmp   = 0;
  int cur_seq    = 0;
  int next_frame = 0; /* animation stuff */

  /*
   * Get the cached textures out of wherever they live
   */
  if (Option_underworld)
  {
    load_uw_texture_cache (level_no);
  }
  else
  {
    load_ss_texture_cache (level_no);
  }

  /*
   * Set up the animation info in the cache
   */
  for (i = 0; i < Num_tex_cache; ++i)
  {
    int texno = Texture_cache [i].subchunk;
    if (Texture_prop_list [texno].anim_group > 0)
    {
      /*
       * It's an animated texture! Check that it's in the proper sequence
       */
      if ((Texture_prop_list [texno].anim_no != next_frame) ||
          (next_frame != 0 && Texture_prop_list [texno].anim_group != cur_seq))
      {
        Cant_happen ("Tex_init_level(): non-sequential texanims");
      }

      if (next_frame == 0) /* first frame this anim */
      {
        base_bmp = Texture_cache [i].bmpno;
        cur_seq  = Texture_prop_list [texno].anim_group;
      }

      /*
       * Give it a texanim
       */
      if (Num_texanims >= MAX_TEXANIM)
      {
        Cant_happen ("Tex_init_level(): too many texanims");
      }
      Texture_anim [Num_texanims].texid      = i;
      Texture_anim [Num_texanims].start_bmp  = base_bmp;
      Texture_anim [Num_texanims].num_frames = Texanim_last [cur_seq] + 1;
      Texture_anim [Num_texanims].frame      = next_frame;

      printf ("Animated texture %d %d %d %d\n", i, base_bmp,
              Texture_anim [Num_texanims].num_frames, next_frame);

      Texture_cache [i].texanim = Num_texanims++;
      Texture_cache [i].flags   = TCF_ANIMATED | TCF_REVERSE;

      ++next_frame;
      if (next_frame > Texanim_last [cur_seq]) next_frame = 0;
    }
    else if (next_frame != 0)
    {
      Cant_happen ("Tex_init_level(): non-sequential texanims");
    }

  }

  if (Num_texanims > 0)
  {
    /*
     * Set up a fake anim_control to handle all the animated textures
     * This gets a NULL object but is kept paused always so it doesn't try
     *  to advance object state
     */
    anim_control *ac = Anim_add (0.5, NULL, 0, AF_PAUSED);
    ac->advance_hook = advance_texanim;
  }

}




/*
 * Find a texture cache ID for a bitmap chunk / subchunk. Allocates a cache
 *  entry and an internal bitmap number if one doesn't yet exist, but doesn't
 *  necessarily load the bitmap
 */
int Tex_get_cache_id (int chunk, int subchunk)
{

  int texid = find_cache_texture (chunk, subchunk, 0);
  if (texid < 0)
  {
    if (Num_tex_cache >= Max_tex_cache)
    {
      Cant_happen ("Tex_get_cache_id(): too many textures");
    }
    texid = Num_tex_cache++;
    Texture_cache [texid].chunk    = chunk;
    Texture_cache [texid].subchunk = subchunk;
    Texture_cache [texid].texanim  = -1;
    Texture_cache [texid].flags    = 0;
    Texture_cache [texid].bmpno    = -1;
    Texture_cache [texid].refresh_hook = NULL;
  }

  return texid;

}




/*
 * Allocate a block of textures starting at the given chunk / subchunk
 * Note that if it finds an existing block it will check all the cache entries
 *  in the block to make sure that they are all present and in order ...
 *  shouldn't be a real problem, just don't go allocating 1000-texture blocks
 *  every frame, ok?
 */
int Tex_get_cache_block (int chunk, int subchunk, int ntexs, int which)
{

  int i;
  int texid;

  if (which == TEXTURE_BLOCK_UNIQUE) /* don't check these */
  {
    chunk = subchunk = -1;
  }
  else
  {

    texid = find_cache_texture (chunk, subchunk, 0);
    while (texid >= 0)
    {
      /*
       * Found a potential block. See if it contains the right chunks
       */
      bool found = true;
      int tchunk = chunk;
      int tsub   = subchunk;
      for (i = 0; i < ntexs; ++i)
      {
        if ((Texture_cache [texid + i].chunk != tchunk) ||
            (Texture_cache [texid + i].subchunk != tsub))
        {
          found = false;
          break;
        }
        if (which == TEXTURE_BLOCK_CHUNK) ++tchunk; else ++tsub;
      }
      if (found) return texid; /* hooray! */
      texid = find_cache_texture (chunk, subchunk, texid);
    }
  }

  /*
   * OK, we don't already have such a block. Allocate it and set the cache
   * entries accordingly
   */
  if (Num_tex_cache + ntexs > Max_tex_cache)
  {
    Cant_happen ("Tex_get_cache_block(): too many textures");
  }
  texid = Num_tex_cache;
  for (i = 0; i < ntexs; ++i)
  {
    Texture_cache [texid + i].chunk    = chunk;
    Texture_cache [texid + i].subchunk = subchunk;
    Texture_cache [texid + i].texanim  = -1;
    Texture_cache [texid + i].flags    = 0;
    Texture_cache [texid + i].bmpno    = -1;
    Texture_cache [texid + i].refresh_hook = NULL;
    if (which == TEXTURE_BLOCK_CHUNK) ++chunk;
    else if (which == TEXTURE_BLOCK_SUBCHUNK) ++subchunk;
  }
  Num_tex_cache += ntexs;
  return texid;
}




/*
 * Set the bitmap for a given texture cache entry to one prepared by the
 *  caller. Caller relinquishes responsibility. This may only be called once
 *  for any given texture id
 */
void Tex_set_bitmap (int texid, ss_bitmap_header *bmp)
{

  int bmpno;

  if (texid < 0 || texid > Num_tex_cache)
  {
    Cant_happen ("Tex_set_bitmap(): bad texture number");
  }

  bmpno = Texture_cache [texid].bmpno;
  if (bmpno >= 0)
  {
    Cant_happen ("Tex_set_bitmap(): tried to set bitmap twice");
  }

  bmpno = new_tex_bitmap ();
  Texture_cache [texid].bmpno = bmpno;
  Texture_bitmap [bmpno] = bmp;

}



/*
 * Add a bitmap prepared by the caller to the texture cache. Caller gives up
 *  responsibility for freeing -- it'll be done on texture cache cleanup
 */
int Tex_add_cache (ss_bitmap_header *bmp)
{
  int texid, bmpno;

  if (Num_tex_cache >= Max_tex_cache)
  {
    Cant_happen ("Tex_get_cache_id(): too many textures");
  }

  bmpno = new_tex_bitmap ();
  Texture_bitmap [bmpno] = bmp;

  texid = Num_tex_cache++;
  Texture_cache [texid].chunk    = -1;
  Texture_cache [texid].subchunk = -1;
  Texture_cache [texid].texanim  = -1;
  Texture_cache [texid].flags    = 0;
  Texture_cache [texid].bmpno    = bmpno;
  Texture_cache [texid].refresh_hook = NULL;

  return texid;
}



/*
 * Get a texture bitmap from its texid. Loads up the bitmap if we haven't got
 *  it yet
 */
ss_bitmap_header *Tex_get_texture (int texno)
{

  int bmpno;
  ss_bitmap_header *bmp;

  if (texno < 0 || texno > Num_tex_cache)
  {
    Cant_happen ("Tex_get_texture(): bad texture number");
  }

  bmpno = Texture_cache [texno].bmpno;
  if (bmpno < 0)
  {
    bmpno = new_tex_bitmap ();
    Texture_cache [texno].bmpno = bmpno;
  }

  bmp = Texture_bitmap [bmpno];
  if (bmp == NULL)
  {
    bmp = Bmp_load_bitmap (Texture_cache [texno].chunk,
                           Texture_cache [texno].subchunk);
    if (bmp == NULL)
    {
      Cant_happen ("Tex_get_texture(): couldn't load texture bitmap");
    }
    Texture_bitmap [bmpno] = bmp;
  }

  if (Texture_cache [texno].flags & TCF_REFRESH)
  {
    /*
     * Need to refresh this bitmap each time it is accessed
     */
    assert (Texture_cache [texno].refresh_hook != NULL);
    Texture_cache [texno].refresh_hook (bmp);
  }
  return bmp;

}


/*
 * Get (wall or floor) texture properties from a texid
 */
const texture_prop *Tex_get_prop (int texid)
{
  int texno;
  if ((texid < 0) || (texid >= Num_tex_cache))
  {
    Cant_happen ("Tex_get_prop(): bad texture id");
  }
  if (Texture_cache [texid].chunk != CHUNK_TEXTURE_128_BASE)
  {
    Cant_happen ("Tex_get_prop(): not a terrain texture");
  }
  texno = Texture_cache [texid].subchunk;
  assert ((texno >= 0) && (texno < Num_textures));
  return &(Texture_prop_list [texno]);
}




/*
 * Set a refresh hook for a texture
 * Also sets the TCF_REFRESH flag
 */
void Tex_set_refresh (int texid, tex_refresh_func hook)
{
  if ((texid < 0) || (texid >= Num_tex_cache))
  {
    Cant_happen ("Tex_set_refresh(): bad texture id");
  }
  Texture_cache [texid].flags |= TCF_REFRESH;
  Texture_cache [texid].refresh_hook = hook;
}



ss_bitmap_header *Tex_load_uw_texture (int texno)
{

  FILE *wallfile;
  ss_bitmap_header *bmp;
  unsigned long offs;

  if (!Option_underworld)
  {
    Cant_happen ("Tex_load_uw_texture(): not Underworld level");
  }

  wallfile = fopen (Underworld2 ? UW2_TEXTURE_FILE : UW_WALL_TEXTURE_FILE,
		    "rb");
  if (wallfile == NULL)
  {
    Cant_happen ("Tex_load_uw_texture(): can't open texture file");
  }

  printf ("%02x ", fgetc (wallfile));
  printf ("%02x ", fgetc (wallfile));
  printf ("%02x ", fgetc (wallfile));
  printf ("%02x ", fgetc (wallfile));

  bmp = Bmp_create (64, 64);

  fseek (wallfile, 4 + texno * 4, SEEK_SET);
  READ_LE_LONG (wallfile, offs);
  fseek (wallfile, offs, SEEK_SET);
  fread (bmp->bits, 1, 64*64, wallfile);

  fclose (wallfile);
  return bmp;

}








/*
 * End of src/texture.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
