/*
 * uwmodel.c -
 *  Interpreter functions for Underworld3D models
 */

/*
 * File:    src/uwmodel.c
 * Purpose: Underworld model file interpreter
 * Prefix:  UW_mod_
 * Note:    
 * Music:   Iron Maiden, "Live After Death CD"
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>

#include "types.h"
#include "main.h"     /* Cant_happen() */
#include "model.h"
#include "resource.h"
#include "shade.h"
#include "uwmodel.h"


/*
 * Debugging
 */
#define UW_MODEL_MESSAGES 1


/*
 * Local function prototypes
 */
static void load_uw_model_node (model *m, model_node *n, FILE *uwexe);



/*
 * #defines and constants
 */

/*
 * For variable-height models
 */
#define VH_THRESHOLD 32768.f     /* max System Shock level height is 16384 */
#define VH_VALUE     40000.f     /* comfortably more than VH_THRESHOLD     */




/*
 * Underworld models are (unfortunately) buried deep inside the executable
 */
struct
{
  unsigned long table_offset;
  unsigned long value;          /* 4 bytes at table_offset */
  unsigned long base_offset;
} UW_model_offset [] =
{
  { 0x0004e910, 0x40064ab6, 0x0004e99e },
  { 0x0004ccd0, 0x40064ab6, 0x0004cd5e }, /* same models, different place */
  { 0x00054cf0, 0x59aa64d4, 0x00054d8a },
  { 0x000550e0, 0x59aa64d4, 0x0005517a }, /* another UW2 build            */
  { 0, 0, 0 }
};

#define READ_UW_M3_FLOAT(f,file) do {        \
  int READ_UW_M3_FLOAT_t;                    \
  READ_LE_SHORT (file, READ_UW_M3_FLOAT_t);  \
  if (READ_UW_M3_FLOAT_t & 0x8000) READ_UW_M3_FLOAT_t -= 0x10000; \
  f = READ_UW_M3_FLOAT_t; } while (0)

#define READ_UW_M3_VERTNO(i,file) do { \
  READ_LE_SHORT (file, i);             \
  i /= 8; } while (0)



model *UW_mod_load (int model_no)
{

  FILE *uwexe;
  model *m;
  int offset;
  int i;
  uint32 test;

  uwexe = fopen (Res_make_path (UW_BASE_PATH, "uw.exe"), "rb");
  if (uwexe == NULL)
    uwexe = fopen (Res_make_path (UW_BASE_PATH, "uw2.exe"), "rb");
  if (uwexe == NULL)
    uwexe = fopen (Res_make_path (UW_BASE_PATH, "ultimau1.exe"), "rb");
  if (uwexe == NULL)
  {
    Cant_happen ("model.c:decode_uw_model(): can't open uw.exe");
  }

  m = malloc (sizeof *m);
  m->num_vertices = 0;
  m->flags        = 0;

  /*
   * Try to find the model in the model table
   */
  for (i = 0; UW_model_offset [i].table_offset != 0; ++i)
  {
    unsigned long check;
    fseek (uwexe, UW_model_offset [i].table_offset, SEEK_SET);
    READ_LE_LONG (uwexe, check);
    if (check == UW_model_offset [i].value)
    {
      /*
       * Found it!
       */
      break;
    }
  }
  if (UW_model_offset [i].table_offset == 0)
  {
    Cant_happen ("model.c:decode_uw_model(): can't find model table");
  }

  fseek (uwexe, UW_model_offset [i].table_offset + 2 * model_no, SEEK_SET);
  READ_LE_SHORT (uwexe, offset);
  fseek (uwexe, UW_model_offset [i].base_offset + offset, SEEK_SET);
# if UW_MODEL_MESSAGES
  printf ("Model %d at offset %04x %08lx\n", model_no, offset,
	  UW_model_offset [i].base_offset + offset);
# endif

  /*
   * Header is 10 bytes, being 32-bit ??diameter?? and an extent vector
   */
  READ_LE_LONG (uwexe, test);
# if UW_MODEL_MESSAGES
  printf ("%08lx\n", test);
# endif
  READ_UW_M3_FLOAT (m->extent.x, uwexe);
  READ_UW_M3_FLOAT (m->extent.z, uwexe);
  READ_UW_M3_FLOAT (m->extent.y, uwexe);
# if UW_MODEL_MESSAGES
  printf ("Extent\t(%g, %g, %g)\n", m->extent.x, m->extent.y, m->extent.z);
# endif

  /*
   * "Centre" point is not defined in the header, but might be specified as a
   *  command later
   */
  m->centre.x = 0.; m->centre.y = 0.; m->centre.z = 0.;

  /*
   * Load the root node, and the others recursively
   */
  load_uw_model_node (m, &(m->root), uwexe);

  fclose (uwexe);
  return m;

}



/*
 * Recursively load a node from the Underworld model file
 */
void load_uw_model_node (model *m, model_node *n, FILE *uwexe)
{

  int left_pos  = -1;
  int right_pos = -1;

  bool in_face = false;
  static int default_shade = 0;
  int face_shade    = default_shade;   /* shade value, current face  */
  int face_length = 0;  /* length of face information */

  n->num_surfaces = 0;
  n->left  = NULL;
  n->right = NULL;

  while (1)
  {
    Vector tv;
    int vertno;   /* vertex no. */
    int command;
    double vert_offs;
    model_surface ts; /* surface for current face   */
    double u [MAX_VERTEX_MODEL];
    double v [MAX_VERTEX_MODEL];

    int i;

    READ_LE_SHORT (uwexe, command);
    if (command == 0) break;
    switch (command)
    {

    case M3_UW_SORT_PLANE: /* split into sub-nodes */
    {
      Vector normal, p;
      READ_UW_M3_FLOAT (normal.x, uwexe);
      READ_UW_M3_FLOAT (p.x, uwexe);
      READ_UW_M3_FLOAT (normal.z, uwexe);
      READ_UW_M3_FLOAT (p.z, uwexe);
      READ_UW_M3_FLOAT (normal.y, uwexe);
      READ_UW_M3_FLOAT (p.y, uwexe);
      vec_normalise (normal);
#     if UW_MODEL_MESSAGES
      printf ("%02x Sort plane ", command);
      printf ("(%g, %g, %g) (%g, %g, %g)\n", normal.x, normal.y, normal.z,
	      p.x, p.y, p.z);
#     endif
      n->sort_plane.normal = normal;
      n->sort_plane.np     = -vec_dot (normal, p);
      READ_LE_SHORT (uwexe, left_pos);
      left_pos  += ftell (uwexe);
      READ_LE_SHORT (uwexe, right_pos);
      right_pos += ftell (uwexe);
      break;
    }

    case M3_UW_SORT_PLANE_ZY:
    {
#     if UW_MODEL_MESSAGES
      printf ("%02x Sort plane\n", command);
#     endif
      fseek (uwexe, 8, SEEK_CUR);
      READ_LE_SHORT (uwexe, left_pos);
      left_pos  += ftell (uwexe);
      READ_LE_SHORT (uwexe, right_pos);
      right_pos += ftell (uwexe);
      break;
    }

    case M3_UW_SORT_PLANE_XY:
    {
#     if UW_MODEL_MESSAGES
      printf ("%02x Sort plane\n", command);
#     endif
      fseek (uwexe, 8, SEEK_CUR);
      READ_LE_SHORT (uwexe, left_pos);
      left_pos  += ftell (uwexe);
      READ_LE_SHORT (uwexe, right_pos);
      right_pos += ftell (uwexe);
      break;
    }

    case M3_UW_SORT_PLANE_XZ:
    {
#     if UW_MODEL_MESSAGES
      printf ("%02x Sort plane\n", command);
#     endif
      fseek (uwexe, 8, SEEK_CUR);
      READ_LE_SHORT (uwexe, left_pos);
      left_pos  += ftell (uwexe);
      READ_LE_SHORT (uwexe, right_pos);
      right_pos += ftell (uwexe);
      break;
    }

    case 0x14: /* don't know what this is, might have to do with colour */
      READ_LE_SHORT (uwexe, vertno);
#     if UW_MODEL_MESSAGES
      printf ("%02x ", command);
      printf ("%04x ", vertno);
      printf ("%02x ", fgetc (uwexe));
      printf ("%02x ", fgetc (uwexe));
#     endif
      READ_LE_SHORT (uwexe, vertno);
#     if UW_MODEL_MESSAGES
      printf ("%04x\n", vertno);
#     endif
      break;

    case 0x2e:
#     if UW_MODEL_MESSAGES
      printf ("%02x\n", command);
#     endif
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      fseek (uwexe, 2, SEEK_CUR);
      face_length -= 4;
      break;

    case 0x40: /* seems to do nothing but introduce a face definition */
      printf ("%02x\n", command);
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      face_length -= 2;
      break;

    case 0x44: /* this one too ... */
      printf ("%02x\n", command);
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      face_length -= 2;
      break;

    case M3_UW_FACE_PLANE:
      READ_LE_SHORT (uwexe, face_length);
      READ_UW_M3_FLOAT (ts.normal.x, uwexe);
      READ_UW_M3_FLOAT (tv.x, uwexe);
      READ_UW_M3_FLOAT (ts.normal.z, uwexe);
      READ_UW_M3_FLOAT (tv.z, uwexe);
      READ_UW_M3_FLOAT (ts.normal.y, uwexe);
      READ_UW_M3_FLOAT (tv.y, uwexe);
      vec_normalise (ts.normal);
      ts.flags = 0;
      in_face = true;
      printf ("%02x Face plane (%g, %g, %g) (%g, %g, %g) %g\n", command,
	      ts.normal.x, ts.normal.y, ts.normal.z,
	      tv.x, tv.y, tv.z, vec_dot (tv, ts.normal));
      break;

    case M3_UW_FACE_PLANE_ZY:
      READ_LE_SHORT (uwexe, face_length);
      ts.normal.x = 0.;
      READ_UW_M3_FLOAT (ts.normal.z, uwexe); fseek (uwexe, 2, SEEK_CUR);
      READ_UW_M3_FLOAT (ts.normal.y, uwexe); fseek (uwexe, 2, SEEK_CUR);
      vec_normalise (ts.normal);
      ts.flags = 0;
      in_face = true;
      printf ("%02x Face plane (%g, %g, %g)\n", command,
	      ts.normal.x, ts.normal.y, ts.normal.z);
      break;

    case M3_UW_FACE_PLANE_XY:
      READ_LE_SHORT (uwexe, face_length);
      READ_UW_M3_FLOAT (ts.normal.x, uwexe); fseek (uwexe, 2, SEEK_CUR);
      READ_UW_M3_FLOAT (ts.normal.y, uwexe); fseek (uwexe, 2, SEEK_CUR);
      ts.normal.z = 0.;
      vec_normalise (ts.normal);
      ts.flags = 0;
      in_face = true;
      printf ("%02x Face plane (%g, %g, %g)\n", command,
	      ts.normal.x, ts.normal.y, ts.normal.z);
      break;

    case M3_UW_FACE_PLANE_XZ:
      READ_LE_SHORT (uwexe, face_length);
      READ_UW_M3_FLOAT (ts.normal.x, uwexe); fseek (uwexe, 2, SEEK_CUR);
      ts.normal.y = 0.;
      READ_UW_M3_FLOAT (ts.normal.z, uwexe); fseek (uwexe, 2, SEEK_CUR);
      vec_normalise (ts.normal);
      ts.flags = 0;
      in_face = true;
      printf ("%02x Face plane (%g, %g, %g)\n", command,
	      ts.normal.x, ts.normal.y, ts.normal.z);
      break;

    case M3_UW_FACE_PLANE_X:
      READ_LE_SHORT (uwexe, face_length);
      in_face = true;
      READ_UW_M3_FLOAT (ts.normal.x, uwexe);
      ts.normal.y = 0.; ts.normal.z = 0.; vec_normalise (ts.normal);
      ts.flags = 0;
      fseek (uwexe, 2, SEEK_CUR); /* don't use face distance ... */
      printf ("%02x Face plane (%g, %g, %g)\n", command,
	      ts.normal.x, ts.normal.y, ts.normal.z);
      break;

    case M3_UW_FACE_PLANE_Y:
      READ_LE_SHORT (uwexe, face_length);
      in_face = true;
      READ_UW_M3_FLOAT (ts.normal.y, uwexe);
      ts.normal.x = 0.; ts.normal.z = 0.; vec_normalise (ts.normal);
      ts.flags = 0;
      fseek (uwexe, 2, SEEK_CUR); /* don't use face distance ... */
      printf ("%02x Face plane (%g, %g, %g)\n", command,
	      ts.normal.x, ts.normal.y, ts.normal.z);
      break;

    case M3_UW_FACE_PLANE_Z:
      READ_LE_SHORT (uwexe, face_length);
      in_face = true;
      READ_UW_M3_FLOAT (ts.normal.z, uwexe);
      ts.normal.x = 0.; ts.normal.y = 0.; vec_normalise (ts.normal);
      ts.flags = 0;
      fseek (uwexe, 2, SEEK_CUR); /* don't use face distance ... */
      printf ("%02x Face plane (%g, %g, %g)\n", command,
	      ts.normal.x, ts.normal.y, ts.normal.z);
      break;

    case M3_UW_ORIGIN:
    {
      int test;
      READ_UW_M3_VERTNO (vertno, uwexe);
      m->centre = m->vertex [vertno];
      m->num_vertices = 0;
      READ_LE_SHORT (uwexe, test); printf ("%d,", test);
      READ_LE_SHORT (uwexe, test); printf ("%d,", test);
      READ_LE_SHORT (uwexe, test); printf ("%d\n", test);
      READ_LE_SHORT (uwexe, test); printf ("%d\n", test);
      printf ("%02x Origin %d\n", command, vertno);
      break;
    }

    case M3_UW_VERTEX:
      READ_UW_M3_FLOAT  (tv.x, uwexe);
      READ_UW_M3_FLOAT  (tv.y, uwexe);
      READ_UW_M3_FLOAT  (tv.z, uwexe);
      READ_UW_M3_VERTNO (vertno, uwexe);
      m->vertex [m->num_vertices++] = tv;
      printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
	      vertno);
      break;

    case M3_UW_FACE_VERTICES:
    {
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      READ_LE_SHORT (uwexe, ts.num_vertices);
      printf ("%02x Face [ ", command);
      face_length -= 4;
      for (i = 0; i < ts.num_vertices; ++i)
      {
	READ_UW_M3_VERTNO (ts.vertex [i], uwexe);
	printf ("%d ", ts.vertex [i]);
	face_length -= 2;
      }
      printf ("]\n");
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("model.c:decode_uw_model(): too many surfaces");
      }
      ts.flags |= M3F_FLAT;
      ts.colour = face_shade;
      printf ("%d\n", ts.colour);
      ts.anchor = m->vertex [ts.vertex [0]];
      n->surface [n->num_surfaces++] = ts;
      break;      
    }

    case M3_UW_VERTICES:
    {
      int nvert;
      READ_LE_SHORT (uwexe, nvert);
      fseek (uwexe, 2, SEEK_CUR); /* don't know what this does */
      for (i = 0; i < nvert; ++i)
      {
	READ_UW_M3_FLOAT  (tv.x, uwexe);
	READ_UW_M3_FLOAT  (tv.y, uwexe);
	READ_UW_M3_FLOAT  (tv.z, uwexe);
	m->vertex [m->num_vertices++] = tv;
	printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
		i);
      }
      break;
    }

    case M3_UW_VERTEX_X:
      READ_UW_M3_VERTNO (vertno,    uwexe);
      READ_UW_M3_FLOAT  (vert_offs, uwexe);
      tv = m->vertex [vertno];
      tv.x += vert_offs;
      m->vertex [m->num_vertices++] = tv;
      READ_UW_M3_VERTNO (vertno,    uwexe);
      printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
	      vertno);
      break;

    case M3_UW_VERTEX_Y:
      READ_UW_M3_VERTNO (vertno,    uwexe);
      READ_UW_M3_FLOAT  (vert_offs, uwexe);
      tv = m->vertex [vertno];
      tv.y += vert_offs;
      m->vertex [m->num_vertices++] = tv;
      READ_UW_M3_VERTNO (vertno,    uwexe);
      printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
	      vertno);
      break;

    case M3_UW_VERTEX_Z:
      READ_UW_M3_VERTNO (vertno,    uwexe);
      READ_UW_M3_FLOAT  (vert_offs, uwexe);
      tv = m->vertex [vertno];
      tv.z += vert_offs;
      m->vertex [m->num_vertices++] = tv;
      READ_UW_M3_VERTNO (vertno,    uwexe);
      printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
	      vertno);
      break;

      /*
       * I have a very sneaky suspicion that this one refers to variable size
       *  things ... that pillar, for example
       */
    case M3_UW_VERTEX_ROOF:
    {
      READ_UW_M3_VERTNO (vertno, uwexe);
      tv = m->vertex [vertno];
      tv.z = VH_VALUE;
      READ_LE_SHORT (uwexe, vertno);
      printf ("%04x ", vertno);
      READ_UW_M3_VERTNO (vertno, uwexe);
      printf ("%02x Vertex (%g, %g, varheight) \t%d\n", command, tv.x, tv.y,
	      vertno);
      m->vertex [m->num_vertices++] = tv;
      m->flags |= M3_VAR_HEIGHT;
      break;
    }

    case M3_UW_VERTEX_XY:   /* vertex offset X, Y        */
    {
      int offs_x;
      int offs_y;
      READ_LE_SHORT (uwexe, offs_x);
      READ_LE_SHORT (uwexe, offs_y);
      if (offs_x & 0x8000) offs_x -= 0x10000;
      if (offs_y & 0x8000) offs_y -= 0x10000;
      READ_UW_M3_VERTNO (vertno, uwexe);
      tv.x = m->vertex [vertno].x + offs_x;
      tv.y = m->vertex [vertno].y + offs_y;
      tv.z = m->vertex [vertno].z;
      m->vertex [m->num_vertices++] = tv;
      READ_UW_M3_VERTNO (vertno,    uwexe);
      printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
	      vertno);
      break;
    }

    case M3_UW_VERTEX_XZ:   /* vertex offset X, Z        */
    {
      int offs_x;
      int offs_z;
      READ_LE_SHORT (uwexe, offs_x);
      READ_LE_SHORT (uwexe, offs_z);
      if (offs_x & 0x8000) offs_x -= 0x10000;
      if (offs_z & 0x8000) offs_z -= 0x10000;
      READ_UW_M3_VERTNO (vertno, uwexe);
      tv.x = m->vertex [vertno].x + offs_x;
      tv.y = m->vertex [vertno].y;
      tv.z = m->vertex [vertno].z + offs_z;
      m->vertex [m->num_vertices++] = tv;
      READ_UW_M3_VERTNO (vertno,    uwexe);
      printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
	      vertno);
      break;
    }

    case M3_UW_VERTEX_YZ:   /* vertex offset Y, Z        */
    {
      int offs_y;
      int offs_z;
      READ_LE_SHORT (uwexe, offs_y);
      READ_LE_SHORT (uwexe, offs_z);
      if (offs_y & 0x8000) offs_y -= 0x10000;
      if (offs_z & 0x8000) offs_z -= 0x10000;
      READ_UW_M3_VERTNO (vertno, uwexe);
      tv.x = m->vertex [vertno].x;
      tv.y = m->vertex [vertno].y + offs_y;
      tv.z = m->vertex [vertno].z + offs_z;
      m->vertex [m->num_vertices++] = tv;
      READ_UW_M3_VERTNO (vertno,    uwexe);
      printf ("%02x Vertex (%g, %g, %g) \t%d\n", command, tv.x, tv.y, tv.z,
	      vertno);
      break;
    }

    case 0xa0: /* shorthand face definition? */
    {
/*        Vector t1_v; */
/*        Vector t2_v; */
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      printf ("%02x Face [ ", command);
      READ_LE_SHORT (uwexe, vertno);
      face_length -= 2;
      ts.num_vertices = 4;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [0] = vertno;
      u [vertno] =   0; v [vertno] =   0;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [1] = vertno;
      u [vertno] = 256; v [vertno] =   0;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [2] = vertno;
      u [vertno] = 256; v [vertno] = 256;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [3] = vertno;
      u [vertno] =   0; v [vertno] = 256;
      face_length -= 4;
      printf ("] %04x\n", vertno);
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("model.c:decode_uw_model(): too many surfaces");
      }
      Mod_uv2anchor_vec (m, &ts, u, v);
      ts.flags |= M3F_TEXTURED | M3F_OBJ_TEXTURE;
      n->surface [n->num_surfaces++] = ts;
      break;      
    }

    case M3_UW_TEXTURE_FACE:
    {
      int tno;
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      READ_LE_SHORT (uwexe, tno);
      READ_LE_SHORT (uwexe, ts.num_vertices);
      printf ("%02x Face %d [ ", command, tno);
      face_length -= 4;
      for (i = 0; i < ts.num_vertices; ++i)
      {
	int vno;
	READ_UW_M3_VERTNO (vno, uwexe);
	printf ("%d ", vno);
	ts.vertex [i] = vno;
	READ_LE_SHORT (uwexe, u [vno]); u [vno] /= 65535./256;
	READ_LE_SHORT (uwexe, v [vno]); v [vno] /= 65535./256;
	printf ("(%g,%g) ", u [vno], v [vno]);
	face_length -= 6;
      }
      printf ("]\n");
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("model.c:decode_uw_model(): too many surfaces");
      }

      Mod_uv2anchor_vec (m, &ts, u, v);
      ts.flags |= M3F_TEXTURED | M3F_OBJ_TEXTURE;
      ts.texture = tno;
/*        ts.anchor = m->vertex [ts.vertex [0]]; */
      n->surface [n->num_surfaces++] = ts;
      break;      
    }

    case 0xb2:
      /*
       * Haven't a clue what this is
       */
      READ_LE_SHORT (uwexe, vertno);
      printf ("%02x %02x\n", command, vertno);
      break;

    case M3_UW_TMAP_VERTICES:
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      READ_LE_SHORT (uwexe, ts.num_vertices);
      printf ("%02x Face [ ", command);
      face_length -= 4;
      for (i = 0; i < ts.num_vertices; ++i)
      {
	int vno;
	READ_UW_M3_VERTNO (vno, uwexe);
	printf ("%d ", vno);
	ts.vertex [i] = vno;
	READ_LE_SHORT (uwexe, u [vno]); u [vno] /= 65535./256;
	READ_LE_SHORT (uwexe, v [vno]); v [vno] /= 65535./256;
	printf ("(%g,%g) ", u [vno], v [vno]);
	face_length -= 6;
      }
      printf ("]\n");
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("model.c:decode_uw_model(): too many surfaces");
      }
      Mod_uv2anchor_vec (m, &ts, u, v);
      ts.flags |= M3F_TEXTURED | M3F_OBJ_TEXTURE;
/*        ts.anchor = m->vertex [ts.vertex [0]]; */
      n->surface [n->num_surfaces++] = ts;
      break;      

    case M3_UW_FACE_SHADE:
    {
      int c1 = fgetc (uwexe);
      int c2 = fgetc (uwexe);
      int c3 = fgetc (uwexe);
      fgetc (uwexe);
      face_shade = c1;
      printf ("%02x Face shade %02x %02x %02x\n", command, c1, c2, c3);
      if (in_face) face_length -= 6;
      break;
    }

    case 0xbe: /* seems to define 2 shades */
      face_shade = fgetc (uwexe);
      fseek (uwexe, 3, SEEK_CUR);
      printf ("%02x %d\n", command, face_shade);
      if (in_face) face_length -= 6;
      break;

    case 0xce: /* yet another texture-mapped face */
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      READ_LE_SHORT (uwexe, ts.num_vertices);
      printf ("%02x Face [ ", command);
      face_length -= 4;
      for (i = 0; i < ts.num_vertices; ++i)
      {
	int vno;
	READ_UW_M3_VERTNO (vno, uwexe);
	printf ("%d ", vno);
	ts.vertex [i] = vno;
	READ_LE_SHORT (uwexe, u [vno]); u [vno] /= 65535./256;
	READ_LE_SHORT (uwexe, v [vno]); v [vno] /= 65535./256;
	printf ("(%g,%g) ", u [vno], v [vno]);
	face_length -= 6;
      }
      printf ("]\n");
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("model.c:decode_uw_model(): too many surfaces");
      }
      Mod_uv2anchor_vec (m, &ts, u, v);
      ts.flags |= M3F_TEXTURED | M3F_OBJ_TEXTURE;
/*        ts.anchor = m->vertex [ts.vertex [0]]; */
      n->surface [n->num_surfaces++] = ts;
      break;      

    case 0xd2: /* shorthand face definition? */
    {
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      printf ("%02x Face [ ", command);
      ts.num_vertices = 4;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [0] = vertno;
      u [vertno] =   0; v [vertno] =   0;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [1] = vertno;
      u [vertno] = 256; v [vertno] =   0;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [2] = vertno;
      u [vertno] = 256; v [vertno] = 256;
      vertno = fgetc (uwexe);
      printf ("%d ", vertno);
      ts.vertex [3] = vertno;
      u [vertno] =   0; v [vertno] = 256;
      face_length -= 4;
      printf ("]\n");
      if (n->num_surfaces >= MAX_SURFACE_NODE)
      {
	Cant_happen ("model.c:decode_uw_model(): too many surfaces");
      }
      Mod_uv2anchor_vec (m, &ts, u, v);
      ts.flags |= M3F_TEXTURED | M3F_OBJ_TEXTURE;
      n->surface [n->num_surfaces++] = ts;
      break;      
    }

    case M3_UW_VERTEX_DARK:
    {
      int nvert, c1, c2, i;
      READ_LE_SHORT (uwexe, nvert);
      c1 = fgetc (uwexe);
      c2 = fgetc (uwexe);
#     if UW_MODEL_MESSAGES
      printf ("%02x Vertex shading %d %02x %02x\n", command, nvert, c1, c2);
#     endif
      for (i = 0; i < nvert; ++i)
      {
	int v;
	READ_UW_M3_VERTNO (v, uwexe);
	assert (v < m->num_vertices);
	m->dark [v] = fgetc (uwexe) * DARK_SHADES / 8.;
      }
      if (nvert & 1) fgetc (uwexe);
      default_shade = c2;
      face_shade = default_shade;
      break;
    }

    case M3_UW_FACE_GOURAUD:
      printf ("%02x Gouraud face\n", command);
      if (!in_face) Cant_happen ("model.c:decode_uw_model(): not in face");
      ts.flags |= M3F_SHADED;
      face_length -= 2;
      break;

    default:
      printf ("%02x\n", command);
      Cant_happen ("model.c:decode_uw_model(): unknown command");
    }

    if (in_face)
    {
      if (face_length < 0)
      {
	Cant_happen ("model.c:decode_uw_model(): face length mismatch");
      }
      else if (face_length == 0)
      {
	/*
	 * Finish off the face
	 */
	in_face = false;
	face_shade = default_shade;
      }
    }

  }

  printf ("NODE %d\n\n", n->num_surfaces);

  /*
   * Partition any non-convex faces into convex sub-polys
   */
  Mod_partition_node (m, n);

  /*
   * Load up the child nodes, if present
   */
  if (left_pos >= 0)
  {
    n->left = malloc (sizeof *n->left);
    if (n->left == NULL)
    {
      Cant_happen ("model.c:load_uw_model_node(): can't allocate child node");
    }
    fseek (uwexe, left_pos, SEEK_SET);
    load_uw_model_node (m, n->left, uwexe);

    n->right = malloc (sizeof *n->right);
    if (n->right == NULL)
    {
      Cant_happen ("model.c:load_uw_model_node(): can't allocate child node");
    }
    fseek (uwexe, right_pos, SEEK_SET);
    load_uw_model_node (m, n->right, uwexe);

  }

}
