/*
 * weapon.c -
 *  Weapons, grenades, bullets, clips
 */

/*
 * File:    src/weapon.c
 * Purpose: Class-specific routines for weaponry, classes 0-3
 * Prefix:  Wep_
 * Note:    This file is responsible for weapons, clips, grenades and tracers
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>

#include "object.h"
#include "resource.h"
#include "weapon.h"


#if __DEBUG__
#define WEAPON_MESSAGES 0
#endif


#define PROJECTILE_NUM_MISSILES 16



/*
 * The master weapon properties for each of the 16 weapons in class 0
 */
weapon_prop Weapon_prop_list [NUM_WEAPONS];


/*
 * The master weapon properties for the 15 clips in class 1
 */
weapon_prop Ammo_prop_list [NUM_AMMO_CLIPS];

/*
 * And for the 8 explosives in class 3
 */
weapon_prop Explosive_prop_list [NUM_EXPLOSIVES];

/*
 * Colour scheme for the projectiles class 2/1
 */
byte Projectile_colour [16] [6];


/*
 * Local function prototypes
 */
void load_weapon_genprop (FILE *propfile, weapon_prop *wp);




/*
 * File offsets for classes 0-3 in objprop.dat
 * These are here so we can make sure we've got the right amount of data
 */
#define OBJPROP_OFFSET_WEAPON     0x0004
#define OBJPROP_OFFSET_AMMO       0x00B0
#define OBJPROP_OFFSET_PROJECTILE 0x0191
#define OBJPROP_OFFSET_EXPLOSIVE  0x0283


/*
 * Load up weapon data from objprop.dat / objects.dat
 * We are passed the file with its pointer at the beginning of the weapon
 *  properties, and must leave it at the end for the next class (patches)
 */
void Wep_load_objprop (FILE *propfile)
{

  int i; /* misc. loop index */

  /*
   * Trust No One ...
   */
  assert (ftell (propfile) == OBJPROP_OFFSET_WEAPON);

  /*
   * First comes the generic weapon properties, 2 bytes per weapon
   */
  for (i = 0; i < NUM_WEAPONS; ++i)
  {
    int c;
    c = fgetc (propfile); /* printf ("%02x ", c); */
    c = fgetc (propfile);
    /*
     * The second byte determines which types of ammo clip the weapon accepts
     */
    Weapon_prop_list [i].clipsub   = c >> 4;
    Weapon_prop_list [i].cliptypes = c & 0xf;
  }
/*    printf ("\n"); */

  /*
   * Weapon properties for subclass 0 (semi-auto) and 1 (full auto). These take
   *  their properties from the ammo, so have no information stored with them,
   *  just a placeholder byte each
   */
  fseek (propfile, 7, SEEK_CUR);

  /*
   * Weapon properties for subclass 2 (slow projectile)
   */
  for (i = 7; i < 9; ++i)
  {
    int c;

    /*
     * Common weapon properties
     */
    load_weapon_genprop (propfile, &(Weapon_prop_list [i]));

    /*
     * Specific to this subclass
     */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    READ_LE_LONG (propfile, Weapon_prop_list [i].projectile);
#   if WEAPON_MESSAGES
    printf ("shoots %lu/%lu/%lu ", Weapon_prop_list [i].projectile >> 16,
	    (Weapon_prop_list [i].projectile & 0xff00) >> 8,
	    Weapon_prop_list [i].projectile & 0xff);
#   endif
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
/*      printf ("\n"); */
  }

  /*
   * Weapon properties for subclass 3 (melee) and 4 (energy beam)
   */
  for (i = 9; i < 14; ++i)
  {
    int c;

    /*
     * Common weapon properties
     */
    load_weapon_genprop (propfile, &(Weapon_prop_list [i]));

    /*
     * Specific to this subclass
     */
    Weapon_prop_list [i].energ_rnds = fgetc (propfile);
#   if WEAPON_MESSAGES
    printf ("energy %d ", Weapon_prop_list [i].energ_rnds);
#   endif
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
/*      printf ("\n"); */
  }

  /*
   * Weapon properties for subclass 5 (energy projectile)
   */
  for (i = 14; i < 16; ++i)
  {
    int c;

    /*
     * Common weapon properties
     */
    load_weapon_genprop (propfile, &(Weapon_prop_list [i]));

    /*
     * Specific to this subclass
     */
    Weapon_prop_list [i].energ_rnds = fgetc (propfile);
#   if WEAPON_MESSAGES
    printf ("energy %d ", Weapon_prop_list [i].energ_rnds);
#   endif
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    READ_LE_SHORT (propfile, Weapon_prop_list [i].stuff2);
#   if WEAPON_MESSAGES
    printf ("??? %d ", Weapon_prop_list [i].stuff2);
#   endif
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    READ_LE_LONG (propfile, Weapon_prop_list [i].projectile);
#   if WEAPON_MESSAGES
    printf ("shoots %lu/%lu/%lu ", Weapon_prop_list [i].projectile >> 16,
	    (Weapon_prop_list [i].projectile & 0xff00) >> 8,
	    Weapon_prop_list [i].projectile & 0xff);
#   endif
    c = fgetc (propfile); /*  printf ("%02x ", c); */
/*      printf ("\n"); */
  }


  /*
   * Now for the ammo clips. This is very similar
   */
  assert (ftell (propfile) == OBJPROP_OFFSET_AMMO);
  for (i = 0; i < NUM_AMMO_CLIPS; ++i)
  {
    int c;

    /*
     * Common weapon properties
     */
    load_weapon_genprop (propfile, &(Ammo_prop_list [i]));

    /*
     * Specific to ammo clips
     */
    Ammo_prop_list [i].energ_rnds = fgetc (propfile); /* rounds per clip */
#   if WEAPON_MESSAGES
    printf ("%d rounds/clip ", Ammo_prop_list [i].energ_rnds);
#   endif
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    READ_LE_SHORT (propfile, Ammo_prop_list [i].stuff2);
#   if WEAPON_MESSAGES
    printf ("??? %d ", Ammo_prop_list [i].stuff2);
#   endif
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
/*      printf ("\n"); */
  }

  /*
   * There is nothing of specific interest in the file for ammo clips, just a
   *  placeholder byte each
   */
  fseek (propfile, 15, SEEK_CUR);

  /*
   * Now bullets, missiles, tracers
   */
  assert (ftell (propfile) == OBJPROP_OFFSET_PROJECTILE);
# if WEAPON_MESSAGES
  printf ("\nProjectiles class 2\n");
# endif

  /*
   * The generic class 2 section holds a byte per projectile, purpose unknown
   *  at present
   */
  for (i = 0; i < 24; ++i)
  {
    int c;

    c = fgetc (propfile); /*  printf ("%02x ", c); */
  }
/*    printf ("\n"); */

  /*
   * There are 6 objects in class 2/0 (tracers), with 20 bytes each reserved
   *  in the objprop file, all zero
   */
  fseek (propfile, 20*6, SEEK_CUR);

  /*
   * Class 2/1 (projectiles) has 16 objects, with 6 bytes each giving the
   *  colour scheme (only used for cyberspace)
   */
  for (i = 0; i < 16; ++i)
  {
    fread (Projectile_colour [i], 1, 6, propfile);
#   if WEAPON_MESSAGES
    printf ("colour %02x %02x %02x %02x %02x %02x\n",
	    Projectile_colour [i] [0], Projectile_colour [i] [1],
	    Projectile_colour [i] [2], Projectile_colour [i] [3],
	    Projectile_colour [i] [4], Projectile_colour [i] [5]);
#   endif
  }

  /*
   * Class 2/2 has no interesting objprop data
   */
  fseek (propfile, 2, SEEK_CUR);

  /*
   * Grenades and explosives, class 3
   */
  assert (ftell (propfile) == OBJPROP_OFFSET_EXPLOSIVE);
# if WEAPON_MESSAGES
  printf ("\nExplosives class 3\n");
# endif
  for (i = 0; i < NUM_EXPLOSIVES; ++i)
  {
    int c;

    load_weapon_genprop (propfile, &(Explosive_prop_list [i]));
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
/*      printf ("\n"); */

    /*
     * The thought strikes me that it might be useful to make an explosive's
     *  "projectile" itself ...
     */
  }

  /*
   * Grenades class 3/0. No useful info here
   */
  fseek (propfile, 5, SEEK_CUR);

  /*
   * Explosives class 3/1. 3 bytes per explosive type, purpose unknown
   */
# if WEAPON_MESSAGES
  for (i = 0; i < 3; ++i)
  {
    int c;

    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
    c = fgetc (propfile); /*  printf ("%02x ", c); */
/*      printf ("\n"); */
  }
# else
  fseek (propfile, 9, SEEK_CUR);
# endif

}




byte *Wep_get_colour (map_object *obj)
{

  /*
   * Only projectiles are of interest here; they are the only weapon types
   *  which exist in cyberspace
   */
  assert (obj->class == OBJCLASS_PROJECTILE);
  if (obj->subclass != PROJECTILE_SUBCLASS_MISSILE) return NULL;
  assert (obj->type < PROJECTILE_NUM_MISSILES);
  return Projectile_colour [obj->type];

}



void load_weapon_genprop (FILE *propfile, weapon_prop *wp)
{

  READ_LE_SHORT (propfile, wp->damage);
  wp->offence = fgetc (propfile);
  wp->type    = fgetc (propfile);
  wp->special = fgetc (propfile);
  READ_LE_SHORT (propfile, wp->stuff1);
  wp->penetration = fgetc (propfile);

# if WEAPON_MESSAGES
  printf ("damage %d offence %d type %d special %02x ??? %04x ap %d\n",
	  wp->damage, wp->offence, wp->type, wp->special, wp->stuff1,
	  wp->penetration);
# endif

}

