/*
 * playlog.c -
 *  System Shock logfile player
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "../src/types.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <unistd.h>

#include "../src/resource.h"


#define DEBUG_SOUND 0

/*
 * This is log (base 2) of the fragment size for the sound driver
 */
#define FRAGMENT_BITS 9
#define NUM_FRAGMENTS 2



int DSP_fd;
int Sample_rate = -1;
int Fragment_size = (1 << FRAGMENT_BITS);


int main (int argc, char *argv [])
{

  int logchunk;
  block_handle bh;
  size_t size;
  size_t offset;
  int i;
  unsigned char *buffer;
  FILE *movfile;

  if ((argc != 2) && (argc != 3))
  {
    fprintf (stderr, "syntax: %s <log chunk no>\n", argv [0]);
    return 1;
  }

  logchunk = strtol (argv [1], NULL, 0);
  if (argc == 3)
  {
    Sample_rate = strtol (argv [2], NULL, 0);
  }

  /*
   * Initialise the resources and find the chunk we are looking for
   */
  Res_init ();
  Res_add_resfile (Res_make_path (CD_DATA_PATH, "citalog.res"));
  Res_add_resfile (Res_make_path (CD_DATA_PATH, "citbark.res"));
  Res_add_resfile (Res_make_path (HD_DATA_PATH, "digifx.res"));
  Res_add_resfile (Res_make_path (CD_DATA_PATH, "lowintr.res"));
/*    Res_print_resfile_info (); */
  bh = Res_load_chunk_by_id (logchunk, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    fprintf (stderr, "%s: no such log chunk\n", argv [0]);
    return 1;
  }

  movfile = fopen ("sound.chk", "wb");
  fwrite (*bh, 1, size, movfile);
  fclose (movfile);

  buffer = *bh;

  /*
   * We ought to test this by checking the chunk type, that's what it's for
   */
  if (strncmp ((char *) buffer, "MOVI", 4) == 0)
  {
    /*
     * Audiolog in some sort of movie format. This is organised into chunks
     *  but I'm not bothering with that for our purposes. The chunkdir is at
     *  offset 0x0400 and the 32-bit word at offset 0x0008 is its length.
     * Skip the header
     */
    offset = 0x400 + buffer [8] + (buffer [9] << 8) +
      (buffer [10] << 16) + (buffer [11] << 24);

    if (Sample_rate < 0)
    {
      /*
       * If sample rate hasn't been explicitly overwritten by the user, set it
       *  from the file. It's the 32-bit word at offset 0x0026
       */
      Sample_rate = buffer [38] + (buffer [39] << 8) +
	(buffer [40] << 16) + (buffer [41] << 24);
    }
    buffer += offset;
    size -= offset;

  }
  else if (strncmp ((char *) buffer, "Crea", 4) == 0)
  {
    /*
     * Digital fx in embedded Creative Labs .voc format. Even naughtier than
     *  above, assume it has only one data block which is sound data and take
     *  the sample rate from there
     */
    if (Sample_rate < 0)
    {
      Sample_rate = 1000000 / (256 - buffer [30]);
      buffer += 32;
    }
  }

  if (Sample_rate < 0)
  {

    /*
     * Default sample rate
     */
    Sample_rate = 22050;
  }

  /*
   * Got the chunk. Now initialise the sound system
   */

  /*
   * Initialise the DSP device
   */
  DSP_fd = open ("/dev/dsp", O_WRONLY);
  if (DSP_fd < 0)
  {
    fprintf (stderr, "%s: could not open /dev/dsp\n", argv [0]);
    return 1;
  }

# if DEBUG_SOUND
  printf ("/dev/dsp opened successfully\n");
# endif

  /*
   * Set the DSP to 8-bit mono, 11025Hz
   */
  i = (NUM_FRAGMENTS << 16) + FRAGMENT_BITS;
  if (ioctl (DSP_fd, SNDCTL_DSP_SETFRAGMENT, &i) < 0)
  {
    fprintf (stderr, "failed to set fragment size\n");
    close (DSP_fd);
    return 1;
  }
  Fragment_size = 1 << (i & 0xFFFF);
# if DEBUG_SOUND
  printf ("fragment size set to %d bytes\n", Fragment_size);
# endif

  i = AFMT_S8;
  if (ioctl (DSP_fd, SNDCTL_DSP_SETFMT, &i) < 0)
  {
    fprintf (stderr, "failed to set format\n");
    close (DSP_fd);
    return 1;
  }
# if DEBUG_SOUND
  printf ("set format to 8-bit signed\n");
# endif

  i = 0;
  if (ioctl (DSP_fd, SNDCTL_DSP_STEREO, &i) < 0)
  {
    fprintf (stderr, "failed to set mono mode\n");
    close (DSP_fd);
    return 1;
  }
# if DEBUG_SOUND
  printf ("set to mono\n");
# endif

  i = Sample_rate;
  if (ioctl (DSP_fd, SNDCTL_DSP_SPEED, &i) < 0)
  {
    fprintf (stderr, "failed to set sample rate\n");
    close (DSP_fd);
    return 1;
  }
  Sample_rate = i;
# if DEBUG_SOUND
  printf ("sample rate set to %dHz\n", Sample_rate);
# endif

  while (size > Fragment_size)
  {
    write (DSP_fd, buffer, Fragment_size);
    size -= Fragment_size;
    buffer += Fragment_size;
  }

  close (DSP_fd);

  return 0;

}



/*
 * Error handling
 */
void Cant_happen (const char *message)
{

  fprintf (stderr, "%s\n\n", message);
  fprintf (stderr, "Salt the Fries!\n\n");
  exit (1);

}
